/*
 * AdvertiseMaterial.java
 *
 * Copyright (c) 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is part of the Janos Bees Software.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */

package apps.authenticator;

import bees.core.Panic;
import bees.core.Assert;
import bees.core.XdrByteArray;

import bees.core.security.PermissionSet;

/**
 * An AdvertiseMaterial object is used to hold the material that is advertised
 * by the discovery protocol.  In this instance, the material is made up of
 * protocol and address information that can be used to create NodeOSNeighbors.
 */
final class AdvertiseMaterial
{
    /**
     * The local addresses that the neighbor will use to send/receive packets.
     */
    private PermissionSet address;

    /**
     * Construct an empty AdvertiseMaterial object.
     */
    AdvertiseMaterial()
    {
    }

    /**
     * @param ps The local addresses that the neighbor will use to send/receive
     * packets.
     */
    void setAddress(PermissionSet ps)
    {
	this.address = ps;
    }

    /**
     * @return The local addresses that the neighbor will use to send/receive
     * packets.
     */
    PermissionSet getAddress()
    {
	return this.address;
    }

    /**
     * @return The number of bytes needed to marshal this object.
     */
    int marshalledLength()
    {
	return this.address.marshalledLength();
    }

    /**
     * Marshal this object into the given destination.
     *
     * @param xba The destination byte array.
     */
    void marshal(XdrByteArray xba)
    {
	this.address.marshal(xba);
    }

    /**
     * Unmarshal this object from the given source.
     *
     * @param xba The source byte array.
     */
    void unmarshal(XdrByteArray xba)
    {
	this.address = PermissionSet.instantiate(xba);
    }

    /**
     * @return A flattened version of this object.
     */
    byte[] toBytes()
    {
	XdrByteArray xba;

	xba = new XdrByteArray(this.marshalledLength());
	this.marshal(xba);
	if( xba.index != xba.len )
	{
	    new Panic("marshalledLength() doesn't match actual length");
	}
	return xba.toBytes();
    }

    /**
     * @param flattened A flattened version of an AdvertiseMaterial object.
     * @return A new AdvertiseMaterial object initialized with the data in
     * "flattened".
     */
    static AdvertiseMaterial instantiate(byte flattened[])
    {
	AdvertiseMaterial retval;
	XdrByteArray xba;

	Assert.require(flattened);
	
	xba = new XdrByteArray(flattened);
	retval = new AdvertiseMaterial();
	retval.unmarshal(xba);
	if( xba.index != xba.len )
	{
	    throw new IllegalArgumentException(
		"Flattened AdvertiseMaterial is short");
	}
	return retval;
    }
    
    /**
     * @see java.lang.Object#toString()
     */
    public String toString()
    {
	return "AdvertiseMaterial[address="
	    + this.address
	    + "]";
    }
}
