/*
 * AuthenticatorUnknownProtocolPresentation.java
 *
 * Copyright (c) 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is part of the Janos Bees Software.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */

package apps.authenticator;

import java.util.Vector;
import java.util.NoSuchElementException;

import bees.core.Node;
import bees.core.Pair;
import bees.core.TypeID;
import bees.core.NodeCache;
import bees.core.ObjectMap;
import bees.core.PairChain;
import bees.core.Capability;
import bees.core.NeighborImpl;
import bees.core.CapabilityEnvelope;
import bees.core.CapabilityReference;
import bees.core.DLProtocolPresentation;
import bees.core.MissingCapabilityException;
import bees.core.UnknownProtocolPresentation;

import bees.core.resource.Logger;
import bees.core.resource.LoggerSpec;

import bees.core.security.PermissionSet;
import bees.core.security.IntegerPermission;

import bees.core.bdomain.CapabilityTree;
import bees.core.bdomain.BDomainCreationException;

import bees.util.MD5;

/**
 * An AuthenticatorUnknownProtocolPresentation is a base class for
 * authenticating unknown protocol flows.
 */
public abstract class AuthenticatorUnknownProtocolPresentation
    extends UnknownProtocolPresentation
{
    /**
     * Timeout for flow identifiers put into the Authenticators NodeCache.
     */
    private static final IntegerPermission FLOW_ID_TIMEOUT =
	new IntegerPermission("interval", 60 * 1000);
    
    /**
     * The Authenticator this presentation is being used by.
     */
    private final Authenticator auth;

    /**
     * Construct an AuthenticatorUnknownProtocolPresentation with the given
     * values.
     *
     * @param auth The Authenticator this presentation is being used by.
     */
    public AuthenticatorUnknownProtocolPresentation(Authenticator auth)
    {
	this.auth = auth;
    }

    /**
     * @return The Authenticator this presentation is being used by.
     */
    public Authenticator getAuthenticator()
    {
	return this.auth;
    }

    /**
     * Compute the Logger prefix for the new protocol flow.
     *
     * @param fp A finger print of the protocol flow.
     * @param pc The PairChain describing the protocol flow.
     * @return The Logger prefix for the given protocol flow.
     */
    protected String computePrefix(TypeID fp, PairChain pc)
    {
	return "DL[FlowID=" + fp.toPlainString() + "]";
    }

    /**
     * Create the Logger object for the new protocol flow.
     *
     * @param fp A finger print of the protocol flow.
     * @param pc The PairChain describing the protocol flow.
     * @return A CapabilityEnvelope wrapping the new Logger.
     */
    protected CapabilityEnvelope createLogger(TypeID fp, PairChain pc)
    {
	return new CapabilityEnvelope(
		"logger",
		this.auth.getLoggerFactory().
		create(Logger.currentLogger(),
		       new LoggerSpec(this.computePrefix(fp, pc), 0),
		       Logger.DEFAULT_PERMISSIONS,
		       this.computePrefix(fp, pc)));
    }

    /**
     * @return A CapabilityTree containing the set of discovered neighbors
     * under the tree "user/neighbors".
     */
    CapabilityTree createNeighborTree()
    {
	return new CapabilityTree("user/neighbors",
				  this.auth.getNeighborManager().
				  namedNeighbors());
    }

    /**
     * @param fp A finger print of the protocol flow.
     * @param pc The PairChain describing the protocol flow.
     * @return A CapabilityTree containing all the capabilities to be passed
     * to the new protocol flow.
     */
    protected CapabilityTree computeCapabilities(TypeID fp, PairChain pc)
    {
	CapabilityTree retval;

	retval = new CapabilityTree(new CapabilityEnvelope[] {
	    this.createLogger(fp, pc),
	}, new CapabilityTree[] {
	    this.auth.getCommonCapabilities(),
	    this.createNeighborTree(),
	    this.computeAuthority(pc),
	});
	return retval;
    }

    /**
     * Find a group of capabilities based on a set of tags in a PairChain.
     * 
     * @param prefix The prefix to use when computing the names of the
     * within the Authenticator's name space.
     * @param pair The pair containing the PairChain of capability names.
     * @return An array of capabilities that correspond to those requested by
     * the PairChain.
     */
    protected Capability[] findCapabilities(String prefix, Pair pair)
    {
	Capability retval[];
	PairChain pc;
	Pair pairs[];
	int lpc;
	
	pc = PairChain.instantiate(pair.getValue());
	retval = new Capability[pc.size()];
	pairs = pc.toArray();
	for( lpc = 0; lpc < pairs.length; lpc++ )
	{
	    String source;
	    
	    if( pairs[lpc].getTag().startsWith("/global") )
	    {
		source = pairs[lpc].getTag();
	    }
	    else
	    {
		source = prefix + "/" + pairs[lpc].getTag();
	    }
	    retval[lpc] = this.auth.getNode().getCapability(source);
	}
	return retval;
    }

    /**
     * Wrap a Capability described by the given Pair in an envelope for
     * delivery to the new protocol flow.
     *
     * @param caps The Vector of capabilities that will be passed to the new flow.
     * @param prefix The prefix to prepend to "pair.getTag()" when looking up
     * the value.
     * @param pair A Pair whose tag names a Capability and value is a PairChain
     * optionally containing, "name", the name to use in the destination flow,
     * "revoke", the PermissionSet to revoke from the capability, and "bound",
     * the PermissionSet to bound the capability by.
     */
    protected void mungeCapability(Vector caps, String prefix, Pair pair)
    {
	PermissionSet revoked = null, bounded = null;
	String name = null, source;
	CapabilityEnvelope retval;
	ObjectMap om = null;
	Capability cap;
	PairChain pc;
	Pair psPair;

	if( pair.getTag().startsWith("/global") )
	{
	    source = pair.getTag();
	}
	else
	{
	    source = prefix + "/" + pair.getTag();
	}
	pc = PairChain.instantiate(pair.getValue());
	if( (psPair = pc.get("name")) != null )
	{
	    name = psPair.getStringValue();
	}
	if( (psPair = pc.get("revoke")) != null )
	{
	    revoked = PermissionSet.instantiate(psPair.getValue());
	}
	if( (psPair = pc.get("bound")) != null )
	{
	    bounded = PermissionSet.instantiate(psPair.getValue());
	}
	try
	{
	    CapabilityEnvelope ce;
	    
	    cap = this.auth.getNode().getCapability(source);
	    ce = new CapabilityEnvelope(name, cap);
	    if( revoked != null )
	    {
		ce.revokePermissions(revoked);
	    }
	    if( bounded != null )
	    {
		ce.boundPermissions(bounded);
		Logger.log("Bounding permissions of: "
			   + cap
			   + "; to: "
			   + bounded
			   + "; in: "
			   + ce);
	    }
	    caps.addElement(ce);
	}
	catch(ClassCastException e)
	{
	    om = this.auth.getNode().getCapabilityMap(source);
	    CapabilityTree.flattenMap(caps, name, om);
	}
    }

    /**
     * Wrap a set of capabilities described by the given Pair.
     *
     * @param prefix The prefix to prepend to "pair.getTag()" when looking up
     * the value.
     * @param pair A Pair whose value contains a PairChain whose elements can
     * be passed to mungeCapability.
     * @return An initialized CapabilityTree containing the capabilities
     * described by the given Pair.
     */
    protected CapabilityTree mungeCapabilities(String prefix, Pair pair)
    {
	Vector caps = new Vector();
	CapabilityEnvelope ce[];
	CapabilityTree retval;
	PairChain pc;
	Pair pairs[];
	int lpc;
	
	pc = PairChain.instantiate(pair.getValue());
	pairs = pc.toArray();
	for( lpc = 0; lpc < pairs.length; lpc++ )
	{
	    try
	    {
		this.mungeCapability(caps, prefix, pairs[lpc]);
	    }
	    catch(NoSuchElementException e)
	    {
		Logger.log(e);
	    }
	}
	ce = new CapabilityEnvelope[caps.size()];
	caps.copyInto(ce);
	return new CapabilityTree(ce);
    }

    /**
     * Callback used to validate a protocol flow description.
     *
     * @param pc The protocol flow description.
     * @return True if the description is valid and should be accepted for
     * installation on this node.
     * @throws Throwable If anything goes wrong.
     */
    protected abstract boolean validate(PairChain pc)
	throws Throwable;

    /**
     * @param pc The protocol flow description.
     * @return An initialized CapabilityTree containing any extra capabilities
     * that should be passed to the flow.
     */
    protected abstract CapabilityTree computeAuthority(PairChain pc);

    /**
     * Call back used to compute the set of initial capabilities for a new
     * BDomain.  Usually, this is just a set of resource meters.
     * 
     * @param pc The flow description PairChain.
     * @return An array of capabilities to be used when creating the new
     * BDomain.
     */
    protected abstract Capability[] computeInitialCapabilities(PairChain pc);

    /**
     * Call back used to compute the resource bounds for a new BDomain.
     * 
     * @param pc The flow description PairChain.
     * @return A PermissionSet containing the resource bounds.
     */
    protected abstract PermissionSet computeResources(PairChain pc);

    /**
     * @see bees.core.UnknownProtocolPresentation#newProtocol(CapabilityReference, byte[])
     */
    public void newProtocol(CapabilityReference neighbor, byte flowDesc[])
    {
	MD5 digest = new MD5();
	NodeCache nc;
	PairChain pc;
	TypeID fp;
	Node node;

	digest.update(flowDesc);
	fp = new TypeID(digest.hash());
	nc = this.auth.getNodeCache();
	synchronized( nc )
	{
	    if( nc.get(fp) == null )
	    {
		/*
		 * XXX We should probably put something in here that kills
		 * the flow if it takes too long to download.
		 */
		nc.put(fp, fp, FLOW_ID_TIMEOUT);
	    }
	    else
	    {
		/* The flow's already been started. */
		Logger.log("Already downloaded flow description.");
		return;
	    }
	}
	
	node = this.auth.getNode();
	pc = PairChain.instantiate(flowDesc);
	Logger.log(Logger.NOTICE, "Got Flow description: " + pc);
	/* Validate the flow before going any further. */
	try
	{
	    if( this.validate(pc) )
	    {
		Logger.log(Logger.INFO, "Valid flow: " + fp);
		try
		{
		    DLProtocolPresentation dpp;

		    dpp = this.auth.createDLProtocolPresentation(fp);
		    /* Register the flow. */
		    node.register(dpp,
				  (NeighborImpl)neighbor.dereference(),
				  pc,
				  this.computeInitialCapabilities(pc),
				  this.computeResources(pc),
				  this.computeCapabilities(fp, pc));
		}
		catch(MissingCapabilityException e)
		{
		    Logger.log(e);
		}
		catch(BDomainCreationException e)
		{
		    Logger.log(e);
		}
		catch(Throwable e)
		{
		    Logger.log(e);
		}
	    }
	    else
	    {
		Logger.log(Logger.INFO, "Invalid flow: " + fp);
	    }
	}
	catch(Throwable th)
	{
	    Logger.log(Logger.INFO, "Invalid flow: " + fp);
	    Logger.log(Logger.INFO, th);
	}
    }

    /**
     * @see java.lang.Object#toString()
     */
    public String toString()
    {
	return "AuthenticatorUnknownProtocolPresentation["
	    + super.toString()
	    + "; auth="
	    + this.auth
	    + "]";
    }
}
