/*
 * FlowBuilder.java
 *
 * Copyright (c) 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is part of the Janos Bees Software.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */

package apps.authenticator;

import java.util.Stack;
import java.util.NoSuchElementException;

import edu.utah.janosvm.kit.comm.jsi.JSIMessage;

import bees.core.Node;
import bees.core.Pair;
import bees.core.ByteArray;
import bees.core.PairChain;
import bees.core.Capability;
import bees.core.Application;
import bees.core.ProtocolFlow;

import bees.core.kernel.KernelCapabilities;

import bees.core.resource.File;
import bees.core.resource.Logger;

import bees.core.security.Signature;
import bees.core.security.PrivateKey;

/**
 * The FlowBuilder application is used to construct protocol flow descriptions.
 * A description is made up of the protocol tree, consisting of the primary
 * protocol and its companions.  After the tree, there can be any number of
 * "pairs", which contain a string identifier and an opaque byte array.  Pairs
 * can then be strung together into a "pair chain", which can be flattened and
 * made into the value part of another pair.
 * 
 * The arguments accepted by the application are as follows:
 *
 * <ul>
 * <li>&lt;Primary protocol name&gt; - The name of the primary protocol.  This
 * protocol is basically the top of the tree, therefore, the flow will be made
 * up of it and all of its companions.
 * <li>[-companion &lt;protocol name&gt;] - Add a companion protocol.
 * <li>[-start-chain &lt;pair chain name&gt;] - Start a new pair chain.
 * <li>[-end-chain &lt;pair chain name&gt;] - End a pair chain.
 * <li>[-push-chain &lt;pair chain name&gt;] - Push an existing chain.
 * <li>[-pop-chain &lt;pair chain name&gt;] - Pop a pushed chain.
 * <li>[-start-sign &lt;private key&gt;;&lt;algorithm&gt;] - Start a signature.
 * <li>[-sign &lt;name&gt;] - Finish and add a signature.
 * <li>[-update-sig &lt;name&gt;] - Update a signature with a pair.
 * <li>[-pair &lt;name&gt;=&lt;value&gt;] - Add a generic pair.
 * <li>[-start-permissions &lt;name&gt;] - Start a PermissionSet.
 * <li>[-end-permissions &lt;name&gt;] - End a PermissionSet.
 * <li>[-permission &lt;name&gt;=&lt;value&gt;] - Add a permission.
 * </ul>
 *
 * The capability arguments are:
 *
 * <ul>
 * <li>file - The file to write the description to.
 * <li>[&lt;private keys&gt;] - Any private keys needed to sign the
 * description.
 * </ul>
 *
 * An example usage would be:
 * <pre>
 * bees.core.FlowBuilder
 *   apps.health.HealthProtocol
 *   -companion bees.core.UnknownProtocol
 *   -start-chain source-auth
 *     -pair source=bob
 *     -pair algorithm=DSA
 *     -start-chain capability.files/foo
 *       -pair destination=user/foo
 *       -start-permissions revoke
 *         -permission file=write
 *       -end-permissions
 *     -end-chain capability.foo
 *   -end-chain source-auth
 *   -start-sign privateKey;DSA
 *   -update-sig primary-protocol
 *   -update-sig companion-protocol
 *   -update-sig source-auth
 *   -sign source-auth-signature
 * </pre>
 *
 * The first argument .
 */
public final class FlowBuilder
    extends Application
{
    /**
     * Construct a FlowBuilder object.
     */
    public FlowBuilder()
    {
    }

    /**
     * Split a string at the first instance of a character.  For example,
     * "split("foo=bar", '=')" will return "new String[] { "foo", "bar" }".
     *
     * @param str The string to split.
     * @param token The token to look for in "str".
     * @return The pair of strings created from the split.
     * @throws IllegalArgumentException If the token cannot be found in the
     * string.
     */
    private String[] split(String str, char token)
	throws IllegalArgumentException
    {
	String retval[] = new String[2];
	int index;
	
	if( (index = str.indexOf(token)) != -1 )
	{
	    retval = new String[] {
		str.substring(0, index),
		str.substring(index + 1)
	    };
	}
	else
	{
	    throw new IllegalArgumentException("Token \""
					       + token
					       + "\" not found in string: "
					       + str);
	}
	return retval;
    }

    /**
     * @return A usage string.
     */
    private String usage()
    {
	return "Usage: " + this.getClass().getName()
	    + " <primary protocol name>\n"
	    + "\t[-companion <protocol name>] - Add a companion protocol.\n"
	    + "\t[-start-chain <pair chain name>] - Start a new pair chain.\n"
	    + "\t[-end-chain <pair chain name>] - End a pair chain.\n"
	    + "\t[-push-chain <pair chain name>] - Push an existing chain.\n"
	    + "\t[-pop-chain <pair chain name>] - Pop a pushed chain.\n"
	    + "\t[-start-sign <private key>;<algorithm>] - Start a signature.\n"
	    + "\t[-sign <name>] - Finish and add a signature.\n"
	    + "\t[-update-sig <name>] - Update a signature with a pair.\n"
	    + "\t[-pair <name>=<value>] - Add a generic pair.\n"
	    + "\t[-start-permissions <name>] - Start a PermissionSet.\n"
	    + "\t[-end-permissions <name>] - End a PermissionSet.\n"
	    + "\t[-permission <name>=<value>] - Add a permission.\n"
	    + "\n"
	    + "Capabilities:\n"
	    + "\tfile - The file to write the description to.\n"
	    + "\t[<private keys>] - Any private keys needed to sign the "
	    + "description.\n";
    }

    /**
     * @see bees.core.Application#start(Node, String[])
     */
    public void start(Node node, String args[])
    {
	try
	{
	    this.build(node, args);
	}
	catch(Throwable th)
	{
	    Logger.log(th);
	    Logger.log(Logger.INFO, this.usage());
	}
	node.shutdown();
    }

    /**
     * Build the protocol flow description from the given arguments.
     *
     * @param node The current node.
     * @param args The application arguments.
     * @throws Throwable If anything goes wrong.
     */
    private void build(Node node, String args[])
	throws Throwable
    {
	PairChain flowChain;
	ProtocolFlow pf;
	File file;

	pf = new ProtocolFlow();
	pf.setPrimaryProtocol(args[0]);
	/* Get the PairChain wrapped by the ProtocolFlow. */
	flowChain = pf.getFlowDesc();
	if( ((args.length - 1) % 2) == 0 )
	{
	    JSIMessage permissions = null;
	    Stack pairStack = new Stack();
	    Signature signer = null;
	    PairChain pc, pc2;
	    String spair[];
	    int lpc;

	    /* Process the rest of the arguments. */
	    pairStack.push(flowChain);
	    for( lpc = 1; lpc < args.length; lpc += 2 )
	    {
		if( args[lpc].equals("-companion") )
		{
		    pf.addCompanion(args[lpc + 1]);
		}
		else if( args[lpc].equals("-start-chain") )
		{
		    pairStack.push(new PairChain());
		}
		else if( args[lpc].equals("-end-chain") )
		{
		    pc = (PairChain)pairStack.pop();
		    pc2 = (PairChain)pairStack.peek();
		    pc2.add(args[lpc + 1], pc.toBytes());
		}
		else if( args[lpc].equals("-push-chain") )
		{
		    pc = (PairChain)pairStack.peek();
		    pairStack.push(PairChain.instantiate(pc.get(args[lpc + 1]).
							 getValue()));
		}
		else if( args[lpc].equals("-pop-chain") )
		{
		    pairStack.pop();
		}
		else if( args[lpc].equals("-start-sign") )
		{
		    PrivateKey pk;
		    
		    spair = this.split(args[lpc + 1], ';');
		    pk = (PrivateKey)node.getCapability(spair[0]);
		    signer = new Signature(spair[1]);
		    signer.initSign(pk);
		}
		else if( args[lpc].equals("-sign") )
		{
		    pc = (PairChain)pairStack.peek();
		    pc.add(args[lpc + 1], signer.sign());
		    signer = null;
		}
		else if( args[lpc].equals("-update-sig") )
		{
		    int instance;
		    Pair pair;
		    
		    pc = (PairChain)pairStack.peek();
		    for( instance = 0;
			 (pair = pc.get(args[lpc + 1], instance)) != null;
			 instance++ )
		    {
			signer.update(pair.getValue());
		    }
		}
		else if( args[lpc].equals("-pair") )
		{
		    spair = this.split(args[lpc + 1], '=');
		    try
		    {
			Capability cap;
			
			cap = node.getCapability(args[1]);
			// value = cap...
		    }
		    catch(NoSuchElementException e)
		    {
		    	/* Ignore me... */
		    }
		    pc = (PairChain)pairStack.peek();
		    pc.add(spair[0], spair[1]);
		}
		else if( args[lpc].equals("-start-permissions") )
		{
		    permissions = new JSIMessage();
		}
		else if( args[lpc].equals("-end-permissions") )
		{
		    pc = (PairChain)pairStack.peek();
		    pc.add(args[lpc + 1],
			   KernelCapabilities.toPermissionSet(permissions).
			   toBytes());
		}
		else if( args[lpc].endsWith("ermission") )
		{
		    permissions.add(args[lpc].substring(1),
				    args[lpc + 1]);
		}
		else
		{
		    throw new IllegalArgumentException(args[lpc]);
		}
	    }
	}
	else
	{
	    throw new IllegalArgumentException("Uneven number of flags.");
	}
	file = (File)node.getCapability("file");
	ByteArray.toFile(file, flowChain.toBytes());
    }

    /**
     * @see java.lang.Object#toString()
     */
    public String toString()
    {
	return "FlowBuilder["
	    + super.toString()
	    + "]";
    }
}
