/*
 * FlowDumper.java
 *
 * Copyright (c) 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is part of the Janos Bees Software.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */

package apps.authenticator;

import java.util.Stack;

import bees.core.Node;
import bees.core.Pair;
import bees.core.ByteArray;
import bees.core.PairChain;
import bees.core.Capability;
import bees.core.Application;
import bees.core.ProtocolType;
import bees.core.CapabilityEnvelope;

import bees.core.resource.File;
import bees.core.resource.Logger;

import bees.core.security.PublicKey;
import bees.core.security.Signature;
import bees.core.security.PermissionSet;

/**
 * The FlowDumper application is used to dump a compiled flow description to
 * the log.
 */
public final class FlowDumper
    extends Application
{
    /**
     * Construct a FlowDumper application.
     */
    public FlowDumper()
    {
    }

    /**
     * Split a string at the first instance of a character.  For example,
     * "split("foo=bar", '=')" will return "new String[] { "foo", "bar" }".
     *
     * @param str The string to split.
     * @param token The token to look for in "str".
     * @return The pair of strings created from the split.
     * @throws IllegalArgumentException If the token cannot be found in the
     * string.
     */
    private String[] split(String str, char token)
	throws IllegalArgumentException
    {
	String retval[] = new String[2];
	int index;
	
	if( (index = str.indexOf(token)) != -1 )
	{
	    retval[0] = str.substring(0, index);
	    retval[1] = str.substring(index + 1);
	}
	else
	{
	    throw new IllegalArgumentException("Token \""
					       + token
					       + "\" not found in string: "
					       + str);
	}
	return retval;
    }

    /**
     * Formats the given byte array into a string.  If the byte array
     * represents printable ASCII characters, then it is turned into a java
     * String and returned.  Otherwise, a string containing the hexadecimal
     * values of the byte array is returned.
     * 
     * @param data The byte array to format.
     * @return A printable string object.
     */
    private String pairValue(byte data[])
    {
	boolean isPrintable = true;
	String retval;
	int lpc;

	for( lpc = 0; (lpc < data.length) && isPrintable; lpc++ )
	{
	    char cval;
		
	    cval = (char)data[lpc];
	    if( !Character.isWhitespace(cval) )
	    {
		switch( Character.getType(cval) )
		{
		case Character.DECIMAL_DIGIT_NUMBER:
		    
		case Character.UPPERCASE_LETTER:
		case Character.LOWERCASE_LETTER:
		case Character.TITLECASE_LETTER:
		case Character.MODIFIER_LETTER:
		case Character.OTHER_LETTER:
		    
		case Character.DASH_PUNCTUATION:
		case Character.START_PUNCTUATION:
		case Character.END_PUNCTUATION:
		case Character.CONNECTOR_PUNCTUATION:
		case Character.OTHER_PUNCTUATION:
		    break;
		default:
		    isPrintable = false;
		    break;
		}
	    }
	}
	if( isPrintable )
	{
	    retval = new String(data);
	}
	else
	{
	    retval = ByteArray.toString(data);
	}
	return retval;
    }

    /**
     * Print an array of Pair objects.
     * 
     * @param root The source of the pair objects.
     * @param pairs The Pair objects to print.
     */
    private void printPairs(String root, Pair pairs[])
    {
	int lpc;

	for( lpc = 0; lpc < pairs.length; lpc++ )
	{
	    if( pairs[lpc].getTag().equals("primary-protocol") ||
		pairs[lpc].getTag().equals("companion-protocol") )
	    {
		ProtocolType pt;

		pt = ProtocolType.instantiate(pairs[lpc].getValue());
		Logger.log(root
			   + "."
			   + pairs[lpc].getTag()
			   + "="
			   + pt);
	    }
	    else
	    {
		try
		{
		    PairChain pc;
		    
		    pc = PairChain.instantiate(pairs[lpc].getValue());
		    if( pc.size() == 0 )
		    {
			Logger.log(root
				   + "."
				   + pairs[lpc].getTag()
				   + "=");
		    }
		    else
		    {
			this.printPairs(root + "." + pairs[lpc].getTag(),
					pc.toArray());
		    }
		}
		catch(Throwable th)
		{
		    try
		    {
			PermissionSet ps;
			
			ps = PermissionSet.instantiate(pairs[lpc].getValue());
			Logger.log(root
				   + "."
				   + pairs[lpc].getTag()
				   + "="
				   + ps);
		    }
		    catch(Throwable th2)
		    {
			Logger.log(root
				   + "."
				   + pairs[lpc].getTag()
				   + "="
				   + this.pairValue(pairs[lpc].getValue()));
		    }
		}
	    }
	}
    }

    /**
     * @see bees.core.Application#start(Node, String[])
     */
    public void start(Node node, String args[])
	throws Throwable
    {
	PairChain pc;
	byte data[];
	File file;

	file = (File)node.getCapability("file");
	data = ByteArray.fromFile(file);
	pc = PairChain.instantiate(data);
	this.printPairs("", pc.toArray());
	if( (args.length % 2) == 0 )
	{
	    Stack pairStack = new Stack();
	    Signature verifier = null;
	    String spair[];
	    int lpc;

	    pairStack.push(pc);
	    for( lpc = 0; lpc < args.length; lpc += 2 )
	    {
		if( args[lpc].equals("-push-chain") )
		{
		    pc = (PairChain)pairStack.peek();
		    pairStack.push(PairChain.instantiate(pc.get(args[lpc + 1]).
							 getValue()));
		}
		else if( args[lpc].equals("-pop-chain") )
		{
		    pairStack.pop();
		}
		else if( args[lpc].equals("-print-cap") )
		{
		    CapabilityEnvelope ce;
		    PairChain capPC;
		    Capability cap;
		    Pair pair;
		    
		    pc = (PairChain)pairStack.peek();
		    cap = node.getCapability(args[lpc + 1]);
		    ce = new CapabilityEnvelope(cap);
		    pair = pc.get(args[lpc + 1]);
		    capPC = PairChain.instantiate(pair.getValue());
		    if( (pair = capPC.get("revoke")) != null )
		    {
			ce.revokePermissions(PermissionSet.
					     instantiate(pair.getValue()));
		    }
		    else if( (pair = capPC.get("bound")) != null )
		    {
			ce.boundPermissions(PermissionSet.
					    instantiate(pair.getValue()));
		    }
		    Logger.log(args[lpc + 1]
			       + "="
			       + ce);
		}
		else if( args[lpc].equals("-start-verify") )
		{
		    PublicKey pk;
		    
		    spair = this.split(args[lpc + 1], '/');
		    pk = (PublicKey)node.getCapability(spair[0]);
		    verifier = new Signature(spair[1]);
		    verifier.initVerify(pk);
		}
		else if( args[lpc].equals("-verify") )
		{
		    pc = (PairChain)pairStack.peek();
		    Logger.log("Signature: "
			       + args[lpc + 1]
			       + "; verifies="
			       + verifier.verify(pc.get(args[lpc + 1]).
						 getValue()));
		    verifier = null;
		}
		else if( args[lpc].equals("-update-sig") )
		{
		    pc = (PairChain)pairStack.peek();
		    verifier.update(pc.get(args[lpc + 1]).getValue());
		}
	    }
	}
    }

    /**
     * @see java.lang.Object#toString()
     */
    public String toString()
    {
	return "FlowDumper["
	    + super.toString()
	    + "]";
    }
}
