/*
 * SourceBased.java
 *
 * Copyright (c) 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is part of the Janos Bees Software.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */

package apps.authenticator;

import java.security.SignatureException;

import bees.core.Node;
import bees.core.Pair;
import bees.core.PairChain;
import bees.core.Capability;
import bees.core.UnknownProtocolPresentation;

import bees.core.resource.Logger;

import bees.core.security.Signature;
import bees.core.security.PublicKey;
import bees.core.security.PermissionSet;

import bees.core.bdomain.CapabilityTree;

/**
 * A source based authenticator that will only allow flows that are from a
 * known source.
 */
public class SourceBased
    extends Authenticator
{
    /**
     * Construct a SourceBased Authenticator.
     */
    public SourceBased()
    {
    }
    
    /**
     * The actual UnknownProtocol authenticator.
     */
    private class MyUnknownProtocolPresentation
	extends AuthenticatorUnknownProtocolPresentation
    {
	/**
	 * Construct a MyUnknownProtocolPresentation with the given values.
	 *
	 * @param auth The Authenticator this presentation is being used by.
	 */
	MyUnknownProtocolPresentation(Authenticator auth)
	{
	    super(auth);
	}

	/**
	 * @param sig The Signature object to update.
	 * @param pc The PairChain containing the Pairs that should be covered
	 * by the signature.
	 * @param name The name of the pairs in "pc" to cover.
	 * @throws SignatureException If there are any problems updating the
	 * signature.
	 */
	private void updateSignature(Signature sig, PairChain pc, String name)
	    throws SignatureException
	{
	    Pair pair;
	    int lpc;

	    for( lpc = 0; (pair = pc.get(name, lpc)) != null; lpc++ )
	    {
		sig.update(pair.getValue());
	    }
	}
	
        /**
         * @see apps.authenticator.AuthenticatorUnknownProtocolPresentation#validate(PairChain)
         */
	protected boolean validate(PairChain pc)
	    throws Throwable
	{
	    boolean retval = false;
	    Signature verifier;
	    String algorithm;
	    PairChain auth;
	    PublicKey pk;
	    Node node;

	    node = super.getAuthenticator().getNode();
	    auth = PairChain.instantiate(pc.expect("source-auth").getValue());
	    algorithm = auth.expect("algorithm").getStringValue();
	    pk = (PublicKey)node.getCapability("users/"
					       + auth.expect("source").
					       getStringValue()
					       + "/keys/"
					       + algorithm
					       + "/public");
	    verifier = new Signature(algorithm);
	    verifier.initVerify(pk);
	    this.updateSignature(verifier, pc, "primary-protocol");
	    this.updateSignature(verifier, pc, "companion-protocol");
	    this.updateSignature(verifier, pc, "source-auth");
	    retval = verifier.verify(pc.get("source-auth-signature").
				     getValue());
	    return retval;
	}

        /**
         * @see apps.authenticator.AuthenticatorUnknownProtocolPresentation#computeAuthority(PairChain)
         */
	protected CapabilityTree computeAuthority(PairChain pc)
	{
	    CapabilityTree retval;
	    
	    pc = PairChain.instantiate(pc.get("source-auth").getValue());
	    retval = super.mungeCapabilities("users/"
					     + pc.get("source").
					     getStringValue(),
					     pc.get("runtime-caps"));
	    return retval;
	}

        /**
         * @see apps.authenticator.AuthenticatorUnknownProtocolPresentation#computeInitialCapabilities(PairChain)
         */
	protected Capability[] computeInitialCapabilities(PairChain pc)
	{
	    CapabilityTree retval;
	    
	    pc = PairChain.instantiate(pc.get("source-auth").getValue());
	    return super.findCapabilities("users/"
					  + pc.get("source").getStringValue(),
					  pc.get("initial-caps"));
	}

        /**
         * @see apps.authenticator.AuthenticatorUnknownProtocolPresentation#computeResources(PairChain)
         */
	protected PermissionSet computeResources(PairChain pc)
	{
	    pc = PairChain.instantiate(pc.get("source-auth").getValue());
	    return PermissionSet.instantiate(pc.get("resources").getValue());
	}
	
        /**
         * @see java.lang.Object#toString()
         */
	public String toString()
	{
	    return "MyUnknownProtocolPresentation["
		+ super.toString()
		+ "]";
	}
    }

    /**
     * @see apps.authenticator.Authenticator#createUnknownProtocolPresentation()
     */
    protected UnknownProtocolPresentation createUnknownProtocolPresentation()
    {
	return new MyUnknownProtocolPresentation(this);
    }

    /**
     * @see java.lang.Object#toString()
     */
    public String toString()
    {
	return "SourceBased["
	    + super.toString()
	    + "]";
    }
}
