/*
 * BenchmarkSession.java
 *
 * Copyright (c) 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is part of the Janos Bees Software.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */

package apps.benchmark;

import java.util.Enumeration;

import java.io.IOException;

import edu.utah.janosvm.kit.comm.jsi.JSIPair;
import edu.utah.janosvm.kit.comm.jsi.JSIHandler;
import edu.utah.janosvm.kit.comm.jsi.JSIMessage;
import edu.utah.janosvm.kit.comm.jsi.JSIPropertyInfo;
import edu.utah.janosvm.kit.comm.jsi.JSIHandlerContext;
import edu.utah.janosvm.kit.comm.jsi.MissingPairException;

import bees.core.Node;
import bees.core.Capsule;
import bees.core.Neighbor;
import bees.core.NeighborImpl;
import bees.core.ExternalCapsule;
import bees.core.EncodingException;
import bees.core.ExtendedProtocolSession;

import bees.core.resource.Logger;
import bees.core.resource.NodeOSNeighbor;

final class BenchmarkSession
    extends ExtendedProtocolSession
    implements JSIHandler
{
    static final int MAX_RUN_LENGTH = 1000000;
    
    static final int VERSION_MAJOR = 0;
    
    static final int VERSION_MINOR = 5;

    static final int STATE_PRIMING = 0;
    static final int STATE_IDLE = 1;
    static final int STATE_TESTING = 2;
    static final int STATE_QUERY = 3;
    
    BenchmarkParameters bp;

    final StatsCapsule lastRun = new StatsCapsule();
    final StatsCapsule currentRun = new StatsCapsule();
    
    long lastStat;
    int intervalCount;

    int state = STATE_PRIMING;

    Neighbor previous;
    Neighbor next;

    boolean source = false;
    private boolean testing = false;
    
    public BenchmarkSession()
    {
    }

    synchronized void waitForTurn()
	throws InterruptedException
    {
	while( this.state != STATE_IDLE )
	{
	    this.wait();
	}
	this.state = STATE_TESTING;
    }

    synchronized void turnOver()
    {
	this.state = STATE_IDLE;
	this.notify();
    }

    public synchronized Neighbor setPrevious(Neighbor n)
    {
	Enumeration enum;
	Neighbor retval;

	if( this.previous == null )
	{
	    this.previous = n;
	}
	enum = super.neighbors.elements();
	while( enum.hasMoreElements() )
	{
	    NeighborImpl ni;

	    ni = (NeighborImpl)enum.nextElement();
	    n = ni.getNeighbor(this);
	    if( n != this.previous )
	    {
		this.next = n;
		break;
	    }
	}
	retval = this.next;
	return( retval );
    }

    public synchronized void attachNeighbor(Node node, Neighbor neighbor)
    {
	super.attachNeighbor(node, neighbor);
	if( this.next == null )
	{
	    this.next = neighbor;
	}
    }

    public synchronized void detachNeighbor(Node node, Neighbor neighbor)
    {
	if( this.next == neighbor )
	{
	    this.next = null;
	}
	super.detachNeighbor(node, neighbor);
    }

    public JSIHandler resolveSpecifier(JSIMessage jm, JSIPair jp)
    {
	return this;
    }

    public void handleMessage(JSIHandlerContext jhc, JSIMessage jm)
	throws InterruptedException,
	       EncodingException,
	       IOException
    {
	JSIPair jp;

	jm.restoreSpecifier();
	jp = jm.popSpecifier();
	switch( jm.getWhat() )
	{
	case JSIMessage.EXECUTE_PROPERTY:
	    if( jp.getTag().equals("send") )
	    {
		int lpc, size = 0, cycles = 100000;
		SendCapsule sc;
		Node node;

		node = this.getCache().getOwnerNode();
		if( (jp = jm.getPair("size")) != null )
		{
		    size = (int)jp.getIntValue();
		}
		if( (jp = jm.getPair("cycles")) != null )
		{
		    cycles = (int)jp.getIntValue();
		}
		try
		{
		    this.waitForTurn();
		    sc = new SendCapsule();
		    sc.setDataLength(size);
		    sc.fromCapsule(this.currentRun);
		    for( lpc = 0; lpc < cycles; lpc++ )
		    {
			this.next.send(sc);
			sc.recordHit(node, this);
		    }
		    jm.clear();
		    jm.setWhat(JSIMessage.REPLY);
		    jm.add(new JSIPair("start", this.currentRun.start));
		    jm.add(new JSIPair("last", this.currentRun.last));
		    jm.add(new JSIPair("packetCount",
				       this.currentRun.packetCount));
		    Logger.log("run = " + this.currentRun);
		    jm.add(new JSIPair("pps",
				       (this.currentRun.packetCount * 1000) /
					(this.currentRun.last -
					 this.currentRun.start)));
		    this.currentRun.reset();
		}
		finally
		{
		    this.turnOver();
		}
	    }
	    else
	    {
		jm.error("Unknown property: " + jp);
	    }
	    break;
	default:
	    throw new UnsupportedOperationException();
	}
    }

    public void getSuites(JSIMessage jm)
    {
    }
    
    public String toString()
    {
	return "BenchmarkSession["
	    + super.toString()
	    + "; bp="
	    + this.bp
	    + "; lastStat="
	    + this.lastStat
	    + "; previous="
	    + this.previous
	    + "; next="
	    + this.next
	    + "]";
    }
}
