/*
 * AdvertiseMaterial.java
 *
 * Copyright (c) 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is part of the Janos Bees Software.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */

package apps.discovery;

import bees.core.Node;
import bees.core.TypeID;
import bees.core.NodeCache;
import bees.core.ProtocolSession;

import bees.core.resource.Logger;

import bees.core.security.IntegerPermission;

/**
 * An AdvertiseMaterial class is used to remove a finger print from the set
 * of discovered material after a certain amount of time has expired.
 *
 * XXX We don't actually store the material in here, should we?
 */
public final class AdvertiseMaterial
    implements NodeCache.EventHandler
{
    /**
     * The finger print to remove after expiration.
     */
    private final TypeID fingerPrint;

    /**
     * Construct an AdvertiseMaterial object with the given values.
     *
     * @param fp The finger print to remove after expiration.
     */
    public AdvertiseMaterial(TypeID fp)
    {
	this.fingerPrint = fp;
    }

    /**
     * @return The finger print to remove after expiration.
     */
    public TypeID getFingerPrint()
    {
	return this.fingerPrint;
    }
    
    /**
     * @see bees.core.NodeCache.EventHandler#removed(Node, ProtocolSession)
     */
    public void removed(Node node, ProtocolSession ps)
    {
	DiscoveryProtocolSession dps;

	Logger.log(Logger.INFO,
		   "Removing stale material: fp=" + this.fingerPrint);
	/*
	 * Remove the finger print from the set of discovered
	 * advertisements.
	 */
	dps = (DiscoveryProtocolSession)ps;
	dps.getDiscoveredMaterial().remove(this.fingerPrint);
    }

    /**
     * @see bees.core.NodeCache.EventHandler#collected(Node, ProtocolSession)
     */
    public IntegerPermission collected(Node node, ProtocolSession ps)
    {
	return null;
    }
    
    /**
     * @see java.lang.Object#toString()
     */
    public String toString()
    {
	return "AdvertiseMaterial[fingerPrint="
	    + this.fingerPrint
	    + "]";
    }
}
