/*
 * DiscoveryApplication.java
 *
 * Copyright (c) 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is part of the Janos Bees Software.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */

package apps.discovery;

import bees.core.Node;
import bees.core.TypeID;
import bees.core.ByteArray;
import bees.core.Application;
import bees.core.EventChannel;
import bees.core.NeighborImpl;
import bees.core.ProtocolFlow;
import bees.core.CapabilityEnvelope;

import bees.core.resource.Logger;

import bees.core.security.Permission;
import bees.core.security.PermissionSet;

import bees.core.bdomain.CapabilityTree;

/**
 * Test case for the DiscoveryProtocol.
 */
public final class DiscoveryApplication
    extends Application
{
    /**
     * Construct an empty DiscoveryApplication.
     */
    public DiscoveryApplication()
    {
    }

/**
 * @see bees.core.Application#start(Node, String[])
 */
    public void start(Node node, String args[])
	throws Exception
    {
	DiscoveryProtocolPresentation dpp;
	byte adMaterial[] = null;
	CapabilityTree ct;
	ProtocolFlow pf;

	/*
	 * Create the discovery presentation object/subclass.  This is used
	 * to communicate with the discovery protocol we're about to register.
	 */
	dpp = new DiscoveryProtocolPresentation() {
		public void found(TypeID fp, byte material[], byte signature[])
		{
		    Logger.log("Found: finger-print="
			       + fp
			       + "; material="
			       + new String(material)
			       + "; signature="
			       + ByteArray.toString(signature));
		}
	    };
	/* Create a description of the flow. */
	pf = new ProtocolFlow();
	pf.setPrimaryProtocol("apps.discovery.DiscoveryProtocol");
	/* Create the tree of caps we want to pass to the protocol */
	ct = new CapabilityTree(new CapabilityEnvelope[] {
	    /*
	     * Pass our node, sans the ability to spawn threads and
	     * protocols.
	     */
	    new CapabilityEnvelope("node", node).
		revokePermissions(new PermissionSet(new Permission[] {
		    Node.THREAD_START,
		    Node.PROTOCOL_REGISTER,
		})),
	    /* A place to log messages. */
	    new CapabilityEnvelope("logger",
				   node.getCapability("protocolLogger")),
	    /* Extra classes. */
	    new CapabilityEnvelope(node.getCapability("protocolClasses")),
	    /* We them to be able to talk to us. */
	    new CapabilityEnvelope(EventChannel.currentChannel()),
	});
	try
	{
	    NeighborImpl bn;
	    byte material[];

	    /* Register the protocol */
	    node.register(dpp, pf, ct);
	    Logger.log("Started DiscoveryProtocol.");
	    /* Tell the protocol that we want to be notified of things */
	    dpp.register();

	    /* Ask the protocol to advertise some data. */
	    material = ("Hello World!  This is " + args[0]).getBytes();
	    bn = (NeighborImpl)node.getCapability("broadcastNeighbor");
	    dpp.addNeighbor(bn);
	    dpp.advertise(bn, material, null);
	}
	catch(Throwable th)
	{
	    Logger.log(th);
	}
    }

/**
 * @see java.lang.Object#toString()
 */
    public String toString()
    {
	return "DiscoveryApplication["
	    + super.toString()
	    + "]";
    }
}
