/*
 * DiscoveryProtocolSession.java
 *
 * Copyright (c) 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is part of the Janos Bees Software.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */

package apps.discovery;

import java.util.Hashtable;

import bees.core.Xdr;
import bees.core.Node;
import bees.core.Strict;
import bees.core.TypeID;
import bees.core.Neighbor;
import bees.core.NodeCache;
import bees.core.NeighborImpl;
import bees.core.ExternalCapsule;
import bees.core.CapabilityReference;
import bees.core.ExtendedProtocolSession;
import bees.core.MissingCapabilityException;

import bees.core.resource.Logger;
import bees.core.resource.NodeOSNeighbor;

/**
 * A DiscoveryProtocolSession tracks information used by the protocol and
 * manages communication with the applications.
 */
public final class DiscoveryProtocolSession
    extends ExtendedProtocolSession
{
    /**
     * Major version of the protocol.
     */
    static final int VERSION_MAJOR = 0;
    
    /**
     * Minor version of the protocol.
     */
    static final int VERSION_MINOR = 5;

    /**
     * Method identifier for advertisement requests.
     */
    static final int METHOD_ADVERTISE = METHOD_USER + 1;
    
    /**
     * Method identifier for received advertisements.
     */
    static final int METHOD_FOUND = METHOD_USER + 100;
    
    /**
     * Method identifier for requests to remove a received advertisement.
     *
     * XXX Still needed?
     */
    static final int METHOD_DIED = METHOD_USER + 101;

    /**
     * The parameters to this protocol flow.
     */
    private DiscoveryParameters dp;

    /**
     * The set of discovered advertisements.
     */
    private final Hashtable discoveredMaterial = new Hashtable();

    /**
     * The set of local advertisements.
     */
    private final Hashtable localMaterial = new Hashtable();

    /**
     * Construct an empty DiscoveryProtocolSession.
     */
    public DiscoveryProtocolSession()
    {
    }

    /**
     * @param dp The parameters to this protocol flow.
     */
    public void setDiscoveryParameters(DiscoveryParameters dp)
    {
	this.dp = dp;
    }

    /**
     * @return The parameters to this protocol flow.
     */
    public DiscoveryParameters getDiscoveryParameters()
    {
	return this.dp;
    }

    /**
     * @return The set of discovered advertisements.
     */
    public Hashtable getDiscoveredMaterial()
    {
	return this.discoveredMaterial;
    }

    /**
     * @return The set of local advertisements.
     */
    public Hashtable getLocalMaterial()
    {
	return this.localMaterial;
    }

    /**
     * @see bees.core.ProtocolSession#attachNeighbor(Node, Neighbor)
     */
    public void attachNeighbor(Node node, Neighbor neighbor)
    {
	super.attachNeighbor(node, new BroadcastNeighbor(neighbor.
							 getNeighborImpl()));
    }

    /**
     * Notify any interested applications of the discovery of new material.
     *
     * @param fp The finger print of the material.
     * @param ac The AdvertiseCapsule that carried the newly discovered
     * material.
     */
    public void foundMaterial(TypeID fp, AdvertiseCapsule ac)
    {
	ExternalCapsule ec;
	byte data[];
	Xdr xdr;

	Logger.log("source = " +
		   ((NodeOSNeighbor)ac.getSource().getNeighborImpl()).toVerboseString());
	ec = new ExternalCapsule(new CapabilityReference[] {
	    new CapabilityReference(ac.getSource().getNeighborImpl())
	}).init();
	xdr = ec.getXDR();
	xdr.PUT(METHOD_FOUND);
	xdr.PUT(fp);
	data = ac.getMaterial();
	xdr.PUT(data.length);
	xdr.PUT(data);
	if( (data = ac.getSignature()) != null )
	{
	    xdr.PUT(data.length);
	    xdr.PUT(data);
	}
	else
	{
	    xdr.PUT(0);
	}
	super.notifyInterests(ec);
    }
    
    /**
     * @see bees.core.ExtendedProtocolSession#userReceive(Node, int, ExternalCapsule)
     */
    public void userReceive(Node node, int method, ExternalCapsule ec)
    {
	Xdr xdr;

	xdr = ec.getXDR();
	switch( method )
	{
	case METHOD_ADVERTISE:
	    try
	    {
		BroadcastNeighbor bn;
		AdvertiseCapsule ac;
		NeighborImpl ni;
		TypeID fp;

		ni = (NeighborImpl)ec.getReferences()[0].dereference();
		bn = (BroadcastNeighbor)ni.getNeighbor(this);
		/* Create a new capsule for the new material. */
		ac = new AdvertiseCapsule(bn,
					  xdr.BYTES(xdr.INT()),
					  xdr.BYTES(xdr.INT()),
					  this.dp.getAdvertiseFrequency());
		fp = ac.computeFingerPrint();
		if( !this.localMaterial.containsKey(fp) )
		{
		    /* Haven't added this material yet. */
		    this.localMaterial.put(fp, ac.getMaterial());
		    /*
		     * Use this external capsule's authorization for further
		     * propagation.
		     */
		    ac.fromExternal(ec);
		    bn.getAdvertisements().add(ac);
		    Logger.log(Logger.INFO, "New advertisement: fp=" + fp);
		    /* Do an initial broadcast. */
		    ac.collected(node, this);
		    /*
		     * Add our capsule timer handler with the given
		     * frequency.
		     */
		    super.getCache().put(ac, ac, ac.getFrequency());
		}
		else
		{
		    Logger.log(Logger.WARNING,
			       "Redundant advertisement: fp="
			       + fp);
		}
	    }
	    catch(MissingCapabilityException e)
	    {
		Logger.log(e);
	    }
	    break;
	case METHOD_DIED:
	    {
		NodeCache nc;
		TypeID fp;

		fp = xdr.TYPEID();
		nc = this.getCache();
		synchronized( nc )
		{
		    nc.remove(fp);
		    Logger.log(Logger.INFO, "Dead neighbor: fp=" + fp);
		}
	    }
	    break;
	default:
	    Strict.fail();
	    break;
	}
	ec.getSource().recycleCapsule(ec);
    }

    /**
     * @see java.lang.Object#toString()
     */
    public String toString()
    {
	return "DiscoveryProtocolSession["
	    + super.toString()
	    + "; dp="
	    + this.dp
	    + "; discoveredMaterial="
	    + this.discoveredMaterial
	    + "; localMaterial="
	    + this.localMaterial
	    + "]";
    }
}
