/*
 * HealthBootCapsule.java
 *
 * Copyright (c) 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is part of the Janos Bees Software.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */

package apps.health;

import bees.core.Xdr;
import bees.core.Node;
import bees.core.Panic;
import bees.core.Capsule;

import bees.core.resource.Logger;

import apps.pathfinder.PathfinderSession;

/**
 * A HealthBootCapsule is used to initialize the health protocol.  It will
 * construct a HealthParameters or HealthExtraParameters object, whichever
 * the flows privileges will allow, configure the pathfinder protocol, and
 * install a NodeHealthCapsule to poll the motherboard sensors.
 *
 * @see apps.health.HealthParameters
 * @see apps.health.HealthExtraParameters
 * @see apps.health.HealthSession
 *
 * @see apps.health.NodeHealthCapsule
 * @see apps.health.NodeHealth
 * @see apps.health.SimpleNodeHealth
 *
 * @see apps.pathfinder.PathfinderSession
 */
public final class HealthBootCapsule
    extends Capsule
{
    /**
     * Construct an empty HealthBootCapsule.
     */
    public HealthBootCapsule()
    {
    }

    /**
     * @see bees.core.Capsule#encode(Xdr)
     */
    public void encode(Xdr xdr)
    {
    }

    /**
     * @see bees.core.Capsule#decode(Xdr)
     */
    public void decode(Xdr xdr)
    {
    }

    /**
     * @see bees.core.Capsule#evaluate(Node)
     */
    public void evaluate(Node node)
    {
	PathfinderSession pfs;
	HealthParameters hp;
	HealthSession hs;
	
	/* Say hi. */
	Logger.log(Logger.INFO,
		   "Health Protocol v"
		   + HealthSession.VERSION_MAJOR
		   + "."
		   + HealthSession.VERSION_MINOR);

	/* Build up the parameters for this flow. */
	hs = (HealthSession)this.getSession();
	try
	{
	    hp = new HealthExtraParameters(hs);
	    hp.initialize(node, hs);
	    /* Monitoring is possible. */
	}
	catch(NoClassDefFoundError e)
	{
	    Logger.log(Logger.DBG, e);
	    /* Just do forwarding. */
	    hp = new HealthParameters(hs);
	    hp.initialize(node, hs);
	}
	hs.setHealthParameters(hp);
	try
	{
	    /* Cache the Pathfinder session object. */
	    pfs = (PathfinderSession)
		this.getSession("apps.pathfinder.PathfinderSession");
	}
	catch(ClassNotFoundException e)
	{
	    throw new Panic(e);
	}
	hs.setPathfinderSession(pfs);
	node.setInactivityTimeout(hp.getFlowTimeout());
	hs.getCache().put(node, node, hp.getFlowTimeout());

	try
	{
	    if( ((HealthExtraParameters)hp).getServerPrivateKey() != null )
	    {
		/* This is the server node. */
		pfs.configureRoot(node, this);

		node.setJSIHandler("client", hs);
		node.setJSIHandler("clients", hs);

		node.beginManagement(hs);
	    }
	}
	catch(NoClassDefFoundError e)
	{
	    Logger.log(e);
	}
	catch(ClassCastException e)
	{
	    Logger.log(e);
	}
	if( !pfs.isRoot() )
	{
  	    /* Client and/or router. */
  	    NodeHealthCapsule nhc;

  	    pfs.configure(node, this);
	    if( (hp instanceof HealthExtraParameters) &&
		(((HealthExtraParameters)hp).getSensors() != null) )
	    {
		nhc = new NodeHealthCapsule(
			new SimpleNodeHealth(node.getName()));
		nhc.fromCapsule(this);
		hs.getCache().put(nhc, nhc, hp.getPollFrequency());
	    }
	}
    }

    /**
     * @see java.lang.Object#toString()
     */
    public String toString()
    {
	return "HealthBootCapsule["
	    + super.toString()
	    + "]";
    }
}
