/*
 * HealthParameters.java
 *
 * Copyright (c) 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is part of the Janos Bees Software.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */

package apps.health;

import java.util.Hashtable;
import java.util.StringTokenizer;
import java.util.NoSuchElementException;

import bees.core.Node;
import bees.core.ProtocolSession;
import bees.core.ProtocolParameters;

import bees.core.resource.Logger;

import bees.core.security.IntegerPermission;

/**
 * HealthParameters holds the unprivileged parameters for a health flow.
 */
public class HealthParameters
    extends ProtocolParameters
{
    /**
     * Default sensor polling frequency.
     */
    static final int DEFAULT_POLL_FREQUENCY = 2 * 1000;

    /**
     * Default retention time for NodeHealth objects.
     */
    static final int DEFAULT_HEALTH_RETENTION = 60 * 1000;

    /**
     * Default time frame for accepting NodeHealth reports.
     */
    static final int DEFAULT_REPORT_FRESHNESS = 5 * 1000;

    /**
     * Default timeout for the flow.
     */
    static final int DEFAULT_TIMEOUT = 20 * 1000;

    /**
     * User specified polling frequency or the default.
     */
    protected IntegerPermission pollFrequency;

    /**
     * User specified health retention time or the default.
     */
    protected IntegerPermission healthRetention;

    /**
     * The number of milliseconds within which we're willing to accept a
     * NodeHealth report.
     */
    protected long reportFreshness;

    /**
     * User specified flow timeout.
     */
    protected IntegerPermission flowTimeout;

    /**
     * Crypto algorithm to use for signing/verifying.
     */
    protected String algorithm;

    /**
     * The names of sensor features which should be ignored because they would
     * cause false alarms.
     */
    protected final Hashtable ignoredFeatures = new Hashtable();

    /**
     * Construct a HealthParameters object for the given ProtocolSession.
     *
     * @param ps The HealthSession object that this parameters object should
     * be bound to.
     */
    public HealthParameters(ProtocolSession ps)
    {
	/*
	 * Initialize the super's Properties object from the included
	 * protocol.properties file.
	 */
	super(ps.getByteArray("apps/health/protocol.properties"));
    }

    /**
     * Initialize the parameters.
     *
     * @param node The current Node object.
     * @param ps The HealthSession object that this parameters object should
     * be bound to.
     */
    public void initialize(Node node, ProtocolSession ps)
    {
	/* Get the poll-frequency value out of the protocol.properties file. */
	this.pollFrequency = new IntegerPermission(
		"interval",
		super.getMillisecondValue("parameter.poll-frequency",
					  DEFAULT_POLL_FREQUENCY));
	/*
	 * Get the health-retention value out of the protocol.properties file.
	 */
	this.healthRetention = new IntegerPermission(
		"interval",
		super.getMillisecondValue("parameter.health-retention",
					  DEFAULT_HEALTH_RETENTION));
	/*
	 * Get the timeout value out of the protocol.properties file.
	 */
	this.flowTimeout = new IntegerPermission(
		"interval",
		super.getMillisecondValue("parameter.timeout",
					  DEFAULT_TIMEOUT));
	/*
	 * Get the report-freshness value out of the protocol.properties file.
	 */
	this.reportFreshness =
		super.getMillisecondValue("parameter.report-freshness",
					  DEFAULT_REPORT_FRESHNESS);
	/* Get the crypto algorithm out of the protocol.properties file. */
	this.algorithm = super.getString("parameter.algorithm", "DSA");

	/*
	 * Get the set of ignored features out of the protocol.properties file.
	 */
	{
	    StringTokenizer st;
	    String str;
	    
	    str = super.getString("parameter.ignored-features", "");
	    /* It should just be a white space separated list. */
	    st = new StringTokenizer(str);
	    try
	    {
		while( true )
		{
		    String feature;

		    feature = st.nextToken();
		    this.ignoredFeatures.put(feature, feature);
		}
	    }
	    catch(NoSuchElementException e)
	    {
		/* Done processing. */
	    }
	    Logger.log(Logger.INFO, "Ignored features: " + str);
	}

	Logger.log(Logger.INFO, "Poll frequency: " + this.pollFrequency);
	Logger.log(Logger.INFO, "Health retention: " + this.healthRetention);
	Logger.log(Logger.INFO, "Flow timeout: " + this.flowTimeout);
	Logger.log(Logger.INFO, "Report freshness: " + this.reportFreshness);
	Logger.log(Logger.INFO, "Algorithm: " + this.algorithm);
    }

    /**
     * @return User specified polling frequency or the default.
     */
    public IntegerPermission getPollFrequency()
    {
	return this.pollFrequency;
    }

    /**
     * @return User specified health retention time or the default.
     */
    public IntegerPermission getHealthRetention()
    {
	return this.healthRetention;
    }

    /**
     * @return User specified health retention time or the default.
     */
    public IntegerPermission getFlowTimeout()
    {
	return this.flowTimeout;
    }

    /**
     * @return The number of milliseconds within which we're willing to accept
     * a NodeHealth report.
     */
    public long getReportFreshness()
    {
	return this.reportFreshness;
    }

    /**
     * @return Crypto algorithm to use for signing/verifying.
     */
    public String getAlgorithm()
    {
	return this.algorithm;
    }

    /**
     * @return The names of sensor features which should be ignored because
     * they would cause false alarms.
     */
    public Hashtable getIgnoredFeatures()
    {
	return this.ignoredFeatures;
    }

    /**
     * @see java.lang.Object#toString()
     */
    public String toString()
    {
	return "HealthParameters[pollFrequency="
	    + this.pollFrequency
	    + "; healthRetention="
	    + this.healthRetention
	    + "; reportFreshness="
	    + this.reportFreshness
	    + "; algorithm="
	    + this.algorithm
	    + "; ignoredFeatures="
	    + this.ignoredFeatures
	    + "]";
    }
}
