/*
 * HealthProtocol.java
 *
 * Copyright (c) 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is part of the Janos Bees Software.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */

package apps.health;

import java.io.FileNotFoundException;

import bees.core.Panic;
import bees.core.Protocol;
import bees.core.DLProtocol;

import bees.core.resource.Logger;

import bees.core.security.PublicKey;
import bees.core.security.Permission;
import bees.core.security.PermissionSet;
import bees.core.security.CountPermission;
import bees.core.security.MinimumIntegerPermission;

import apps.pathfinder.PathfinderProtocol;

/**
 * The HealthProtocol attempts to implement a healthd like protocol.  It will
 * spawn onto all available neighbors and, capabilities permitting, poll and
 * report sensor data to a central server.
 */
public class HealthProtocol
    extends Protocol
{
    /**
     * Construct a HealthProtocol description.
     */
    public HealthProtocol()
    {
	try
	{
	    super.startProtocolDefn();
	    {
		PublicKey pk;

		/* Add the protocol.properties config file. */
		this.addFile("apps/health/protocol.properties");

		/*
		 * Add the server's public key, this will be used to verify
		 * any requests to the nodes made by the server.  We're
		 * expecting the authenticator to verify the flow description
		 * and hence, this value, to prevent spoofing.
		 *
		 * Normally, you wouldn't do this since the key would change
		 * over time.  But, its convenient and a nice demonstration of
		 * adding a byte array to the protocol description.
		 */
		pk = (PublicKey)
		    super.getNode().getCapability("serverKeyPair/public");
		this.addByteArray("publicKey", pk.toBytes());

		/*
		 * The NodeHealthCapsule carries NodeHealth object(s) between
		 * nodes.
		 */
		this.addCapsule("apps.health.NodeHealthCapsule",
				new PermissionSet(new Permission[] {
				    new CountPermission("initial", 1),
				    new CountPermission("forwards", 1),
				    new CountPermission("neighborhood", 5),

				    new CountPermission("cached", 1),
				    new MinimumIntegerPermission("interval",
								 500),
				    
				    new CountPermission("threadCount", 1),
				    new CountPermission("bufferCount", 2),
				}));

		/*
		 * The HealthBootCapsule will be received when the flow is
		 * finished being bootstrapped.
		 */
		this.setBootStrap("apps.health.HealthBootCapsule",
				  new PermissionSet(new Permission[] {
				      new CountPermission(
					"transformTo."
					+ "apps.health.NodeHealthCapsule",
					1),
				      new CountPermission(
					"transformTo.apps.pathfinder."
					+ "ConfigurationCapsule",
					1),
				      
				      new CountPermission("threadCount", 0),
				      new CountPermission("bufferCount", 0),
				  }));

		/*
		 * Set the class that should be used for our
		 * ProtocolSession.
		 */
		this.setSession("apps.health.HealthSession");
		
		this.addCode("apps.health.NodeHealth");
		this.addCode("apps.health.SimpleNodeHealth");
		this.addCode("apps.health.HealthParameters");
		this.addCode("apps.health.HealthExtraParameters");

		/*
		 * Add the DLProtocol so the protocol can be downloaded to
		 * other nodes.
		 */
		this.addCompanion(new DLProtocol());

		/*
		 * Note:  We don't include IDTranslationProtocol since we use
		 * Pathfinder to decide which nodes to spawn onto.
		 */

		/**
		 * Add the PathfinderProtocol to handle routing between the
		 * clients and the server.
		 */
		this.addCompanion(new PathfinderProtocol());
	    }
	    super.endProtocolDefn();
	}
	catch(FileNotFoundException e)
	{
	    Logger.log(e);
	    throw new Panic("Unable to access file", e);
	}
	catch(ClassNotFoundException e)
	{
	    Logger.log(e);
	    throw new Panic("Unable to access classes", e);
	}
	catch(Throwable th)
	{
	    Logger.log(th);
	    throw new Panic("Big problem", th);
	}
    }

    /**
     * @see java.lang.Object#toString()
     */
    public String toString()
    {
	return "HealthProtocol["
	    + super.toString()
	    + "]";
    }
}
