/*
 * HealthSession.java
 *
 * Copyright (c) 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is part of the Janos Bees Software.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */

package apps.health;

import java.util.NoSuchElementException;

import edu.utah.janosvm.kit.comm.jsi.JSIPair;
import edu.utah.janosvm.kit.comm.jsi.JSIHandler;
import edu.utah.janosvm.kit.comm.jsi.JSIMessage;
import edu.utah.janosvm.kit.comm.jsi.JSIPropertyInfo;
import edu.utah.janosvm.kit.comm.jsi.JSIHandlerContext;
import edu.utah.janosvm.kit.comm.jsi.MissingPairException;

import bees.core.Node;
import bees.core.NeighborImpl;
import bees.core.ExternalCapsule;
import bees.core.ExtendedProtocolSession;

import apps.pathfinder.PathfinderSession;

/**
 * A HealthSession manages the HealthProtocol specific properties in a flow.
 */
public final class HealthSession
    extends ExtendedProtocolSession
    implements JSIHandler
{
    /**
     * Name prefix for NodeHealth objects added to the NodeCache.
     */
    public static final String CLIENT_PREFIX = "client.";
    
    /**
     * Major version of the protocol.
     */
    static final int VERSION_MAJOR = 0;
    
    /**
     * Minor version of the protocol.
     */
    static final int VERSION_MINOR = 5;

    /**
     * The HealthParameters for this flow.
     */
    private HealthParameters parameters;

    /**
     * A cached reference to the Pathfinder protocol session object.
     */
    private PathfinderSession pfs;

    /**
     * Construct an empty HealthSession object.
     */
    public HealthSession()
    {
    }

    /**
     * @return The HealthParameters for this flow.
     */
    HealthParameters getHealthParameters()
    {
	return this.parameters;
    }

    /**
     * @param hp The HealthParameters for this flow.
     */
    void setHealthParameters(HealthParameters hp)
    {
	this.parameters = hp;
    }

    /**
     * @param pfs The PathfinderSession for this flow.
     */
    void setPathfinderSession(PathfinderSession pfs)
    {
	this.pfs = pfs;
    }

    /**
     * @return A cached reference to the Pathfinder protocol session object.
     */
    PathfinderSession getPathfinderSession()
    {
	return this.pfs;
    }

    /**
     * @param node The current Node object.
     * @param ec The ExternalCapsule that requested the addition of the
     * neighbor, can be null.
     * @param ni The new NeighborImpl capability.
     */
    public void addNeighbor(Node node, ExternalCapsule ec, NeighborImpl ni)
    {
	super.addNeighbor(node, ec, ni);
    }

    /**
     * @param node The current Node object.
     * @param ec The ExternalCapsule that requested the removal of the
     * neighbor, can be null.
     * @param ni The old NeighborImpl capability.
     */
    public void removeNeighbor(Node node, ExternalCapsule ec, NeighborImpl ni)
    {
	super.removeNeighbor(node, ec, ni);
    }

    /**
     * @see edu.utah.janosvm.kit.comm.jsi.JSIHandler#resolveSpecifier(JSIMessage, JSIPair)
     */
    public JSIHandler resolveSpecifier(JSIMessage jm, JSIPair jp)
	throws Throwable
    {
	JSIHandler retval;
	String name;
	Object obj;

	name = jp.getStringValue();
	if( name.equals("") )
	{
	    retval = this;
	}
	else if( (obj = this.getCache().get(CLIENT_PREFIX + name)) != null )
	{
	    if( obj instanceof JSIHandler )
	    {
		retval = (JSIHandler)obj;
		if( (jp = jm.popSpecifier()) != null )
		{
		    retval = retval.resolveSpecifier(jm, jp);
		}
	    }
	    else
	    {
		throw new NoSuchElementException(name);
	    }
	}
	else
	{
	    throw new NoSuchElementException(name);
	}
	return retval;
    }

    /**
     * @see edu.utah.janosvm.kit.comm.jsi.JSIHandler#handleMessage(JSIHandlerContext, JSIMessage)
     */
    public void handleMessage(JSIHandlerContext jhc, JSIMessage jm)
    {
	switch( jm.getWhat() )
	{
	case JSIMessage.LIST_PROPERTIES:
	    {
		Object keys[];
		int lpc;

		jm.clear();
		jm.setWhat(JSIMessage.REPLY);
		keys = this.getCache().keys();
		for( lpc = 0; lpc < keys.length; lpc++ )
		{
		    if( keys[lpc] instanceof String )
		    {
			String str;

			str = (String)keys[lpc];
			if( str.startsWith(CLIENT_PREFIX) )
			{
			    jm.add("result",
				   str.substring(CLIENT_PREFIX.length()));
			}
		    }
		}
	    }
	    break;
	default:
	    throw new UnsupportedOperationException();
	}
    }

    /**
     * @see edu.utah.janosvm.kit.comm.jsi.JSIHandler#getSuites(JSIMessage)
     */
    public void getSuites(JSIMessage jm)
    {
    }
    
    /**
     * @see java.lang.Object#toString()
     */
    public String toString()
    {
	return "HealthSession["
	    + super.toString()
	    + "; parameters="
	    + this.parameters
	    + "; pfs="
	    + this.pfs
	    + "]";
    }
}
