/*
 * NodeHealth.java
 *
 * Copyright (c) 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is part of the Janos Bees Software.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */

package apps.health;

import edu.utah.janosvm.kit.comm.jsi.JSIPair;
import edu.utah.janosvm.kit.comm.jsi.JSIHandler;
import edu.utah.janosvm.kit.comm.jsi.JSIMessage;
import edu.utah.janosvm.kit.comm.jsi.JSIPropertyInfo;
import edu.utah.janosvm.kit.comm.jsi.JSIHandlerContext;
import edu.utah.janosvm.kit.comm.jsi.MissingPairException;

import bees.core.Xdr;
import bees.core.Node;
import bees.core.Assert;
import bees.core.NodeCache;
import bees.core.XdrByteArray;
import bees.core.ProtocolSession;

import bees.core.security.IntegerPermission;

/**
 * A NodeHealth object is used to store the health state of a node in the
 * network.  The implementation of this class only maintains coarse grained
 * state, however, it should be simple enough to extend if more detail needs
 * to be transmitted.
 */
public class NodeHealth
    implements NodeCache.EventHandler, JSIHandler
{
    /**
     * The minimum state value.
     */
    public static final int STATE_MIN = 0;

    /**
     * The node is fine, all sensors are within limits.
     */
    public static final int STATE_FINE		= 1;

    /**
     * The node is not alright, one or more sensors are outside acceptable
     * limits.
     */
    public static final int STATE_ON_FIRE	= 2;

    /**
     * The node is assumed to be ok, there hasn't been a recent report, but
     * its not long enough to start worrying.
     */
    public static final int STATE_ASSUMED_OK	= 3;

    /**
     * The node state cannot be polled.
     */
    public static final int STATE_UNKNOWN	= 4;

    /**
     * The node is unreachable through the network.
     */
    public static final int STATE_UNREACHABLE	= 5;

    /**
     * The node is turned off, don't expect any reports.
     */
    public static final int STATE_OFF		= 6;
    
    /**
     * The maximum state value.
     */
    public static final int STATE_MAX		= 7;

    /**
     * The human readable names of the state values.
     */
    public static final String STATE_NAMES[] = {
	"",
	"fine",
	"on-fire",
	"assumed-ok",
	"unknown",
	"unreachable",
	"off",
    };

    /**
     * The name of the node this health information pertains to.
     */
    private String name;

    /**
     * The health state of the node.
     */
    protected int state = STATE_OFF;

    /**
     * Construct a NodeHealth object with the given values.
     *
     * @param name The name of the node this health information pertains to.
     */
    public NodeHealth(String name)
    {
	this.name = name;
    }

    /**
     * @return The name of the node this health information pertains to.
     */
    public String getName()
    {
	return this.name;
    }

    /**
     * @param state The health state of the node.
     */
    public void setState(int state)
    {
	Assert.require(state > STATE_MIN);
	Assert.require(state < STATE_MAX);
	
	this.state = state;
    }

    /**
     * @return The health state of the node.
     */
    public int getState()
    {
	return this.state;
    }

    public int marshalledLength()
    {
	return Xdr.STRING(this.name) + Xdr.INT;
    }

    public void marshal(XdrByteArray xba)
    {
	xba.PUT(this.name);
	xba.PUT(this.state);
    }

    public void unmarshal(XdrByteArray xba)
    {
	this.state = xba.INT();
    }

    /**
     * Poll the node's health and update this object's state.  This is a dummy
     * implementation it will always set the state to STATE_UNKNOWN and return
     * the poll frequency permission.
     *
     * @param hs The HealthSession for the current flow.
     * @return An IntegerPermission containing the millisecond delay to the
     * next polling interval.
     */
    public IntegerPermission poll(HealthSession hs)
    {
	this.state = STATE_UNKNOWN;
	return ((HealthExtraParameters)hs.getHealthParameters()).
	    getPollFrequency();
    }

    /**
     * @see bees.core.NodeCache.EventHandler#removed(Node, ProtocolSession)
     */
    public void removed(Node node, ProtocolSession ps)
    {
    }

    /**
     * @see bees.core.NodeCache.EventHandler#collected(Node, ProtocolSession)
     */
    public IntegerPermission collected(Node node, ProtocolSession ps)
    {
	return null;
    }

    /**
     * @see edu.utah.janosvm.kit.comm.jsi.JSIHandler#resolveSpecifier(JSIMessage, JSIPair)
     */
    public JSIHandler resolveSpecifier(JSIMessage jm, JSIPair jp)
    {
	return this;
    }

    /**
     * @see edu.utah.janosvm.kit.comm.jsi.JSIHandler#handleMessage(JSIHandlerContext, JSIMessage)
     */
    public void handleMessage(JSIHandlerContext jhc, JSIMessage jm)
    {
	JSIPair jp;
	
	switch( jm.getWhat() )
	{
	case JSIMessage.GET_PROPERTY:
	    jm.restoreSpecifier();
	    if( (jp = jm.popSpecifier()) != null )
	    {
		if( jp.getTag().equals("state") )
		{
		    jm.reply(STATE_NAMES[this.state]);
		}
		else
		{
		    jm.error("Unknown property: " + jp.getTag());
		}
	    }
	    else
	    {
		jm.reply(this.toString());
	    }
	    break;
	default:
	    throw new UnsupportedOperationException();
	}
    }

    /**
     * @see edu.utah.janosvm.kit.comm.jsi.JSIHandler#getSuites(JSIMessage)
     */
    public void getSuites(JSIMessage jm)
    {
    }
    
    /**
     * @see java.lang.Object#toString()
     */
    public String toString()
    {
	return "NodeHealth[name="
	    + this.name
	    + "; state="
	    + STATE_NAMES[this.state]
	    + "]";
    }
}
