/*
 * NodeHealthCapsule.java
 *
 * Copyright (c) 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is part of the Janos Bees Software.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */

package apps.health;

import java.util.NoSuchElementException;

import java.security.SignatureException;
import java.security.InvalidKeyException;
import java.security.NoSuchAlgorithmException;

import bees.core.Xdr;
import bees.core.Node;
import bees.core.NodeCache;
import bees.core.ProtocolSession;
import bees.core.SignedXdrByteArray;

import bees.core.resource.Logger;

import bees.core.security.PublicKey;
import bees.core.security.IntegerPermission;

import apps.pathfinder.SinkCapsule;
import apps.pathfinder.PathfinderSession;

/**
 * The NodeHealthCapsule is used to poll and haul a NodeHealth object back to
 * the server.
 * <p>
 * The NodeHealth and NodeHealthCapsule classes were expressly designed so
 * that, with some modification, more than one NodeHealth object could be
 * carried in a capsule.  This would allow packet fusion at router nodes and
 * hopefully better overall performance.
 *
 * @see bees.core.SignedXdrByteArray
 *
 * @see apps.health.HealthSession
 */
public class NodeHealthCapsule
    extends SinkCapsule
    implements NodeCache.EventHandler
{
    /**
     * The NodeHealth object to be polled when this object is being used as the
     * client's source capsule.
     */
    protected final NodeHealth nh;
    
    /**
     * A signed byte array which holds the node name, sensor name, feature
     * name, and the out of bounds value.
     */
    private SignedXdrByteArray signedData;

    /**
     * Construct a NodeHealthCapsule with the given value.
     *
     * @param nh The NodeHealth object to poll when this capsule is being
     * collected by the NodeCache.
     */
    public NodeHealthCapsule(NodeHealth nh)
    {
	this.nh = nh;
    }

    /**
     * Construct an empty NodeHealthCapsule.
     */
    public NodeHealthCapsule()
    {
	this(null);
    }

    /**
     * @see bees.core.Capsule#reset()
     */
    public void reset()
    {
	this.signedData = null;
    }

    /**
     * @see bees.core.NodeCache.EventHandler#removed(Node, ProtocolSession)
     */
    public void removed(Node node, ProtocolSession ps)
    {
    }

    /**
     * @see bees.core.NodeCache.EventHandler#collected(Node, ProtocolSession)
     */
    public IntegerPermission collected(Node node, ProtocolSession ps)
    {
	IntegerPermission retval;
	HealthSession hs;

	hs = (HealthSession)ps;
	/* Poll the node's sensors. */
	retval = this.nh.poll(hs);
	/* Use the Pathfinder to send the capsule towards the root node. */
	super.evaluate(node, hs.getPathfinderSession());
	return retval;
    }
    
    /**
     * @see bees.core.Capsule#decode(Xdr)
     */
    public void decode(Xdr xdr)
    {
	HealthParameters hp;
	HealthSession hs;

	hs = (HealthSession)this.getSession();
	hp = hs.getHealthParameters();
	/* Decode into a signed data array for future decoding. */
	this.signedData = new SignedXdrByteArray(xdr.BYTES(xdr.INT()),
						 xdr.BYTES(xdr.INT()),
						 hp.getAlgorithm());
    }

    /**
     * @see bees.core.Capsule#encode(Xdr)
     */
    public void encode(Xdr xdr)
	throws SignatureException,
	       InvalidKeyException,
	       NoSuchAlgorithmException
    {
	byte ba[];
	
	if( this.nh != null )
	{
	    HealthExtraParameters hp;
	    HealthSession hs;

	    /*
	     * We're the source capsule on the client, build a
	     * SignedXdrByteArray.
	     */
	    hs = (HealthSession)this.getSession();
	    hp = (HealthExtraParameters)hs.getHealthParameters();
	    this.signedData =
		new SignedXdrByteArray(this.nh.marshalledLength() +
				       SignedXdrByteArray.TIMESTAMP,
				       hp.getAlgorithm());
	    this.signedData.PUT(hp.getClock());
	    this.nh.marshal(this.signedData);
	    this.signedData.sign(hp.getNodePrivateKey(), hs);
	}
	ba = this.signedData.buf;
	xdr.PUT(ba.length);
	xdr.PUT(ba);
	ba = this.signedData.getSignature();
	xdr.PUT(ba.length);
	xdr.PUT(ba);
    }

    /**
     * @see bees.core.Capsule#evaluate(Node)
     */
    public void evaluate(Node node)
    {
	PathfinderSession pfs;
	HealthSession hs;

	hs = (HealthSession)this.getSession();
	pfs = hs.getPathfinderSession();
	/* Ask Pathfinder to forward this capsule if we're not the root. */
	if( super.evaluate(node, pfs) )
	{
	    HealthExtraParameters hp;

	    /* We're the root, do something with the report(s)... */
	    hp = (HealthExtraParameters)hs.getHealthParameters();
	    try
	    {
		if( this.signedData.TIMESTAMP(hp.getClock(),
					      hp.getReportFreshness()) )
		{
		    String nodeName;
		    PublicKey pk;
		    
		    /* Get the name of the node this report is for. */
		    nodeName = this.signedData.STRING();
		    if( (pk = hp.getNodePublicKey(nodeName)) != null )
		    {
			if( this.signedData.verify(pk, hs) )
			{
			    NodeCache nc;
			    
			    nc = hs.getCache();
			    synchronized( nc )
			    {
				NodeHealth nh;

				nodeName = HealthSession.CLIENT_PREFIX +
				    nodeName;
				/*
				 * Unmarshal the report into our local object,
				 * or make a new one.
				 */
				if( (nh = (NodeHealth)nc.get(nodeName))
				    == null )
				{
				    nh = new NodeHealth(nodeName);
				}
				nh.unmarshal(this.signedData);
				Logger.log(Logger.INFO,
					   "Health report: "
					   + nh);
				nc.put(nodeName, nh, hp.getHealthRetention());
			    }
			}
			else
			{
			    Logger.log(Logger.NOTICE,
				       "NodeHealthCapsule, "
				       + this
				       + ", doesn't verify!");
			}
		    }
		    else
		    {
			Logger.log(Logger.NOTICE,
				   "NodeHealthCapsule, "
				   + this
				   + ", couldn't find "
				   + hp.getAlgorithm()
				   + " key for node: "
				   + nodeName);
		    }
		}
		else
		{
		    Logger.log(Logger.NOTICE,
			       "NodeHealthCapsule, "
			       + this
			       + ", is stale.");
		}
	    }
	    catch(SignatureException e)
	    {
		Logger.log(e);
	    }
	    catch(InvalidKeyException e)
	    {
		Logger.log(e);
	    }
	    catch(NoSuchAlgorithmException e)
	    {
		Logger.log(e);
	    }
	    catch(NoSuchElementException e)
	    {
		Logger.log("Unknown node: " + e.getMessage());
		Logger.log(e);
	    }
	}
	this.addToPool();
    }
    
    /**
     * @see java.lang.Object#toString()
     */
    public String toString()
    {
	return "NodeHealthCapsule["
	    + super.toString()
	    + "; nh="
	    + this.nh
	    + "; signedData="
	    + this.signedData
	    + "]";
    }
}
