/*
 * LivenessProtocol.java
 *
 * Copyright (c) 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is part of the Janos Bees Software.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */

package apps.liveness;

import java.io.FileNotFoundException;

import java.util.NoSuchElementException;

import bees.core.Node;
import bees.core.Panic;
import bees.core.Protocol;

import bees.core.resource.Logger;

import bees.core.security.Permission;
import bees.core.security.PermissionSet;
import bees.core.security.VersionString;
import bees.core.security.CountPermission;
import bees.core.security.MinimumIntegerPermission;

/**
 * The LivenessProtocol can be used to test for responsiveness of a node's
 * neighbors.
 */
public final class LivenessProtocol
    extends Protocol
{
    /**
     * Construct the LivenessProtocol description.
     */
    public LivenessProtocol()
    {
	try
	{

	    this.startProtocolDefn();
	    {
		VersionString vs = null;
		Node node;

		node = super.getNode();
		try
		{
		    /* Get the version string to use with the capsules. */
		    vs = (VersionString)node.getCapability("liveVersion");
		}
		catch(NoSuchElementException e)
		{
		    Logger.log(Logger.WARNING,
			       "No \"liveVersion\" capability.  Using byte "
			       + "code to compute capsule identifiers.");
		}
		/* Add our properties file. */
		this.addFile("apps/liveness/protocol.properties");
		/* The main capsule for testing neighbors. */
		this.addCapsule("apps.liveness.PokeCapsule",
				vs,
				new PermissionSet(new Permission[] {
				    /*
				     * We're sending to a group of
				     * neighbors.
				     */
				    new CountPermission("neighborhood", 1),
				    /* The poke replies come right back. */
				    new CountPermission("returns", 1),
				    
				    /* We put the primary one in the cache */
				    new CountPermission("cached", 1),

				    /* Only need one thread to handle these. */
				    new CountPermission("threadCount", 1),
				    /* Shouldn't need more than one buffer. */
				    new CountPermission("bufferCount", 1),

				    /*
				     * Minimum interval between timer
				     * replenishes.
				     */
				    new MinimumIntegerPermission("interval",
								 500),
				}));
		/* Class used to track parameters. */
		this.addCode("apps.liveness.LivenessParameters");
		/* Class used to track neighbors. */
		this.addCode("apps.liveness.LiveNeighbor");
		this.setSession("apps.liveness.LivenessSession");
		this.setBootStrap("apps.liveness.LivenessBootCapsule",
				  new PermissionSet(new Permission[] {
				      /*
				       * The first PokeCapsule is authorized
				       * from this capsule.
				       */
				      new CountPermission(
					"transformTo"
					+ ".apps.liveness.PokeCapsule", 1),
				      
				      /* This doesn't go over the wire. */
				      new CountPermission("threadCount", 0),
				      new CountPermission("bufferCount", 0),
				  }));
	    }
	    this.endProtocolDefn();
	}
	catch(FileNotFoundException e)
	{
	    Logger.log(e);
	    throw new Panic("Unable to access file", e);
	}
	catch(ClassNotFoundException e)
	{
	    Logger.log(e);
	    throw new Panic("Unable to access classes", e);
	}
    }

    /**
     * @see java.lang.Object#toString()
     */
    public String toString()
    {
	return "LivenessProtocol["
	    + super.toString()
	    + "]";
    }
}
