/*
 * PokeCapsule.java
 *
 * Copyright (c) 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is part of the Janos Bees Software.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */

package apps.liveness;

import bees.core.Xdr;
import bees.core.Node;
import bees.core.Capsule;
import bees.core.NodeCache;
import bees.core.ProtocolSession;
import bees.core.EncodingException;

import bees.core.resource.Logger;

import bees.core.security.IntegerPermission;
import bees.core.security.PermissionException;

/**
 * A PokeCapsule is used to probe a neighbor for liveness/responsiveness.
 */
public final class PokeCapsule
    extends Capsule
    implements NodeCache.EventHandler
{
    /**
     * The recorded start time of this capsule on the sending host, time is in
     * milliseconds.
     */
    private long startTime;

    /**
     * Indicates whether or not the capsule is returning to the originator.
     */
    private boolean returning = false;

    /**
     * Construct an empty PokeCapsule.
     */
    public PokeCapsule()
    {
    }

    /**
     * @return The recorded start time of this capsule on the sending host,
     * time is in milliseconds.
     */
    long getStartTime()
    {
	return this.startTime;
    }

    /**
     * @see bees.core.NodeCache.EventHandler#removed(Node, ProtocolSession)
     */
    public void removed(Node node, ProtocolSession ps)
    {
    }

    /**
     * @see bees.core.NodeCache.EventHandler#collected(Node, ProtocolSession)
     */
    public IntegerPermission collected(Node node, ProtocolSession ps)
    {
	LivenessParameters lp;
	LivenessSession ls;

	ls = (LivenessSession)ps;
	lp = ls.getLivenessParameters();
	this.startTime = lp.getClock().current();
	try
	{
	    Logger.log(Logger.NOTICE, "Sending poke.");
	    ls.getNeighborhood().send(this);
	}
	catch(EncodingException e)
	{
	    Logger.log(e);
	}
	return lp.getPokeFrequency();
    }

    /**
     * @see bees.core.Capsule#decode(Xdr)
     */
    public void decode(Xdr xdr)
    {
	this.startTime = xdr.LONG();
	this.returning = xdr.BOOLEAN();
    }

    /**
     * @see bees.core.Capsule#encode(Xdr)
     */
    public void encode(Xdr xdr)
    {
	xdr.PUT(this.startTime);
	xdr.PUT(this.returning);
    }

    /**
     * @see bees.core.Capsule#evaluate(Node)
     */
    public void evaluate(Node node)
    {
	if( this.returning )
	{
	    LivenessSession ls;
	    long diff;

	    /* A returning capsule. */
	    ls = (LivenessSession)this.getSession();
	    diff = ls.getLivenessParameters().getClock().current()
		- this.startTime;

	    /* Touch the source. */
	    ((LiveNeighbor)this.getSource()).touch();
	    
	    Logger.log(Logger.NOTICE,
		       "Neighbor: "
		       + this.getSource().getNeighborImpl().getFingerPrint()
		       + "; Elapsed Time: "
		       + diff
		       + "ms");
	}
	else
	{
	    try
	    {
		this.returning = true;
		this.getSource().send(this);
	    }
	    catch(PermissionException e)
	    {
		Logger.log(e);
	    }
	    catch(Exception e)
	    {
		Logger.log(Logger.ERROR,
			   "Unable to send to neighbor: " + this.getSource());
		Logger.log(e);
	    }
	}
	this.addToPool();
    }

    /**
     * @see java.lang.Object#toString()
     */
    public String toString()
    {
	return "PokeCapsule["
	    + super.toString()
	    + "; startTime="
	    + this.startTime
	    + "; returning="
	    + this.returning
	    + "]";
    }
}
