/*
 * PathfinderParameters.java
 *
 * Copyright (c) 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is part of the Janos Bees Software.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */

package apps.pathfinder;

import bees.core.Node;
import bees.core.Assert;
import bees.core.ProtocolParameters;

import bees.core.resource.Logger;

import bees.core.security.IntegerPermission;

/**
 * PathfinderParameters holds any arguments to the pathfinder flow.
 */
public final class PathfinderParameters
    extends ProtocolParameters
{
    /**
     * Default frequency used when sending out configuration capsules.
     */
    private static final long DEFAULT_UPDATE_FREQUENCY = 5 * 1000;

    /**
     * Default timeout for the root neighbor.
     */
    private static final long DEFAULT_ROOT_TIMEOUT = 30 * 1000;

    /**
     * The frequency used when sending out configuration capsules.
     */
    private IntegerPermission updateFrequency;

    /**
     * The timeout for the root neighbor.
     */
    private IntegerPermission rootTimeout;

    /**
     * Construct a PathfinderParameters object with the given values.
     *
     * @param ps The PathfinderSession that these parameters are for.
     */
    public PathfinderParameters(PathfinderSession ps)
    {
	super(ps.getByteArray("apps/pathfinder/protocol.properties"));
    }

    /**
     * Initialize the values in the object.
     *
     * @param node The current Node object.
     */
    public void initialize(Node node)
    {
	Assert.require(node);

	/* Get the update frequency out of the protocol.properties file. */
	this.updateFrequency = new IntegerPermission(
		"interval",
		super.getMillisecondValue("parameter.update-frequency",
					  DEFAULT_UPDATE_FREQUENCY));
	
	/* Get the root timeout out of the protocol.properties file. */
	this.rootTimeout = new IntegerPermission(
		"interval",
		super.getMillisecondValue("parameter.root-timeout",
					  DEFAULT_ROOT_TIMEOUT));
	
	Logger.log(Logger.INFO,
		   "Pathfinder update frequency: "
		   + this.updateFrequency);
	Logger.log(Logger.INFO,
		   "Pathfinder root timeout: "
		   + this.rootTimeout);
    }

    /**
     * @return The frequency used when sending out configuration capsules.
     */
    public IntegerPermission getUpdateFrequency()
    {
	return this.updateFrequency;
    }
    
    /**
     * @return The timeout for the root neighbor.
     */
    public IntegerPermission getRootTimeout()
    {
	return this.rootTimeout;
    }
    
    /**
     * @see java.lang.Object#toString()
     */
    public String toString()
    {
	return "PathfinderParameters[updateFrequency="
	    + this.updateFrequency
	    + "; rootTimeout="
	    + this.rootTimeout
	    + "]";
    }
}
