/*
 * SinkCapsule.java
 *
 * Copyright (c) 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is part of the Janos Bees Software.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */

package apps.pathfinder;

import java.io.IOException;

import bees.core.Node;
import bees.core.Capsule;
import bees.core.Neighbor;
import bees.core.EncodingException;

import bees.core.resource.Logger;

/**
 * A SinkCapsule is a base class for Capsule types that originate at the leafs
 * of the pathfinder tree and are directed toward the root.
 */
public abstract class SinkCapsule
    extends Capsule
{
    /**
     * Construct an empty SinkCapsule object.
     */
    public SinkCapsule()
    {
    }

    /**
     * Evaluate a Capsule and possibly forward it toward the root.
     *
     * @param node The current Node object.
     * @param pfs The PathfinderSession this capsule belongs to.
     * @return True if this is the root and the Capsule should be processed or
     * false if this is an intermediate node.
     */
    public boolean evaluate(Node node, PathfinderSession pfs)
    {
	boolean retval = false;
	Neighbor neighbor;
	
	if( pfs.isRoot() )
	{
	    /* We're the root, let the subclass do its thing. */
	    retval = true;
	}
	else if( (neighbor = pfs.getRootNeighbor()) != null )
	{
	    /* We're not the root and there is a path, send it that way. */
	    try
	    {
		neighbor.send(this);
	    }
	    catch(EncodingException e)
	    {
		Logger.log(e);
	    }
	    catch(IOException e)
	    {
		Logger.log(e);
	    }
	}
	return retval;
    }
    
    /**
     * @see java.lang.Object#toString()
     */
    public String toString()
    {
	return "SinkCapsule["
	    + super.toString()
	    + "]";
    }
}
