/*
 * Assert.java
 *
 * Copyright (c) 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is part of the Janos Bees Software.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */

package bees.core;

/**
 * Provides a number of static functions for verifying state in the system.
 * The basic <code>check()</code> function tests its argument and throws an
 * exception if it's false/NULL.  The <code>require()</code> and
 * <code>ensure()</code> methods are simply aliases for <code>check()</code>
 * and used to imply pre/post conditions for functions.
 *
 * @author <a href="mailto:janos-discuss@flux.cs.utah.edu">Utah Janos Team</a>
 */
public final class Assert
{
    /**
     * Indicates whether checks are checked or not.
     */
    private static final boolean ENABLED = true;
    
    /**
     * A run-time Error is thrown if an check fails.
     * All AssertFailures must have a descriptive message.
     */
    public static class AssertFailure
	extends Error
    {
        /**
         * @see java.lang.Throwable#Throwable(String)
         */
	public AssertFailure(String msg)
	{
	    super(msg);
	}
    }
    
    /**
     * Throw an AssertFailure error with the given message.
     *
     * @param msg A description of the failed test.
     * @throws AssertFailure always.
     */
    private static void failure(String msg)
	throws AssertFailure
    {
	Error er;
	
	System.out.println("FAILURE:" +msg);
	er = new AssertFailure(msg);
	er.printStackTrace();
	throw er;
    }
    
    /**
     * Check a boolean is true.
     *
     * @param test The boolean to test.
     * @throws AssertFailure if test if false.
     */
    public static void check(boolean test)
	throws AssertFailure
    {
	if (Assert.ENABLED)
	{
	    if (!test)
	    {
		failure("check test");
	    }
	}
    }
    
    /**
     * Check a boolean is true.
     *
     * @param test The boolean to test.
     * @param msg A description of the tested value.
     * @throws AssertFailure if test if false.
     */
    public static void check(boolean test, String msg)
	throws AssertFailure
    {
	if (Assert.ENABLED)
	{
	    if (!test)
	    {
		failure("check test: " +msg);
	    }
	}
    }
    
    /**
     * Check that an object reference is not null.
     *
     * @param obj The object reference to check.
     * @throws AssertFailure if obj is null.
     */
    public static void check(Object obj)
	throws AssertFailure
    {
	if (Assert.ENABLED)
	{
	    if (obj == null)
	    {
		failure("check <obj> != null");
	    }
	}
    }
    
    /**
     * Require a boolean to be true.
     *
     * @param test The boolean to test.
     * @throws AssertFailure if test if false.
     */
    public static void require(boolean test)
	throws AssertFailure
    {
	if (Assert.ENABLED)
	{
	    if (!test)
	    {
		failure("require test");
	    }
	}
    }
    
    /**
     * Require an object reference to be non-null.
     *
     * @param obj The object reference to check.
     * @throws AssertFailure if obj is null
     */
    public static void require(Object obj)
	throws AssertFailure
    {
	if (Assert.ENABLED)
	{
	    if (obj == null)
	    {
		failure("require <obj> != null");
	    }
	}
    }
    
    /**
     * Ensure a boolean is true.
     *
     * @param test The boolean to test.
     * @throws AssertFailure if test if false.
     */
    public static void ensure(boolean test)
	throws AssertFailure
    {
	if (Assert.ENABLED)
	{
	    if (!test)
	    {
		failure("ensure test");
	    }
	}
    }

    /**
     * Ensure an object reference is not null.
     *
     * @param obj The object reference to check.
     * @throws AssertFailure if obj is null.
     */
    public static void ensure(Object obj)
	throws AssertFailure
    {
	if (Assert.ENABLED)
	{
	    if (obj == null)
	    {
		failure("ensure <obj> != null");
	    }
	}
    }
    
    /**
     * Private constructor prevents instances of this class from
     * being created.
     */
    private Assert()
    {
    }
}
