/*
 * AuthorizationCoroutine.java
 *
 * Copyright (c) 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is part of the Janos Bees Software.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */

package bees.core;

import bees.core.resource.Logger;

import bees.core.bdomain.DeadBDomainException;

/**
 * A Coroutine for interacting with the Auth App.  This coroutine can be used
 * by a downloaded protocol to dynamically request capabilities from the Auth
 * App.  It can also be used to inform the Auth App that the protocol wishes to
 * be "managed."  A protocol becomes "managed" when its ProtocolSession or
 * ExtendedProtocolSession is connected to a ProtocolPresentation in the Auth
 * App.  Thus, in the case of the extended version, the protocol will be
 * notified of newly discovered or lost neighbors.
 */
public class AuthorizationCoroutine
    extends ResponderCoroutine
{
    /**
     * Indicates the requesting Coroutine is waiting for a response.
     */
    public static final int STATE_REQUESTING = STATE_USER + 0;

    /**
     * Indicates the coroutine is connected to a peer.
     */
    public static final int STATE_CONNECTED = STATE_USER + 1;

    /**
     * The requested Capabilities, if authorization was successful.
     */
    private CapabilityEnvelope ce[];

    /**
     * An EventChannel connected to the peer domain.
     */
    private EventChannel ec;

    /**
     * The peer coroutine identifier.
     */
    private long peerID;

    /**
     * Construct an AuthorizationCoroutine with the given arguments.
     *
     * @param sr A StateResponder that will handle state transitions in this
     * Coroutine.
     * @param ac A connected AuthorizationCoroutine or null if a new connection
     * is to be made.
     */
    public AuthorizationCoroutine(StateResponder sr, AuthorizationCoroutine ac)
    {
	super(sr);

	if( ac != null )
	{
	    this.ec = ac.ec;
	    this.peerID = ac.peerID;
	}
    }
    
    /**
     * Construct an AuthorizationCoroutine with the given argument.
     *
     * @param ac A connected AuthorizationCoroutine or null if a new connection
     * is to be made.
     */
    public AuthorizationCoroutine(AuthorizationCoroutine ac)
    {
	this(null, ac);
    }

    /**
     * Construct an empty AuthorizationCoroutine.
     */
    public AuthorizationCoroutine()
    {
	this(null, null);
    }

    /**
     * @return The requested Capabilities, if authorization was successful.
     */
    public CapabilityEnvelope[] getEnvelopes()
    {
	return this.ce;
    }

    /**
     * @return The peer coroutine identifier.
     */
    public long getPeerID()
    {
	return this.peerID;
    }

    /**
     * An EventChannel connected to the peer domain.
     */
    public EventChannel getEventChannel()
    {
	return this.ec;
    }

    /**
     * Request a capability from the Authenticator.
     *
     * @param ec An EventChannel connected to the Authenticator.
     * @param src The name of the capability, relative to the Authenticator's
     * internal ObjectTree.
     * @param dst The name to be placed in the CapabilityEnvelopes.
     * @param auth Authorization data/signatures.
     */
    public synchronized void requestCapability(String src,
					       String dst,
					       byte auth[])
    {
	try
	{
	    AuthorizationEvent ae;

	    ae = new AuthorizationEvent(this,
					AuthorizationEvent.METHOD_AUTHORIZE,
					EventChannel.currentChannel(),
					src, dst, auth);
	    this.ec.sendEvent(ae);
	    this.setState(STATE_REQUESTING);
	}
	catch(DeadBDomainException e)
	{
	    this.setState(STATE_ERROR);
	}
	catch(Throwable th)
	{
	    super.setThrowable(th);
	}
    }

    /**
     * Tell the authorization application that a protocol wishes to be
     * "managed" using a basic ProtocolPresentation.
     *
     * @param ec The authorizer's event channel.
     * @param sessionID The protocol session identifier that wishes to be
     * managed.
     */
    public synchronized void beginManagement(EventChannel ec, TypeID sessionID)
    {
	try
	{
	    AuthorizationEvent ae;

	    this.ec = ec;
	    ae = new AuthorizationEvent(this,
					AuthorizationEvent.METHOD_BEGIN,
					EventChannel.currentChannel(),
					sessionID);
	    this.ec.sendEvent(ae);
	    this.setState(STATE_REQUESTING);
	}
	catch(DeadBDomainException e)
	{
	    this.setState(STATE_ERROR);
	}
	catch(Throwable th)
	{
	    super.setThrowable(th);
	}
    }

    /**
     * Tell the authorization application that a protocol wishes to be
     * "managed" using an ExtendedProtocolPresentation.  In other words, the
     * authorizer will tell the protocol about new neighbors, lost neighbors,
     * and other messages supported by the extended presentation.
     *
     * @param ec The authorizer's event channel.
     * @param sessionID The protocol session identifier that wishes to be
     * managed.
     */
    public synchronized void beginExtendedManagement(EventChannel ec,
						     TypeID sessionID)
    {
	try
	{
	    AuthorizationEvent ae;

	    this.ec = ec;
	    ae = new AuthorizationEvent(this,
					AuthorizationEvent.
					METHOD_BEGIN_EXTENDED,
					EventChannel.currentChannel(),
					sessionID);
	    this.ec.sendEvent(ae);
	    this.setState(STATE_REQUESTING);
	}
	catch(DeadBDomainException e)
	{
	    this.setState(STATE_ERROR);
	}
	catch(Throwable th)
	{
	    super.setThrowable(th);
	}
    }

    /**
     * Call back used to handle a beginManagement event.
     *
     * @param ce An envelope containing the event channel of the requester.
     * @param sessionID The protocol session identifier that wishes to be
     * managed.
     */
    protected void beginManagement(CapabilityEnvelope ce, TypeID sessionID)
    {
    }
    
    /**
     * Call back used to handle a beginExtendedManagement event.
     *
     * @param ce An envelope containing the event channel of the requester.
     * @param sessionID The protocol session identifier that wishes to be
     * managed.
     */
    protected void beginExtendedManagement(CapabilityEnvelope ce,
					   TypeID sessionID)
    {
    }

    /**
     * Call back used to handle an authorize event.
     *
     * @param requester The envelope containing the event channel.
     * @param src The name of the capability, relative to the Authenticator's
     * internal ObjectTree.
     * @param dst The name to be placed in the CapabilityEnvelopes.
     * @param auth Authorization data/signatures.
     * @return An array of envelopes containing the requested capabilities.
     */
    protected CapabilityEnvelope[] authorize(CapabilityEnvelope requester,
					     String src,
					     String dst,
					     byte auth[])
    {
	return null;
    }

    /**
     * @see bees.core.Coroutine#handle(BeesEvent)
     */
    public synchronized void handle(BeesEvent ae)
    {
	switch( this.getState() )
	{
	case STATE_NEWBORN:
	    if( (ae instanceof AuthorizationEvent) && !ae.isReply() )
	    {
		AuthorizationEvent authe = (AuthorizationEvent)ae;

		try
		{
		    this.peerID = authe.getClientCoroutineID();
		    switch( authe.getMethod() )
		    {
		    case AuthorizationEvent.METHOD_BEGIN:
			this.beginManagement(authe.getChannel(),
					     authe.getSessionID());
			super.setState(STATE_CONNECTED);
			break;
		    case AuthorizationEvent.METHOD_BEGIN_EXTENDED:
			this.beginExtendedManagement(authe.getChannel(),
						     authe.getSessionID());
			super.setState(STATE_CONNECTED);
			break;
		    default:
			throw new IllegalStateException();
		    }
		}
		catch(Throwable th)
		{
		    Logger.log(th);
		    authe.setThrowable(th);
		}
		try
		{
		    authe.sendReply();
		}
		catch(DeadBDomainException e)
		{
		    super.setThrowable(e);
		}
	    }
	    else
	    {
		Strict.fail();
	    }
	    break;
	case STATE_REQUESTING:
	    if( (ae instanceof AuthorizationEvent) && ae.isReply() )
	    {
		AuthorizationEvent authe = (AuthorizationEvent)ae;
		Throwable th;
		
		if( (th = authe.getThrowable()) != null )
		{
		    super.setThrowable(th);
		}
		else if( (authe.getMethod() == AuthorizationEvent.METHOD_BEGIN)
			 || (authe.getMethod() ==
			     AuthorizationEvent.METHOD_BEGIN_EXTENDED) )
		{
		    this.peerID = authe.getServerCoroutineID();
		    this.setState(STATE_CONNECTED);
		}
		else
		{
		    Strict.fail();
		}
	    }
	    else
	    {
		Strict.fail();
	    }
	    break;
	case STATE_CONNECTED:
	    if( (ae instanceof AuthorizationEvent) && !ae.isReply() )
	    {
		AuthorizationEvent authe = (AuthorizationEvent)ae;

		try
		{
		    switch( authe.getMethod() )
		    {
		    case AuthorizationEvent.METHOD_AUTHORIZE:
			authe.setEnvelopes(this.authorize(
				authe.getChannel(),
				authe.getSource(),
				authe.getDestination(),
				authe.getAuthorization()));
			break;
		    default:
			throw new IllegalStateException();
		    }
		}
		catch(Throwable th)
		{
		    Logger.log(th);
		    authe.setThrowable(th);
		}
		try
		{
		    authe.sendReply();
		}
		catch(DeadBDomainException e)
		{
		    super.setThrowable(e);
		}
	    }
	    else if( (ae instanceof AuthorizationEvent) && ae.isReply() )
	    {
		AuthorizationEvent authe = (AuthorizationEvent)ae;
		Throwable th;
		
		if( (th = authe.getThrowable()) != null )
		{
		    super.setThrowable(th);
		}
		else
		{
		    this.ce = authe.getEnvelopes();
		    this.setState(STATE_COMPLETE);
		}
	    }
	    else
	    {
		Strict.fail();
	    }
	    break;
	default:
	    throw new Panic("Invalid state for coroutine: " + this);
	}
    }

    /**
     * @see java.lang.Object#toString()
     */
    public String toString()
    {
	return "AuthorizationCoroutine["
	    + super.toString()
	    + "; ce="
	    + this.ce
	    + "; peerID="
	    + this.peerID
	    + "]";
    }
}
