/*
 * AuthorizationEvent.java
 *
 * Copyright (c) 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is part of the Janos Bees Software.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */

package bees.core;

import bees.core.bdomain.DeadBDomainException;

/**
 * Event used by AuthorizationCoroutines.
 */
final class AuthorizationEvent
    extends TwoWayBeesEvent
{
    /**
     * The beginManagement message type.
     */
    static final int METHOD_BEGIN = 0;
    
    /**
     * The beginExtendedManagement message type.
     */
    static final int METHOD_BEGIN_EXTENDED = 1;
    
    /**
     * The authorize message type.
     */
    static final int METHOD_AUTHORIZE = 2;
    
    /**
     * The endManagement message type.
     */
    static final int METHOD_END = 3;

    /**
     * The message type.
     */
    private int method;

    /**
     * The name of the capability, relative to the Authenticator's internal
     * ObjectTree.
     */
    private String src;
    
    /**
     * The name to be placed in the CapabilityEnvelopes.
     */
    private String dst;
    
    /**
     * Authorization data/signatures.
     */
    private byte auth[];

    /**
     * An envelope containing the current channel for a request.  A set of
     * envelopes resulting from a successful authorize request for a reply.
     */
    private CapabilityEnvelope ce[];

    /**
     * The protocol session identifier that wishes to be managed.
     */
    private TypeID sessionID;

    /**
     * Construct an authorize event type.
     *
     * @param co The AuthorizationCoroutine this event is from.
     * @param method The message type.
     * @param ec The replyTo event channel.
     * @param src The name of the capability, relative to the Authenticator's
     * internal ObjectTree.
     * @param dst The name to be placed in the CapabilityEnvelopes.
     * @param auth Authorization data/signatures.
     */
    AuthorizationEvent(Coroutine co, int method,
		       EventChannel ec, String src, String dst, byte auth[])
    {
	super(co);

	this.method = method;
	this.ce = new CapabilityEnvelope[] { new CapabilityEnvelope(ec) };
	this.src = src;
	this.dst = dst;
	this.auth = auth;
    }

    /**
     * Construct a management event type.
     *
     * @param co The AuthorizationCoroutine this event is from.
     * @param method The message type.
     * @param ec The replyTo event channel.
     * @param sessionID The protocol session identifier that wishes to be
     * managed.
     */
    AuthorizationEvent(Coroutine co, int method,
		       EventChannel ec, TypeID sessionID)
    {
	super(co);

	this.method = method;
	this.ce = new CapabilityEnvelope[] { new CapabilityEnvelope(ec) };
	this.sessionID = sessionID;
    }

    /**
     * @return The message type.
     */
    int getMethod()
    {
	return this.method;
    }

    /**
     * @return The protocol session identifier that wishes to be managed.
     */
    TypeID getSessionID()
    {
	return this.sessionID;
    }
    
    /**
     * @return The envelope containing the event channel.
     */
    CapabilityEnvelope getChannel()
    {
	return this.ce[0];
    }

    /**
     * @return The name of the capability, relative to the Authenticator's
     * internal ObjectTree.
     */
    String getSource()
    {
	return this.src;
    }

    /**
     * @return The name to be placed in the CapabilityEnvelopes.
     */
    String getDestination()
    {
	return this.dst;
    }

    /**
     * @return Authorization data/signatures.
     */
    byte[] getAuthorization()
    {
	return this.auth;
    }

    /**
     * @param ce A set of envelopes resulting from a successful authorize
     * request for a reply.
     */
    void setEnvelopes(CapabilityEnvelope ce[])
    {
	this.ce = ce;
    }

    /**
     * @return A set of envelopes resulting from a successful authorize request
     * for a reply.
     */
    CapabilityEnvelope[] getEnvelopes()
    {
	return this.ce;
    }

    /**
     * @see bees.core.BeesEvent#sendReply()
     */
    protected void sendReply()
	throws DeadBDomainException
    {
	this.src = null;
	this.dst = null;
	this.auth = null;
	this.sessionID = null;
	super.sendReply();
    }

    /**
     * @see java.lang.Object#clone()
     */
    public Object clone()
    {
	Object retval = super.clone();
	AuthorizationEvent ae;

	ae = (AuthorizationEvent)retval;
	if( this.src != null )
	{
	    ae.src = CloneFactory.clone(this.src);
	}
	if( this.dst != null )
	{
	    ae.dst = CloneFactory.clone(this.dst);
	}
	if( this.auth != null )
	{
	    ae.auth = (byte[])this.auth.clone();
	}
	if( this.sessionID != null )
	{
	    ae.sessionID = (TypeID)this.sessionID.clone();
	}
	if( this.ce != null )
	{
	    int lpc;
	    
	    ae.ce = new CapabilityEnvelope[this.ce.length];
	    for( lpc = 0; lpc < ae.ce.length; lpc++ )
	    {
		ae.ce[lpc] = (CapabilityEnvelope)this.ce[lpc].clone();
	    }
	}
	return retval;
    }
    
    /**
     * @see java.lang.Object#toString()
     */
    public String toString()
    {
	return "AuthorizationEvent["
	    + super.toString()
	    + "; src="
	    + this.src
	    + "; dst="
	    + this.dst
	    + "; auth="
	    + this.auth
	    + "; ce="
	    + this.ce
	    + "]";
    }
}
