/*
 * BeesEvent.java
 *
 * Copyright (c) 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is part of the Janos Bees Software.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */

package bees.core;

import edu.utah.janosvm.sys.DeadTeamException;
import edu.utah.janosvm.sys.TeamEjectionException;

import edu.utah.janosvm.kit.comm.amiga.Facsimile;

import bees.core.bdomain.BDomain;
import bees.core.bdomain.DeadBDomainException;

/**
 * Base class for generic events that can be sent between coroutines in two
 * domains.
 */
public class BeesEvent
    extends Facsimile
{
    /**
     * The id of the coroutine that sent this event.  Only needs to be set if a
     * reply is expected.
     */
    private long clientCoroutineID = -1;

    /**
     * The id of the coroutine that should handle the initial event.
     */
    private long serverCoroutineID = -1;

    /**
     * The coroutine type this event is bound to.
     */
    private Class coroutineClass;

    /**
     * Construct an empty BeesEvent object.
     *
     * @param coroutine The coroutine thats using this event.
     */
    protected BeesEvent(Coroutine coroutine)
    {
	if( coroutine != null )
	{
	    this.clientCoroutineID = coroutine.getID();
	    this.coroutineClass = coroutine.getClass();
	}
    }

    /**
     * Construct an empty BeesEvent object.
     */
    protected BeesEvent()
    {
	this(null);
    }

    /**
     * Use the domain's main port for the event's reply port.
     */
    public final void useBDomainReplyPort()
    {
	this.setReplyPort(BDomain.currentBDomain().getPort());

	Assert.ensure(this.getReplyPort());
    }

    /**
     * @param id The id of the coroutine that sent this event.  Only needs to
     * be set if a reply is expected.
     */
    public void setClientCoroutineID(long id)
    {
	this.clientCoroutineID = id;
    }

    /**
     * @return The id of the coroutine that sent this event.  Only needs to be
     * set if a reply is expected.
     */
    public long getClientCoroutineID()
    {
	return this.clientCoroutineID;
    }

    /**
     * @param id The id of the coroutine that should handle the initial event.
     */
    public void setServerCoroutineID(long id)
    {
	this.serverCoroutineID = id;
    }

    /**
     * @return The id of the coroutine that should handle the initial event.
     */
    public long getServerCoroutineID()
    {
	return this.serverCoroutineID;
    }

    /**
     * @param cl The coroutine type this event is bound to.
     */
    protected void setCoroutineClass(Class cl)
    {
	this.coroutineClass = cl;
    }

    /**
     * @return The coroutine type this event is bound to.
     */
    protected Class getCoroutineClass()
    {
	return this.coroutineClass;
    }

    /**
     * @see edu.utah.janosvm.kit.comm.amiga.Facsimile#reply()
     */
    protected final void reply()
	throws DeadTeamException, TeamEjectionException
    {
	super.reply();
    }

    /**
     * Send a reply to the event.
     * 
     * @throws DeadBDomainException If the destination domain is unreachable
     */
    protected void sendReply()
	throws DeadBDomainException
    {
	try
	{
	    this.reply();
	}
	catch(DeadTeamException e)
	{
	    throw new DeadBDomainException();
	}
	catch(TeamEjectionException e)
	{
	    throw new DeadBDomainException();
	}
    }

    /**
     * @see java.lang.Object#toString()
     */
    public String toString()
    {
	return "BeesEvent["
	    + super.toString()
	    + "; clientCoroutineID="
	    + this.clientCoroutineID
	    + "; serverCoroutineID="
	    + this.serverCoroutineID
	    + "; coroutineClass="
	    + this.coroutineClass
	    + "]";
    }
}
