/*
 * ByteArrayType.java
 *
 * Copyright (c) 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is part of the Janos Bees Software.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */

package bees.core;

import java.util.Hashtable;

import bees.core.security.VersionString;

import bees.core.resource.Logger;

import bees.core.bdomain.BDomainLocal;

import bees.util.MD5;

/**
 * A ByteArrayType is used store type information for Capsules.
 */
class ByteArrayType
{
    /**
     * A BDomainLocal containing the mapping of code identifiers to
     * ByteArrayTypes.
     */
    private static final BDomainLocal BDOMAIN_BYTE_ARRAY = new BDomainLocal() {
	    protected Object initialValue()
	    {
		return new Hashtable();
	    }
	};

    /**
     * @return The BDomainLocal mapping table.
     */
    static Hashtable getBDomainByteArray()
    {
	return (Hashtable)BDOMAIN_BYTE_ARRAY.get();
    }

    /**
     * @param ct The ByteArrayType to put in the BDomainLocal mapping.
     * @return The actual ByteArrayType in the mapping, if one was already in
     * place.
     */
    static ByteArrayType putBDomainByteArray(ByteArrayType ct)
    {
	Hashtable ht = (Hashtable)BDOMAIN_BYTE_ARRAY.get();
	ByteArrayType retval;
	
	Assert.require(ct.byteArrayID);
	
	synchronized( ht )
	{
	    if( (retval = (ByteArrayType)ht.get(ct.byteArrayID)) == null )
	    {
		ht.put(ct.byteArrayID, ct);
		retval = ct;
	    }
	}
	return retval;
    }

    /**
     * @param baID The unique code identifier to lookup in the BDomainLocal
     * mapping.
     * @return The ByteArrayType with the matching code identifier value or
     * NULL if it doesn't exist.
     */
    static ByteArrayType getBDomainByteArrayFromID(TypeID baID)
    {
	Hashtable ht = (Hashtable)BDOMAIN_BYTE_ARRAY.get();
	
	return (ByteArrayType)ht.get(baID);
    }

    /**
     * Indicates a new ByteArrayType object.
     */
    static final short STATE_NEW = 0;

    /**
     * The code has been added to the flow's ClassLoader.
     */
    static final short STATE_LOADED = 1;

    /**
     * The class has been loaded and resolved.
     */
    static final short STATE_BOUND = 2;

    /**
     * The class is built into Bees.
     */
    static final short STATE_BUILTIN = 3;

    /**
     * The minimum valid type value.
     */
    static final short TYPE_MIN = 0;
    
    /**
     * An opaque byte array.
     */
    static final short TYPE_OPAQUE = 1;

    /**
     * The class is just regular code.
     */
    static final short TYPE_CODE = 2;

    /**
     * The class is a capsule.
     */
    static final short TYPE_CAPSULE = 3;

    /**
     * The class is a session.
     */
    static final short TYPE_SESSION = 4;

    /**
     * The maximum valid type value.
     */
    static final short TYPE_MAX = 5;

    /**
     * The type of class.
     */
    protected short type = TYPE_MIN;

    /**
     * The current state of the object.
     */
    private short state = STATE_NEW;

    /**
     * The name of the byte array.
     */
    private String byteArrayName;

    /**
     * The byte code for the class.
     */
    private byte byteArray[];

    /**
     * The unique code identifier.
     */
    private TypeID byteArrayID;

    /**
     * Construct an empty ByteArrayType object.
     */
    ByteArrayType()
    {
    }

    /**
     * @return The current state of the object.
     */
    synchronized short getState()
    {
	return this.state;
    }
    
    /**
     * @param state The new state for the object.
     */
    synchronized void setState(short state)
    {
    	this.state = state;
    }

    /**
     * @param byteArrayName The name of the class.
     * @return this
     */
    ByteArrayType setByteArrayName(String byteArrayName)
    {
	this.byteArrayName = byteArrayName;
	return this;
    }

    /**
     * @return The name of the class.
     */
    String getByteArrayName()
    {
	return this.byteArrayName;
    }

    /**
     * @param type The type of code in the object.
     * @return this
     */
    ByteArrayType setType(short type)
    {
	this.type = type;
	return this;
    }

    /**
     * @return The type of code in the object.
     */
    short getType()
    {
	return this.type;
    }

    /**
     * @return The byte code for the class.
     */
    byte[] getByteArray()
    {
	return this.byteArray;
    }

    /**
     * @param byteArrayID The unique identifier for this ByteArrayType.
     * @return this
     */
    ByteArrayType setByteArrayID(TypeID byteArrayID)
    {
	this.byteArrayID = byteArrayID;
	return this;
    }

    /**
     * @return The unique identifier for this ByteArrayType.
     */
    TypeID getByteArrayID()
    {
	return this.byteArrayID;
    }

    /**
     * Call back used to add more data to an MD5 digest.
     *
     * @param digest The digest to update with additional data.
     */
    void extraDigest(MD5 digest)
    {
    }

    /**
     * Import a byte array into the system.  If this byte array is new and has
     * no ID, then a new one is computed.  Otherwise, the 
     *
     * @param byteArray The byte array to import.
     * @param versionID ...
     * @return this
     */
    synchronized ByteArrayType importByteArray(byte byteArray[],
    					       VersionString versionID)
    {
	if( this.byteArray == null )
	{
	    MD5 digest = new MD5();
	    TypeID computedID;
	    
	    Assert.require(this.state == STATE_NEW);

	    if( versionID != null )
	    {
		digest.update(this.byteArrayName
			      + " - "
			      + versionID.getValue());
	    }
	    else
	    {
		digest.update(byteArray);
	    }
	    this.extraDigest(digest);
	    computedID = new TypeID(digest.hash());
	    if( this.byteArrayID != null )
	    {
		if( !this.byteArrayID.equals(computedID) )
		{
		    throw new SecurityException("ByteArray("
						+ this.byteArrayName
						+ ") has invalid computed ID: "
						+ computedID
						+ ".  It should be: "
						+ this.byteArrayID);
		}
	    }
	    else
	    {
		this.byteArrayID = computedID;
	    }
	    this.byteArray = byteArray;
	}
	return this;
    }

    /**
     * @see #importByteArray(byte[], VersionString)
     */
    ByteArrayType importByteArray(byte byteArray[])
    {
	return this.importByteArray(byteArray, null);
    }

    /**
     * @return The number of bytes needed to marshal this object.
     */
    int marshalledLength()
    {
	return (Xdr.SHORT
		+ Xdr.STRING(this.byteArrayName)
		+ Xdr.BOOLEAN
		+ Xdr.TYPEID);
    }

    /**
     * @param xba The byte array to marshal the ByteArrayType into.
     */
    void marshal(XdrByteArray xba)
    {
    	xba.PUT(this.type);
	xba.PUT(this.byteArrayName);
	xba.PUT(this.state == STATE_BUILTIN);
	xba.PUT(this.byteArrayID);
    }
    
    /**
     * @param xba The source of the flattened version of this ByteArrayType.
     */
    void unmarshal(XdrByteArray xba)
    {
	this.byteArrayName = xba.STRING();
	if( xba.BOOLEAN() )
	{
	    this.state = STATE_BUILTIN;
	}
	this.byteArrayID = xba.TYPEID();
    }

    /**
     * @param xba The byte array containing the flattened ByteArrayType.
     * @return A new ByteArrayType.
     */
    static ByteArrayType instantiate(XdrByteArray xba)
    {
    	ByteArrayType bat;
    	short type;
    	
    	type = xba.SHORT();
    	switch( type )
    	{
  	case TYPE_OPAQUE:
	    bat = new ByteArrayType();
	    break;
	case TYPE_CODE:
	case TYPE_CAPSULE:
	case TYPE_SESSION:
	    bat = new CodeType();
	    break;
	default:
	    throw new IllegalArgumentException("Bad ByteArrayType.type value");
    	}
    	bat.type = type;
	bat.unmarshal(xba);
	return putBDomainByteArray(bat);
    }

    /**
     * @see java.lang.Object#toString()
     */
    public String toString()
    {
	return "ByteArrayType[state="
	    + this.state
	    + "; type="
	    + this.type
	    + "; byteArrayName="
	    + this.byteArrayName
	    + "; byteArray="
	    + this.byteArray
	    + "; byteArrayID="
	    + this.byteArrayID
	    + "]";
    }
}
