/*
 * Capability.java
 *
 * Copyright (c) 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is part of the Janos Bees Software.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */

package bees.core;

import edu.utah.janosvm.kit.util.SparseArray;

import bees.core.kernel.KernelObjectCoroutine;

import bees.core.security.Permission;
import bees.core.security.PermissionSet;
import bees.core.security.PermissionException;

import bees.core.bdomain.BDomain;

/**
 * Base class for capabilities in the system.  A "capability" in the context
 * of Bees is an object with the following attributes:
 *
 * <ul>
 * <li>Controls access to node resources - The main purpose of an Bees
 * capability is to control the effects untrusted code can have on a node.
 * <li>Shareable between domains - Domains isolate individual Java objects, a
 * capability provides a means to uniquely reference objects in other domains.
 * <li>Serializable - Each capability must be able to flatten its state for
 * storage on disk and transmission to other domains.
 * </ul>
 */
public abstract class Capability
{
    /**
     * Permission to delete a capability.
     */
    public static final Permission DELETE_PERMISSION =
	new Permission("delete", "execute");

    /**
     * An invalid capability identifier.
     */
    public static final long INVALID_ID = SparseArray.INVALID_ID;
    
    /**
     * The permissions attached to this capability.
     */
    private PermissionSet ps;

    /**
     * The unique kernel object store identifier for this capability.
     */
    protected long id = INVALID_ID;
    
    /**
     * Construct an empty capability.
     */
    protected Capability()
    {
    }

    /**
     * Clear the internal bits binding this capability to the kernel object.
     */
    protected void clear()
    {
	this.ps = null;
	this.id = INVALID_ID;
    }

    /**
     * @return True if this capability is still valid.
     */
    protected final boolean valid()
    {
	return (this.id != INVALID_ID);
    }

    /**
     * @param ps The permissions attached to this capability.
     */
    void setPermissions(PermissionSet ps)
    {
	this.ps = ps;
    }

    /**
     * @return The permissions attached to this capability.
     */
    PermissionSet getPermissions()
    {
	return this.ps;
    }

    /**
     * @param id The unique kernel object store identifier for this capability.
     */
    void setID(long id)
    {
	this.id = id;
    }

    /**
     * @return The unique kernel object store identifier for this capability.
     */
    long getID()
    {
	return this.id;
    }

    /**
     * Bind this capability to the given object.
     *
     * @param flattened The object this capability gives access too.
     * @throws Exception if anything goes wrong during initialization.
     */
    protected synchronized void fromBytes(byte flattened[])
	throws Exception
    {
	if( this.id == INVALID_ID )
	{
	    throw new SecurityException(
		"Attempt to initialize invalid capability");
	}
    }

    /**
     * Bind this capability to the given object.
     *
     * @param flattened The object this capability gives access too.
     * @param sub Null or any sub-capabilities.
     * @throws Exception if anything goes wrong during initialization.
     */
    protected synchronized void fromBytes(byte flattened[],
					  CapabilityEnvelope sub[])
	throws Exception
    {
	this.fromBytes(flattened);
    }

    /**
     * Revoke access to the capability.
     */
    public synchronized void revoke()
    {
	throw new UnsupportedOperationException();
    }

    /**
     * Check if this capability has the permissions to do something.
     *
     * @param what The permission to check for.
     * @throws PermissionException if the capability doesn't
     * has the permission.
     */
    public final synchronized void checkPermission(Permission what)
	throws PermissionException
    {
	if( (this.ps == null) || this.ps.hasPermission(what) == null )
	{
	    throw new PermissionException("Invalid access:  Capability, "
					  + this
					  + ", doesn't have permission, "
					  + what.toString());
	}
    }

    /**
     * @param what The permission to clone.
     * @return A clone of the capabilities Permission that matches the given
     * one.
     * @throws PermissionException if the capability doesn't
     * has the permission.
     */
    public final synchronized Permission clonePermission(Permission what)
	throws PermissionException
    {
	Permission retval;
	
	if( (this.ps == null) ||
	    (retval = this.ps.hasPermission(what)) == null )
	{
	    throw new PermissionException("Invalid access:  Capability, "
					  + this
					  + ", doesn't have permission, "
					  + what.toString());
	}
	return (Permission)retval.clone();
    }

    /**
     * @param what The Permission to revoke.
     */
    public final synchronized void revokePermission(Permission what)
    {
	this.ps.revokePermission(what);
    }

    /**
     * @param ps The set of Permissions to revoke.
     */
    public final synchronized void revokePermissions(PermissionSet ps)
    {
	this.ps.revokePermissions(ps);
    }

    /**
     * Callback...
     *
     * XXX Implement me.
     */
    public synchronized void revokedPermission(Permission what)
    {
    }

    /**
     * Delete a capability.
     */
    public final void deleteCapability()
    {
	long id;
	
	this.checkPermission(DELETE_PERMISSION);

	id = this.id;
	this.revoke();
	KernelObjectCoroutine.delete(id,
				     BDomain.currentBDomain().
				     getBDomainCoroutine());
    }

    /**
     * @see java.lang.Object#equals(Object)
     */
    public boolean equals(Object obj)
    {
	boolean retval = false;

	if( (obj != null) && (obj.getClass() == super.getClass()) )
	{
	    retval = (this.id == ((Capability)obj).id);
	}
	return retval;
    }

    /**
     * @see java.lang.Object#toString()
     */
    public String toString()
    {
	return "Capability[ps="
	    + this.ps
	    + "; id="
	    + this.id
	    + "]";
    }
}
