/*
 * CapabilityReference.java
 *
 * Copyright (c) 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is part of the Janos Bees Software.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */

package bees.core;

import edu.utah.janos.nodeos.Domain;

import bees.core.security.PermissionSet;

/**
 * A lightweight class for addressing capabilities shared by multiple domains.
 * It is expected that both domains have already been given the Capability
 * through a CapabilityEnvelope.
 *
 * @see Capability
 * @see CapabilityEnvelope
 */
public final class CapabilityReference
    implements Cloneable
{
    /**
     * Encode a capability in an xdr.
     *
     * @param xdr The xdr that will hold the capability reference.
     * @param cap The capability to encode.
     */
    public static void encode(Xdr xdr, Capability cap)
    {
	Assert.require(xdr);
	Assert.require(cap);

	synchronized( cap )
	{
	    if( cap.getID() == Capability.INVALID_ID )
	    {
		throw new IllegalArgumentException("Invalid capability");
	    }
	    xdr.PUT(cap.getID());
	    xdr.PUT(cap.getClass().getName());
	}
    }

    /**
     * Decode a capability reference.
     * <p>
     * XXX Should it just return the Capability?
     *
     * @param xdr The xdr that holds the capability reference.
     * @return A new CapabilityReference.
     * @throws InvalidCapabilityReferenceException if the encoded reference is
     * invalid.
     * @throws ClassNotFoundException if the capability class wasn't found.
     */
    public static CapabilityReference decode(Xdr xdr)
	throws InvalidCapabilityReferenceException,
	       ClassNotFoundException
    {
	CapabilityReference retval = null;
	Domain domain = Domain.currentDomain();
	Class cl;
	long id;

	Assert.require(xdr);

	id = xdr.LONG();
	if( id == Capability.INVALID_ID )
	{
	    throw new InvalidCapabilityReferenceException();
	}
	cl = domain.loadClass(xdr.STRING());
	retval = new CapabilityReference(id, cl);
	return retval;
    }

    /**
     * The capability identifier.
     */
    private final long id;

    /**
     * The capability class.
     */
    private final Class cl;

    /**
     * Construct a CapabilityReference with the given values.
     *
     * @param id The capability id.
     * @param cl The capability class.
     */
    CapabilityReference(long id, Class cl)
    {
	Assert.require(id >= 0);
	Assert.require(cl);

	if( id < 0 )
	{
	    throw new IllegalArgumentException("Invalid capability id");
	}
	this.id = id;
	this.cl = cl;
    }

    /**
     * Construct a CapabilityReference referring to the given capability.
     *
     * @param cap The capability to enclose.
     */
    public CapabilityReference(Capability cap)
    {
	this(cap.getID(), cap.getClass());
    }

    /**
     * Encode the reference in an xdr.
     *
     * @param xdr The xdr to encode the reference into.
     */
    public void encode(Xdr xdr)
    {
	Assert.require(xdr);
	
	xdr.PUT(this.id);
	xdr.PUT(this.cl.getName());
    }

    /**
     * Dereference this object to get the referred to capability.
     *
     * @return The referenced capability.
     * @throws MissingCapabilityException if the referenced Capability is not
     * in this domain or has been revoked.
     */
    public synchronized Capability dereference()
	throws MissingCapabilityException
    {
	CapabilityTracker ct = CapabilityTracker.current();
	Capability retval;

	if( (retval = ct.get(this.id, this.cl)) == null )
	{
	    throw new MissingCapabilityException(id);
	}
	return retval;
    }

    /**
     * @see java.lang.Object#clone()
     */
    public synchronized Object clone()
    {
	try
	{
	    CapabilityReference cr;
	    PermissionSet ps;
	    Object retval;

	    retval = super.clone();
	    cr = (CapabilityReference)retval;
	    return retval;
	}
	catch(CloneNotSupportedException e)
	{
	    throw new Panic(e);
	}
    }

    /**
     * @see java.lang.Object#toString()
     */
    public String toString()
    {
	return "CapabilityReference[id="
	    + this.id
	    + "; cl="
	    + this.cl
	    + "]";
    }
}
