/*
 * CapsuleType.java
 *
 * Copyright (c) 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is part of the Janos Bees Software.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */

package bees.core;

import java.util.Hashtable;

import bees.core.resource.Logger;

import bees.core.security.Permission;
import bees.core.security.PermissionSet;
import bees.core.security.CountPermission;

import bees.core.bdomain.BDomainLocal;

import bees.util.MD5;

/**
 * A CapsuleType is used store type information for Capsules.  This is a
 * "run-time" type, it is constructed out of the Flow/Protocol/ByteArray type
 * objects.
 */
final class CapsuleType
{
    /**
     * The NULL specializer identifier.
     */
    static final TypeID NULL_SPECIALIZER =
	new TypeID(new byte[MD5.HASH_LENGTH]);

    /**
     * Index for the "grouped" permission type.  This is used to indicate the
     * maximum number of capsules of a given type can be placed in
     * CapsuleGroups.
     */
    static final int GLOBAL_GROUPED = 0;
    
    /**
     * Index for the "cached" permission type.  This is used to indicate the
     * maximum number of capsules of a given type can be placed in NodeCaches.
     */
    static final int GLOBAL_CACHED = 1;

    /**
     * The maximum number of global permission types.
     */
    static final int GLOBAL_MAX = 2;

    /**
     * The set of global permissions.
     */
    static final Permission GLOBAL_PERMISSIONS[] = new Permission[] {
	    new Permission("grouped"),
	    new Permission("cached"),
    };
    
    /**
     * A BDomainLocal containing the tree of Flow/Code:Class/Capsule mappings.
     */
    private static final BDomainLocal BDOMAIN_CAPSULES = new BDomainLocal() {
	    protected Object initialValue()
	    {
		return new Hashtable();
	    }
	};

    /**
     * Put a CapsuleType into the BDomainLocal tree.
     *
     * @param ct The CapsuleType to add to the tree.
     */
    static void putBDomainCapsuleType(CapsuleType ct)
    {
	Hashtable ht = (Hashtable)BDOMAIN_CAPSULES.get();

	synchronized( ht )
	{
	    Hashtable sub, sub2;
	    Object other = null;
	    TypeID specializer;

	    if( (sub = (Hashtable)ht.get(ct.ft.getFlowID())) == null )
	    {
		sub = new Hashtable();
		other = ht.put(ct.ft.getFlowID(), sub);

		Assert.ensure(other == null);
	    }
	    if( (sub2 = (Hashtable)sub.get(ct.ct.getByteArrayID())) == null )
	    {
		sub2 = new Hashtable();
		other = sub.put(ct.ct.getByteArrayID(), sub2);
		Assert.ensure(other == null);
		other = sub.put(ct.ct.getCodeClass(), sub2);
		Assert.ensure(other == null);
	    }
	    if( (specializer = ct.specializedCapsuleID) == null )
	    {
		specializer = NULL_SPECIALIZER;
	    }
	    other = sub2.put(specializer, ct);

	    Assert.ensure(other == null);
	}
    }

    /**
     * @param ct The CapsuleType to check for.
     * @return True if the BDomainLocal tree has the given CapsuleType.
     */
    static boolean containsBDomainCapsuleType(CapsuleType ct)
    {
	boolean retval = false;
	TypeID specializer;
	
	if( (specializer = ct.specializedCapsuleID) == null )
	{
	    specializer = NULL_SPECIALIZER;
	}
	if( getBDomainCapsuleType(ct.ft.getFlowID(),
				  ct.ct.getByteArrayID(),
				  specializer) != null )
	{
	    retval = true;
	}
	return retval;
    }
    
    /**
     * Lookup a CapsuleType in the BDomainLocal tree using the given
     * identifiers.
     *
     * @param flowID The flow identifier.
     * @param codeID The implementation class' identifier.
     * @param specializerID The Capsule specializer identifier.
     * @return The corresponding CapsuleType or null if it was not found.
     */
    static CapsuleType getBDomainCapsuleType(TypeID flowID,
					     TypeID codeID,
					     TypeID specializerID)
    {
	Hashtable ht = (Hashtable)BDOMAIN_CAPSULES.get();
	CapsuleType retval = null;

	Assert.require(flowID);
	Assert.require(codeID);
	Assert.require(specializerID);
	
	synchronized( ht )
	{
	    if( (ht = (Hashtable)ht.get(flowID)) != null )
	    {
		if( (ht = (Hashtable)ht.get(codeID)) != null )
		{
		    retval = (CapsuleType)ht.get(specializerID);
		}
	    }
	}
	return retval;
    }

    /**
     * Lookup a CapsuleType in the BDomainLocal tree using the given
     * identifiers and the NULL_SPECIALIZER.
     *
     * @param flowID The flow identifier.
     * @param codeID The implementation class' identifier.
     * @return The corresponding CapsuleType or null if it was not found.
     */
    static CapsuleType getBDomainCapsuleType(TypeID flowID,
					  TypeID codeID)
    {
	return getBDomainCapsuleType(flowID, codeID, NULL_SPECIALIZER);
    }

    /**
     * Lookup a CapsuleType in the BDomainLocal tree using the given
     * identifiers.
     *
     * @param flowID The flow identifier.
     * @param cl The implementation class.
     * @param specializerID The Capsule specializer identifier.
     * @return The corresponding CapsuleType or null if it was not found.
     */
    static CapsuleType getBDomainCapsuleType(TypeID flowID,
					     Class cl,
					     TypeID specializerID)
    {
	Hashtable ht = (Hashtable)BDOMAIN_CAPSULES.get();
	CapsuleType retval = null;

	synchronized( ht )
	{
	    if( (ht = (Hashtable)ht.get(flowID)) != null )
	    {
		if( (ht = (Hashtable)ht.get(cl)) != null )
		{
		    retval = (CapsuleType)ht.get(specializerID);
		}
	    }
	}
	return retval;
    }

    /**
     * Lookup a CapsuleType in the BDomainLocal tree using the given
     * identifiers and the NULL_SPECIALIZER.
     *
     * @param flowID The flow identifier.
     * @param cl The implementation class.
     * @return The corresponding CapsuleType or null if it was not found.
     */
    static CapsuleType getBDomainCapsuleType(TypeID flowID, Class cl)
    {
	return getBDomainCapsuleType(flowID, cl, NULL_SPECIALIZER);
    }

    /**
     * The FlowType that this CapsuleType belongs to.
     *
     * NOTE: A single ProtocolType can be shared between many flows so we must
     * track both types.
     */
    private FlowType ft;

    /**
     * The ProtocolType that this CapsuleType belongs to.
     */
    private ProtocolType pt;

    /**
     * The CodeType that implements this CapsuleType.
     */
    private CodeType ct;

    private InternalPacketAuthorization ipa;

    /**
     * The unique Capsule identifier.
     */
    private TypeID capsuleID;

    /**
     * The companion specialized Capsule identifier.
     */
    private TypeID specializedCapsuleID;

    /**
     * TTLs for controlling the global number of capsules that can gain
     * authorization.
     */
    private int globalCapsules[] = new int[GLOBAL_MAX];

    /**
     * Construct an empty CapsuleType.
     */
    CapsuleType()
    {
    }

    /**
     * Decrement a global permission value.
     *
     * @param type One of the GLOBAL_* enumeration values.
     * @throws SecurityException if there are too many capsules in groups or
     * caches.
     */
    synchronized void decGlobal(int type)
    	throws SecurityException
    {
	Assert.require(type >= 0);
	Assert.require(type < GLOBAL_MAX);
	
	if( this.globalCapsules[type] <= 0 )
	{
	    Assert.require(this.globalCapsules[type] == 0);
	    
	    throw new SecurityException("Too many global capsules for: "
					+ GLOBAL_PERMISSIONS[type]);
	}
	this.globalCapsules[type] -= 1;
    }

    /**
     * Increment a global permission value.
     *
     * @param type One of the GLOBAL_* enumeration values.
     */
    synchronized void incGlobal(int type)
    {
	Assert.require(type >= 0);
	Assert.require(type < GLOBAL_MAX);
	
	this.globalCapsules[type] += 1;
    }

    /**
     * @param ft The FlowType that this CapsuleType belongs to.
     * @return this
     */
    CapsuleType setFlowType(FlowType ft)
    {
	this.ft = ft;
	return this;
    }

    /**
     * @return The FlowType that this CapsuleType belongs to.
     */
    FlowType getFlowType()
    {
	return this.ft;
    }

    /**
     * @param pt The ProtocolType that this CapsuleType belongs to.
     * @return this
     */
    CapsuleType setProtocolType(ProtocolType pt)
    {
	this.pt = pt;
	return this;
    }

    /**
     * @return The ProtocolType that this CapsuleType belongs to.
     */
    ProtocolType getProtocolType()
    {
	return this.pt;
    }

    /**
     * @param ct The CodeType that implements this CapsuleType.
     * @return this
     */
    CapsuleType setCodeType(CodeType ct)
    {
	PermissionSet ps;
	int lpc;
	
	this.ct = ct;
	ps = this.ct.getPermissions();
	for( lpc = 0; lpc < GLOBAL_MAX; lpc++ )
	{
	    CountPermission cp;

	    try
	    {
		if( (cp = (CountPermission)ps.
		     hasPermission(GLOBAL_PERMISSIONS[lpc])) != null )
		{
		    this.globalCapsules[lpc] = cp.getMaxCount();
		}
	    }
	    catch(ClassCastException e)
	    {
		/* Ignore non-CountPermissions in ps. */
		Logger.log(Logger.WARNING,
			   "Invalid permission in Capsule permission set: "
			   + ps.hasPermission(GLOBAL_PERMISSIONS[lpc]));
	    }
	}
	this.ipa = new InternalPacketAuthorization(ps.subset(new Permission() {
		public boolean implies(Permission p)
		{
		    boolean retval = false;
		    
		    if( p.getName().startsWith(InternalPacketAuthorization.
					       TRANSFORM_PREFIX) )
		    {
			retval = true;
		    }
		    return retval;
		}
	    }));
	return this;
    }

    /**
     * @return The CodeType that implements this CapsuleType.
     */
    CodeType getCodeType()
    {
	return this.ct;
    }

    InternalPacketAuthorization getInternalPacketAuthorization()
    {
	return this.ipa;
    }

    /**
     * @param capsuleID The unique Capsule identifier.
     * @return this
     */
    CapsuleType setCapsuleID(TypeID capsuleID)
    {
	this.capsuleID = capsuleID;
	return this;
    }

    /**
     * @return The unique Capsule identifier.
     */
    TypeID getCapsuleID()
    {
	return this.capsuleID;
    }

    /**
     * @param id The companion specialized Capsule identifier.
     * @return this
     */
    CapsuleType setSpecializedCapsuleID(TypeID id)
    {
	this.specializedCapsuleID = id;
	return this;
    }

    /**
     * @return The companion specialized Capsule identifier.
     */
    TypeID getSpecializedCapsuleID()
    {
	return this.specializedCapsuleID;
    }

    /**
     * Snapshot the state of the CapsuleType.  This is done after the structure
     * has been completely filled out in order to manufacture/verify the
     * Capsule's unique identifier.
     *
     * XXX I don't think the verification version actually gets used.
     */
    void snapshotCapsuleType()
    {
	TypeID computedCapsuleID;
	MD5 digest = new MD5();

	if( (this.pt.getFlags() & ProtocolType.FLAG_COMPANION_FLOW_INDEPENDENT)
	    == 0 )
	{
	    digest.update(this.ft.getFlowID().getHash());
	}
	digest.update(this.pt.getProtocolID().getHash());
	digest.update(this.ct.getByteArrayID().getHash());
	computedCapsuleID = new TypeID(digest.hash());
	if( this.capsuleID != null )
	{
	    if( !this.capsuleID.equals(computedCapsuleID) )
	    {
		throw new SecurityException("Computed CapsuleID "
					    + computedCapsuleID
					    + " != "
					    + this.capsuleID
					    + " for "
					    + this);
	    }
	}
	else
	{
	    this.capsuleID = computedCapsuleID;
	}
    }

    /**
     * @see java.lang.Object#toString()
     */
    public String toString()
    {
	return "CapsuleType[ft="
	    + this.ft
	    + "; pt="
	    + this.pt
	    + "; ct="
	    + this.ct
	    + "; capsuleID="
	    + this.capsuleID
	    + "; specializedCapsuleID="
	    + this.specializedCapsuleID
	    + "]";
    }
}
