/*
 * CipheredXdrByteArray.java
 *
 * Copyright (c) 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is part of the Janos Bees Software.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */

package bees.core;

import java.security.InvalidKeyException;
import java.security.NoSuchAlgorithmException;
import java.security.InvalidAlgorithmParameterException;

import javax.crypto.BadPaddingException;
import javax.crypto.NoSuchPaddingException;
import javax.crypto.IllegalBlockSizeException;

import javax.crypto.spec.IvParameterSpec;

import bees.core.resource.Clock;

import bees.core.security.Cipher;
import bees.core.security.KeySpec;
import bees.core.security.PrivateKeySpec;

/**
 * A CipheredXdrByteArray is a convenience for encrypting/decrypting an
 * XdrByteArray.
 */
public class CipheredXdrByteArray
    extends XdrByteArray
{
    /**
     * The name of the cryptographic algorithm to use.
     */
    private String algorithm;
    
    /**
     * The iv value.
     */
    private byte iv[];

    /**
     * Construct a CipheredXdrByteArray with the given values.
     *
     * @param len The length of the byte array.
     * @param algorithm The name of the cryptographic algorithm to use.
     */
    public CipheredXdrByteArray(int len, String algorithm)
    {
	super(len);

	Assert.require(algorithm);

	this.algorithm = algorithm;
    }

    /**
     * Construct a CipheredXdrByteArray with the given values.
     *
     * @param buf The byte array to wrap.
     * @param algorithm The name of the cryptographic algorithm to use.
     */
    public CipheredXdrByteArray(byte buf[], String algorithm)
    {
	super(buf);
	
	Assert.require(algorithm);

	this.algorithm = algorithm;
    }

    /**
     * Construct a CipheredXdrByteArray with the given values.
     *
     * @param buf The byte array to wrap.
     * @param iv The initialization vector for "buf".
     * @param algorithm The name of the cryptographic algorithm to use.
     */
    public CipheredXdrByteArray(byte buf[], byte iv[], String algorithm)
    {
	super(buf);
	
	Assert.require(algorithm);

	this.iv = iv;
	this.algorithm = algorithm;
    }

    /**
     * @return The name of the cryptographic algorithm to use.
     */
    public String getAlgorithm()
    {
	return this.algorithm;
    }

    /**
     * @return The initialization vector.
     */
    public byte[] getIV()
    {
	return this.iv;
    }

    /**
     * @param ks The KeySpec to use when encrypting this byte array.
     * @throws IllegalBlockSizeException
     * @throws NoSuchAlgorithmException
     * @throws NoSuchPaddingException
     * @throws BadPaddingException
     * @throws InvalidKeyException
     */
    public void encrypt(KeySpec ks)
	throws IllegalBlockSizeException,
	       NoSuchAlgorithmException,
	       NoSuchPaddingException,
	       BadPaddingException,
	       InvalidKeyException
    {
	Cipher cipher;

	Assert.require(ks);

	cipher = new Cipher(this.algorithm);
	cipher.init(Cipher.ENCRYPT_MODE, ks);
	super.buf = cipher.doFinal(super.buf);
	super.offset = 0;
	super.len = super.buf.length;
	this.iv = cipher.getIV();
    }

    /**
     * @param ks The KeySpec to use when verifying this byte array.
     * @throws InvalidAlgorithmParameterException
     * @throws IllegalBlockSizeException
     * @throws NoSuchAlgorithmException
     * @throws NoSuchPaddingException
     * @throws BadPaddingException
     * @throws InvalidKeyException
     */
    public void decrypt(KeySpec ks)
	throws InvalidAlgorithmParameterException,
	       IllegalBlockSizeException,
	       NoSuchAlgorithmException,
	       NoSuchPaddingException,
	       BadPaddingException,
	       InvalidKeyException
    {
	if( this.iv != null )
	{
	    Cipher cipher;
	    
	    Assert.require(ks);
	    
	    cipher = new Cipher(this.algorithm);
	    cipher.init(Cipher.DECRYPT_MODE, ks, new IvParameterSpec(this.iv));
	    super.buf = cipher.doFinal(super.buf);
	    super.offset = 0;
	    super.len = super.buf.length;
	    this.iv = null;
	}
    }

    /**
     * @see java.lang.Object#toString()
     */
    public String toString()
    {
	return "CipheredXdrByteArray["
	    + super.toString()
	    + "; algorithm="
	    + this.algorithm
	    + "; iv="
	    + ByteArray.toString(this.iv)
	    + "]";
    }
}
