/*
 * ConnectSessionEvent.java
 *
 * Copyright (c) 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is part of the Janos Bees Software.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */

package bees.core;

import bees.core.bdomain.DeadBDomainException;

/**
 * A ConnectSessionEvent is used to connect two ExternalCoroutines in domain's
 * that wish to communicate.
 */
final class ConnectSessionEvent
    extends TwoWayBeesEvent
{
    /**
     * The EventChannel that will be used to carry messages.
     */
    private CapabilityReference eventChannel;

    /**
     * The identifier for a particular protocol session that should receive
     * ExternalCapsules sent through the coroutines.
     */
    private TypeID sessionID;

    /**
     * Construct a ConnectSessionEvent with the given values.
     *
     * @param ec The client ExternalCoroutine.
     * @param eventChannel The channel that will be the communication medium
     * for the domains.
     * @param sessionID The protocol session identifier.
     */
    ConnectSessionEvent(ExternalCoroutine ec,
			EventChannel eventChannel,
			TypeID sessionID)
    {
	super(ec);

	this.eventChannel = new CapabilityReference(eventChannel);
	this.sessionID = sessionID;
    }

    /**
     * @return The channel that will be the communication medium for the
     * domains.
     * @throws MissingCapabilityException if the EventChannel has since been
     * deleted.
     */
    EventChannel getEventChannel()
	throws MissingCapabilityException
    {
	return (EventChannel)this.eventChannel.dereference();
    }

    /**
     * @return The protocol session identifier.
     */
    TypeID getSessionID()
    {
	return this.sessionID;
    }

    /**
     * @see bees.core.BeesEvent#sendReply()
     */
    protected void sendReply()
	throws DeadBDomainException
    {
	this.sessionID = null;
	super.sendReply();
    }

    /**
     * @see java.lang.Object#clone()
     */
    public Object clone()
    {
	Object retval = super.clone();
	ConnectSessionEvent cse;

	cse = (ConnectSessionEvent)retval;
	if( this.eventChannel != null )
	{
	    cse.eventChannel = (CapabilityReference)this.eventChannel.clone();
	}
	if( this.sessionID != null )
	{
	    cse.sessionID = (TypeID)this.sessionID.clone();
	}
	return retval;
    }

    /**
     * @see java.lang.Object#toString()
     */
    public String toString()
    {
	return "ConnectSessionEvent["
	    + super.toString()
	    + "; eventChannel="
	    + this.eventChannel
	    + "; sessionID="
	    + this.sessionID
	    + "]";
    }
}
