/*
 * DLProtocolPresentation.java
 *
 * Copyright (c) 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is part of the Janos Bees Software.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */

package bees.core;

import bees.core.resource.Logger;

/**
 * A DLProtocolPresentation is used to communicate with the DLProtocolSession
 * in a protocol domain.
 */
public class DLProtocolPresentation
    extends ExtendedProtocolPresentation
{
    /**
     * The download has completed.
     *
     * XXX Be more explicit with the state here.
     */
    private int state = DLProtocolSession.STATE_UNKNOWN;

    private SessionType userSession;

    protected DLProtocolPresentation()
    {
    }

    void setSessionType(SessionType userSession)
    {
	this.userSession = userSession;
    }

    /**
     * Start the download.
     */
    void start()
    {
	ExternalCapsule ec = new ExternalCapsule().init();
	Xdr xdr;
	
	xdr = ec.getXDR();
	xdr.PUT(DLProtocolSession.METHOD_START);
	this.sendCapsule(ec);
    }

    public void bootstrap()
    {
	ExternalCapsule ec = new ExternalCapsule().init();
	Xdr xdr;
	
	xdr = ec.getXDR();
	xdr.PUT(DLProtocolSession.METHOD_BOOT_CAPSULE);
	this.sendCapsule(ec);
    }

    /**
     * Wait for the download to complete.
     *
     * @param timeout The time to wait before timing out.
     */
    public synchronized boolean waitForCompletion(long timeout)
    {
	long start, curr, end;

	start = curr = System.currentTimeMillis();
	end = start + timeout;
	while( ((timeout == 0) || (curr < end)) &&
	       (this.state == DLProtocolSession.STATE_UNKNOWN) )
	{
	    try
	    {
		this.wait(timeout);
		curr = System.currentTimeMillis();
	    }
	    catch(InterruptedException e)
	    {
	    }
	}
	return (this.state > DLProtocolSession.STATE_UNKNOWN);
    }

    /**
     * Wait indefinitely for the download to complete.
     */
    public synchronized boolean waitForCompletion()
    {
	return this.waitForCompletion(0);
    }

    public ProtocolPresentation createBasicPresentation()
    {
	return new ProtocolPresentation() {
		public void receive(Node node, ExternalCapsule ec)
		{
		}
	    };
    }

    public ExtendedProtocolPresentation createExtendedPresentation()
    {
	return new ExtendedProtocolPresentation() {
		public void receive(Node node, ExternalCapsule ec)
		{
		}
	    };
    }
    
    public void downloadComplete(boolean success)
    {
	if( success )
	    this.bootstrap();
    }
    
    public void bootComplete(ProtocolPresentation pp)
    {
    }
    
    public synchronized void userReceive(Node node,
					 int method,
					 ExternalCapsule ec)
    {
	Xdr xdr = ec.getXDR();
	
	if( (method & ExtendedProtocolSession.METHOD_REPLY) != 0 )
	{
	    switch( method & ~ExtendedProtocolSession.METHOD_REPLY )
	    {
	    case DLProtocolSession.METHOD_START:
		this.downloadComplete(xdr.BOOLEAN());
		break;
	    case DLProtocolSession.METHOD_BOOT_CAPSULE:
		if( xdr.BOOLEAN() )
		{
		    ProtocolPresentation pp = null;
		    ExternalCoroutine eco;
		    
		    switch( (this.state = xdr.INT()) )
		    {
		    case DLProtocolSession.STATE_FAILURE:
			break;
		    case DLProtocolSession.STATE_BASIC:
			pp = this.createBasicPresentation();
			break;
		    case DLProtocolSession.STATE_EXTENDED:
			pp = this.createExtendedPresentation();
			break;
		    }
		    if( pp != null )
		    {
			ProtocolType pt;

			pt = new ProtocolType().
			    setProtocolID(xdr.TYPEID());
			this.userSession.setProtocolType(pt).
			    snapshotSessionType();
			eco = new ExternalCoroutine();
			eco.connect(pp,
				    super.getCoroutine().getChannel(),
				    this.userSession);
			/* XXX bleh, don't wait here. */
			if( eco.waitForConnection() )
			{
			    pp.setCoroutine(eco);
			}
			else
			{
			    pp = null;
			}
		    }
		    this.bootComplete(pp);
		}
		else
		{
		    this.bootComplete(null);
		}
		this.notifyAll();
		break;
	    default:
		Strict.fail();
		break;
	    }
	}
	else
	{
	    Strict.fail();
	}
    }

    public String toString()
    {
	return "DLProtocolPresentation["
	    + super.toString()
	    + "; state="
	    + this.state
	    + "]";
    }
}
