/*
 * DLProtocolProgress.java
 *
 * Copyright (c) 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is part of the Janos Bees Software.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */

package bees.core;

import java.io.IOException;

import java.util.Vector;
import java.util.Hashtable;
import java.util.Enumeration;

import bees.core.resource.Logger;

/**
 * A DLProtocolProgress is used to track the progress of code downloading
 * between nodes.
 */
final class DLProtocolProgress
{
    static final String PREVIOUS_NEIGHBOR = "tmp/previousNeighbor";
    
    /**
     * Mapping of FlowCode identifiers to FlowCodeTypes.  The mappings are
     * progressively removed as the byte code for each FlowByteArrayType is
     * received.
     */
    private final Hashtable code = new Hashtable();

    /**
     * The neighbor where we can find the code.
     */
    private Neighbor previousNeighbor;

    /**
     * Construct a DLProtocolProgress object that will be used to download
     * the given FlowType.
     *
     * @param ft The FlowType being downloaded.
     */
    DLProtocolProgress(FlowType ft)
    {
	Assert.require(ft);
	
	this.absorb(ft, ft.getPrimaryProtocolType());
    }

    /**
     * @return The neighbor where we can find the code.
     */
    Neighbor getPreviousNeighbor()
    {
	return this.previousNeighbor;
    }

    /**
     * Walk the ProtocolType tree in a FlowType.
     *
     * @param ft The FlowType being downloaded.
     * @param pt The sub-ProtocolType that contains CodeTypes to be downloaded.
     */
    private void absorb(FlowType ft, ProtocolType pt)
    {
	int lpc, size;
	Vector cts;

	Assert.require(ft);
	Assert.require(pt);

	cts = pt.getByteArrayTypes();
	size = cts.size();
	for( lpc = 0; lpc < size; lpc++ )
	{
	    ByteArrayType bat = (ByteArrayType)cts.elementAt(lpc);

	    /* Only need to download non-builtin CodeTypes. */
	    if( bat.getState() == ByteArrayType.STATE_NEW )
	    {
		FlowByteArrayType fbat;
		
		fbat = FlowByteArrayType.
		    getBDomainFlowByteArray(ft.getFlowID(), bat.getByteArrayID());
		if( fbat == null )
		    throw new Panic("Couldn't find: " + bat);
		this.code.put(fbat.getFlowByteArrayID(), fbat);
	    }
	}
	cts = pt.getCompanions();
	size = cts.size();
	for( lpc = 0; lpc < size; lpc++ )
	{
	    pt = (ProtocolType)cts.elementAt(lpc);
	    this.absorb(ft, pt);
	}
    }

    synchronized void setPreviousNeighbor(Node node, ProtocolSession ps)
	throws IOException
    {
	if( this.previousNeighbor == null )
	{
	    NeighborImpl ni;
	
	    ni = (NeighborImpl)node.getSystemCapability(PREVIOUS_NEIGHBOR);
	    this.previousNeighbor = new Neighbor(ni);
	    this.previousNeighbor.attachReceivers(ps);
	}
    }
    
    /**
     * Send ByteArrayRequestCapsules to the previous neighbor.
     *
     * @param ec The ExternalCapsule used to authorize the creation of
     * CodeRequestCapsules or null if cap is given.
     * @param cap A Capsule used to authorize the creation of
     * CodeRequestCapsules or null if ec is given.
     */
    synchronized void requestCode(ExternalCapsule ec, Capsule cap)
    {
	Enumeration enum;
	
	enum = this.code.keys();
	if( enum.hasMoreElements() )
	{
	    ByteArrayRequestCapsule crc;
	    TypeID fcid;
	    
	    fcid = (TypeID)enum.nextElement();
	    crc = new ByteArrayRequestCapsule(fcid);
	    if( ec != null )
	    {
		crc.fromExternal(ec, fcid);
	    }
	    else
	    {
		crc.fromCapsule(cap, fcid);
	    }
	    crc.getSession().getCache().
		put(fcid, crc, ReliableDataRequestCapsule.
		    DEFAULT_RETRY_INTERVAL);
	    try
	    {
		this.previousNeighbor.send(crc);
	    }
	    catch(EncodingException e)
	    {
		Logger.log(e);
	    }
	    catch(IOException e)
	    {
		Logger.log(e);
	    }
	}
    }

    /**
     * @return True if all the code has been downloaded.
     */
    synchronized boolean isComplete()
    {
	return (this.code.size() == 0);
    }

    /**
     * Signal the receipt of all the byte code for a FlowByteArrayType.
     *
     * @param cap A Capsule used to authorize the creation of
     * CodeRequestCapsules or null if ec is given.
     * @param flowCodeID The FlowByteArrayType identifier that has been
     * completed.
     * @param byteCode The byte code for the FlowByteArrayType.
     * @return True if all of the FlowCodeTypes have been received.
     */
    boolean completedCode(Capsule cap, TypeID flowCodeID, byte byteCode[])
    {
	FlowByteArrayType fct = null;
	boolean retval = false;

	Assert.require(flowCodeID);
	Assert.require(byteCode);

	synchronized( this )
	{
	    fct = (FlowByteArrayType)this.code.remove(flowCodeID);
	    retval = this.isComplete();
	}
	if( fct != null )
	{
	    ByteArrayType bat;
	    
	    bat = fct.getByteArrayType();
	    if( bat instanceof CodeType )
	    {
	    	CodeType ct;
	    	
	    	ct = (CodeType)bat;
	    	ct.importClass(byteCode);
	    }
	    else
	    {
		bat.importByteArray(byteCode);
	    }
	}
	requestCode(null, cap);
	return retval;
    }

    public String toString()
    {
	return "DLProtocolProgress[code="
	    + this.code
	    + "; previousNeighbor="
	    + this.previousNeighbor
	    + "]";
    }
}
