/*
 * DataAggregator.java
 *
 * Copyright (c) 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is part of the Janos Bees Software.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */

package bees.core;

import bees.core.resource.Logger;

/**
 * A DataAggregator is used to collect and redistribute byte arrays.
 *
 * @see DataRequestCapsule
 * @see DataResponseCapsule
 * @see ReliableDataRequestCapsule
 */
public final class DataAggregator
{
    /**
     * The size of a window, where each element is a "chunk".
     */
    static final int WINDOW_SIZE = 64;

    /**
     * The number of bytes in a chunk.
     */
    static final int CHUNK_SIZE = 64;

    /**
     * The byte array used to collect data.
     */
    private byte bytes[];

    /**
     * The current window into the byte array.
     */
    private int currentWindow;

    /**
     * The state of the current window, a set bit represents the receipt of
     * that data chunk.
     */
    private long windowChunks;

    /**
     * Construct a DataAggregator with the given byte array.
     *
     * @param bytes The initial byte array.
     */
    public DataAggregator(byte bytes[])
    {
	this.bytes = bytes;
    }

    /**
     * Construct a DataAggregator that will receive a byte array of the given
     * length.
     *
     * @param length The expected length of the byte array.
     */
    public DataAggregator(int length)
    {
	this(new byte[length]);
    }

    /**
     * @return The byte array used to collect data.
     */
    public byte[] getBytes()
    {
	return this.bytes;
    }

    /**
     * @return The expected length of the byte array.
     */
    public int getLength()
    {
	return this.bytes.length;
    }

    /**
     * @param cw The current window into the byte array.
     */
    public void setCurrentWindow(int cw)
    {
	this.currentWindow = cw;
    }
    
    /**
     * @return The current window into the byte array.
     */
    public int getCurrentWindow()
    {
	return this.currentWindow;
    }

    /**
     * @param windowChunks The state of the current window.
     */
    public void setWindowChunks(long windowChunks)
    {
	this.windowChunks = windowChunks;
    }

    /**
     * @return The state of the current window.
     */
    public long getWindowChunks()
    {
	return this.windowChunks;
    }

    /**
     * Fill in the enclosed byte array with the contents of the given window.
     *
     * @param buffer The buffer holding the window data.
     * @param fromWindow The window where the data was collected from.
     * @param fromChunks The set of chunks the data in "buffer" represents.
     * @return True if byte array is now complete.
     */
    public synchronized boolean fromWindow(byte buffer[],
					   int fromWindow,
					   long fromChunks)
    {
	Assert.require(buffer != null);
	
	if( this.currentWindow == fromWindow )
	{
	    int bufferCursor, windowCursor, lpc;

	    bufferCursor = 0;
	    windowCursor = (this.currentWindow * WINDOW_SIZE * CHUNK_SIZE);
	    for( lpc = 0;
		 (bufferCursor < buffer.length) && (lpc < WINDOW_SIZE);
		 lpc++ )
	    {
		if( (((1L << lpc) & fromChunks) != 0) &&
		    (((1L << lpc) & this.windowChunks) == 0) )
		{
		    int chunkSize = CHUNK_SIZE;

		    if( (windowCursor + CHUNK_SIZE) > this.bytes.length )
		    {
			chunkSize = this.bytes.length - windowCursor;
			if( chunkSize == 0 )
			{
			    break;
			}
		    }
		    System.arraycopy(buffer, bufferCursor,
				     this.bytes, windowCursor,
				     chunkSize);
		    bufferCursor += CHUNK_SIZE;
		}
		windowCursor += CHUNK_SIZE;
	    }
	    this.windowChunks |= fromChunks;
	    if( this.windowChunks == ~0 )
	    {
		this.currentWindow++;
		this.windowChunks = 0;
	    }
	}
	else
	{
	    Logger.log("Received chunks that are outside the current "
		       + "window. [fromWindow="
		       + fromWindow
		       + "; fromChunks="
		       + Long.toHexString(fromChunks)
		       + "; this.bytes.length="
		       + this.bytes.length
		       + "; this.currentWindow="
		       + this.currentWindow
		       + "; this.windowChunks="
		       + Long.toHexString(this.windowChunks)
		       + "]");
	}
	return this.isComplete();
    }

    /**
     * @return True if the byte array is now completed.
     */
    public synchronized boolean isComplete()
    {
	return (this.currentWindow * WINDOW_SIZE * CHUNK_SIZE) >=
	    this.bytes.length;
    }

    /**
     * @see java.lang.Object#toString()
     */
    public String toString()
    {
	return "DataAggregator[bytes="
	    + ByteArray.toString(this.bytes)
	    + "; currentWindow="
	    + this.currentWindow
	    + "; windowChunks=0x"
	    + Long.toHexString(this.windowChunks)
	    + "]";
    }

    /**
     * Limit the byte size of a set of chunks.
     *
     * @param windowChunks The desired set of chunks to be transmitted.
     * @param windowLength ...
     * @param maxSize The maximum number of bytes that can be transmitted.
     * @return The subset of "windowChunks" that doesn't violate the maximum
     * size value.
     */
    public static long limitWindow(long windowChunks,
				   int windowLength,
				   int maxSize)
    {
	long retval = 0;
	int size, lpc;

	for( size = 0, lpc = 0;
	     (lpc < WINDOW_SIZE) && (size < maxSize);
	     lpc++ )
	{
	    if( size >= windowLength )
	    {
		retval |= (1L << lpc);
	    }
	    else if( ((1L << lpc) & windowChunks) != 0 )
	    {
		retval |= (1L << lpc);
		size += CHUNK_SIZE;
	    }
	}
	return retval;
    }

    /**
     * Produce a byte array that holds the bytes selected by the given values.
     *
     * @param bytes The source bytes.
     * @param currentWindow The window within "bytes" from which data is to
     * be pulled.
     * @param windowChunks The set of chunks to be placed in the returned
     * array.
     * @return A new byte array containing the selected chunks from "bytes".
     */
    public static byte[] fillWindow(byte bytes[],
				    int currentWindow,
				    long windowChunks)
    {
	int bufCursor = 0, windowCursor, lpc;
	byte retval[];

	windowCursor = (currentWindow * WINDOW_SIZE * CHUNK_SIZE);
	if( (bytes.length - windowCursor) < (WINDOW_SIZE * CHUNK_SIZE) )
	{
	    retval = new byte[bytes.length - windowCursor];
	}
	else
	{
	    retval = new byte[WINDOW_SIZE * CHUNK_SIZE];
	}
	for( lpc = 0;
	     (windowCursor < bytes.length) && (lpc < WINDOW_SIZE);
	     lpc++ )
	{
	    if( ((1L << lpc) & windowChunks) != 0 )
	    {
		int chunkSize = CHUNK_SIZE;

		if( (windowCursor + CHUNK_SIZE) > bytes.length )
		{
		    chunkSize = bytes.length - windowCursor;
		    if( chunkSize == 0 )
		    {
			break;
		    }
		}
		System.arraycopy(bytes, windowCursor,
				 retval, bufCursor,
				 chunkSize);
		bufCursor += CHUNK_SIZE;
	    }
	    windowCursor += CHUNK_SIZE;
	}
	return retval;
    }
}
