/*
 * EventChannel.java
 *
 * Copyright (c) 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is part of the Janos Bees Software.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */

package bees.core;

import edu.utah.janosvm.sys.DeadTeamException;
import edu.utah.janosvm.sys.TeamEjectionException;

import edu.utah.janosvm.kit.comm.amiga.Facsimile;
import edu.utah.janosvm.kit.comm.amiga.FaxPortHandle;

import edu.utah.janosvm.kit.comm.jsi.JSIPair;
import edu.utah.janosvm.kit.comm.jsi.JSIHandler;
import edu.utah.janosvm.kit.comm.jsi.JSIMessage;
import edu.utah.janosvm.kit.comm.jsi.JSIHandlerRoot;
import edu.utah.janosvm.kit.comm.jsi.JSIPropertyInfo;
import edu.utah.janosvm.kit.comm.jsi.JSIHandlerContext;
import edu.utah.janosvm.kit.comm.jsi.JSIEncodingException;
import edu.utah.janosvm.kit.comm.jsi.MissingPairException;

import edu.utah.janos.nodeos.Domain;
import edu.utah.janos.nodeos.ThreadLocal;

import bees.core.kernel.KernelPanic;
import bees.core.kernel.JSICoroutine;
import bees.core.kernel.ConnectBDomainCoroutine;

import bees.core.security.Permission;
import bees.core.security.PermissionSet;

import bees.core.bdomain.BDomain;
import bees.core.bdomain.BDomainLocal;
import bees.core.bdomain.DeadBDomainException;

/**
 * An EventChannel is used to send events to another domain in the system.
 */
public final class EventChannel
    extends Capability
    implements JSIHandler
{
    public static final JSIHandlerRoot sub = new JSIHandlerRoot();

    static {
	Domain domain = Domain.currentDomain();
	String classes[] = {
	    "bees.core.InheritedCapabilityCoroutine",
	    "bees.core.kernel.JSICoroutine",
	};
	int lpc;

	for( lpc = 0; lpc < classes.length; lpc++ )
	{
	    try
	    {
		domain.loadClass(classes[lpc]);
	    }
	    catch(ClassNotFoundException e)
	    {
		e.printStackTrace();
	    }
	}
    }

    private static final BDomainLocal domainChannel = new BDomainLocal();

    public static void setCurrentChannel()
    {
	EventChannel ec = new EventChannel();
	
	try
	{
	    CapabilityFactory.create(ec,
				     BDomain.currentBDomain().
				     getBDomainCoroutine().getStoreID(),
				     DEFAULT_PERMISSIONS);
	    domainChannel.set(ec);
	}
	catch(MissingCapabilityException e)
	{
	    throw new Panic(e);
	}
	catch(InstantiationException e)
	{
	    throw new Panic(e);
	}
    }

    public static EventChannel currentChannel()
    {
	return (EventChannel)domainChannel.get();
    }

    private static final ThreadLocal channel = new ThreadLocal();

    public static final EventChannel NULL_CHANNEL = new EventChannel();

    public static EventChannel threadChannel()
    {
	EventChannel retval;

	retval = (EventChannel)channel.get();
	channel.set(NULL_CHANNEL);
	return retval;
    }
    
    private static final Permission SEND_PERMISSION =
	new Permission("channel", "write");

    public static final PermissionSet DEFAULT_PERMISSIONS =
	new PermissionSet(new Permission[] {
	    new Permission("channel", "write")
	});

    private static final int STATE_NEW_BORN = 0;
    private static final int STATE_IN_PROGRESS = 1;
    private static final int STATE_FOUND = 2;
    private static final int STATE_DEAD = 3;
    private static final int STATE_ERROR = 4;

    /**
     * The current state, used during initialization.
     */
    private int state;

    /**
     * The FaxPortHandle connected to a FaxPort in a remote domain.
     */
    private FaxPortHandle port;

    /**
     * Construct an empty EventChannel.
     */
    public EventChannel()
    {
    }

    /**
     * Construct an EventChannel with the given value.
     *
     * @param port A connected FaxPortHandle.
     */
    EventChannel(FaxPortHandle port)
    {
	this.state = STATE_FOUND;
	this.port = port;
    }

    /**
     * A simple StateResponder used when connecting the FaxPortHandle.
     */
    final class Responder
	extends StateResponder
    {
	Responder()
	{
	    super(Coroutine.STATE_COMPLETE);
	}

	public void transition(Coroutine coroutine, int newState)
	{
	    switch( newState )
	    {
	    case Coroutine.STATE_COMPLETE:
		synchronized( EventChannel.this )
		{
		    if( EventChannel.this.state == STATE_IN_PROGRESS )
		    {
			EventChannel.this.port =
			    ((ConnectBDomainCoroutine)coroutine).getPort();
			EventChannel.this.state = STATE_FOUND;
			EventChannel.this.notifyAll();
		    }
		}
		break;
	    default:
		synchronized( EventChannel.this )
		{
		    coroutine.getThrowable().printStackTrace();
		    EventChannel.this.state = STATE_ERROR;
		    EventChannel.this.revoke();
		}
		break;
	    }
	}

	public String toString()
	{
	    return "EventChannel$Responder["
		+ super.toString()
		+ "]";
	}
    }

    protected synchronized final void fromBytes(byte flattened[])
	throws Exception
    {
	BDomain bdomain = BDomain.currentBDomain();
	ConnectBDomainCoroutine csc;
	
	super.fromBytes(flattened);

	if( this.getID() == bdomain.getBDomainCoroutine().getStoreID() )
	{
	    this.state = STATE_FOUND;
	}
	else
	{
	    csc = new ConnectBDomainCoroutine(new Responder());
	    csc.connect(bdomain.getBDomainCoroutine(), flattened);
	    this.state = STATE_IN_PROGRESS;
	}
    }

    public synchronized void revoke()
    {
	super.clear();

	this.state = STATE_ERROR;
	this.port = null;
	this.notifyAll();
    }

    public synchronized void waitForDeath()
    {
	while( (this.state != STATE_DEAD) && (this.state != STATE_ERROR) )
	{
	    try
	    {
		this.wait();
	    }
	    catch(InterruptedException e)
	    {
	    }
	}
    }

    public synchronized void sendEvent(BeesEvent ae)
	throws DeadBDomainException
    {
	Assert.require(ae);

	this.checkPermission(SEND_PERMISSION);

	while( this.state == STATE_IN_PROGRESS )
	{
	    try
	    {
		this.wait();
	    }
	    catch(InterruptedException e)
	    {
	    }
	}
	try
	{
	    this.port.putFacsimile(ae);
	}
	catch(DeadTeamException e)
	{
	    throw new DeadBDomainException();
	}
	catch(TeamEjectionException e)
	{
	    throw new DeadBDomainException();
	}
    }

    public JSIHandler resolveSpecifier(JSIMessage jm, JSIPair jp)
	throws Throwable
    {
	JSIHandler retval;

	channel.set(this);
	retval = sub.resolveSpecifier(jm, jp);
	return retval;
    }

    public void handleMessage(JSIHandlerContext jhc, JSIMessage jm)
    {
	switch( jm.getWhat() )
	{
	case JSIMessage.GET_PROPERTY:
	    break;
	case JSIMessage.LIST_PROPERTIES:
	    sub.handleMessage(jhc, jm);
	    break;
	default:
	    throw new UnsupportedOperationException(
		"What: " +
		JSIMessage.MESSAGE_TYPE_STRINGS[jm.getWhat()]);
	}
    }

    public void getSuites(JSIMessage jm)
    {
    }

    public boolean equals(Object obj)
    {
	boolean retval = false;

	if( this == obj )
	{
	    retval = true;
	}
	else if( obj instanceof EventChannel )
	{
	    EventChannel ec = (EventChannel)obj;

	    if( this.port.equals(ec.port) )
		retval = true;
	}
	return retval;
    }

    public String toString()
    {
	return "EventChannel["
	    + super.toString()
	    + "; port="
	    + this.port
	    + "]";
    }
}
