/*
 * ExternalCapsule.java
 *
 * Copyright (c) 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is part of the Janos Bees Software.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */

package bees.core;

import edu.utah.janos.nodeos.Address;
import edu.utah.janos.nodeos.BufferHandle;

import bees.core.bdomain.BDomain;

/**
 * An ExternalCapsule is a class used to carry packets between protocols and
 * applications.  Unlike the Capsule class, an ExternalCapsule doesn't have any
 * typing applied to it, the body of the capsule is a raw byte array.  In
 * addition, a set of CapabilityEnvelopes and CapabilityReferences can be
 * attached to the ExternalCapsule.  These functionality can be used to
 * dynamically transmit new capabilities and reference existing ones.
 */
public final class ExternalCapsule
{
    /**
     * The Xdr object used to marshal the body.
     */
    private final Xdr xdr = new Xdr();

    /**
     * The source of the capsule.
     */
    private ExternalChannel source;

    /**
     * The SessionType that classifies the capsule for a protocol flow.
     */
    private SessionType st;

    /**
     * A set of CapabilityEnvelopes to pass to the remote domain.
     */
    private CapabilityEnvelope ce[];
    
    /**
     * A set of CapabilityReferences to pass to the remote domain.
     */
    private CapabilityReference cr[];
    
    /**
     * The BufferHandle containing the body of the message.
     */
    private BufferHandle bh;

    private boolean used;

    /**
     * Construct an ExternalCapsule with the given values.
     *
     * @param ce The set of CapabilityEnvelopes to pass to the remote domain.
     * @param cr The set of CapabilityReferences to pass to the remote domain.
     */
    public ExternalCapsule(CapabilityEnvelope ce[], CapabilityReference cr[])
    {
	this.ce = ce;
	this.cr = cr;
    }

    /**
     * Construct an ExternalCapsule with the given values.
     *
     * @param ce The set of CapabilityEnvelopes to pass to the remote domain.
     */
    public ExternalCapsule(CapabilityEnvelope ce[])
    {
	this(ce, null);
    }

    /**
     * Construct an ExternalCapsule with the given values.
     *
     * @param cr The set of CapabilityReferences to pass to the remote domain.
     */
    public ExternalCapsule(CapabilityReference cr[])
    {
	this(null, cr);
    }

    /**
     * Construct an empty ExternalCapsule.
     */
    public ExternalCapsule()
    {
	this(null, null);
    }

    /**
     * @return The source of this capsule.
     */
    public ExternalChannel getSource()
    {
	return this.source;
    }

    /**
     * @param source The source of this capsule.
     */
    void setSource(ExternalChannel source)
    {
	this.source = source;
    }

    SessionType getSessionType()
    {
	return this.st;
    }

    void setSessionType(SessionType st)
    {
	this.st = st;
    }

    void consume()
    {
	if( this.used )
	{
	    throw new SecurityException();
	}
	this.used = true;
    }

    public ProtocolSession getSession()
    {
	try
	{
	    return this.st.getSession();
	}
	catch(Throwable th)
	{
	    throw new Panic(th);
	}
    }

    /**
     * @return The Xdr object used to marshal the body.
     */
    public synchronized Xdr getXDR()
    {
	return this.xdr;
    }

    /**
     * @return The BufferHandle containing the body of the message.
     */
    synchronized BufferHandle getBuffer()
    {
	if( this.bh == null )
	{
	    this.bh = this.xdr.getBuffer();
	}
	return this.bh;
    }

    /**
     * @return A set of CapabilityEnvelopes to pass to the remote domain.
     */
    public synchronized CapabilityEnvelope[] getEnvelopes()
    {
	return this.ce;
    }

    /**
     * @return A set of CapabilityReferencess to pass to the remote domain.
     */
    public synchronized CapabilityReference[] getReferences()
    {
	return this.cr;
    }
    
    void fromEvent(ExternalCoroutine source, ExternalCapsuleEvent ece)
    {
	Assert.ensure(source);
	
	this.source = source;
	this.st = source.getSessionType();
	this.xdr.readReset(ece.getBody());
    }

    /**
     * Initialize a new born capsule.
     *
     * @return this
     */
    public ExternalCapsule init()
    {
	this.xdr.writeReset(BufferHandle.create());
	return this;
    }

    public String toString()
    {
	return "ExternalCapsule[xdr="
	    + this.xdr
	    + "; source="
	    + this.source
	    + "; st="
	    + this.st
	    + "; ce="
	    + this.ce
	    + "; cr="
	    + this.cr
	    + "; bh="
	    + this.bh
	    + "]";
    }
}
