/*
 * Neighbor.java
 *
 * Copyright (c) 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is part of the Janos Bees Software.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */

package bees.core;

import java.util.Vector;
import java.util.Hashtable;
import java.util.Enumeration;

import java.io.IOException;

import java.security.InvalidKeyException;
import java.security.NoSuchAlgorithmException;

import bees.core.security.SecretKey;

/**
 * A Neighbor is used to "activate" a NeighborImpl for a protocol flow.  In
 * this context, "activation" involves attaching handlers to receive and
 * send capsules.
 *
 * @see bees.core.NeighborImpl
 * @see bees.core.Neighborhood
 */
public class Neighbor
{
    /**
     * Object used to encode capsules for transmission.
     */
    private final CapsuleEncoder ce = new CapsuleEncoder();

    /**
     * The current Node object for this domain.
     */
    private final Node node;

    /**
     * The neighbor implementation that this neighor uses to send and
     * receive capsules.
     */
    private final NeighborImpl ni;

    /**
     * Construct a Neighbor that uses the given capability for communication.
     *
     * @param ni The neighbor capability to use for communication.
     */
    public Neighbor(NeighborImpl ni)
    {
	this.node = Node.currentNode();
	this.ni = ni;
    }

    /**
     * @return The neighbor implementation that this neighor uses to send and
     * receive capsules.
     */
    public final NeighborImpl getNeighborImpl()
    {
	return this.ni;
    }

    /**
     * Enqueue a key for checking hop-by-hop integrity.
     *
     * @param sk The SecretKey to use when computing MACs.
     * @param algorithm The MAC algorithm to use with the key.
     * @throws NoSuchAlgorithmException If the algorithm value is unknown.
     * @throws InvalidKeyException If the given key is invalid.
     */
    private void internalEnqueueKey(SecretKey sk, String algorithm)
	throws NoSuchAlgorithmException,
	       InvalidKeyException
    {
	this.ni.enqueueKey(bees.core.security.Leak.get(sk), algorithm);
    }

    /**
     * Enqueue a key for checking hop-by-hop integrity.
     *
     * @param ce An envelope containing the secret key to use when computing
     * MACs.
     * @param algorithm The MAC algorithm to use with the key.
     * @throws MissingCapabilityException If the capability has since been
     * deleted.
     * @throws InstantiationException If there was an error constructing the
     * SecretKey capability.
     * @throws NoSuchAlgorithmException If the algorithm value is unknown.
     * @throws InvalidKeyException If the given key is invalid.
     */
    public void enqueueKey(CapabilityEnvelope ce, String algorithm)
	throws MissingCapabilityException,
	       NoSuchAlgorithmException,
	       InstantiationException,
	       InvalidKeyException
    {
	this.internalEnqueueKey((SecretKey)ce.unseal(), algorithm);
    }

    /**
     * Enqueue a key for checking hop-by-hop integrity.
     *
     * @param sk The SecretKey to use when computing MACs.
     * @param algorithm The MAC algorithm to use with the key.
     * @throws NoSuchAlgorithmException If the algorithm value is unknown.
     * @throws InvalidKeyException If the given key is invalid.
     */
    public void enqueueKey(SecretKey sk, String algorithm)
	throws NoSuchAlgorithmException,
	       InvalidKeyException
    {
	this.internalEnqueueKey(sk, algorithm);
    }

    /**
     * Dequeue a key previously enqueued.  This method has the effect of
     * invalidating any capsules that have a MAC based on this key.
     */
    public void dequeueKey()
    {
	this.ni.dequeueKey();
    }

    /**
     * Attach the receiver hooks for a set of capsule types.
     *
     * XXX Not sure about this method, should it be called by the constructor?
     *
     * @param ps A ProtocolSession whose capsule types should have receivers
     * attached.
     */
    public final void attachReceivers(ProtocolSession ps)
	throws IOException
    {
	Hashtable ht;

	ht = CapsulePool.getBDomainCapsulePools();
	// XXX I don't think this sync is necessary
	synchronized( ht )
	{
	    Enumeration elem;

	    elem = ht.elements();
	    synchronized( this.ni )
	    {
		while( elem.hasMoreElements() )
		{
		    CapsulePool cp;
		    CapsuleType ct;
		    SessionType st;
		    
		    /* XXX gross. */
		    cp = (CapsulePool)elem.nextElement();
		    ct = cp.getCapsuleType();
		    st = SessionType.getBDomainSession(
			ct.getFlowType().getFlowID(),
			ct.getProtocolType().getProtocolID());
		    if( st == ps.getSessionType() )
		    {
			this.ni.receiveBuffer(new CapsuleReceiver(this,
								  cp,
								  ps));
		    }
		}
		this.ni.putNeighbor(ps.getSessionType(), this);
	    }
	}
    }

    /**
     * Detach the receiver hooks for a set of capsule types.
     *
     * @param ps A ProtocolSession whose capsule types should have receivers
     * detached.
     */
    public final void detachReceivers(ProtocolSession ps)
    {
	Hashtable ht;

	ht = CapsulePool.getBDomainCapsulePools();
	// XXX I don't think this sync is necessary
	synchronized( ht )
	{
	    Enumeration elem;

	    elem = ht.elements();
	    synchronized( ni )
	    {
		while( elem.hasMoreElements() )
		{
		    CapsulePool cp;
		    CapsuleType ct;
		    SessionType st;
		    
		    /* XXX gross. */
		    cp = (CapsulePool)elem.nextElement();
		    ct = cp.getCapsuleType();
		    st = SessionType.getBDomainSession(
			ct.getFlowType().getFlowID(),
			ct.getProtocolType().getProtocolID());
		    if( st == ps.getSessionType() )
		    {
			this.ni.cancelReceivers(this, ct.getCapsuleID());
		    }
		}
		this.ni.removeNeighbor(ps.getSessionType());
	    }
	}
    }

    /**
     * Send a Capsule to this neighbor.
     *
     * @param cap The Capsule to send.
     * @exception Throws an EncodingException if there was an error while
     * marshalling the capsule.
     * @exception Throws an IOException if there was an error during
     * transmission.
     */
    public synchronized void send(Capsule cap)
	throws EncodingException,
	       IOException
    {
	Neighbor source;

	source = cap.getSource();
	if( source == null )
	{
	    cap.pa.consume(PacketAuthorization.VALUE_INITIAL);
	}
	else if( source == this )
	{
	    cap.pa.consume(PacketAuthorization.VALUE_RETURNS);
	}
	else
	{
	    cap.pa.consume(PacketAuthorization.VALUE_FORWARDS);
	}
	this.node.used();
	this.ni.sendBuffer(this.ce.encode(cap));
    }

    /**
     * Send a group of capsule to this neighbor.
     *
     * @param cg The CapsuleGroup to send.
     * @exception Throws an EncodingException if there was an error while
     * marshalling the capsule(s).
     * @exception Throws an IOException if there was an error during
     * transmission.
     */
    public void send(CapsuleGroup cg)
	throws EncodingException,
	       IOException
    {
	synchronized( cg )
	{
	    Vector group = cg.getCapsules();
	    int lpc, size;

	    size = group.size();
	    for( lpc = 0; lpc < size; lpc++ )
	    {
		this.send((Capsule)group.elementAt(lpc));
	    }
	}
    }

    /**
     * @see java.lang.Object#toString()
     */
    public String toString()
    {
	return "Neighbor[ce="
	    + this.ce
	    + "; ni="
	    + this.ni
	    + "]";
    }
}
