/*
 * Neighborhood.java
 *
 * Copyright (c) 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is part of the Janos Bees Software.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */

package bees.core;

import java.util.Vector;

import java.io.IOException;

import bees.core.resource.Logger;

import bees.core.security.PermissionException;

/**
 * A Neighborhood is a collection of neighbors that a single capsule can be
 * sent to.
 */
public abstract class Neighborhood
{
    /**
     * The current node.
     */
    protected final Node node;
    
    /**
     * The set of neighbors to send capsules to.
     */
    protected final Vector neighbors = new Vector();

    /**
     * The CapsuleEncoder used to marshal capsules.
     */
    protected final CapsuleEncoder ce = new CapsuleEncoder();

    /**
     * Construct an empty Neighborhood.
     */
    public Neighborhood()
    {
	this.node = Node.currentNode();
    }

    /**
     * @param neighbor The neighbor to add to the neighborhood.
     */
    public synchronized void addNeighbor(NeighborImpl neighbor)
    {
	if( !this.neighbors.contains(neighbor) )
	{
	    this.neighbors.addElement(neighbor);
	}
    }

    /**
     * @param neighbor The neighbor to remove from the neighborhood.
     */
    public synchronized void removeNeighbor(NeighborImpl neighbor)
    {
	this.neighbors.removeElement(neighbor);
    }

    /**
     * @return An array containing all the NeighborImpl's in the neighborhood.
     */
    public synchronized NeighborImpl[] elements()
    {
	NeighborImpl retval[];
	int lpc, size;

	size = this.neighbors.size();
	retval = new NeighborImpl[size];
	this.neighbors.copyInto(retval);
	return retval;
    }

    /**
     * @return The number of neighbors in the neighborhood.
     */
    public synchronized int size()
    {
	return this.neighbors.size();
    }

    /**
     * @param cap The capsule to send to the elements of the Neighborhood.
     * @throws EncodingException if there was a problem encoding the Capsule.
     * @throws IOException if there was a problem sending the Capsule.
     */
    public abstract void send(Capsule cap)
	throws EncodingException, IOException;

    /**
     * Call back used to handle a revoked neighbor.
     *
     * @param e The received exception.
     */
    public void handleRevokedNeighbor(NeighborImpl ni)
    {
    }
    
    /**
     * Call back used to handle IOException during a send.  The default simply
     * prints out the exception and moves on.
     *
     * @param e The received exception.
     */
    public void handleSendException(IOException e)
    {
	Logger.log(e);
    }
    
    /**
     * Call back used to handle a PermissionException during a send.  The
     * default simply prints out the exception and moves on.
     *
     * @param e The received exception.
     */
    public void handleSendException(PermissionException e)
    {
	Logger.log(e);
    }
    
    /**
     * @see java.lang.Object#toString()
     */
    public String toString()
    {
	return "Neighborhood[neighbors="
	    + this.neighbors
	    + "; ce="
	    + this.ce
	    + "]";
    }
}
