/*
 * Node.java
 *
 * Copyright (c) 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is part of the Janos Bees Software.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */

package bees.core;

import java.util.Vector;
import java.util.Hashtable;
import java.util.Enumeration;
import java.util.NoSuchElementException;

import edu.utah.janosvm.kit.comm.jsi.JSIPair;
import edu.utah.janosvm.kit.comm.jsi.JSIHandler;
import edu.utah.janosvm.kit.comm.jsi.JSIMessage;
import edu.utah.janosvm.kit.comm.jsi.JSIPropertyInfo;
import edu.utah.janosvm.kit.comm.jsi.JSIHandlerContext;
import edu.utah.janosvm.kit.comm.jsi.MissingPairException;

import edu.utah.janos.nodeos.Domain;

import bees.core.kernel.BeesKernel;
import bees.core.kernel.JSICoroutine;
import bees.core.kernel.KernelObjectStore;
import bees.core.kernel.CreateBDomainCoroutine;

import bees.core.resource.Clock;
import bees.core.resource.Logger;

import bees.core.security.Permission;
import bees.core.security.PermissionSet;
import bees.core.security.IntegerPermission;
import bees.core.security.PermissionException;

import bees.core.bdomain.BDomain;
import bees.core.bdomain.BDomainLocal;
import bees.core.bdomain.CapabilityTree;
import bees.core.bdomain.DeadBDomainException;
import bees.core.bdomain.BDomainCreationException;

/**
 * The Node class provides access to node global functionality.
 */
public final class Node
    extends NodeSpec
    implements NodeCache.EventHandler,
	       JSIHandler
{
    /**
     * JSI Property info.
     */
    private static final JSIPropertyInfo JSI_SUB_PROPERTIES[] = {
	new JSIPropertyInfo("name",
			    (1 << JSIMessage.GET_PROPERTY),
			    "Get the name of this node.\n"),
	new JSIPropertyInfo("version",
			    (1 << JSIMessage.GET_PROPERTY),
			    "Get the version of this node.\n"),
    };
    
    /**
     * Permission for reading the node name.
     */
    public static final Permission READ_NODE_NAME =
	new Permission("name", "read");

    /**
     * Permission for reading the node version info.
     */
    public static final Permission READ_NODE_VERSION =
	new Permission("version", "read");

    /**
     * Permission for starting a thread.
     */
    public static final Permission THREAD_START =
	new Permission("threadStart", "execute");

    /**
     * Permission for registering a protocol.
     */
    public static final Permission PROTOCOL_REGISTER =
	new Permission("protocolRegister", "execute");

    /**
     * Permission to register a JSI handler.
     */
    public static final Permission JSI_REGISTER =
	new Permission("jsiRegister", "execute");

    /**
     * The default permissions for a new Node object.
     */
    public static final PermissionSet DEFAULT_PERMISSIONS =
	new PermissionSet(new Permission[] {
	    new Permission("name", "read"),
	    new Permission("version", "read"),
	    new Permission("jsiRegister", "execute"),
	    new Permission("threadStart", ""),
	    new Permission("protocolRegister", ""),
	});

    /**
     * The Node object for the current domain.
     */
    private static final BDomainLocal domainNode = new BDomainLocal() {
	    protected Object initialValue()
	    {
		BDomain bdomain = BDomain.currentBDomain();
		Node retval;

		retval = (Node)bdomain.getCapabilities().get("node");
		retval.start();
		return retval;
	    }
	};

    /**
     * @return The current Node object.
     */
    static Node currentNode()
    {
	return (Node)domainNode.get();
    }

    /**
     * Cached reference to the current domain.
     */
    private BDomain bdomain;

    /**
     * Cached reference to this domain's capabilities.
     */
    private ObjectTree capabilities;

    /**
     * The garbage collector for all the NodeCache's in this domain.
     */
    private NodeCacheCollector ncc;

    /**
     * True if this domain has done "something" and then reset by the
     * collected() method.
     */
    private boolean used = false;

    /**
     * The clock to use in the time() method.
     */
    private Clock clock;

    /**
     * Slot to hold this domain's "authorizer" handler.
     */
    private Authorizer authorizer;

    /**
     * If non-null, a connected NodeAuthorization coroutine.
     */
    private NodeAuthorization na;

    /**
     * Permission that indicates the number of milliseconds this node should
     * live for without any activity.
     */
    private IntegerPermission inactivity;

    /**
     * Construct an empty Node object.
     */
    public Node()
    {
    }

    /**
     * @see bees.core.Capability#fromBytes(byte[], CapabilityEnvelope[])
     */
    protected synchronized void fromBytes(byte flattened[],
					  CapabilityEnvelope ce[])
	throws Exception
    {
	super.fromBytes(flattened);

	super.specFromBytes(flattened);

	this.bdomain = BDomain.currentBDomain();
	this.capabilities = this.bdomain.getCapabilities();
	if( ce != null )
	{
	    this.clock = (Clock)ce[0].unseal();
	}
    }

    /**
     * @see bees.core.Capability#revoke()
     */
    public synchronized void revoke()
    {
	super.clear();

	this.bdomain = null;
	this.capabilities = null;
	this.ncc = null;
	this.inactivity = null;
    }

    /**
     * Start the NodeCacheCollector thread for this domain/node.
     */
    private void start()
    {
	this.ncc = new NodeCacheCollector(this);
	this.systemThreadStart(this.ncc,
			       "NodeCache Garbage Collector - "
			       + this.bdomain.getName());
    }

    /**
     * @return The clock to use in the time() method.
     */
    Clock getClock()
    {
	return this.clock;
    }

    /**
     * @return Cached reference to this domain's capabilities.
     */
    ObjectTree getCapabilities()
    {
	return this.capabilities;
    }

    /**
     * Get a system capability.
     *
     * @param name The name of the system capability.
     * @return The capability matching the given name.
     * @throws NoSuchElementException if the capability couldn't be found.
     */
    Capability getSystemCapability(String name)
	throws NoSuchElementException
    {
	return (Capability)this.capabilities.get(name);
    }

    /**
     * @return This node's "authorizer" handler.
     */
    public Authorizer getAuthorizer()
    {
	return this.authorizer;
    }

    /**
     * @param authorizer This node's "authorizer" handler.
     */
    public void setAuthorizer(Authorizer authorizer)
    {
	this.authorizer = authorizer;
    }

    /**
     * Get a user subtree from the capability tree.
     *
     * @param name The name of the subtree.
     * @return The ObjectMap matching the given name.
     * @throws NoSuchElementException if the subtree couldn't be found.
     */
    public ObjectMap getCapabilityMap(String name)
	throws NoSuchElementException
    {
	/* XXX better sub name */
	return (ObjectMap)this.capabilities.get("user/" + name);
    }

    /**
     * Get a user capability from the capability tree.
     *
     * @param name The name of the capability.
     * @return The Capability matching the given name.
     * @throws NoSuchElementException if the capability couldn't be found.
     */
    public Capability getCapability(String name)
	throws NoSuchElementException
    {
	/* XXX better sub name */
	return (Capability)this.capabilities.get("user/" + name);
    }

    /**
     * @return The name of this node.
     */
    public String getName()
    {
	super.checkPermission(READ_NODE_NAME);

	return super.getName();
    }

    /**
     * @return The the version of Bees on this node.
     */
    public String getVersion()
    {
	super.checkPermission(READ_NODE_VERSION);

	return super.getVersion();
    }

    /**
     * @return The garbage collector for all the NodeCache's in this domain.
     */
    NodeCacheCollector getCacheCollector()
    {
	return this.ncc;
    }

    /**
     * @param ip Permission that indicates the number of milliseconds this node
     * should live for without any activity.
     */
    public void setInactivityTimeout(IntegerPermission ip)
    {
	this.inactivity = ip;
    }

    /**
     * @return Permission that indicates the number of milliseconds this node
     * should live for without any activity.
     */
    public IntegerPermission getInactivityTimeout()
    {
	return this.inactivity;
    }

    /**
     * Set the "used" value of this node.
     */
    void used()
    {
	this.used = true;
    }
    
    /**
     * @see bees.core.NodeCache.EventHandler#removed(Node, ProtocolSession)
     */
    public final void removed(Node node, ProtocolSession ps)
    {
    }

    /**
     * @see bees.core.NodeCache.EventHandler#collected(Node, ProtocolSession)
     */
    public final IntegerPermission collected(Node node, ProtocolSession ps)
    {
	if( this.used )
	{
	    this.used = false;
	}
	else
	{
	    Logger.log("Inactivity timeout reached.  Shutting down...");
	    this.shutdown();
	}
	return this.inactivity;
    }

    /**
     * Send a request to the Auth Application asking for it to manage the
     * given protocol.
     *
     * XXX This should probably be moved into ProtocolSession.
     *
     * @param ps The ProtocolSession to be managed.
     * @throws NoSuchElementException If the "authenticator" capability could
     * not be found.
     */
    public synchronized void beginManagement(ProtocolSession ps)
	throws NoSuchElementException
    {
	NodeAuthorization na_tmp;
	TypeID sessionID;
	EventChannel ec;

	if( this.na != null )
	{
	    /* Already connected. */
	    return;
	}
	
	ec = (EventChannel)this.getSystemCapability("authenticator");
	na_tmp = new NodeAuthorization();
	sessionID = ps.getSessionType().getSessionID();
	if( ps instanceof ExtendedProtocolSession )
	{
	    na_tmp.beginExtendedManagement(ec, sessionID);
	}
	else
	{
	    na_tmp.beginManagement(ec, sessionID);
	}
	if( na_tmp.waitForState(NodeAuthorization.STATE_CONNECTED) ==
	    NodeAuthorization.STATE_CONNECTED )
	{
	    this.na = na_tmp;
	}
    }

    /**
     * Request a capability from the Auth Application.
     *
     * @param src The name of the capability in the Auth Applications name
     * space.
     * @param dst The name of the capability in this domain's name space.
     * @param auth The Auth data required to gain access to the requested
     * capabilities.
     * @return An array of CapabilityEnvelopes containing the requested
     * capabilities or null if the capabilities were not available.
     * @throws IllegalStateException If beginManagement was not called before
     * requesting the capability.
     */
    public CapabilityEnvelope[] requestCapability(String src,
						  String dst,
						  byte auth[])
	throws IllegalStateException
    {
	CapabilityEnvelope retval[] = null;
	NodeAuthorization na;

	if( this.na == null )
	{
	    throw new IllegalStateException(
		"Capability request before management request.");
	}
	
	na = new NodeAuthorization(this.na);
	na.requestCapability(src, dst, auth);
	if( na.waitForCompletion() )
	{
	    retval = na.getEnvelopes();
	}
	return retval;
    }

    public synchronized boolean pass(ProtocolPresentation pp,
				     CapabilityTree ct)
    {
	InheritedCapabilityCoroutine icc;
	EventChannel ec;
	boolean retval;

	ec = pp.getCoroutine().getChannel();
	icc = new InheritedCapabilityCoroutine();
	icc.initial(ec, ct.flatten());
	retval = icc.waitForCompletion();
	return retval;
    }

    /**
     * Create a new domain and load a protocol from the local disk.
     *
     * @param pp The ProtocolPresentation to connect to the protocol in the
     * new domain.
     * @param pf The description of the flow.
     * @param initialCaps The resource capabilities required to get the domain
     * going.
     * @param ps The resource limits for the new domain.
     * @param ct The capabilities to pass to the new domain.
     */
    public synchronized void register(ProtocolPresentation pp,
				      ProtocolFlow pf,
				      Capability initialCaps[],
				      PermissionSet ps,
				      CapabilityTree ct)
	throws BDomainCreationException,
	       ClassNotFoundException,
	       InstantiationException,
	       IllegalAccessException,
	       Throwable
    {
	FlowType ft = pf.getFlowType();
	CreateBDomainCoroutine csc;
	FlowIdentifier fid;
	TypeID flowID;

	flowID = ft.getFlowID();
	fid = FlowIdentifierFactory.create(this.bdomain,
					   flowID,
					   "FlowID: " + flowID.toString());
	csc = new CreateBDomainCoroutine();
	csc.create(this.bdomain.getBDomainCoroutine(),
		   "Registered Domain - "
		   + pf.getPrimaryProtocol(),
		   "bees.core.ProtocolBDomain",
		   null,
		   initialCaps,
		   /*
		   new Capability[] {
		       fid,
		       },*/
		   ps);
	if( csc.waitForState(CreateBDomainCoroutine.STATE_ALIVE) ==
	    CreateBDomainCoroutine.STATE_ALIVE )
	{
	    try
	    {
		InheritedCapabilityCoroutine icc;
		EventChannel ec;

		ec = new EventChannel();
		CapabilityFactory.create(ec,
					 csc.getStoreID(),
					 EventChannel.DEFAULT_PERMISSIONS);
		icc = new InheritedCapabilityCoroutine();
		icc.initial(ec, ct.flatten());
		if( icc.waitForCompletion() )
		{
		    LocalFlowCoroutine lfc;
		    
		    lfc = new LocalFlowCoroutine(ec);
		    lfc.loadFlow(pf);
		    if( lfc.waitForCompletion() )
		    {
			ExternalCoroutine eco;
			SessionType st;

			st = new SessionType().
			    setFlowType(ft).
			    setProtocolType(ft.getPrimaryProtocolType()).
			    snapshotSessionType();
			eco = new ExternalCoroutine();
			eco.connect(pp, ec, st);
			if( eco.waitForConnection() )
			{
			    pp.setCoroutine(eco);
			}
			else
			{
			    System.out.println("eco = " + eco);
			    throw new BDomainCreationException();
			}
		    }
		    else
		    {
			switch( lfc.getState() )
			{
			case Coroutine.STATE_ERROR:
			    throw lfc.getThrowable();
			default:
			    System.out.println("error " + lfc);
			    break;
			}
		    }
		}
		else
		{
		    throw icc.getThrowable();
		}
	    }
	    catch(MissingCapabilityException e)
	    {
		throw new BDomainCreationException();
	    }
	}
	else
	{
	    throw new BDomainCreationException(csc.getThrowable());
	}
    }
    
    /**
     * Create a new domain and load the protocol from a neighbor.
     *
     * @param retval The DLProtocolPresentation to connect to the DLProtocol
     * in the new domain or null if one should be created.
     * @param ni The neighbor to download the flow from.
     * @param flowDesc The flow tuple.
     * @param initialCaps The resource capabilities required to get the domain
     * going.
     * @param transfer The resource limits for the new domain.
     * @param ct The capabilities to pass to the new domain.
     * @return A DLProtocolPresentation that can be used to control the
     * download process.
     */
    public DLProtocolPresentation register(DLProtocolPresentation retval,
					   NeighborImpl ni,
					   PairChain flowDesc,
					   Capability initialCaps[],
					   PermissionSet transfer,
					   CapabilityTree ct)
	throws BDomainCreationException,
	       Throwable
    {
	CreateBDomainCoroutine csc;

	ct = new CapabilityTree(new CapabilityEnvelope[] {
	    new CapabilityEnvelope("tmp/previousNeighbor", ni),
	},
				new CapabilityTree[] {
				    ct
				});

	if( initialCaps == null )
	{
	    initialCaps = new Capability[0];
	}
	csc = new CreateBDomainCoroutine();
	csc.create(this.bdomain.getBDomainCoroutine(),
		   "Downloaded Flow - " /* XXX Add Something */,
		   "bees.core.ProtocolBDomain",
		   null,
		   initialCaps,
		   transfer);
	if( csc.waitForState(CreateBDomainCoroutine.STATE_ALIVE) ==
	    CreateBDomainCoroutine.STATE_ALIVE )
	{
	    try
	    {
		InheritedCapabilityCoroutine icc;
		EventChannel ec;

		ec = new EventChannel();
		CapabilityFactory.create(ec,
					 csc.getStoreID(),
					 EventChannel.DEFAULT_PERMISSIONS);
		icc = new InheritedCapabilityCoroutine();
		icc.initial(ec, ct.flatten());
		if( icc.waitForCompletion() )
		{
		    RemoteFlowCoroutine rfc;

		    rfc = new RemoteFlowCoroutine(ec);
		    rfc.downloadFlow(flowDesc);
		    if( rfc.waitForCompletion() )
		    {
			ExternalCoroutine eco;
			SessionType st, ust;
			ProtocolType pt;
			TypeID flowID;
			DLProtocol dp;
			FlowType ft;
			
			dp = new DLProtocol();
			pt = dp.getProtocolType();
			ft = new FlowType(flowDesc);
			ft.snapshotFlowType();
			st = new SessionType().
			    setFlowType(ft).
			    setProtocolType(pt).
			    snapshotSessionType();
			ust = new SessionType().
			    setFlowType(ft);
			if( retval == null )
			{
			    retval = new DLProtocolPresentation();
			}
			retval.setSessionType(ust);
			eco = new ExternalCoroutine();
			eco.connect(retval, ec, st);
			if( eco.waitForConnection() )
			{
			    retval.setCoroutine(eco);
			    retval.register();
			    retval.start();
			}
			else
			{
			    System.out.println("eco = " + eco);
			    throw new BDomainCreationException();
			}
		    }
		    else
		    {
			System.out.println("error " + rfc);
			switch( rfc.getState() )
			{
			case Coroutine.STATE_ERROR:
			    throw rfc.getThrowable();
			default:
			    System.out.println("error " + rfc);
			    break;
			}
		    }
		}
		else
		{
		    System.out.println("error " + icc);
		    switch( icc.getState() )
		    {
		    case Coroutine.STATE_ERROR:
			throw icc.getThrowable();
		    default:
			System.out.println("error " + icc);
			break;
		    }
		}
	    }
	    catch(MissingCapabilityException e)
	    {
		throw new BDomainCreationException();
	    }
	    catch(InstantiationException e)
	    {
		throw new BDomainCreationException();
	    }
	}
	else
	{
	    System.out.println("error " + csc);
	    switch( csc.getState() )
	    {
	    case Coroutine.STATE_ERROR:
		throw csc.getThrowable();
	    default:
		System.out.println("error " + csc);
		break;
	    }
	}
  	return retval;
    }
    
    /**
     * Create a new domain and load the protocol from a neighbor.
     *
     * @param ni The neighbor to download the flow from.
     * @param flowDesc The flow tuple.
     * @param initialCaps The resource capabilities required to get the domain
     * going.
     * @param transfer The resource limits for the new domain.
     * @param ct The capabilities to pass to the new domain.
     * @return A DLProtocolPresentation that can be used to control the
     * download process.
     */
    public DLProtocolPresentation register(NeighborImpl ni,
					   PairChain flowDesc,
					   Capability initialCaps[],
					   PermissionSet transfer,
					   CapabilityTree ct)
	throws BDomainCreationException,
	       Throwable
    {
	return this.register(null,
			     ni,
			     flowDesc,
			     initialCaps,
			     transfer,
			     ct);
    }

    /**
     * Connect the given presentation to a protocol session in another domain.
     *
     * @param pp The ProtocolPresentation to connect.
     * @param ce An envelope containing an EventChannel capability.
     * @param sessionID The protocol session identifier to connect to in the
     * domain.
     */
    public void register(ProtocolPresentation pp,
			 CapabilityEnvelope ce,
			 TypeID sessionID)
	throws MissingCapabilityException,
	       InstantiationException
    {
	ExternalCoroutine eco;
	SessionType st;

	st = new SessionType().
	    setSessionID(sessionID);
	eco = new ExternalCoroutine();
	eco.connect(pp, (EventChannel)ce.unseal(), st);
	if( eco.waitForConnection() )
	{
	    pp.setCoroutine(eco);
	}
	else
	{
	    /* XXX */
	    Logger.log("fuck");
	}
    }
    
    /**
     * Create and start a new thread.
     *
     * @param r The Runnable to begin executing in a new thread.
     * @param threadName The name of the newly created thread.
     * @throws PermissionException If this node does not have the THREAD_START
     * permission.
     */
    public void threadStart(Runnable r, String threadName)
	throws PermissionException
    {
	Assert.require(r);
	Assert.require(Domain.currentDomain() != null);

	this.checkPermission(THREAD_START);

	systemThreadStart(r, threadName);
    }
    
    /**
     * Create and start a new thread with a default name.
     *
     * @param r The Runnable to begin executing in a new thread.
     * @throws PermissionException If this node does not have the THREAD_START
     * permission.
     */
    public void threadStart(Runnable r)
	throws PermissionException
    {
	Assert.require(r);
	Assert.require(Domain.currentDomain() != null);
	
	threadStart(r, "Another Underappreciated Thread: " + r.toString());
    }

    /**
     * Internal method used to start threads.  This method is not bound by the
     * THREAD_START permission.
     *
     * @param r The Runnable to begin executing in a new thread.
     * @param threadName The name of the newly created thread.
     */
    /*PACKAGE*/ void systemThreadStart(Runnable r, String threadName)
    {
	Assert.require(r);
	Assert.require(Domain.currentDomain() != null);
	
	Domain.currentDomain().threadStart(r, threadName);
    }

    /**
     * Set the JSIHandler for a given subproperty.
     *
     * @param name The property name that should resolve to the given
     * JSIHandler.
     * @param jh The JSIHandler that implements a property.
     * @throws PermissionException If this node does not have the JSI_REGISTER
     * property.
     */
    public void setJSIHandler(String name, JSIHandler jh)
	throws PermissionException
    {
	super.checkPermission(JSI_REGISTER);
	
	JSICoroutine.addRootHandler(name, jh);
    }

    /**
     * Shutdown the BDomain containing this node.
     */
    public synchronized void shutdown()
    {
	BDomain.currentBDomain().shutdown();
    }

    /**
     * Put the current thread to sleep for the given number of milliseconds.
     *
     * @param millis The number of milliseconds to sleep.
     * @throws InterruptedException If another thread uses interrupt() on the
     * current thread.
     */
    public void sleep(long millis)
	throws InterruptedException
    {
	Domain f = Domain.currentDomain();
	
	edu.utah.janos.nodeos.Thread.sleep(millis, 0);
    }

    /**
     * @return The current time in milliseconds.
     */
    long time()
    {
	return this.clock.current();
    }

    /**
     * @see java.lang.System#arraycopy(Object, int, Object, int, int)
     */
    public void arraycopy(Object src, int src_position,
			  Object dst, int dst_position, int length)
    {
	System.arraycopy(src, src_position, dst, dst_position, length);
    }

    /**
     * @return True if the node is on fire.
     */
    public boolean isOnFire()
    {
	return false;
    }

    /**
     * @see edu.utah.janosvm.kit.comm.jsi.JSIHandler#resolveSpecifier(JSIMessage, JSIPair)
     */
    public JSIHandler resolveSpecifier(JSIMessage jm, JSIPair jp)
    {
	return this;
    }
    
    /**
     * @see edu.utah.janosvm.kit.comm.jsi.JSIHandler#handleMessage(JSIHandlerContext, JSIMessage)
     */
    public void handleMessage(JSIHandlerContext jhc, JSIMessage jm)
    {
	JSIPair jp;

	switch( jm.getWhat() )
	{
	case JSIMessage.GET_PROPERTY:
	    jm.restoreSpecifier();
	    if( (jp = jm.popSpecifier()) != null )
	    {
		if( jp.getTag().equals("name") )
		{
		    jm.reply(this.getName());
		}
		else if( jp.getTag().equals("version") )
		{
		    jm.reply(this.getVersion());
		}
		else
		{
		    jm.error("Unknown Node property: " + jp.getTag());
		}
	    }
	    else
	    {
		jm.reply(this.toString());
	    }
	    break;
	case JSIMessage.GET_SUITES:
	    jm.clear();
	    jm.setWhat(JSIMessage.REPLY);
	    jm.add("result", JSI_SUB_PROPERTIES);
	    break;
	default:
	    throw new UnsupportedOperationException(
		"What: " + JSIMessage.MESSAGE_TYPE_STRINGS[jm.getWhat()]);
	}
    }
    
    /**
     * @see edu.utah.janosvm.kit.comm.jsi.JSIHandler#getSuites(JSIMessage)
     */
    public void getSuites(JSIMessage jm)
    {
    }
    
    /**
     * @return An abbreviated description of this object.
     */
    public String toShortString()
    {
	return "Node["
	    + super.toString()
	    + "]";
    }
    
    public String toString()
    {
	return "Node["
	    + super.toString()
	    + "; capabilities="
	    // XXX recursive + this.capabilities
	    + "; ncc="
	    + this.ncc
	    + "; used="
	    + this.used
	    + "; authorizer="
	    + this.authorizer
	    + "; inactivity="
	    + this.inactivity
	    + "]";
    }
}
