/* 
 * NodeCache.Java
 *  
 * Copyright (c) 1999, 2000, 2001, 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Bees distribution.
 * 
 * Permission to use, copy, modify, distribute, and sell this
 * software and its documentation for any purpose is hereby granted
 * without fee, provided that (i) the above copyright notice(s)
 * appear in all copies and (ii) that both the copyright notice(s)
 * and this permission notice appear in supporting documentation,
 * and (iii) that the name(s) of the copyright holders not be used
 * in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.  The
 * copyright holders make no representations about the suitability
 * of this software for any purpose.  It is provided "as is"
 * without express or implied warranty.
 * 
 * The maintainers of this code ask, but do not require, that
 * improvements and bug fixes be emailed to csl-dist@cs.utah.edu
 *
 *
 *  Copyright 1996 Massachusetts Institute of Technology
 * 
 *  Permission to use, copy, modify, distribute, and sell this software and
 *  its documentation for any purpose is hereby granted without fee, provided
 *  that the above copyright notice appear in all copies and that both that
 *  copyright notice and this permission notice appear in supporting
 *  documentation, and that the name of M.I.T. not be used in advertising or
 *  publicity pertaining to distribution of the software without specific,
 *  written prior permission.  M.I.T. makes no representations about the
 *  suitability of this software for any purpose.  It is provided "as is"
 *  without express or implied warranty.
 * 
 */
package bees.core;

import edu.utah.janos.util.LinkedList;

import bees.core.resource.Logger;

import bees.core.security.IntegerPermission;

/**
 * Node cache of managed soft-state.
 * <p>
 * Combines hashtable and doubly linked list to implement fixed size table
 * with LRU replacement.  all public entry points are synchronized.
 * a background gc thread cleans entries that have expired.
 * <p>
 * XXX the background GC thread is probably unnecessary since this is
 * only per-domain state.  At the least, the background thread should
 * be configurable by the domain (e.g., its collection parameters).
 * <p>
 * XXX should this class be final?  Should more methods be marked
 * final? (e.g., get)
 * <p>
 * This differs from the original version in that it does not have a maximum
 * time limit or a maximum number of cached items, both of which are unneeded
 * because of per-domain resource limits.
 *
 * @author David Wetherall
 * @author <a href="mailto:janos-discuss@flux.cs.utah.edu">Utah Janos Team</a>
 */
public final class NodeCache
{
    /**
     * Entries in the hashtable and time list.
     * Can be nested under NodeCache in 1.1
     */
    private final class NCElement
	extends LinkedList.Node
    {
	/**
	 * Link to next hash table elments
	 */
	NCElement next;
	
	/**
	 * Key for this node
	 */
	Object key;
	
	/**
	 * Hash code for this key, faster to compare than the actual key
	 */
	int hash;
	
	/**
	 * Stored value for the key
	 */
	Object value;
	
	/**
	 * Lifetime of the key/value pair
	 */
	long expiration;

        /**
         * @see java.lang.Object#toString()
         */
	public String toString()
	{
	    return "NCElement[hash="
		+ this.hash
		+ "; expiration="
		+ this.expiration
		+ "]";
	}
    }
    
    /**
     * Elements of this type in the node cache will receive calls when
     * one of a set of events occurs.
     * <p>
     * XXX Not sure if this is useful or not, and it can probably be done
     * better
     */
    public static interface EventHandler
    {
	/**
	 * This is called when the element is manually removed from
	 * from the cache.
	 * 
	 * @param node The current Node object.
	 * @param ps The ProtocolSession that the NodeCache is 
	 */
	void removed(Node node, ProtocolSession ps);
	
	/**
	 * This is called when this element is up for collection,
	 * the return value specifies the new expiration time.  Null
	 * is returned if the element should really be collected.
	 * 
	 * @param node The current Node object.
	 * @param ps The ProtocolSession that the NodeCache is attached to.
	 * @return Null if the element should be collected or an
	 * IntegerPermission specifying the number of milliseconds to delay
	 * before this element should be collected again.
	 */
	IntegerPermission collected(Node node, ProtocolSession ps);
    }
    
    /**
     * Reference to whichever node owns this cache
     */
    private final Node owner;

    /**
     * The ProtocolSession this NodeCache is attached to.
     */
    private final ProtocolSession ps;
    
    /**
     * Hash table for quick access to the NCElements
     */
    private NCElement[] hash;
    
    private final LinkedList timeOrdered = new LinkedList();
    
    /**
     * Tracks the number of elements;
     */
    private int elementCount;

    public final Node getOwnerNode()
    {
	return this.owner;
    }
    
    /**
     * Get the bucket for a specific hash code
     *
     * @param hcode A hash code, most likely produced by
     * java.lang.Object.hashCode()
     */
    private final int hashBucket( int hcode )
    {
	int retval;
	
	retval = (hcode & 0x7FFFFFF) % this.hash.length;
	return( retval );
    }
    
    /**
     * Find the NCElement with the given key.  The NCElement will also
     * be moved to the front of the hash table list.
     *
     * @param key The key to match on
     * @return The NCElement matching the given key
     */
    private synchronized NCElement find(Object key)
    {
	int hcode = key.hashCode();
	int bucket = this.hashBucket( hcode );
	NCElement nce, pred = null;
	NCElement retval = null;
	
	for( nce = this.hash[bucket];
	     (nce != null) && (retval == null);
	     pred = nce, nce = nce.next )
	{
	    /*
	     * Compare the hash code first so we don't have to
	     * compare the whole key.  (Questionable whether this
	     * is a good thing or not)
	     */
	    if( (nce.hash == hcode) && nce.key.equals(key) )
	    {
		retval = nce;
		/*
		 * Push the hash element to the front for faster access
		 */
		if( pred != null )
		{
		    pred.next = nce.next;
		    nce.next = this.hash[bucket];
		    this.hash[bucket] = nce;
		}
	    }
	}
	return retval;
    }

    private void timeInsert(NCElement nce)
    {
	LinkedList.Node curr;

	curr = this.timeOrdered.getHead();
	while( (curr.getSucc() != null) &&
	       (((NCElement)curr).expiration < nce.expiration) )
	{
	    curr = curr.getSucc();
	}
	curr.prepend(nce);
    }
    
    /**
     * Return the value paired with a certain key
     *
     * @param key The key to search for
     * @return The value associated with the key
     */
    public final synchronized Object get(Object key)
    {
	Object retval = null;
	NCElement nce;
	
	nce = this.find(key);
	if( nce != null )
	{
	    retval = nce.value;
	}
	return retval;
    }
    
    /**
     * Add a key/value pair to the cache with a given expiration time.
     * If the key is already present in the cache then it will be replaced
     * with the parameters and the old value returned.
     *
     * @param key The key for identifying the value
     * @param value The value to be tracked
     * @param time The expiration time for this association
     * @return Old value from a prior association, if it existed
     */
    public final synchronized Object put(Object key,
					 Object value,
					 IntegerPermission time)
    {
	Object retval = null;
	int hcode = key.hashCode();
	int bucket = this.hashBucket( hcode );
	NCElement nce;
	
	nce = this.find(key);

	if( value instanceof Capsule )
	{
	    Capsule cap;
	    
	    cap = (Capsule)value;
	    cap.checkPermission(time);
	    cap.getCapsuleType().decGlobal(CapsuleType.GLOBAL_CACHED);
	}
	if( nce == null )
	{
	    /* This is a new key, so we create an element for it */
	    nce = new NCElement();
	    nce.hash = hcode;
	    nce.key = key;
	    nce.next = this.hash[bucket];
	    this.hash[bucket] = nce;
	    this.elementCount++;
	}
	else
	{
	    retval = nce.value;
	    nce.remove();
	}
	nce.value = value;
	nce.expiration = this.owner.time() + time.getValue();
	this.timeInsert(nce);
	/*
	 * If the expiration for this association happens before the
	 * current minimum then we need to wake up the gc thread and
	 * have it wait for the new time.
	 */
	if( this.timeOrdered.getHead() == nce )
	{
	    synchronized(this.owner.getCacheCollector())
	    {
		this.owner.getCacheCollector().notify();
	    }
	}
	return retval;
    }
    
    /**
     *
     */
    public synchronized boolean refresh(Object key, IntegerPermission time)
    {
	int hcode = key.hashCode();
	int bucket = this.hashBucket( hcode );
	boolean retval = false;
	NCElement nce;
	
	nce = this.find(key);
	if( nce != null )
	{
	    if( nce.value instanceof Capsule )
	    {
		Capsule cap;
		
		cap = (Capsule)nce.value;
		cap.checkPermission(time);
	    }
	    retval = true;
	    nce.expiration = this.owner.time() + time.getValue();
	    nce.remove();
	    this.timeInsert(nce);
	    if( this.timeOrdered.getHead() == nce )
	    {
		synchronized(this.owner.getCacheCollector())
		{
		    this.owner.getCacheCollector().notify();
		}
	    }
	}
	return retval;
    }
    
    /**
     * Remove an association from the cache.
     *
     * @param key Key for the association to remove
     * @return The value from the association
     */
    public synchronized Object remove(Object key)
    {
	Object retval = null;
	
	int hcode = key.hashCode();
	int bucket = this.hashBucket( hcode );
	
	NCElement e = this.hash[bucket];
	NCElement prev = null;
	
	while( (e != null) && (retval == null) )
	{
	    if( (e.hash == hcode) && e.key.equals(key) )
	    {
		if (prev != null)
		{
		    prev.next = e.next;
		}
		else
		{
		    this.hash[bucket] = e.next;
		}
		e.next = null;
		retval = e.value;
		e.remove();
		this.elementCount--;
		if( e.value instanceof Capsule )
		{
		    Capsule cap;

		    cap = (Capsule)e.value;
		    cap.getCapsuleType().incGlobal(CapsuleType.GLOBAL_GROUPED);
		}
		if( e.value instanceof EventHandler )
		{
		    try
		    {
			((EventHandler)e.value).removed(this.owner, this.ps);
		    }
		    catch(Throwable th)
		    {
			Logger.log(th);
		    }
		}
		/*
		 * Its not a big deal if this object has the
		 * minimum time used for the gc thread.  It
		 * just won't collect anything, and anyways
		 * it will find the new minTime from the scan.
		 */
	    }
	    else
	    {
		prev = e;
		e = e.next;
	    }
	}
	return retval;
    }
    
    /**
     * Overloaded get method for integer keys.
     */
    public final synchronized Object get(int i)
    {
	return this.get(new Integer(i));
    }
    
    /**
     * Overloaded put method for integer keys.
     */
    public final synchronized Object put(int i,
					 Object value,
					 IntegerPermission time)
    {
	return this.put(new Integer(i), value, time);
    }
    
    /**
     * Overloaded get method for long keys.
     */
    public final synchronized Object get(long l)
    {
	return this.get(new Long(l));
    }
    
    /**
     * Overloaded put method for long keys.
     */
    public final synchronized Object put(long l,
					 Object value,
					 IntegerPermission time)
    {
	return this.put(new Long(l), value, time);
    }

    /**
     * @return The set of keys currently in the cache.
     */
    public final synchronized Object[] keys()
    {
	Object retval[] = new Object[this.elementCount];
	int lpc, pos = 0;

	for( lpc = 0; lpc < this.hash.length; lpc++ )
	{
	    NCElement nce;

	    nce = this.hash[lpc];
	    while( nce != null )
	    {
		retval[pos] = nce.key;
		pos += 1;
		nce = nce.next;
	    }
	}
	return retval;
    }
    
    /**
     * Scans the list of associations and garbage collects based on
     * expiration dates.
     */
    synchronized long clean(long now)
    {
	boolean done = false;
	long retval;
	
	/*
	 * We have to determine a new min time since we might be
	 * collecting the current one
	 */
	retval = Long.MAX_VALUE;
	while( !done )
	{
	    LinkedList.Node head;
	    
	    head = this.timeOrdered.getHead();
	    if( !(head instanceof NCElement) )
	    {
		done = true;
	    }
	    else
	    {
		NCElement nce;

		nce = (NCElement)head;
		if( nce.expiration <= now )
		{
		    /* Expired */
		    if( nce.value instanceof EventHandler )
		    {
			IntegerPermission time;
			long newtime = 0;
			
			if( nce.value instanceof Capsule )
			{
			    Capsule cap;
			    
			    cap = (Capsule)nce.value;
			    cap.getPacketAuthorization().replenish();
			}
			try
			{
			    if( (time = ((EventHandler)nce.value).
				 collected(this.owner, this.ps)) != null )
			    {
				newtime = time.getValue();
				if( nce.value instanceof Capsule )
				{
				    Capsule cap;
				    
				    cap = (Capsule)nce.value;
				    cap.checkPermission(time);
				}
			    }
			}
			catch(ThreadDeath td)
			{
			    throw td;
			}
			catch(Throwable th)
			{
			    Logger.log(th);
			    newtime = 0;
			}
			nce.expiration = (now + newtime);
			if( newtime <= 0 )
			{
			    this.remove(nce.key);
			}
			else
			{
			    nce.remove();
			    this.timeInsert(nce);
			}
		    }
		    else
		    {
			this.remove(nce.key);
		    }
		}
		else
		{
		    retval = nce.expiration - now;
		    done = true;
		}
	    }
	}
	return retval;
    }
    
    /**
     * Construct a new NodeCache.
     *
     * @param owner The Node that owns this NodeCache instance
     * @param ps The ProtocolSession this NodeCache should be bound to.
     * @param size The size of the hash table
     */
    NodeCache(Node owner, ProtocolSession ps, int size)
    {
	this.owner = owner;
	this.ps = ps;
	this.hash = new NCElement[size];
	
	this.owner.getCacheCollector().add(this);
    }

    /**
     * Construct a NodeCache with the given values.
     *
     * @param owner The Node that owns this NodeCache instance
     * @param size The size of the hash table
     */
    public NodeCache(Node owner, int size)
    {
	this(owner, null, size);
    }

    /**
     * @see java.lang.Object#toString()
     */
    public String toString()
    {
	return "NodeCache[elementCount="
	    + this.elementCount
	    + "]";
    }
}
