/*
 * PacketAuthorization.java
 *
 * Copyright (c) 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is part of the Janos Bees Software.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */

package bees.core;

import bees.core.resource.Logger;

import bees.core.security.Permission;
import bees.core.security.PermissionSet;
import bees.core.security.CountPermission;
import bees.core.security.PermissionException;

/**
 * A PacketAuthorization tracks the propagation values for a capsule in the
 * system.
 */
final class PacketAuthorization
    implements Cloneable
{
    static final int VALUE_MIN = 0;
    static final int VALUE_INITIAL = 0;
    static final int VALUE_RETURNS = 1;
    static final int VALUE_FORWARDS = 2;
    static final int VALUE_EXTERNALS = 3;
    static final int VALUE_NEIGHBORHOOD = 4;
    static final int VALUE_GROUPS = 5;
    static final int VALUE_MAX = 6;

    static final Permission VALUE_PERMISSIONS[] = new Permission[] {
	new Permission("initial"),
	new Permission("returns"),
	new Permission("forwards"),
	new Permission("externals"),
	new Permission("neighborhood"),
	new Permission("groups"),
    };

    /**
     * The cached maximum values for each type of propagation.
     *
     * XXX This should be stored in the capsule type or something, its
     * redundant store it here for every capsule.
     */
    private final int maxValues[];

    /**
     * The current propagation values.
     */
    private final int values[];

    /**
     * Internal PacketAuthorization constructor.
     *
     * @param maxValues The cached maximum values for each type of propagation.
     * @param values The current propagation values.
     */
    private PacketAuthorization(int maxValues[], int values[])
    {
	this.maxValues = (int[])maxValues.clone();
	this.values = (int[])values.clone();
    }
    
    /**
     * Construct a PacketAuthorization with the given values.
     *
     * @param ps A PermissionSet containing a number of CountPermissions
     * corresponding to the maximum propagation values.
     */
    PacketAuthorization(PermissionSet ps)
    {
	int lpc;

	this.maxValues = new int[VALUE_MAX];
	this.values = new int[VALUE_MAX];
	for( lpc = 0; lpc < VALUE_MAX; lpc++ )
	{
	    CountPermission cp;

	    try
	    {
		if( (cp = (CountPermission)ps.
		     hasPermission(VALUE_PERMISSIONS[lpc])) != null )
		{
		    this.maxValues[lpc] = cp.getMaxCount();
		}
	    }
	    catch(ClassCastException e)
	    {
		/* Ignore non-CountPermissions in ps. */
		Logger.log(Logger.WARNING,
			   "Invalid permission in Capsule permission set: "
			   + ps.hasPermission(VALUE_PERMISSIONS[lpc]));
	    }
	}
    }

    /**
     * Consume a propagation value by dropping its value by one.
     *
     * @param type The type of propagation to check/decrement.
     * @exception Throws a PermissionException if the packet isn't authorized
     * to perform this propagation.
     */
    void consume(int type)
    {
	// Assert.require(type >= VALUE_MIN);
	// Assert.require(type < VALUE_MAX);
	
	if( this.values[type] > 0 )
	{
	    this.values[type] -= 1;
	}
	else
	{
	    throw new PermissionException("Invalid access:  Packet "
					  + "doesn't have permission, "
					  + VALUE_PERMISSIONS[type]
					  + "; "
					  + this);
	}
    }

    /**
     * Reset the propagation values to their maximum values.
     */
    void replenish()
    {
	int lpc;

	for( lpc = 0; lpc < VALUE_MAX; lpc++ )
	{
	    this.values[lpc] = this.maxValues[lpc];
	}
    }

    /**
     * Set all the propagation values to zero.
     */
    void revoke()
    {
	int lpc;

	for( lpc = 0; lpc < VALUE_MAX; lpc++ )
	{
	    this.values[lpc] = 0;
	}
    }

    public Object clone()
    {
	return new PacketAuthorization(this.maxValues, this.values);
    }
    
    public String toString()
    {
	String retval = "PacketAuthorization[maxValues={";
	int lpc;

	for( lpc = 0; lpc < VALUE_MAX; lpc++ )
	{
	    retval += " " + this.maxValues[lpc];
	}
	retval += " }; values={";
	for( lpc = 0; lpc < VALUE_MAX; lpc++ )
	{
	    retval += " " + this.values[lpc];
	}
	retval += " }]";
	return retval;
    }
}
