/*
 * PairChain.java
 *
 * Copyright (c) 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is part of the Janos Bees Software.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */

package bees.core;

import java.util.Vector;

/**
 * A PairChain manages an ordered collection of Pair objects.  PairChains are
 * used to carry untyped byte arrays between nodes in the system.
 *
 * @see Pair
 */
public final class PairChain
    implements Cloneable
{
    /**
     * The array of Pairs.
     */
    private final Vector chain = new Vector();

    /**
     * The total marshalled size of this PairChain.
     */
    private int totalSize = Xdr.INT;

    /**
     * Construct a PairChain with the given values.
     *
     * @param pairs An array of Pairs to add to the new chain.
     */
    public PairChain(Pair pairs[])
    {
	if( pairs != null )
	{
	    int lpc;
	    
	    for( lpc = 0; lpc < pairs.length; lpc++ )
	    {
		if( pairs[lpc] != null )
		{
		    this.add(pairs[lpc]);
		}
	    }
	}
    }

    /**
     * Construct an empty PairChain.
     */
    public PairChain()
    {
	this(null);
    }

    /**
     * @param pairs The pairs to add to this chain.
     * @return this
     */
    public PairChain from(Pair pairs[])
    {
	if( pairs != null )
	{
	    int lpc;
	    
	    for( lpc = 0; lpc < pairs.length; lpc++ )
	    {
		if( pairs[lpc] != null )
		{
		    this.add(pairs[lpc]);
		}
	    }
	}
	return this;
    }

    /**
     * @param pc The pairs to add to this chain.
     * @return this
     */
    public PairChain from(PairChain pc)
    {
	if( pc != null )
	{
	    this.from(pc.toArray());
	}
	return this;
    }

    /**
     * @param pair A pair to add to this chain.
     */
    public synchronized PairChain add(Pair pair)
    {
	Assert.require(pair);
	
	this.chain.addElement(pair);
	this.totalSize += pair.marshalledLength();
	return this;
    }
 
    /**
     * @param tag The name part of the pair.
     * @param value The value part of the pair.
     * @return this
     */
    public PairChain add(String tag, String value)
    {
	Assert.require(tag);
	Assert.require(value);
	
	this.add(new Pair(tag, value));
	return this;
    }
    
    /**
     * @param tag The name part of the pair.
     * @param value The value part of the pair.
     * @return this
     */
    public PairChain add(String tag, byte value[])
    {
	Assert.require(tag);
	Assert.require(value);
	
	this.add(new Pair(tag, value));
	return this;
    }

    /**
     * @param tag The tag to look for.
     * @param instance The instance of "tag" to look for.
     * @return The index of the pair in this.chain.
     */
    private int findPair(String tag, int instance)
    {
	int retval = -1;
	int lpc, size;

	Assert.require(tag);

	size = this.chain.size();
	if( instance >= 0 )
	{
	    for( lpc = 0; (lpc < size) && (retval == -1); lpc++ )
	    {
		Pair pair = (Pair)this.chain.elementAt(lpc);

		if( pair.getTag().equals(tag) )
		{
		    if( instance == 0 )
		    {
			retval = lpc;
		    }
		    else
		    {
			instance -= 1;
		    }
		}
	    }
	}
	else
	{
	    for( lpc = size - 1; (lpc >= 0) && (retval == -1); lpc-- )
	    {
		Pair pair = (Pair)this.chain.elementAt(lpc);
		
		if( pair.getTag().equals(tag) )
		{
		    if( instance == -1 )
		    {
			retval = lpc;
		    }
		    else
		    {
			instance += 1;
		    }
		}
	    }
	}
	return retval;
    }
    
    /**
     * @param tag The tag to look for.
     * @param instance The instance of "tag" to look for.
     * @return The matching Pair or NULL.
     */
    public synchronized Pair get(String tag, int instance)
    {
	Pair retval = null;
	int index;

	Assert.require(tag);
	
	if( (index = this.findPair(tag, instance)) >= 0 )
	{
	    retval = (Pair)this.chain.elementAt(index);
	}
	return retval;
    }
    
    /**
     * @param tag The tag to look for.
     * @return The matching Pair or NULL.
     */
    public Pair get(String tag)
    {
	return this.get(tag, 0);
    }

    public Pair expect(String tag, int instance)
	throws MissingPairException
    {
	Pair retval;

	if( (retval = this.get(tag, instance)) == null )
	{
	    throw new MissingPairException(tag, instance);
	}
	return retval;
    }

    public Pair expect(String tag)
	throws MissingPairException
    {
	return this.expect(tag, 0);
    }

    /**
     * @param tag The tag to remove.
     * @param instance The instance of "tag" to remove.
     */
    public synchronized void remove(String tag, int instance)
    {
	int index;
	
	Assert.require(tag);
	
	if( (index = this.findPair(tag, instance)) >= 0 )
	{
	    Pair pair = (Pair)this.chain.elementAt(index);
	    
	    this.chain.removeElementAt(index);
	    this.totalSize -= pair.marshalledLength();
	}
    }
    
    /**
     * @param tag The tag to remove.
     */
    public void remove(String tag)
    {
	this.remove(tag, 0);
    }

    /**
     * @return The number of Pairs in this chain.
     */
    public synchronized int size()
    {
	return this.chain.size();
    }

    /**
     * @return An array containing all of this chain's Pairs.
     */
    public synchronized Pair[] toArray()
    {
	Pair retval[];

	retval = new Pair[this.chain.size()];
	this.chain.copyInto(retval);
	return retval;
    }

    /**
     * @return The number of bytes needed to flatten this chain.
     */
    public int marshalledLength()
    {
	return this.totalSize;
    }

    /**
     * @param xdr The object to marshal this chain to.
     */
    public void marshal(XdrByteArray xdr)
    {
	int lpc, size;

	Assert.require(xdr);
	
	size = this.chain.size();
	xdr.PUT(size);
	for( lpc = 0; lpc < size; lpc++ )
	{
	    Pair pair = (Pair)this.chain.elementAt(lpc);

	    pair.marshal(xdr);
	}
    }

    /**
     * @param xdr The object to unmarshal this chain from.
     */
    public void unmarshal(XdrByteArray xdr)
    {
	int lpc, size;
	
	Assert.require(xdr);
	
	size = xdr.INT();
	if( size < 0 )
	{
	    throw new IllegalArgumentException("Bogus size in PairChain: "
					       + size);
	}
	for( lpc = 0; lpc < size; lpc++ )
	{
	    this.add(Pair.unmarshal(xdr));
	}
    }

    /**
     * @return A flattened version of this chain.
     */
    public synchronized byte[] toBytes()
    {
	XdrByteArray xdr;

	xdr = new XdrByteArray(this.marshalledLength());
	this.marshal(xdr);
	if( xdr.index != xdr.len )
	{
	    new Panic("marshalledLength() doesn't match actual length");
	}
	return xdr.toBytes();
    }

    /**
     * @param pc A flattened PairChain object.
     * @return A new PairChain object initialized with the values in "pc".
     */
    public static PairChain instantiate(byte pc[])
    {
	PairChain retval;

	Assert.require(pc);
	
	retval = new PairChain();
	if( pc.length > 0 )
	{
	    XdrByteArray xdr;
	    
	    xdr = new XdrByteArray(pc);
	    retval.unmarshal(xdr);
	    if( xdr.index != xdr.len )
	    {
		throw new IllegalArgumentException(
			"Flattened PairChain is short");
	    }
	}
	return retval;
    }

    /**
     * @see java.lang.Object#clone()
     */
    public synchronized Object clone()
    {
	PairChain retval = new PairChain();
	int lpc, size;

	size = this.chain.size();
	for( lpc = 0; lpc < size; lpc++ )
	{
	    Pair pair = (Pair)this.chain.elementAt(lpc);

	    retval.add((Pair)pair.clone());
	}
	return retval;
    }
    
    /**
     * @see java.lang.Object#toString()
     */
    public String toString()
    {
	return "PairChain[chain="
	    + this.chain
	    + "; totalSize="
	    + this.totalSize
	    + "]";
    }
}
