/*
 * ProtocolParameters.java
 *
 * Copyright (c) 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is part of the Janos Bees Software.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */

package bees.core;

import java.io.IOException;
import java.io.ByteArrayInputStream;

import java.util.Properties;
import java.util.StringTokenizer;

import edu.utah.janos.nodeos.TimeSpec;

import bees.core.resource.Logger;

/**
 * ProtocolParameters is a base class that provides basic configuration
 * functionality.
 */
public class ProtocolParameters
{
    /**
     * Load the properties specified in given byte array.
     *
     * @param props Null or the Properties object to initialize.
     * @param propertiesFile Null or a properties text file.
     * @return A Properties object that was initialized from the given
     * file.
     */
    public static Properties loadProperties(Properties props,
					    byte propertiesFile[])
    {
	Properties retval = props;
	
	if( retval == null )
	{
	    retval = new Properties();
	}
	if( propertiesFile != null )
	{
	    try
	    {
		retval.load(new ByteArrayInputStream(propertiesFile));
	    }
	    catch(IOException e)
	    {
		Logger.log(e);
	    }
	}
	return retval;
    }
    
    /**
     * Load the properties specified in given byte array.
     *
     * @param propertiesFile Null or a properties text file.
     * @return A Properties object that was initialized from the given
     * file.
     */
    public static Properties loadProperties(byte propertiesFile[])
    {
	return loadProperties(null, propertiesFile);
    }

    /**
     * The properties for this protocol.
     */
    protected final Properties properties;

    /**
     * Construct a ProtocolParameters object initialized with the given values.
     *
     * @param properties The protocol configuration properties.
     */
    public ProtocolParameters(Properties properties)
    {
	Assert.require(properties);
	
	this.properties = properties;
    }

    /**
     * Construct a ProtocolParameters object initialized with the given values.
     *
     * @param propertiesFile A fully loaded properties text file that should be
     * used to initialize a Properties object.
     */
    public ProtocolParameters(byte propertiesFile[])
    {
	this(loadProperties(propertiesFile));
    }

    /**
     * Get a millisecond value out of the properties object.  This tries to be
     * intelligent about the value in the file, for example, if it ends in "s",
     * it will assume it is a seconds value and should be converted to millis.
     *
     * @param name The name of the millisecond property.
     * @param defaultValue The defaultValue to be returned if the value isn't
     * in the properties object or it cannot be parsed.
     * @return The millisecond value specified in the file or defaultValue.
     */
    public long getMillisecondValue(String name, long defaultValue)
    {
	long retval = defaultValue;

	try
	{
	    String str;

	    if( (str = this.properties.getProperty(name)) != null )
	    {
		StringTokenizer st;
		String num;
		
		st = new StringTokenizer(str, " \tumsh");
		num = st.nextToken();
		retval = Long.decode(num).longValue();
		if( str.endsWith("us") )
		{
		    retval = (retval * TimeSpec.MILLI) / TimeSpec.MICRO;
		}
		else if( str.endsWith("ms") )
		{
		}
		else if( str.endsWith("s") )
		{
		    retval = retval * TimeSpec.MILLI;
		}
		else if( str.endsWith("m") )
		{
		    retval = retval * 60 * TimeSpec.MILLI;
		}
		else if( str.endsWith("h") )
		{
		    retval = retval * 60 * 60 * TimeSpec.MILLI;
		}
		else
		{
		    Logger.log("Unknown suffix on millisecond value: " + str);
		    retval = defaultValue;
		}
	    }
	    else
	    {
		Logger.log(Logger.WARNING, "Couldn't find property: " + name);
	    }
	}
	catch(NumberFormatException e)
	{
	    Logger.log(Logger.WARNING, e);
	}
	return retval;
    }

    /**
     * Get an integer value out of the properties object.
     *
     * @param name The name of the integer property.
     * @param defaultValue The defaultValue to be returned if the value isn't
     * in the properties object or it cannot be parsed.
     * @return The integer value specified in the file or defaultValue.
     */
    public int getInteger(String name, int defaultValue)
    {
	int retval = defaultValue;

	try
	{
	    retval = Integer.decode(this.properties.getProperty(name)).
		intValue();
	}
	catch(NullPointerException e)
	{
	    Logger.log(Logger.WARNING, "Couldn't find property: " + name);
	}
	catch(NumberFormatException e)
	{
	    Logger.log(Logger.WARNING, e);
	}
	return retval;
    }

    /**
     * Get an long value out of the properties object.
     *
     * @param name The name of the long property.
     * @param defaultValue The defaultValue to be returned if the value isn't
     * in the properties object or it cannot be parsed.
     * @return The long value specified in the file or defaultValue.
     */
    public long getLong(String name, long defaultValue)
    {
	long retval = defaultValue;

	try
	{
	    retval = Long.decode(this.properties.getProperty(name)).
		longValue();
	}
	catch(NullPointerException e)
	{
	    Logger.log(Logger.WARNING, "Couldn't find property: " + name);
	}
	catch(NumberFormatException e)
	{
	    Logger.log(Logger.WARNING, e);
	}
	return retval;
    }
    
    /**
     * Get an string value out of the properties object.
     *
     * @param name The name of the string property.
     * @param defaultValue The defaultValue to be returned if the value isn't
     * in the properties object or it cannot be parsed.
     * @return The string value specified in the file or defaultValue.
     */
    public String getString(String name, String defaultValue)
    {
	String retval;

	if( (retval = this.properties.getProperty(name)) == null )
	{
	    Logger.log(Logger.WARNING, "Couldn't find property: " + name);
	    retval = defaultValue;
	}
	return retval;
    }

    public String toString()
    {
	return "ProtocolParameters[properties="
	    + this.properties
	    + "]";
    }
}
