/*
 * ProtocolSession.java
 *
 * Copyright (c) 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is part of the Janos Bees Software.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */

package bees.core;

import java.io.IOException;

import java.util.Hashtable;
import java.util.Enumeration;
import java.util.NoSuchElementException;

import edu.utah.janosvm.sys.DeadTeamException;
import edu.utah.janosvm.sys.TeamEjectionException;

import edu.utah.janosvm.kit.comm.amiga.FaxPortHandle;

import edu.utah.janos.nodeos.Domain;

import bees.core.resource.Logger;

import bees.core.bdomain.BDomainLocal;

/**
 * A ProtocolSession is an intermediary between a Protocol and an application.
 * In addition, it can be used to maintain Protocol global state. 
 */
public class ProtocolSession
    implements ExternalCapsuleReceiver
{
    /**
     * The NodeCache for this session.
     */
    private final NodeCache nc;

    /**
     * This session type, basically, the flow context it is operating in.
     */
    private SessionType st;

    /**
     * The set of neighbors added to the protocol.
     */
    protected final Hashtable neighbors = new Hashtable();

    /**
     * Construct ProtocolSession object and a NodeCache.
     */
    public ProtocolSession()
    {
	Node node;

	node = Node.currentNode();
	this.nc = new NodeCache(node, this, 15);
    }

    /**
     * @return The NodeCache for this session.
     */
    public NodeCache getCache()
    {
	return this.nc;
    }

    /**
     * @param st This session type, basically, the flow context it is operating
     * in.
     * @return this
     */
    ProtocolSession setSessionType(SessionType st)
    {
	this.st = st;
	return this;
    }

    /**
     * @return This session type, basically, the flow context it is operating
     * in.
     */
    SessionType getSessionType()
    {
	return this.st;
    }

    /**
     * Get a raw byte array that was added to the Protocol.
     *
     * @param name The name of the byte array as given to the Protocol.
     * @return The byte array matching the given name or null if it couldn't
     * be found.
     */
    public byte[] getByteArray(String name)
    {
	byte retval[] = null;
	ByteArrayType bat;
	
	if( (bat = this.st.getProtocolType().getByteArrayType(name)) != null )
	{
	    retval = bat.getByteArray();
	}
	return retval;
    }

    /**
     * Call back used to attach receivers for this protocol's Capsules.
     *
     * @param node The current Node object.
     * @param neighbor A standard neighbor object, which can be ignored if this
     * protocol requires a special one.
     */
    public void attachNeighbor(Node node, Neighbor neighbor)
    {
	try
	{
	    neighbor.attachReceivers(this);
	}
	catch(IOException e)
	{
	    Logger.log(e);
	}
    }

    /**
     * Call back used to detach receivers for this protocol's Capsules.
     *
     * @param node The current Node object.
     * @param neighbor The neighbor created for or used by this session.
     */
    public void detachNeighbor(Node node, Neighbor neighbor)
    {
	neighbor.detachReceivers(this);
    }

    /**
     * Add a neighbor to this session and any others in this flow.
     *
     * @param node The current Node object.
     * @param ni The NeighborImpl to be added to this flow.
     */
    public void addNeighbor(Node node, NeighborImpl ni)
    {
	synchronized( this.neighbors )
	{
	    if( this.neighbors.containsKey(ni) )
	    {
		ni = null;
	    }
	    else
	    {
		this.neighbors.put(ni, ni);
	    }
	}
	if( ni != null )
	{
	    Hashtable flowSessions;
	    Neighbor neighbor;
	    Enumeration enum;
	    FlowType ft;
	    
	    neighbor = new Neighbor(ni);
	    ft = this.st.getFlowType();
	    flowSessions = (Hashtable)SessionType.getBDomainFlows().
		get(ft.getFlowID());
	    enum = flowSessions.elements();
	    while( enum.hasMoreElements() )
	    {
		SessionType st;
		
		st = (SessionType)enum.nextElement();
		try
		{
		    ProtocolSession ps;
		    
		    ps = st.getSession();
		    ps.attachNeighbor(node, neighbor);
		}
		catch(Throwable th)
		{
		    Logger.log(th);
		}
	    }
	}
    }

    /**
     * Remove a neighbor from this session and any others in this flow.
     *
     * @param node The current Node object.
     * @param ni The NeighborImpl to be removed from this flow.
     */
    public void removeNeighbor(Node node, NeighborImpl ni)
    {
	if( this.neighbors.remove(ni) != null )
	{
	    Hashtable flowSessions;
	    Neighbor neighbor;
	    Enumeration enum;
	    FlowType ft;
	    
	    ft = this.st.getFlowType();
	    flowSessions = (Hashtable)SessionType.getBDomainFlows().
		get(ft.getFlowID());
	    enum = flowSessions.elements();
	    while( enum.hasMoreElements() )
	    {
		SessionType st;
		
		st = (SessionType)enum.nextElement();
		try
		{
		    ProtocolSession ps;
		    
		    ps = st.getSession();
		    if( (neighbor = ni.getNeighbor(st)) != null )
		    {
			ps.detachNeighbor(node, neighbor);
		    }
		}
		catch(Throwable th)
		{
		    Logger.log(th);
		}
	    }
	}
    }

    public void receive(Node node, ExternalCapsule ec)
    {
	ec.getSource().recycleCapsule(ec);
    }

    public String toString()
    {
	return "ProtocolSession[nc="
	    + this.nc
	    + "; st="
	    + this.st
	    + "]";
    }
}
