/*
 * ProtocolType.java
 *
 * Copyright (c) 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is part of the Janos Bees Software.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */

package bees.core;

import java.util.Vector;
import java.util.Hashtable;

import java.io.FileNotFoundException;

import edu.utah.janos.nodeos.Domain;
import edu.utah.janos.nodeos.BufferHandle;

import bees.util.MD5;

import bees.core.resource.Logger;

import bees.core.bdomain.BDomainLocal;

/**
 * A ProtocolType holds the information about a protocol and the Capsules/Code
 * it uses.
 *
 * XXX prolly shouldn't be public
 */
public final class ProtocolType
{
    private static final BDomainLocal domainProtocols = new BDomainLocal() {
	    protected Object initialValue()
	    {
		return new Hashtable();
	    }
	};

    static Hashtable getBDomainProtocols()
    {
	return (Hashtable)domainProtocols.get();
    }

    static void putBDomainProtocol(ProtocolType pt)
    {
	Hashtable ht = (Hashtable)domainProtocols.get();

	Assert.require(pt.protocolID);

	synchronized( ht )
	{
	    if( !ht.containsKey(pt.protocolID) )
		ht.put(pt.protocolID, pt);
	}
    }

    static ProtocolType getBDomainProtocol(TypeID protocolID)
    {
	Hashtable ht = (Hashtable)domainProtocols.get();
	ProtocolType retval;

	retval = (ProtocolType)ht.get(protocolID);
	return retval;
    }

    /**
     * The protocol is not a Companion.
     */
    static final byte COMPANION_NO = 0;
    
    /**
     * The protocol is a companion for ByteArrayTypes.
     */
    static final byte COMPANION_BYTEARRAY = 1;
    
    /**
     * The protocol is a companion for CapsuleTypes.
     */
    static final byte COMPANION_CAPSULE = 2;
    
    /**
     * The protocol is a companion for FlowTypes.
     */
    static final byte COMPANION_FLOW = 3;
    
    /**
     * XXX The protocol is a companion for generic strings?
     */
    static final byte COMPANION_STRING = 4;

    static final byte FLAG_COMPANION_FLOW_INDEPENDENT = (byte)0x80;

    private ProtocolType parent;

    private byte flags;

    /**
     * The ByteArrays used by this protocol.
     */
    private final Vector byteArrayTypes = new Vector();
    
    /**
     * The Companions of this protocol.
     */
    private final Vector companions = new Vector();
    
    /**
     * The type of Companion protocol this is, if any.
     */
    private byte companionType = COMPANION_NO;
    
    /**
     * The name of this protocol's ProtocolSession class.
     */
    private String sessionClassName;

    /**
     * The name of this protocol's boot strap capsule class.
     */
    private String bootStrapClassName;

    /**
     * The unique identifier for this protocol
     */
    private TypeID protocolID;

    /**
     * Construct an empty ProtocolType.
     */
    ProtocolType()
    {
    }

    ProtocolType getParent()
    {
	return this.parent;
    }

    ProtocolType setFlags(byte flags)
    {
	this.flags = flags;
	return this;
    }

    byte getFlags()
    {
	return this.flags;
    }

    /**
     * @param ct The ByteArrayType to add to this protocol.
     * @return this
     */
    synchronized ProtocolType addByteArrayType(ByteArrayType ct)
    {
	this.byteArrayTypes.addElement(ct);
	return this;
    }

    /**
     * @return A Vector of all this protocol's ByteArrayTypes.
     */
    Vector getByteArrayTypes()
    {
	return this.byteArrayTypes;
    }

    synchronized ByteArrayType getByteArrayType(String name)
    {
	ByteArrayType retval = null;
	int lpc, size;

	size = this.byteArrayTypes.size();
	for( lpc = 0; (retval == null) && (lpc < size); lpc++ )
	{
	    ByteArrayType bat;

	    bat = (ByteArrayType)this.byteArrayTypes.elementAt(lpc);
	    if( bat.getByteArrayName().equals(name) )
	    {
		retval = bat;
	    }
	}
	return retval;
    }

    /**
     * @param pt The Companion protocol to add.
     * @return this
     */
    synchronized ProtocolType addCompanion(ProtocolType pt)
    {
	this.companions.addElement(pt);
	pt.parent = this;
	return this;
    }

    synchronized ProtocolType getCompanion(String sessionName)
    {
	ProtocolType retval = null;
	int lpc, size;

	size = this.companions.size();
	for( lpc = 0; (lpc < size) && (retval == null); lpc++ )
	{
	    ProtocolType pt;

	    pt = (ProtocolType)this.companions.elementAt(lpc);
	    if( (pt.sessionClassName != null) &&
		pt.sessionClassName.equals(sessionName) )
	    {
		retval = pt;
	    }
	}
	return retval;
    }

    /**
     * @return A Vector of all this protocol's companion ProtocolTypes.
     */
    Vector getCompanions()
    {
	return this.companions;
    }

    /**
     * @param ct This protocol's companion type.
     * @return this
     */
    ProtocolType setCompanionType(byte ct)
    {
	this.companionType = ct;
	return this;
    }

    /**
     * @return This protocol's companion type.
     */
    byte getCompanionType()
    {
	return this.companionType;
    }

    /**
     * @return True if this ProtocolType describes a companion protocol.
     */
    boolean isCompanion()
    {
	return (this.companionType == COMPANION_NO);
    }

    /**
     * @param sessionClassName The name of the ProtocolSession class for this
     * protocol.
     * @return this
     */
    ProtocolType setSessionClassName(String sessionClassName)
    {
	this.sessionClassName = sessionClassName;
	return this;
    }

    /**
     * @return The name of the ProtocolSession class for this protocol.
     */
    String getSessionClassName()
    {
	return this.sessionClassName;
    }

    /**
     * @param bootStrapClassName The name of the boot strap capsule class for
     * this protocol.
     * @return this
     */
    ProtocolType setBootStrapClassName(String bootStrapClassName)
    {
	this.bootStrapClassName = bootStrapClassName;
	return this;
    }

    /**
     * @return The name of the boot strap capsule class for this protocol.
     */
    String getBootStrapClassName()
    {
	return this.bootStrapClassName;
    }

    /**
     * @param protocolID The unique identifier for this protocol.
     * @return this
     */
    ProtocolType setProtocolID(TypeID protocolID)
    {
	this.protocolID = protocolID;
	return this;
    }

    /**
     * @return The unique identifier for this protocol.
     */
    TypeID getProtocolID()
    {
	return this.protocolID;
    }

    void loadByteArrays()
	throws ClassNotFoundException,
	       FileNotFoundException
    {
	Domain domain = Domain.currentDomain();
	int lpc, size;

	size = this.companions.size();
	for( lpc = 0; lpc < size; lpc++ )
	{
	    ProtocolType pt;
	    
	    pt = (ProtocolType)this.companions.elementAt(lpc);
	    pt.loadByteArrays();
	}
	size = this.byteArrayTypes.size();
	for( lpc = 0; lpc < size; lpc++ )
	{
	    ByteArrayType bat;
			
	    bat = (ByteArrayType)this.byteArrayTypes.elementAt(lpc);
	    switch( bat.getState() )
	    {
	    case ByteArrayType.STATE_NEW:
		if( bat instanceof CodeType )
		{
		    CodeType ct;
		    
		    ct = (CodeType)bat;
		    ct.importClass(
			domain.readByteCode(bat.getByteArrayName()));
		}
		else
		{
		    bat.importByteArray(
			domain.readClassPathFile(bat.getByteArrayName()));
		}
		break;
	    default:
		break;
	    }
	}
    }

    /**
     * Compute/verify this protocol's unique identifier.  This object should be
     * completely filled out, with the possible exception of the ProtocolID
     * before calling this method.
     */
    void snapshotProtocolType()
    {
	TypeID computedProtocolID;
	MD5 digest = new MD5();
	int lpc, size;

	/* Cover the ByteArrayTypes. */
	size = this.byteArrayTypes.size();
	for( lpc = 0; lpc < size; lpc++ )
	{
	    ByteArrayType bat;

	    bat = (ByteArrayType)this.byteArrayTypes.elementAt(lpc);
	    digest.update(bat.getByteArrayID().getHash());
	}
	/* Cover the companion ProtocolTypes. */
	size = this.companions.size();
	for( lpc = 0; lpc < size; lpc++ )
	{
	    ProtocolType pt = (ProtocolType)this.companions.elementAt(lpc);

	    digest.update(pt.getProtocolID().getHash());
	}
	computedProtocolID = new TypeID(digest.hash());
	if( this.protocolID != null )
	{
	    /* Verify mode. */
	    if( !this.protocolID.equals(computedProtocolID) )
	    {
		throw new SecurityException("Computed ProtocolID "
					    + computedProtocolID
					    + " != "
					    + this.protocolID
					    + " for "
					    + this);
	    }
	}
	else
	{
	    /* Initial construction. */
	    this.protocolID = computedProtocolID;
	}
    }

    /**
     * @return The marshalled length of the type.
     */
    synchronized int marshalledLength()
    {
	int lpc, size, retval;

	retval = Xdr.INT;
	size = this.byteArrayTypes.size();
	for( lpc = 0; lpc < size; lpc++ )
	{
	    ByteArrayType bat;

	    bat = (ByteArrayType)this.byteArrayTypes.elementAt(lpc);
	    retval += bat.marshalledLength();
	}
	retval += Xdr.INT;
	size = this.companions.size();
	for( lpc = 0; lpc < size; lpc++ )
	{
	    ProtocolType pt = (ProtocolType)this.companions.elementAt(lpc);
	    
	    retval += pt.marshalledLength();
	}
	retval += Xdr.BYTE;
	retval += Xdr.BYTE;
	retval += Xdr.STRING(this.sessionClassName);
	retval += Xdr.STRING(this.bootStrapClassName);
	retval += Xdr.TYPEID;
	return retval;
    }

    /**
     * @param xdr The object to marshal this type to.
     */
    synchronized void marshal(XdrByteArray xdr)
    {
	int lpc, size;

	size = this.byteArrayTypes.size();
	xdr.PUT(size);
	for( lpc = 0; lpc < size; lpc++ )
	{
	    ByteArrayType bat;

	    bat = (ByteArrayType)this.byteArrayTypes.elementAt(lpc);
	    bat.marshal(xdr);
	}
	size = this.companions.size();
	xdr.PUT(size);
	for( lpc = 0; lpc < size; lpc++ )
	{
	    ProtocolType pt = (ProtocolType)this.companions.elementAt(lpc);
	    
	    pt.marshal(xdr);
	}
	xdr.PUT(this.flags);
	xdr.PUT(this.companionType);
	xdr.PUT(this.sessionClassName);
	xdr.PUT(this.bootStrapClassName);
	xdr.PUT(this.protocolID);
    }

    /**
     * @param xdr The object to unmarshal a type from.
     */
    static ProtocolType unmarshal(XdrByteArray xdr)
    {
	ProtocolType retval = new ProtocolType();
	int lpc, size;

	size = xdr.INT();
	for( lpc = 0; lpc < size; lpc++ )
	{
	    retval.addByteArrayType(ByteArrayType.instantiate(xdr));
	}
	size = xdr.INT();
	for( lpc = 0; lpc < size; lpc++ )
	{
	    ProtocolType pt = ProtocolType.unmarshal(xdr);

	    retval.addCompanion(pt);
	}
	retval.flags = xdr.BYTE();
	retval.companionType = xdr.BYTE();
	retval.sessionClassName = xdr.STRING();
	retval.bootStrapClassName = xdr.STRING();
	retval.protocolID = xdr.TYPEID();
	return retval;
    }

    /**
     * @return A flattened version of this type.
     */
    byte[] toBytes()
    {
	XdrByteArray xdr;

	xdr = new XdrByteArray(this.marshalledLength());
	this.marshal(xdr);
	if( xdr.index != xdr.len )
	    new Panic("marshalledLength() doesn't match actual length");
	return xdr.toBytes();
    }

    /**
     * @param bytes A flattened ProtocolType object.
     * @return A new ProtocolType object initialized with the values in
     * "bytes".
     */
    public static ProtocolType instantiate(byte bytes[])
    {
	XdrByteArray xdr = new XdrByteArray(bytes);
	ProtocolType retval;
	
	retval = unmarshal(xdr);
	if( xdr.index != xdr.len )
	    throw new IllegalArgumentException("Flattened Protocol is short");
	return retval;
    }
    
    public String toString()
    {
	return "ProtocolType[flags="
	    + this.flags
	    + "; byteArrayTypes="
	    + this.byteArrayTypes
	    + "; companions="
	    + this.companions
	    + "; companionType="
	    + this.companionType
	    + "; sessionClassName="
	    + this.sessionClassName
	    + "; bootStrapClassName="
	    + this.bootStrapClassName
	    + "; protocolID="
	    + this.protocolID
	    + "]";
    }
}
