/*
 * RedundantNeighborhood.java
 *
 * Copyright (c) 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is part of the Janos Bees Software.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */

package bees.core;

import java.io.IOException;

import edu.utah.janos.nodeos.BufferHandle;

import bees.core.security.PermissionException;

/**
 * A RedundantNeighborhood is a Neighborhood that attempts to send a single
 * Capsule through one of its neighbors.
 */
public final class RedundantNeighborhood
    extends Neighborhood
{
    /**
     * Construct an empty RedundantNeighborhood.
     */
    public RedundantNeighborhood()
    {
    }
    
    /**
     * Attempt to send a capsule to one of the member neighbors.  If the
     * neighborhood is able to send the Capsule with the neighbor throwing an
     * exception, the operation will be declared a success and return control.
     * Note: If the given Capsule implements ReencodableCapsule it will be
     * reencod()'ed before it is sent to the neighbor.
     *
     * @param cap The Capsule to send.
     * @throws EncodingException if there is a problem encoding the given
     * Capsule.
     * @throws IOException if there are no reachable neighbors.
     */
    public synchronized void send(Capsule cap)
	throws EncodingException,
	       IOException
    {
	ReencodableCapsule rc = null;
	NeighborImpl source = null;
	boolean success = false;
	BufferHandle bh;
	int lpc, size;

	if( cap.getSource() != null )
	{
	    source = cap.getSource().getNeighborImpl();
	}
	cap.getPacketAuthorization().
	    consume(PacketAuthorization.VALUE_NEIGHBORHOOD);
	if( cap instanceof ReencodableCapsule )
	{
	    rc = (ReencodableCapsule)cap;
	}
	
	super.node.used();
	bh = this.ce.encode(cap);
	size = this.neighbors.size();
	for( lpc = 0; (lpc < size) && !success; lpc++ )
	{
	    NeighborImpl ni = (NeighborImpl)this.neighbors.elementAt(lpc);

	    try
	    {
		if( ni == source )
		{
		    cap.getPacketAuthorization().
			consume(PacketAuthorization.VALUE_RETURNS);
		}
		if( rc != null )
		{
		    this.ce.reencode(rc, ni);
		}
		ni.sendBuffer(bh);
		success = true;
	    }
	    catch(PermissionException e)
	    {
		this.handleSendException(e);
	    }
	    catch(SecurityException e)
	    {
		this.neighbors.remove(ni);
		lpc -= 1;
		size -= 1;
		this.handleRevokedNeighbor(ni);
	    }
	    catch(IOException e)
	    {
		this.handleSendException(e);
	    }
	}
	if( !success )
	{
	    throw new IOException("No reachable neighbors");
	}
    }
    
    /**
     * @see java.lang.Object#toString()
     */
    public String toString()
    {
	return "RedundantNeighborhood["
	    + super.toString()
	    + "]";
    }
}
