/*
 * SignedXdrByteArray.java
 *
 * Copyright (c) 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is part of the Janos Bees Software.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */

package bees.core;

import java.security.SignatureException;
import java.security.InvalidKeyException;
import java.security.NoSuchAlgorithmException;

import bees.core.resource.Clock;

import bees.core.security.Signature;
import bees.core.security.PublicKeySpec;
import bees.core.security.PrivateKeySpec;

/**
 * A SignedXdrByteArray is a convenience for signing/verifying an XdrByteArray.
 *
 * @see bees.core.security.Signature
 */
public class SignedXdrByteArray
    extends XdrByteArray
{
    /**
     * The length of an encoded TIMESTAMP element.
     */
    public static final int TIMESTAMP = Xdr.LONG;

    /**
     * The name of the cryptographic algorithm to use.
     */
    private String algorithm;
    
    /**
     * The signature value.
     */
    private byte signature[];

    /**
     * Construct a SignedXdrByteArray with the given values.
     *
     * @param len The length of the byte array.
     * @param algorithm The name of the cryptographic algorithm to use.
     */
    public SignedXdrByteArray(int len, String algorithm)
    {
	super(len);

	Assert.require(algorithm);

	this.algorithm = algorithm;
    }

    /**
     * Construct a SignedXdrByteArray with the given values.
     *
     * @param buf The byte array to wrap.
     * @param algorithm The name of the cryptographic algorithm to use.
     */
    public SignedXdrByteArray(byte buf[], String algorithm)
    {
	super(buf);
	
	Assert.require(algorithm);

	this.algorithm = algorithm;
    }

    /**
     * Construct a SignedXdrByteArray with the given values.
     *
     * @param buf The byte array to wrap.
     * @param signature The signature covering "buf".
     * @param algorithm The name of the cryptographic algorithm to use.
     */
    public SignedXdrByteArray(byte buf[], byte signature[], String algorithm)
    {
	super(buf);
	
	Assert.require(algorithm);

	this.signature = signature;
	this.algorithm = algorithm;
    }

    /**
     * @return The name of the cryptographic algorithm to use.
     */
    public String getAlgorithm()
    {
	return this.algorithm;
    }

    /**
     * @return The signature covering "buf".
     */
    public byte[] getSignature()
    {
	return this.signature;
    }

    /**
     * @param pk The PrivateKey to use when signing this byte array.
     * @param ps A ProtocolSession from the flow this signature should be bound
     * to or null if this domain is an application.
     * @throws SecurityException if "ps" is null and this domain is not an
     * application.
     * @throws SignatureException if there was a problem signing the byte
     * array.
     * @throws InvalidKeyException if the given key is invalid.
     * @throws NoSuchAlgorithmException if the algorithm given in the
     * constructor is invalid.
     */
    public void sign(PrivateKeySpec pk, ProtocolSession ps)
	throws SecurityException,
	       SignatureException,
	       InvalidKeyException,
	       NoSuchAlgorithmException
    {
	Signature signer;

	Assert.require(pk);

	signer = new Signature(this.algorithm);
	signer.initSign(pk, ps);
	signer.update(super.buf);
	this.signature = signer.sign();
    }

    /**
     * @param pk The PrivateKey to use when signing this byte array.
     * @throws SecurityException if this domain is not an application.
     * @throws SignatureException if there was a problem signing the byte
     * array.
     * @throws InvalidKeyException if the given key is invalid.
     * @throws NoSuchAlgorithmException if the algorithm given in the
     * constructor is invalid.
     */
    public void sign(PrivateKeySpec pk)
	throws SecurityException,
	       SignatureException,
	       InvalidKeyException,
	       NoSuchAlgorithmException
    {
	this.sign(pk, null);
    }
    
    /**
     * @param pk The PublicKeySpec to use when verifying this byte array.
     * @param ps A ProtocolSession from the flow this signature was bound to or
     * null if it was not bound to a flow.
     * @return True if the signature verified against the byte array.
     * @throws SignatureException if there was a problem signing the byte
     * array.
     * @throws InvalidKeyException if the given key is invalid.
     * @throws NoSuchAlgorithmException if the algorithm given in the
     * constructor is invalid.
     */
    public boolean verify(PublicKeySpec pk, ProtocolSession ps)
	throws SignatureException,
	       InvalidKeyException,
	       NoSuchAlgorithmException
    {
	boolean retval = false;
	
	if( this.signature != null )
	{
	    Signature verifier;
	    
	    Assert.require(pk);
	    
	    verifier = new Signature(this.algorithm);
	    verifier.initVerify(pk, ps);
	    verifier.update(super.buf);
	    retval = verifier.verify(this.signature);
	}
	return retval;
    }
    
    /**
     * @param pk The PublicKeySpec to use when verifying this byte array.
     * @return True if the signature verified against the byte array.
     * @throws SignatureException if there was a problem signing the byte
     * array.
     * @throws InvalidKeyException if the given key is invalid.
     * @throws NoSuchAlgorithmException if the algorithm given in the
     * constructor is invalid.
     */
    public boolean verify(PublicKeySpec pk)
	throws SignatureException,
	       InvalidKeyException,
	       NoSuchAlgorithmException
    {
	return this.verify(pk, null);
    }

    public void PUT(Clock clock)
    {
	super.PUT(clock.current());
    }

    public boolean TIMESTAMP(Clock clock, long tolerance)
    {
	boolean retval = false;
	long time, current;

	if( tolerance < 0 )
	{
	    throw new IllegalArgumentException("Negative tolerance value.");
	}
	current = clock.current();
	time = super.LONG() - current;
	if( (-tolerance < time) && (time < tolerance) )
	{
	    retval = true;
	}
	return retval;
    }

    public String toString()
    {
	return "SignedXdrByteArray["
	    + super.toString()
	    + "; algorithm="
	    + this.algorithm
	    + "; signature="
	    + ByteArray.toString(this.signature)
	    + "]";
    }
}
