/*
 * Strict.java
 *
 * Copyright (c) 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is part of the Janos Bees Software.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */

package bees.core;

import edu.utah.janos.nodeos.Node;

import bees.core.kernel.KernelPanic;

/**
 * The Strict class provides some static functions for checking state that
 * can vary depending on how strict the checking should be.  For example, if
 * something is very unlikely to happen, but still possible, we can't use an
 * assert statement or similar hard check.  However, leaving a condition
 * unchecked might mask a real problem.  So, we define a few "strictness" modes
 * and handle failures depending on the current mode.
 */
public final class Strict
{
    /**
     * Development mode, throw a Panic if a test fails.
     */
    private static final int MODE_DEVELOPMENT = 0;

    /**
     * Testing mode, print a warning if a test fails.
     */
    private static final int MODE_TESTING = 1;

    /**
     * Deployed mode, ignore the failure.  XXX Might want to log it...
     */
    private static final int MODE_DEPLOYED = 2;

    /**
     * String names of all the modes.
     */
    private static final String modeNames[] = {
	"development",
	"testing",
	"deployed"
    };

    /**
     * Default to development mode.
     */
    private static final int mode = MODE_DEVELOPMENT;

    static
    {
	if( mode != MODE_DEPLOYED )
	{
	    Node.println("Strict mode: " + modeNames[mode]);
	}
    }

    /**
     * Test a value.
     *
     * @param value The value to check, if false, the action taken will depend
     * on the mode.
     */
    public static void test(boolean value)
    {
	switch( mode )
	{
	case MODE_DEVELOPMENT:
	    if( !value )
	    {
		throw new Panic("Strict test failed");
	    }
	    break;
	case MODE_TESTING:
	    if( !value )
	    {
		Node.println("Warning: Strict test failed");
		Node.printStackTrace();
	    }
	    break;
	case MODE_DEPLOYED:
	    /* Ignore */
	    break;
	}
    }

    /**
     * Unlikely code is being executed, fail depending on the mode.
     */
    public static void fail()
    {
	Strict.test(false);
    }

    /**
     * Unlikely code is being executed, fail depending on the mode and print
     * a message.
     *
     * @param msg A description of the problem.
     */
    public static void fail(String msg)
    {
	switch( mode )
	{
	case MODE_TESTING:
	case MODE_DEVELOPMENT:
	    Node.println(msg);
	    break;
	case MODE_DEPLOYED:
	    /* Ignore */
	    break;
	}
	Strict.test(false);
    }

    /**
     * Private constructor, this is a static only class.
     */
    private Strict()
    {
    }
}
