/*
 * UnknownCapsuleReceiver.java
 *
 * Copyright (c) 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is part of the Janos Bees Software.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */

package bees.core;

import java.io.IOException;

import java.util.Hashtable;

import edu.utah.janosvm.sys.List;

import edu.utah.janos.nodeos.Domain;
import edu.utah.janos.nodeos.InChannel;
import edu.utah.janos.nodeos.OutChannel;
import edu.utah.janos.nodeos.DeliverFunc;
import edu.utah.janos.nodeos.BufferHandle;
import edu.utah.janos.nodeos.DemultiplexKey;

import bees.util.MD5;

import bees.core.anep.ANEPVersion;

import bees.core.resource.Logger;

import bees.core.security.Permission;
import bees.core.security.PermissionSet;
import bees.core.security.CountPermission;

class UnknownCapsuleReceiver
    extends CapsuleReceiver
{
    private static TypeID unknownID = new TypeID(new byte[MD5.HASH_LENGTH]);
    
    static long ID_TIMEOUT = 30 * 1000;
    static int MAX_IDS = 64;

    private final Node node = Node.currentNode();
    private final Hashtable recentIDs = new Hashtable();
    private final List lruID = new List();
    private final UnknownCapsule uc;
    private final byte hash[] = new byte[MD5.HASH_LENGTH];
    private final TypeID capsuleID = new TypeID();
    private final Xdr xdr = new Xdr();
    private final Neighbor neighbor;
    private final Capsule cap;
    
    UnknownCapsuleReceiver(Neighbor neighbor, Capsule cap, ProtocolSession ps)
    {
	super(neighbor, null, ps);
	
	this.uc = new UnknownCapsule();
	this.uc.fromCapsule(cap);
	this.uc.setPacketAuthorization(new PacketAuthorization(
		new PermissionSet(new Permission[] {
		})));
	this.uc.setInternalPacketAuthorization(new InternalPacketAuthorization(
		new PermissionSet(new Permission[] {
		    new CountPermission("transformTo."
					+ "bees.core.IDTranslatorCapsule", 1),
		})));
	this.neighbor = neighbor;
	this.capsuleID.setHash(this.hash);
	this.cap = cap;
    }
    
    public String getCapsuleName()
    {
	return "unknown";
    }

    public TypeID getCapsuleID()
    {
	return unknownID;
    }

    public TypeID getSpecializedCapsuleID()
    {
	return null;
    }

    public int getThreadCount()
    {
	return 1;
    }

    public int getBufferCount()
    {
	return 5;
    }

    public DemultiplexKey createDemultiplexKey()
    {
	Domain domain = Domain.currentDomain();
	DemultiplexKey retval;

	retval = new DemultiplexKey(domain);
	retval.addANEP(0,
		       ANEPVersion.PACKET_VERSION,
		       BeesVersion.ANEP_TYPE);
	retval.add(BeesVersion.CAPSULE_VERSION_BYTES.length,
		   BeesVersion.CAPSULE_VERSION_BYTES,
		   null);
	return retval;
    }

    private void flushLeastRecentlyUsed(int count)
    {
	int lpc;

	Assert.require(count > 0);

	count = this.recentIDs.size() + count - MAX_IDS;
	for( lpc = 0; lpc < count; lpc++ )
	{
	    TranslatedID lru;
	    
	    lru = (TranslatedID)this.lruID.remHead();
	    this.recentIDs.remove(lru.getCapsuleID());
	}
    }

    private synchronized boolean isRecent(TypeID capsuleID)
    {
	TranslatedID tid = (TranslatedID)this.recentIDs.get(capsuleID);
	boolean retval = false;

	if( tid == null )
	{
	    /* New value. */
	    this.flushLeastRecentlyUsed(1);
	    tid = new TranslatedID().
		setCapsuleID(capsuleID).
		setExpiration(this.node.time() + ID_TIMEOUT);
	    this.recentIDs.put(tid.getCapsuleID(), tid);
	}
	else if( node.time() > tid.getExpiration() )
	{
	    /* Expired, but still in the cache. */
	    tid.setExpiration(this.node.time() + ID_TIMEOUT);
	    tid.remove();
	    retval = false;
	}
	else
	{
	    tid.remove();
	    retval = true;
	}
	this.lruID.addTail(tid);
	return retval;
    }

    synchronized void cacheRecentTranslation(TypeID capsuleID, TypeID flowID)
    {
	TranslatedID tid;

	if( (tid = (TranslatedID)this.recentIDs.get(capsuleID)) != null )
	{
	    tid.setFlowID(flowID);
	}
	else
	{
	    /*
	     * The TranslatedID added by isRecent has fallen out...
	     *
	     * XXX Should we create a new tid here?
	     */
	}
    }

    public Capsule receive(BufferHandle bh, int offset)
    {
	try
	{
	    this.xdr.readReset(bh);
	    this.xdr.SKIP(offset + Xdr.INT);
	    bh.getBytes(this.xdr.getReadCursor(),
			this.hash, 0,
			MD5.HASH_LENGTH);
	    this.xdr.SKIP(MD5.HASH_LENGTH);
	    this.uc.fromNeighbor(this.neighbor);
	}
	catch(Throwable th)
	{
	    Logger.log(th);
	}
	return this.uc;
    }

    public void evaluate(Capsule cap)
    {
	if( this.isRecent(this.capsuleID) )
	{
	    Logger.log("Suppress: " + this.capsuleID);
	}
	else
	{
	    this.cap.fromCapsule(this.uc, this.capsuleID);
	    try
	    {
		Logger.log("Asking for mapping! " + this.capsuleID);
		this.neighbor.send(this.cap);
	    }
	    catch(EncodingException e)
	    {
		Logger.log(e);
	    }
	    catch(IOException e)
	    {
		Logger.log(e);
	    }
	}
    }

    public String toString()
    {
	return "UnknownCapsuleReceiver[recentIDs="
	    + this.recentIDs
	    + "; lruID="
	    + this.lruID
	    + "; hash="
	    + ByteArray.toString(this.hash)
	    + "; capsuleID="
	    + this.capsuleID
	    + "; xdr="
	    + this.xdr
	    + "; cap="
	    + this.cap
	    + "]";
    }
}
