/*
 * EventHandler.java
 *
 * Copyright (c) 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is part of the Janos Bees Software.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */

package bees.core.bdomain;

import java.util.Enumeration;

import edu.utah.janosvm.kit.comm.amiga.FaxPort;
import edu.utah.janosvm.kit.comm.amiga.Facsimile;
import edu.utah.janosvm.kit.comm.amiga.SignalSet;
import edu.utah.janosvm.kit.comm.amiga.FaxPortHandle;

import edu.utah.janosvm.kit.comm.jsi.JSIPair;
import edu.utah.janosvm.kit.comm.jsi.JSIHandler;
import edu.utah.janosvm.kit.comm.jsi.JSIMessage;
import edu.utah.janosvm.kit.comm.jsi.JSIPropertyInfo;
import edu.utah.janosvm.kit.comm.jsi.JSIHandlerContext;

import edu.utah.janos.nodeos.Domain;
import edu.utah.janos.nodeos.Node;
import edu.utah.janos.nodeos.ThreadLocal;

import bees.core.Panic;
import bees.core.Assert;
import bees.core.Strict;
import bees.core.BeesEvent;
import bees.core.Coroutine;
import bees.core.Capability;
import bees.core.ExternalCapsule;
import bees.core.ProtocolSession;
import bees.core.ProtocolPresentation;

import bees.core.kernel.KernelPanic;
import bees.core.kernel.BDomainEndingEvent;

/**
 * A class for handling events received by a domain.
 */
public final class EventHandler
    implements Runnable
{
    /**
     * A ThreadLocal which holds a pointer to the EventHandler's port.
     */
    private static final ThreadLocal threadPort = new ThreadLocal();

    /**
     * @return The current EventHandler's port.
     */
    public static FaxPort getHandlerPort()
    {
	FaxPort retval = null;

	try
	{
	    retval = (FaxPort)threadPort.get();
	}
	catch(NullPointerException e)
	{
	}
	return retval;
    }

    public static boolean isEventHandler()
    {
	return getHandlerPort() != null;
    }
    
    /**
     * The SignalSet the handler will block on.
     */
    private final SignalSet ss = new SignalSet();

    /**
     * The FaxPort used to send/receive events.
     */
    private final FaxPort port = new FaxPort(this.ss);

    /**
     * The signal number of the exit signal.
     */
    private final int exitSignal;
    
    /**
     * Construct an EventHandler.
     */
    public EventHandler(BDomain bdomain)
    {
	Assert.require(bdomain);
	
	this.exitSignal = this.ss.allocate();
	bdomain.pushCleanupHandler(new Runnable() {
		public void run()
		{
		    EventHandler.this.ss.signal(EventHandler.this.exitSignal);
		}
	    });
    }

    /**
     * @return The port the handler listens on.
     */
    public FaxPort getPort()
    {
	return this.port;
    }

    /**
     * Handle an event received by the looper.
     *
     * @param ae The received event.
     * @return True if the looper should stop waiting for events, false
     * otherwise.
     */
    public boolean handleEvent(BeesEvent ae)
    {
	boolean retval = false;
	Coroutine coroutine;

	if( (coroutine = Coroutine.getCoroutine(ae)) != null )
	{
	    coroutine.handle(ae);
	}
	else if( !(ae instanceof BDomainEndingEvent) )
	{
	    Strict.fail("Stale event: " + ae);
	}
	return retval;
    }

    public void run()
    {
	try
	{
	    this.realRun();
	}
	catch(InterruptedException e)
	{
	    /* Ignore... */
	}
	catch(Throwable th)
	{
	    Node.println("EventHandler died");
	    Node.printStackTrace(th);
	}
    }

    /**
     * The main looper function.
     */
    private void realRun()
	throws InterruptedException
    {
	Domain domain = Domain.currentDomain();
	BDomain bdomain = BDomain.currentBDomain();
	long waitMask, recvMask;
	boolean done = false;
	Facsimile fax;

	threadPort.set(this.port);
	waitMask = ((1L << this.exitSignal) |
		    (1L << this.port.getSignalNumber()));
	while( !done )
	{
	    recvMask = this.ss.waitFor(waitMask, 15 * 1000);
	    if( ((recvMask & (1L << this.exitSignal)) != 0) ||
		domain.destroyed() /* XXX */ )
	    {
		done = true;
	    }
	    else if( (recvMask & (1L << this.port.getSignalNumber())) != 0 )
	    {
		while( (fax = this.port.getFacsimile()) != null )
		{
		    if( fax instanceof BeesEvent )
		    {
			try
			{
			    this.handleEvent((BeesEvent)fax);
			}
			catch(ThreadDeath td)
			{
			    done = true;
			}
			catch(Panic panic)
			{
			    Node.println("Internal error! " + domain);
			    Node.printStackTrace(panic);
			    done = true;
			}
			catch(Throwable th)
			{
			    Node.println("Exception while handling event: "
					 + fax);
			    Node.printStackTrace(th);
			}
		    }
		    else
		    {
			Node.println("Received unknown fax type: " + fax);
		    }
		}
	    }
	}
	this.port.close();
    }

    public String toString()
    {
	return "EventHandler[ss="
	    + this.ss
	    + "; port="
	    + this.port
	    + "; exitSignal="
	    + this.exitSignal
	    + "]";
    }
}
