/*
 * BDomainCoroutine.java
 *
 * Copyright (c) 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is part of the Janos Bees Software.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */

package bees.core.kernel;

import java.util.Vector;
import java.util.Hashtable;
import java.util.Enumeration;
import java.util.NoSuchElementException;

import edu.utah.janosvm.sys.DeadTeamException;
import edu.utah.janosvm.sys.TeamEjectionException;

import edu.utah.janosvm.kit.comm.amiga.FaxPort;
import edu.utah.janosvm.kit.comm.amiga.Facsimile;
import edu.utah.janosvm.kit.comm.amiga.FaxPortHandle;

import edu.utah.janosvm.kit.comm.commspace.CommSpaceHandle;
import edu.utah.janosvm.kit.comm.commspace.CommSpaceElement;

import edu.utah.janos.nodeos.Node;
import edu.utah.janos.nodeos.Domain;
import edu.utah.janos.nodeos.MemSpec;
import edu.utah.janos.nodeos.ChanStats;
import edu.utah.janos.nodeos.DomainStats;
import edu.utah.janos.nodeos.ThreadStats;
import edu.utah.janos.nodeos.DomainHandle;
import edu.utah.janos.nodeos.Credentials;
import edu.utah.janos.nodeos.DomainClassSpec;
import edu.utah.janos.nodeos.DomainStartSpec;
import edu.utah.janos.nodeos.ThreadPoolSpec;
import edu.utah.janos.nodeos.DomainStartException;

import bees.core.Assert;
import bees.core.Strict;
import bees.core.BeesEvent;
import bees.core.Coroutine;
import bees.core.Capability;
import bees.core.StateResponder;
import bees.core.CapabilityFactory;
import bees.core.ResponderCoroutine;
import bees.core.StateResponderList;
import bees.core.RevokeCapabilityCoroutine;

import bees.core.resource.Meter;
import bees.core.resource.Logger;

import bees.core.security.Permission;
import bees.core.security.PermissionSet;
import bees.core.security.CountPermission;
import bees.core.security.PermissionException;

import bees.core.bdomain.DeadBDomainException;

/**
 * The kernel object that refers to the actual domain.
 */
public final class BDomainCoroutine
    extends ResponderCoroutine
{
    /**
     * Array of package names that are considered part of the Bees system.
     */
    private static final String beesPackages[] = {
	"bees.core",
	"bees.core.kernel",
	"bees.core.resource",
	"bees.core.security",
	"bees.core.bdomain",
	"bees.util",
	
	"edu.utah.janosvm.kit.comm.jsi",
    };
    
    /**
     * Array of classes that should be exported to downloaded code. Note
     * that Bees applications are not subject to this restricted list
     * and may access any visible class.  We really should provide (looser)
     * restrictions on applications, though.
     */
    private static final String exportedBeesClasses[] = {
	/* bees.core.* */
	"bees.core.Assert",
	"bees.core.Strict",
	"bees.core.BeesVersion",
	"bees.core.ByteArray",
	"bees.core.ByteArrayRequestCapsule",
	"bees.core.ByteArrayResponseCapsule",
	"bees.core.Capsule",
	"bees.core.CapsuleCompanionProtocol",
	"bees.core.CapsuleGroup",
	"bees.core.Capability",
	"bees.core.DLProtocol",
	"bees.core.DLProtocolSession",
	"bees.core.EncodingException",
	"bees.core.EventChannel",
	"bees.core.ExtendedProtocolSession",
	"bees.core.ExternalCapsule",
	"bees.core.ExternalCapsuleEvent",
	"bees.core.ExternalChannel",
	"bees.core.FlowIdentifier",
	"bees.core.FlowRequestCapsule",
	"bees.core.FlowResponseCapsule",
	"bees.core.IDTranslationProtocol",
	"bees.core.IDTranslatorBootCapsule",
	"bees.core.IDTranslatorCapsule",
	"bees.core.IDTranslatorSession",
	"bees.core.MulticastNeighborhood",
	"bees.core.MissingCapabilityException",
	"bees.core.Neighbor",
	"bees.core.Neighborhood",
	"bees.core.NeighborImpl",
	"bees.core.Node",
	"bees.core.NodeCache",
	"bees.core.NodeCache$EventHandler",
	"bees.core.ObjectMap",
	"bees.core.ObjectTree",
	"bees.core.Pair",
	"bees.core.PairChain",
	"bees.core.Panic",
	"bees.core.Protocol",
	"bees.core.ProtocolParameters",
	"bees.core.ProtocolSession",
	"bees.core.ProtocolBDomain",
	"bees.core.RedundantNeighborhood",
	"bees.core.ReencodableCapsule",
	"bees.core.SignedXdrByteArray",
	"bees.core.TypeID",
	"bees.core.UnknownCapsule",
	"bees.core.UnknownProtocol",
	"bees.core.UnknownProtocolSession",
	"bees.core.Xdr",
	"bees.core.XdrByteArray",

	/* bees.core.anep.* */
	"bees.core.anep.ANEPVersion",

	/* bees.core.resource.* */
	"bees.core.resource.Clock",
	"bees.core.resource.Logger",
	"bees.core.resource.SystemClass",
	"bees.core.resource.SystemPackage",

	/* bees.core.security.* */
	"bees.core.security.KeySpec",
	"bees.core.security.CountPermission",
	"bees.core.security.IntegerPermission",
	"bees.core.security.MinimumIntegerPermission",
	"bees.core.security.MaximumIntegerPermission",
	"bees.core.security.RangePermission",
	"bees.core.security.Permission",
	"bees.core.security.PermissionSet",
	"bees.core.security.PermissionException",
	"bees.core.security.PrivateKey",
	"bees.core.security.PrivateKeySpec",
	"bees.core.security.PublicKey",
	"bees.core.security.PublicKeySpec",
	"bees.core.security.Signature",
	"bees.core.security.VersionString",

	/* bees.util.* */
	"bees.util.MD5",

	/*
	 * Several miscellaneous JNodeOS classes that *protocols*
	 * should be able to see.
	 *
	 * XXX Are these really needed anymore?
	 */
	"edu.utah.janos.nodeos.Address",
	"edu.utah.janos.nodeos.AddressFormatException",
	"edu.utah.janos.nodeos.NoSuchNodeAddress",

	/* edu.utah.janosvm.kit.comm.jsi.* */
	"edu.utah.janosvm.kit.comm.jsi.JSIPair",
	"edu.utah.janosvm.kit.comm.jsi.JSIHandler",
	"edu.utah.janosvm.kit.comm.jsi.JSIMessage",
	"edu.utah.janosvm.kit.comm.jsi.JSIPropertyInfo",
	"edu.utah.janosvm.kit.comm.jsi.JSIHandlerContext",
	"edu.utah.janosvm.kit.comm.jsi.MissingPairException",
    };

    private static final DomainClassSpec minClassSpec = new DomainClassSpec();
    
    static {
	Domain domain = Domain.currentDomain();
	int lpc;

	minClassSpec.exportServerClasses("bees");
	
	minClassSpec.absorb(DomainClassSpec.getDefault());
	minClassSpec.addServerName("bees");
	for( lpc = 0; lpc < beesPackages.length; lpc++ )
	{
	    minClassSpec.addSystemPackage(beesPackages[lpc]);
	    minClassSpec.sealPackage(beesPackages[lpc]);
	}
	/* Export classes needed by the protocols. */
	for( lpc = 0; lpc < exportedBeesClasses.length; lpc++ )
	{
	    minClassSpec.exportClass(exportedBeesClasses[lpc]);
	    try
	    {
		domain.loadClass(exportedBeesClasses[lpc]);
	    }
	    catch(ClassNotFoundException e)
	    {
		Node.println("Warning: Cannot find exported class "
			     + exportedBeesClasses[lpc]);
		Node.printStackTrace(e);
	    }
	}
    }

    private static final String emptyBDomainArgs[] = new String[0];

    /**
     * The default MemSpec for the domain if one isn't specified.
     */
    private static final MemSpec defaultMemSpec = new MemSpec(1024 * 1024);

    /**
     * The default ThreadPoolSpec for the domain if one isn't specified.
     */
    private static final ThreadPoolSpec DEFAULT_THREADPOOL_SPEC =
	new ThreadPoolSpec(3,
			   32 * 1024,
			   (int)(ThreadPoolSpec.getMaxRate() / 100),
			   null);

    private static final int STACK_SIZE = 32 * 1024;
    private static final int PAGE_SIZE = 4 * 1024;

    public static final int STATE_CREATED = STATE_USER;
    public static final int STATE_CONNECTING = STATE_USER + 1;
    public static final int STATE_CONNECTED = STATE_USER + 2;

    /**
     * The unique identifier of the domain.
     */
    private BDomainIdentifier si;

    private BDomainRegister sr;

    /**
     * The NodeOS domain the domain is executing in.
     */
    private DomainHandle domain;

    private final PortPair pp = new PortPair();

    private Meter meters[];

    private Meter capabilityMeter;
    
    private PermissionSet meterSpec;

    private long peerID;

    private long storeID;

    private final Vector creators = new Vector();

    /**
     * Construct an empty BDomainCoroutine.
     */
    public BDomainCoroutine()
    {
	super(new StateResponderList());
    }

    void setBDomainRegister(BDomainRegister sr)
    {
	this.sr = sr;
    }

    /**
     * @param si The unique identifier of the domain.
     */
    void setBDomainIdentifier(BDomainIdentifier si)
    {
	this.si = si;
    }

    /**
     * @return The unique identifier of the domain.
     */
    public BDomainIdentifier getBDomainIdentifier()
    {
	return this.si;
    }

    /**
     * @param fh The DomainHandle of the domain.
     */
    void setDomainHandle(DomainHandle fh)
    {
	this.domain = fh;
    }

    /**
     * @return The DomainHandle of the domain.
     */
    DomainHandle getDomainHandle()
    {
	return this.domain;
    }

    public PortPair getPortPair()
    {
	return this.pp;
    }

    public void setLocalPort(FaxPort port)
    {
	this.pp.setLocal(port);
    }

    public FaxPortHandle getLocalPort()
    {
	return this.pp.getLocal();
    }

    public void setRemotePort(FaxPortHandle port)
    {
	this.pp.setRemote(port);
    }

    public FaxPortHandle getRemotePort()
    {
	return this.pp.getRemote();
    }

    public void setPeerID(long peerID)
    {
	this.peerID = peerID;
    }

    public long getPeerID()
    {
	return this.peerID;
    }

    void setStoreID(long id)
    {
	this.storeID = id;
    }

    public long getStoreID()
    {
	return this.storeID;
    }

    public void addResponder(StateResponder sr)
    {
	Assert.require(sr);

	((StateResponderList)this.getResponder()).addResponder(this, sr);
    }

    void addCreator(CreateBDomainCoroutine csc, StateResponder sr)
    {
	this.creators.addElement(csc);
	this.addResponder(sr);
    }

    protected void setState(int newState)
    {
	super.setState(newState);
	if( (this.sr != null) &&
	    ((newState <= STATE_ERROR) || (newState == STATE_COMPLETE)) )
	{
	    int lpc, size;
	    
	    this.sr.remove(this.si);
	    size = this.creators.size();
	    for( lpc = 0; lpc < size; lpc++ )
	    {
		CreateBDomainCoroutine csc;

		csc = (CreateBDomainCoroutine)this.creators.elementAt(lpc);
		csc.finished();
	    }
	    if( this.meters != null )
	    {
		for( lpc = 0; lpc < this.meters.length; lpc++ )
		{
		    CapabilityFactory.grant(this.meters[lpc], this.meterSpec);
		}
	    }
	    if( this.domain != null )
	    {
		this.domain.destroy();
	    }
	}
    }

    private ChanStats[] createChanStats(int count)
    {
	ChanStats retval[];
	int lpc;

	retval = new ChanStats[count];
	for( lpc = 0; lpc < count; lpc++ )
	{
	    retval[lpc] = new ChanStats();
	}
	return retval;
    }
    
    private ChanStats[] shrinkChanStats(ChanStats cs[], int count)
    {
	ChanStats retval[];
	int lpc;

	retval = new ChanStats[count];
	System.arraycopy(cs, 0, retval, 0, count);
	return retval;
    }
    
    private ThreadStats[] createThreadStats(int count)
    {
	ThreadStats retval[];
	int lpc;

	retval = new ThreadStats[count];
	for( lpc = 0; lpc < count; lpc++ )
	{
	    retval[lpc] = new ThreadStats();
	}
	return retval;
    }
    
    private ThreadStats[] shrinkThreadStats(ThreadStats cs[], int count)
    {
	ThreadStats retval[];
	int lpc;

	retval = new ThreadStats[count];
	System.arraycopy(cs, 0, retval, 0, count);
	return retval;
    }
    
    public synchronized void fillEvent(BDomainQueryEvent aqe)
    {
	if( this.domain != null )
	{
	    ThreadStats ts[];
	    DomainStats ds;
	    ChanStats cs[];
	    int actual;
	    
	    ds = new DomainStats();
	    this.domain.getStats(ds);
	    aqe.setDomainStats(ds);
	    /* XXX */
	    ts = createThreadStats(24);
	    actual = this.domain.getThreadStats(ts);
	    aqe.setThreadStats(shrinkThreadStats(ts, actual));
	    cs = createChanStats(24);
	    actual = this.domain.getInChannelStats(cs);
	    aqe.setInChanStats(shrinkChanStats(cs, actual));
	    cs = createChanStats(24);
	    actual = this.domain.getOutChannelStats(cs);
	    aqe.setOutChanStats(shrinkChanStats(cs, actual));
	}
    }

    private static final Permission CAPABILITY_MEMORY_PERMISSION =
	new Permission("capability.memory");

    private Meter metersToMeter(Meter meters[], Permission perm)
	throws PermissionException
    {
	int lpc;

	for( lpc = 0; lpc < meters.length; lpc++ )
	{
	    try
	    {
		meters[lpc].checkPermission(perm);
		return meters[lpc];
	    }
	    catch(PermissionException e)
	    {
	    }
	}
	return null;
    }
    
    private static final Permission CPU_THREADS_PERMISSION =
	new Permission("cpu.threads");

    private static final Permission CPU_SHARES_PERMISSION =
	new Permission("cpu.shares");

    private ThreadPoolSpec meterToThreadPool(Meter meters[], PermissionSet ps)
	throws PermissionException
    {
	CountPermission threads, shares;
	ThreadPoolSpec retval;
	int lpc;

	threads = (CountPermission)ps.hasPermission(CPU_THREADS_PERMISSION);
	shares = (CountPermission)ps.hasPermission(CPU_SHARES_PERMISSION);
	for( lpc = 0; lpc < meters.length; lpc++ )
	{
	    try
	    {
		synchronized( meters[lpc] )
		{
		    meters[lpc].checkPermission(threads);
		    meters[lpc].checkPermission(shares);
		    retval = new ThreadPoolSpec((int)threads.getMaxCount(),
						STACK_SIZE,
						(int)shares.getMaxCount(),
						null);
		    meters[lpc].revokePermission(threads);
		    meters[lpc].revokePermission(shares);
		    return retval;
		}
	    }
	    catch(PermissionException e)
	    {
	    }
	}
	throw new PermissionException("No CPU shares.");
    }

    private static final Permission MEM_PAGES_PERMISSION =
	new Permission("mem.pages");

    private MemSpec meterToMem(Meter meters[], PermissionSet ps)
    {
	CountPermission pages;
	MemSpec retval;
	int lpc;

	pages = (CountPermission)ps.hasPermission(MEM_PAGES_PERMISSION);
	for( lpc = 0; lpc < meters.length; lpc++ )
	{
	    try
	    {
		synchronized( meters[lpc] )
		{
		    meters[lpc].checkPermission(pages);
		    retval = new MemSpec((int)pages.getMaxCount() * PAGE_SIZE);
		    meters[lpc].revokePermission(pages);
		    return retval;
		}
	    }
	    catch(PermissionException e)
	    {
	    }
	}
	throw new PermissionException("No memory pages.");
    }

    void createDomain(KernelString selector,
		      String args[],
		      Meter meters[],
		      PermissionSet meterSpec)
    {
	try
	{
	    Domain kernelDomain = Domain.currentDomain();
	    CommSpaceHandle csh;
	    DomainStartSpec fss;
	    ThreadPoolSpec tps;
	    Credentials cred;
	    Runnable th;
	    MemSpec ms;
	    
	    Assert.require(selector);
	    Assert.require(this.sr);

	    this.meters = meters;
	    this.meterSpec = meterSpec;
	    this.capabilityMeter = metersToMeter(this.meters,
						 CAPABILITY_MEMORY_PERMISSION);
	    tps = meterToThreadPool(this.meters, this.meterSpec);
	    ms = meterToMem(this.meters, this.meterSpec);
	    if( args == null )
		args = emptyBDomainArgs;
	    cred = new Credentials("bees-domain:" + si);
	    /*
	     * Setup the CommSpace so the domain knows who it is and what it's
	     * supposed to do.
	     */
	    csh = CommSpaceHandle.newCommSpace(5);
	    csh.write(0, new CommSpaceElement(
		new KernelPortHandle(this.pp.getLocal())));
	    csh.write(1, new CommSpaceElement(this.getID()));
	    csh.write(2, new CommSpaceElement(this.storeID));
	    csh.write(3, new CommSpaceElement(selector.getString()));
	    csh.write(4, new CommSpaceElement(this.si.getDescription().
					      getString()));
	    fss = new DomainStartSpec("bees.core.bdomain.BDomain", args, csh);
	    th = new Runnable() {
		    public void run()
		    {
		    }
		};
	    this.domain = kernelDomain.newDomain(cred,
						 tps,
						 ms,
						 fss,
						 th,
						 minClassSpec);
	    this.setState(STATE_CREATED);
	}
	catch(Throwable th)
	{
	    super.setThrowable(th);
	}
    }

    public synchronized void initKernel(FaxPort fp)
    {
	Assert.require(fp);
	
	this.peerID = this.getID();
	this.pp.setLocal(fp);
	this.pp.setRemote(fp);
	this.setState(STATE_CONNECTED);
    }
    
    public synchronized void connectToKernel(FaxPort localPort,
					     CommSpaceHandle csh)
    {
	try
	{
	    BDomainRunningEvent sre;

	    Assert.require(this.getState() == STATE_NEWBORN);
	    Assert.require(localPort);
	    Assert.require(csh);
	    
	    this.pp.setLocal(localPort);
	    this.pp.setRemote(((KernelPortHandle)csh.take(0).
			       getCommSpaceable()).getPort());
	    this.peerID = csh.take(1).getLong();
	    this.storeID = csh.take(2).getLong();
	    
	    sre = new BDomainRunningEvent(this, localPort);
	    sre.setServerCoroutineID(this.peerID);
	    sre.setReplyPort(this.pp.getLocal());
	    this.pp.getRemote().putFacsimile(sre);
	    this.setState(STATE_CONNECTING);
	}
	catch(DeadTeamException e)
	{
	    this.setState(STATE_ERROR_CAGED);
	}
	catch(TeamEjectionException e)
	{
	    this.setState(STATE_ERROR_CAGED);
	}
	catch(Throwable th)
	{
	    super.setThrowable(th);
	}
    }

    public synchronized void shutdown()
    {
	try
	{
	    BDomainEndingEvent aee;

	    aee = new BDomainEndingEvent(this);
	    aee.setServerCoroutineID(this.peerID);
	    this.pp.getRemote().putFacsimile(aee);
	    this.setState(STATE_COMPLETE);
	}
	catch(DeadTeamException e)
	{
	    this.setState(STATE_ERROR_CAGED);
	}
	catch(TeamEjectionException e)
	{
	    this.setState(STATE_ERROR_CAGED);
	}
	catch(Throwable th)
	{
	    super.setThrowable(th);
	}
    }

    public synchronized void handle(BeesEvent ae)
    {
	switch( this.getState() )
	{
	case STATE_CREATED:
	    if( (ae instanceof BDomainRunningEvent) && !ae.isReply() )
	    {
		BDomainRunningEvent sre = (BDomainRunningEvent)ae;
		FaxPortHandle port;
		
		if( (port = sre.getPort()) != null )
		{
		    this.pp.setRemote(port);
		    try
		    {
			sre.sendReply();
			this.setState(STATE_CONNECTED);
		    }
		    catch(DeadBDomainException e)
		    {
			this.setState(STATE_ERROR_CAGED);
		    }
		}
		else
		{
		    this.setState(STATE_ERROR);
		}
	    }
	    else
	    {
		Strict.fail();
	    }
	    break;
	case STATE_CONNECTING:
	    if( (ae instanceof BDomainRunningEvent) && ae.isReply() )
	    {
		this.setState(STATE_CONNECTED);
	    }
	    else
	    {
		Strict.fail();
	    }
	    break;
	case STATE_CONNECTED:
	    if( (ae instanceof KernelObjectEvent) && !ae.isReply() )
	    {
		KernelObjectStore kos = KernelObjectStore.getPrimary();
		KernelObjectEvent koe;

		koe = (KernelObjectEvent)ae;
		try
		{
		    switch( koe.getMethod() )
		    {
		    case KernelObjectEvent.METHOD_CREATE:
			if( this.peerID == this.getID() )
			{
			    /* Kernel */
			}
			else if( this.capabilityMeter != null )
			{
			    CountPermission cp;
			    PermissionSet ps;
			    int size;

			    ps = koe.getPermissionSet();
			    size = koe.getData().length;
			    if( ps != null )
			    {
				size += ps.marshalledLength();
			    }
			    cp = new CountPermission("capability.memory",
						     size);
			    synchronized( this.capabilityMeter )
			    {
				this.capabilityMeter.checkPermission(cp);
				this.capabilityMeter.revokePermission(cp);
			    }
			}
			else
			{
			    koe.setID(Capability.INVALID_ID);
			    koe.setData(null);
			    throw new PermissionException(
				"No capability pages");
			}
			koe.setID(kos.put(koe.getData(),
					  koe.getSubcapabilities(),
					  koe.getPermissionSet(),
					  this.capabilityMeter,
					  koe.getComment().getString()));
			koe.setData(null);
			break;
		    case KernelObjectEvent.METHOD_LOOKUP:
			{
			    KernelObject ko;
			    
			    ko = kos.get(koe.getID());
			    koe.setData(ko.getData());
			    koe.setSubcapabilities(ko.getSubcapabilities());
			    koe.setPermissionSet(ko.getPermissionSet());
			}
			break;
		    case KernelObjectEvent.METHOD_DELETE:
			kos.remove(koe.getID());
			/* XXX we need to yank it from the domains */
			break;
		    }
		}
		catch(Throwable th)
		{
		    koe.setThrowable(th);
		}
		switch( koe.getMethod() )
		{
		case KernelObjectEvent.METHOD_DELETE:
		    break;
		default:
		    try
		    {
			koe.sendReply();
		    }
		    catch(DeadBDomainException e)
		    {
			this.setState(STATE_COMPLETE); // XXX
		    }
		    break;
		}
	    }
	    else if( (ae instanceof ConnectBDomainEvent) && !ae.isReply() )
	    {
		ConnectBDomainEvent cse = (ConnectBDomainEvent)ae;

		try
		{
		    BDomainIdentifier si;
		    BDomainCoroutine sc;
		    BDomainRegister sr;
		    
		    si = new BDomainIdentifier(cse.getID());
		    sr = (BDomainRegister)BDomainRegister.registry.get();
		    if( (sc = sr.get(si)) != null )
		    {
			cse.setPort(sc.getRemotePort());
		    }
		    else
		    {
			throw new NoSuchElementException(si.toString());
		    }
		}
		catch(Throwable th)
		{
		    cse.setThrowable(th);
		}
		try
		{
		    cse.sendReply();
		}
		catch(DeadBDomainException e)
		{
		    this.setState(STATE_COMPLETE); // XXX
		}
	    }
	    else if( (ae instanceof BDomainEndingEvent) && !ae.isReply() )
	    {
		this.setState(STATE_COMPLETE);
	    }
	    else
	    {
		Strict.fail();
	    }
	    break;
	case STATE_COMPLETE:
	    throw new KernelPanic("You're supposed to be dead!");
	default:
	    throw new KernelPanic("BDomain is in an unknown state: " + this);
	}
    }
    
    public String toString()
    {
	return "BDomainCoroutine["
	    + super.toString()
	    + "; si="
	    + this.si
	    + "; domain="
	    + this.domain
	    + "; pp="
	    + this.pp
	    + "; peerID="
	    + this.peerID
	    + "; storeID="
	    + this.storeID
	    + "]";
    }
}
