/*
 * BDomainIdentifier.java
 *
 * Copyright (c) 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is part of the Janos Bees Software.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */

package bees.core.kernel;

import java.util.Hashtable;

import edu.utah.janos.nodeos.Debug;

import bees.core.ByteArray;

/**
 * Object used to identify/describe a BDomain.
 */
public final class BDomainIdentifier
    implements Cloneable
{
    /**
     * The globally unique identifier of the BDomain.
     */
    private final byte id[];

    /**
     * The description of the BDomain.
     */
    private final KernelString description;

    /**
     * Construct a BDomain with the given values.
     *
     * @param id The unique byte array identifier.
     * @param description The description of the BDomain.
     */
    BDomainIdentifier(byte id[], KernelString description)
    {
	this.id = id;
	this.description = description;
    }
    
    /**
     * Construct a BDomain with the given values.
     *
     * @param id The unique integer identifier.
     * @param description The description of the BDomain.
     */
    BDomainIdentifier(long id, KernelString description)
    {
	this(new byte[] {
	    (byte)((id >> 56) & 0xff),
	    (byte)((id >> 48) & 0xff),
	    (byte)((id >> 40) & 0xff),
	    (byte)((id >> 32) & 0xff),
	    (byte)((id >> 24) & 0xff),
	    (byte)((id >> 16) & 0xff),
	    (byte)((id >>  8) & 0xff),
	    (byte)((id      ) & 0xff),
	},
	     description);
    }

    /**
     * Construct a BDomain with the given value and an empty description.
     *
     * @param id The unique integer identifier.
     */
    BDomainIdentifier(long id)
    {
	this(id, new KernelString(""));
    }

    /**
     * Construct a BDomain with the given value and an empty description.
     *
     * @param id The unique byte array identifier.
     */
    BDomainIdentifier(byte id[])
    {
	this(id, new KernelString(""));
    }

    /**
     * @return The unique byte array identifier.
     */
    byte[] getID()
    {
	return this.id;
    }

    /**
     * @return The unique integer identifier.
     */
    long getIntegerID()
    {
	return (((this.id[0] & 0xFF) << 56) |
		((this.id[1] & 0xFF) << 48) |
		((this.id[2] & 0xFF) << 40) |
		((this.id[3] & 0xFF) << 32) |
		((this.id[4] & 0xFF) << 24) |
		((this.id[5] & 0xFF) << 16) |
		((this.id[6] & 0xFF) <<  8) |
		((this.id[7] & 0xFF)      ));
    }

    /**
     * @return The description of the BDomain.
     */
    public KernelString getDescription()
    {
	return this.description;
    }

    /**
     * @param id The byte array to compare against this unique identifier.
     * @return True if the given identifier matches this one.
     */
    boolean matches(byte id[])
    {
	boolean retval = true;
	int lpc;
	
	for( lpc = 0; (lpc < this.id.length) && retval; lpc++ )
	{
	    retval = (this.id[lpc] == id[lpc]);
	}
	return retval;
    }

    /**
     * @param id The integer to compare against this unique identifier.
     * @return True if the given identifier matches this one.
     */
    boolean matches(long id)
    {
	return this.matches(new byte[] {
	    (byte)((id >> 56) & 0xff),
	    (byte)((id >> 48) & 0xff),
	    (byte)((id >> 40) & 0xff),
	    (byte)((id >> 32) & 0xff),
	    (byte)((id >> 24) & 0xff),
	    (byte)((id >> 16) & 0xff),
	    (byte)((id >>  8) & 0xff),
	    (byte)((id      ) & 0xff),
	});
    }

    /**
     * @param si The BDomainIdentifier to compare against.
     * @return True if the given identifier matches this one.
     */
    public boolean matches(BDomainIdentifier si)
    {
	return this.matches(si.id);
    }

    public int hashCode()
    {
	int retval = 0;
	
	switch( this.id.length )
	{
	case 0:
	    break;
	case 1:
	    retval = this.id[0] & 0xFF;
	    break;
	case 2:
	    retval = (((this.id[0] & 0xFF) << 8) |
		      ((this.id[1] & 0xFF)     ));
	    break;
	case 3:
	    retval = (((this.id[0] & 0xFF) << 16) |
		      ((this.id[1] & 0xFF) <<  8) |
		      ((this.id[2] & 0xFF)      ));
	    break;
	default:
	    retval = (((this.id[0] & 0xFF) << 24) |
		      ((this.id[1] & 0xFF) << 16) |
		      ((this.id[2] & 0xFF) <<  8) |
		      ((this.id[3] & 0xFF)      ));
	    break;
	}
	return retval;
    }

    public boolean equals(Object obj)
    {
	boolean retval = false;

	if( obj instanceof BDomainIdentifier )
	{
	    BDomainIdentifier si = (BDomainIdentifier)obj;

	    if( this.id.length == si.id.length )
	    {
		retval = this.matches(si.id);
	    }
	}
	return retval;
    }

    protected Object clone()
    {
	KernelString clonedDescription = null;
	byte clonedID[];
	
	clonedID = (byte[])this.id.clone();
	if( this.description != null )
	    clonedDescription = (KernelString)this.description.clone();
	return new BDomainIdentifier(clonedID, clonedDescription);
    }

    public String toString()
    {
	return "BDomainIdentifier[id="
	    + ByteArray.toString(this.id)
	    + "; description="
	    + this.description
	    + "]";
    }
}
