/*
 * BDomainRegisterHandler.java
 *
 * Copyright (c) 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is part of the Janos Bees Software.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */

package bees.core.kernel;

import java.util.Vector;
import java.util.Hashtable;
import java.util.Enumeration;
import java.util.NoSuchElementException;

import edu.utah.janosvm.kit.comm.amiga.FaxPortHandle;

import edu.utah.janosvm.kit.comm.jsi.JSIPair;
import edu.utah.janosvm.kit.comm.jsi.JSIHandler;
import edu.utah.janosvm.kit.comm.jsi.JSIMessage;
import edu.utah.janosvm.kit.comm.jsi.JSIPropertyInfo;
import edu.utah.janosvm.kit.comm.jsi.JSIHandlerContext;
import edu.utah.janosvm.kit.comm.jsi.MissingPairException;

import edu.utah.janos.nodeos.Domain;
import edu.utah.janos.nodeos.ChanStats;
import edu.utah.janos.nodeos.DomainStats;
import edu.utah.janos.nodeos.ThreadStats;
import edu.utah.janos.nodeos.ThreadLocal;

import bees.core.ByteArray;
import bees.core.Coroutine;
import bees.core.EventChannel;
import bees.core.StateResponder;
import bees.core.CapabilityFactory;

import bees.core.bdomain.BDomain;
import bees.core.bdomain.BDomainLocal;
import bees.core.bdomain.EventHandler;

public final class BDomainRegisterHandler
    implements JSIHandler
{
    private static final JSIPropertyInfo properties[] = {
	new JSIPropertyInfo("bdomainRegister",
			    (1 << JSIMessage.CREATE_PROPERTY),
			    "name:string - The name of the BDomainRegister.\n"
			    + "\n"
			    + "Create a new BDomainRegister object.\n"),
	new JSIPropertyInfo("bdomainRegister",
			    (1 << JSIMessage.DELETE_PROPERTY),
			    "name:string",
			    "Delete a BDomainRegister object.\n"),
	new JSIPropertyInfo("bdomainRegister",
			    (1 << JSIMessage.LIST_PROPERTIES),
			    "List the set of BDomainRegister objects.\n"),
    };

    private static final JSIPropertyInfo subProperties[] = {
	new JSIPropertyInfo("bdomain",
			    (1 << JSIMessage.CREATE_PROPERTY),
			    "selector:string - The main class.\n"
			    + "arg:string - An argument to the main method.\n"
			    + "capability:string - A kernel capability to "
			    + "pass to the domain.\n"
			    + "\n"
			    + "Create a new domain with the given values.\n"),
	new JSIPropertyInfo("bdomain",
			    (1 << JSIMessage.GET_PROPERTY),
			    "Get a domains statistics.\n"),
	new JSIPropertyInfo("bdomain",
			    (1 << JSIMessage.LIST_PROPERTIES),
			    "List the set of domains managed by this "
			    + "register.\n"),
    };
    
    private static final BDomainLocal registries = new BDomainLocal() {
	    public Object initialValue()
	    {
		return new Hashtable();
	    }
	};

    private static Hashtable getRegistries()
    {
	return (Hashtable)registries.get();
    }

    /**
     * XXX Kind of a hack
     */
    private static final ThreadLocal current = new ThreadLocal() {
	    protected Object initialValue()
	    {
		throw new NullPointerException();
	    }
	};

    static FaxPortHandle currentPort()
    {
	return (FaxPortHandle)current.get();
    }

    private final Domain domain = Domain.currentDomain();

    private final Hashtable nameToCoroutine = new Hashtable();
    private final Hashtable bdomains = new Hashtable();
    
    public BDomainRegisterHandler()
    {
    }

    public JSIHandler resolveSpecifier(JSIMessage jm, JSIPair jp)
    {
	return this;
    }

    public void handleMessage(JSIHandlerContext jhc, JSIMessage jm)
	throws MissingPairException,
	       InstantiationException,
	       Throwable
    {
	JSIPair jp, registerPair;
	FaxPortHandle fp = null;
	
	jm.restoreSpecifier();
	registerPair = jm.popSpecifier();
	if( (registerPair.getValue().length > 0) &&
	    ((fp = (FaxPortHandle)getRegistries().
	      get(registerPair.getStringValue())) == null) )
	{
	    throw new NoSuchElementException("Couldn't find BDomainRegister: "
					     + registerPair.getStringValue());
	}
	switch( jm.getWhat() )
	{
	case JSIMessage.CREATE_PROPERTY:
	    if( (jp = jm.popSpecifier()) == null )
	    {
		if( registerPair.getValue().length == 0 )
		{
		    BDomainRegisterCoroutine src;
		    EventHandler eh;
		    String name;
		    
		    name = jm.expect("name");
		    eh = new EventHandler(BDomain.currentBDomain());
		    domain.threadStart(eh, "BDomainRegister " + name);
		    fp = eh.getPort();
		    src = new BDomainRegisterCoroutine();
		    src.create(fp, fp);
		    if( src.waitForState(BDomainRegisterCoroutine.
					 STATE_ALIVE) ==
			BDomainRegisterCoroutine.STATE_ALIVE )
		    {
			this.nameToCoroutine.put(name, src);
			getRegistries().put(name, fp);
			jm.reply("ok");
		    }
		    else
		    {
			switch( src.getState() )
			{
			case Coroutine.STATE_ERROR_TIMEOUT:
			    jm.error("Timeout");
			    break;
			case Coroutine.STATE_ERROR_CAGED:
			case Coroutine.STATE_ERROR_INTERNAL:
			    throw new KernelPanic("We're dead?");
			case Coroutine.STATE_ERROR:
			    throw src.getThrowable();
			}
		    }
		}
		else
		{
		    jm.error("No specified property to create.");
		}
	    }
	    else if( fp != null )
	    {
		if( jp.getTag().equals("bdomain") )
		{
		    Vector argVector = new Vector();
		    Vector capVector = new Vector();
		    final String name, selector;
		    CreateBDomainCoroutine csc;
		    String cap, arg, args[];
		    int lpc;

		    name = jm.expect("name");
		    selector = jm.expect("selector");
		    
		    for( lpc = 0;
			 (arg = jm.get("arg", lpc, null)) != null;
			 lpc++ )
		    {
			argVector.addElement(arg);
		    }
		    args = new String[lpc];
		    argVector.copyInto(args);
		    
		    for( lpc = 0;
			 (cap = jm.get("capability", lpc, null)) != null;
			 lpc++ )
		    {
			capVector.addElement(KernelCapabilities.getTree().
					     get(cap));
		    }
		    
		    csc = new CreateBDomainCoroutine(
			new StateResponder(Coroutine.STATE_COMPLETE) {
				public void transition(Coroutine coroutine,
						       int newState)
				{
				    BDomainRegisterHandler.this.
					nameToCoroutine.remove(name);
				    BDomainRegisterHandler.this.
					bdomains.remove(name);
				    KernelCapabilities.getTree().
					remove(name);
				}
			});
		    csc.create(fp,
			       fp,
			       jm.get("description", name),
			       selector,
			       args,
			       capVector,
			       KernelCapabilities.toPermissionSet(jm));
		    if( csc.waitForState(CreateBDomainCoroutine.STATE_ALIVE) ==
			CreateBDomainCoroutine.STATE_ALIVE )
		    {
			KernelObject ko;

			this.bdomains.put(name, csc);
			current.set(fp);
			ko = KernelObjectStore.getPrimary().
			    get(csc.getStoreID());
			KernelCapabilities.getTree().
			    put(name,
				CapabilityFactory.
				instantiate(new EventChannel(),
					    csc.getStoreID(),
					    ko.getData(),
					    EventChannel.DEFAULT_PERMISSIONS));
			jm.reply("ok");
		    }
		    else if( csc.getThrowable() != null )
		    {
			csc.getThrowable().printStackTrace();
			jm.error(csc.getThrowable().toString());
		    }
		    else
		    {
			jm.error("Timeout");
		    }
		}
		else
		{
		    jm.error("No such property: " + jp.getTag());
		}
	    }
	    else
	    {
		jm.error("No specified BDomainRegister");
	    }
	    break;
	case JSIMessage.DELETE_PROPERTY:
	    if( fp != null )
	    {
		if( (jp = jm.popSpecifier()) != null )
		{
		    if( jp.getTag().equals("bdomain") )
		    {
			CreateBDomainCoroutine cac;

			if( (cac = (CreateBDomainCoroutine)
			     this.bdomains.get(jp.getStringValue())) != null )
			{
			    cac.finished();
			}
			else
			{
			    BDomainRegisterCoroutine arc;
			    BDomainIdentifier ai = null;

			    ai = new BDomainIdentifier(
				jp.getByteArrayFromHexValue());
			    arc = (BDomainRegisterCoroutine)
				this.nameToCoroutine.get(registerPair.
							 getStringValue());
			    arc.deleteDomain(fp, fp, ai);
			}
			jm.reply("ok");
		    }
		    else
		    {
			jm.error("Unknown property: " + jp.getTag());
		    }
		}
		else
		{
		    BDomainRegisterCoroutine src;
		    
		    src = (BDomainRegisterCoroutine)
			this.nameToCoroutine.
			get(registerPair.getStringValue());
		    src.delete(fp, fp);
		    if( src.waitForCompletion() )
		    {
			jm.reply("ok");
		    }
		    else
		    {
			switch( src.getState() )
			{
			case Coroutine.STATE_ERROR_TIMEOUT:
			    jm.error("Timeout");
			    break;
			case Coroutine.STATE_ERROR_CAGED:
			case Coroutine.STATE_ERROR_INTERNAL:
			    throw new KernelPanic("We're dead?");
			case Coroutine.STATE_ERROR:
			    throw src.getThrowable();
			}
		    }
		}
	    }
	    else
	    {
		jm.error("No BDomainRegister specified");
	    }
	    break;
	case JSIMessage.GET_PROPERTY:
	    if( fp != null )
	    {
		if( (jp = jm.popSpecifier()) != null )
		{
		    if( jp.getTag().equals("bdomain") )
		    {
			BDomainQueryEvent aqe = null;
			CreateBDomainCoroutine csc;
			
			if( (csc = (CreateBDomainCoroutine)
			     this.bdomains.get(jp.getStringValue())) != null )
			{
			    csc.queryStats(fp, fp);
			    synchronized( csc )
			    {
				while( ((aqe =
					 csc.takeQueryEvent()) == null) &&
				       (csc.getState() > Coroutine.STATE_ANY) )
				{
				    csc.wait();
				}
			    }
			}
			else
			{
			    BDomainRegisterCoroutine arc;
			    BDomainIdentifier ai = null;

			    ai = new BDomainIdentifier(
				jp.getByteArrayFromHexValue());
			    arc = (BDomainRegisterCoroutine)
				this.nameToCoroutine.get(registerPair.
							 getStringValue());
			    arc.queryStats(fp, fp, ai);
			    synchronized( arc )
			    {
				while( ((aqe =
					 arc.takeQueryEvent(ai)) == null) &&
				       (arc.getState() > Coroutine.STATE_ANY) )
				{
				    arc.wait();
				}
			    }
			}
			if( (aqe != null) && (aqe.getDomainStats() != null) )
			{
			    ChanStats ics[], ocs[];
			    ThreadStats ts[];
			    DomainStats ds;
			    int lpc;
			    
			    ds = aqe.getDomainStats();
			    ts = aqe.getThreadStats();
			    ics = aqe.getInChanStats();
			    ocs = aqe.getOutChanStats();
			    jm.clear();
			    jm.setWhat(JSIMessage.REPLY);
			    jm.add(new JSIPair("domain.thread-count",
					       ds.getThreadCount()));
			    jm.add(new JSIPair("domain.max-threads",
					       ds.getMaxThreads()));
			    for( lpc = 0; lpc < ts.length; lpc++ )
			    {
				jm.add(new JSIPair(
					"thread.name",
					ts[lpc].getName()));
				jm.add(new JSIPair(
					"thread.cpu-usage",
					ts[lpc].getCPUUsage()));
			    }
			    for( lpc = 0; lpc < ics.length; lpc++ )
			    {
				jm.add(new JSIPair(
					"in.protocol",
					ics[lpc].getProtocolSpec()));
				jm.add(new JSIPair(
					"in.address",
					ics[lpc].getAddressSpec()));
				jm.add(new JSIPair(
					"in.packets",
					ics[lpc].getInPackets()));
				jm.add(new JSIPair(
					"in.bytes",
					ics[lpc].getInBytes()));
				jm.add(new JSIPair(
					"in.drop-packets",
					ics[lpc].getDropPackets()));
				jm.add(new JSIPair(
					"in.drop-no-bufs",
					ics[lpc].getDropNoBufs()));
				jm.add(new JSIPair(
					"in.drop-failed-buf",
					ics[lpc].getDropFailedBuf()));
				jm.add(new JSIPair(
					"in.drop-bad-packet",
					ics[lpc].getDropBadPacket()));
				jm.add(new JSIPair(
					"in.drop-bad-mac",
					ics[lpc].getDropBadMac()));
				jm.add(new JSIPair(
					"in.drop-bytes",
					ics[lpc].getDropBytes()));
				jm.add(new JSIPair(
					"in.buf-queued",
					ics[lpc].getInBufQueued()));
				jm.add(new JSIPair(
					"in.buf-copies",
					ics[lpc].getInBufCopies()));
				jm.add(new JSIPair(
					"in.max-threads",
					ics[lpc].getMaxThreads()));
			    }
			    for( lpc = 0; lpc < ocs.length; lpc++ )
			    {
				jm.add(new JSIPair(
					"out.protocol",
					ics[lpc].getProtocolSpec()));
				jm.add(new JSIPair(
					"out.address",
					ics[lpc].getAddressSpec()));
				jm.add(new JSIPair(
					"out.packets",
					ocs[lpc].getOutPackets()));
				jm.add(new JSIPair(
					"out.bytes",
					ocs[lpc].getOutBytes()));
				jm.add(new JSIPair(
					"out.buf-copies",
					ocs[lpc].getOutBufCopies()));
			    }
			}
			else
			{
			    jm.error("Unknown domain: " + jp.getStringValue());
			}
		    }
		    else
		    {
			jm.error("Unknown property: " + jp.getTag());
		    }
		}
		else
		{
		    jm.error("No property to execute");
		}
	    }
	    else
	    {
		jm.error("No BDomainRegister specified");
	    }
	    break;
	case JSIMessage.LIST_PROPERTIES:
	    if( (jp = jm.popSpecifier()) == null )
	    {
		if( registerPair.getValue().length == 0 )
		{
		    Enumeration enum;
		    
		    jm.clear();
		    jm.setWhat(JSIMessage.REPLY);
		    enum = getRegistries().keys();
		    while( enum.hasMoreElements() )
		    {
			jm.add("result", (String)enum.nextElement());
		    }
		}
		else
		{
		    jm.error("No specified property to list.");
		}
	    }
	    else if( fp != null )
	    {
		if( jp.getTag().equals("bdomain") )
		{
		    ListBDomainCoroutine lsc;
		    
		    jm.setWhat(JSIMessage.REPLY);
		    lsc = new ListBDomainCoroutine();
		    lsc.list(fp, fp);
		    if( lsc.waitForCompletion() )
		    {
			BDomainIdentifier si[];
			long id[];
			int lpc;
			
			jm.clear();
			jm.setWhat(JSIMessage.REPLY);
			id = lsc.getStoreIDs();
			si = lsc.getBDomainIdentifiers();
			for( lpc = 0; lpc < si.length; lpc++ )
			{
			    jm.add(new JSIPair("id", id[lpc]));
			    jm.add(new JSIPair(
				"flowID",
				ByteArray.toPlainString(si[lpc].getID())));
			    jm.add(new JSIPair("description", si[lpc].
					       getDescription().getString()));
			}
		    }
		    else
		    {
			jm.error("I dunno captain");
		    }
		}
		else
		{
		    jm.error("No such property: " + jp.getTag());
		}
	    }
	    else
	    {
		jm.error("Unknown property: " + jp.getTag());
	    }
	    break;
	case JSIMessage.EXECUTE_PROPERTY:
	    if( fp != null )
	    {
		if( (jp = jm.popSpecifier()) != null )
		{
		    if( jp.getTag().equals("bdomain") )
		    {
			CreateBDomainCoroutine csc;
			
			if( (csc = (CreateBDomainCoroutine)
			     this.bdomains.get(jp.getStringValue())) != null )
			{
			    if( (jp = jm.popSpecifier()) != null )
			    {
				if( jp.getTag().equals("wait") )
				{
				    if( csc.waitForCompletion(10 * 60 * 1000) )
				    {
					jm.reply("ok");
				    }
				    else
				    {
					jm.error("Timeout");
				    }
				}
				else
				{
				    jm.error("Unknown property: "
					     + jp.getTag());
				}
			    }
			    else
			    {
				jm.error("No property to execute");
			    }
			}
			else
			{
			    jm.error("Unknown domain: " + jp.getStringValue());
			}
		    }
		    else
		    {
			jm.error("Unknown property: " + jp.getTag());
		    }
		}
		else
		{
		    jm.error("No property to execute");
		}
	    }
	    else
	    {
		jm.error("No specified BDomainRegister");
	    }
	    break;
	case JSIMessage.GET_SUITES:
	    jm.reply(subProperties);
	    break;
	default:
	    throw new UnsupportedOperationException();
	}
    }
    
    public void getSuites(JSIMessage jm)
    {
	jm.add("result", properties);
    }

    public String toString()
    {
	return "BDomainRegisterHandler["
	    + "]";
    }
}
