/*
 * BeesKernel.java
 *
 * Copyright (c) 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is part of the Janos Bees Software.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */

package bees.core.kernel;

import java.util.Properties;

import java.io.IOException;
import java.io.FileNotFoundException;

import edu.utah.janos.nodeos.Node;
import edu.utah.janos.nodeos.Domain;

import edu.utah.janosvm.sys.TeamLocal;

import edu.utah.janosvm.kit.comm.jsi.JSIPair;
import edu.utah.janosvm.kit.comm.jsi.JSIChannel;
import edu.utah.janosvm.kit.comm.jsi.JSIMessage;
import edu.utah.janosvm.kit.comm.jsi.JSIHandler;
import edu.utah.janosvm.kit.comm.jsi.JSIProvider;
import edu.utah.janosvm.kit.comm.jsi.TCPProvider;
import edu.utah.janosvm.kit.comm.jsi.JSIHandlerRoot;
import edu.utah.janosvm.kit.comm.jsi.JSIPropertyInfo;
import edu.utah.janosvm.kit.comm.jsi.JSIHandlerContext;
import edu.utah.janosvm.kit.comm.jsi.NativePlexProvider;
import edu.utah.janosvm.kit.comm.jsi.JSIConnectionHandler;
import edu.utah.janosvm.kit.comm.jsi.JSIEncodingException;
import edu.utah.janosvm.kit.comm.jsi.MissingPairException;

import bees.core.Assert;
import bees.core.ObjectTree;
import bees.core.ProtocolParameters;

import bees.core.bdomain.BDomainLocal;

/**
 * The main class for booting Bees, currently, it just adds some basic JSI
 * handlers and stops.
 */
public final class BeesKernel
{
    private static final Properties DEFAULT_INFO = new Properties();
    
    static {
	DEFAULT_INFO.setProperty("bees.home", ".");
    }
    
    /**
     * The JSIHandlerRoot that is used when dispatching received messages.
     */
    private static final TeamLocal jsiRoot = new TeamLocal() {
	    public Object initialValue()
	    {
		return new JSIHandlerRoot();
	    }
	};
    
    public static JSIHandlerRoot getJSIRoot()
    {
	return (JSIHandlerRoot)jsiRoot.get();
    }
    
    private static final TeamLocal infoTree = new TeamLocal() {
	    public Object initialValue()
	    {
		return new Properties(DEFAULT_INFO);
	    }
	};
    
    public static Properties getInfoTree()
    {
	return (Properties)infoTree.get();
    }
    
    /**
     * Startup a TCPProvider on the given port.
     *
     * @param domain The current domain pointer.
     * @param port The TCP port number to bind to.
     * @exception Throws an IOException if its unable to bind to the port.
     */
    private static void startTCPProvider(final Domain domain, int port)
	throws IOException
    {
	final TCPProvider tp;

	Assert.require(domain);
	Assert.require(port > 0);
	Assert.require(port < ((Short.MAX_VALUE + 1) * 2));
	
	tp = new TCPProvider(port);
	domain.threadStart(new Runnable() {
		public void run()
		{
		    tp.handle(new JSIConnectionHandler() {
			    public void handle(JSIProvider jp, JSIChannel jc)
			    {
				domain.threadStart(new JSIHandlerThread(jc),
						   "JSI connection");
			    }
			});
		}
	    },
			 "JSI provider");
    }

    private static void startPlexProvider(final Domain domain, String spec)
	throws IOException
    {
	final NativePlexProvider npp;

	Assert.require(domain);
	
	npp = new NativePlexProvider(spec);
	domain.threadStart(new Runnable() {
		public void run()
		{
		    npp.handle(new JSIConnectionHandler() {
			    public void handle(JSIProvider jp, JSIChannel jc)
			    {
				domain.threadStart(new JSIHandlerThread(jc),
						 "JSI connection");
			    }
			});
		}
	    },
			 "JSI provider");
    }

    /**
     * Parse the arguments passed to main().
     *
     * @param domain The current domain pointer.
     * @param args The arguments passed to main().
     */
    private static void parseArguments(final Domain domain, String args[])
    {
	int lpc;
	
	Assert.require(domain);
	Assert.require(args);
	
	for( lpc = 0; lpc < args.length; lpc += 2 )
	{
	    if( args[lpc].equals("-tcpport") )
	    {
		try
		{
		    int port;
		    
		    port = Integer.parseInt(args[lpc + 1]);
		    if( (port > 0) && (port < ((Short.MAX_VALUE + 1) * 2)) )
		    {
			startTCPProvider(domain, port);
		    }
		    else
		    {
			Node.println("Port number for `-tcpport', "
				     + args[lpc + 1]
				     + ", is out of range.");
		    }
		}
		catch(NumberFormatException e)
		{
		    Node.println("Argument to `-tcpport', "
				 + args[lpc + 1]
				 + ", isn't a number.");
		}
		catch(IOException e)
		{
		    Node.println("Unable to bind to TCP port: "
				 + args[lpc + 1]);
		    Node.printStackTrace(e);
		}
	    }
	    else if( args[lpc].equals("-plexport") )
	    {
		try
		{
		    startPlexProvider(domain, args[lpc + 1]);
		}
		catch(IOException e)
		{
		}
	    }
	    else
	    {
		Node.println("Invalid argument: " + args[lpc]);
	    }
	}
    }

    /**
     * The JSIHandler which begats most other handlers.
     */
    private static class LoadClassHandler
	implements JSIHandler
    {
	/**
	 * The root level properties for loadHandler.
	 */
	private static final JSIPropertyInfo properties[] = {
	    new JSIPropertyInfo("handler",
				(1 << JSIMessage.CREATE_PROPERTY),
				"name:string - The name the "
				+ "JSIHandler should be registered under.\n"
				+ "from:string - The name of an existing "
				+ "handler.\n"
				+ " - or - \n"
				+ "class:string - The handler class name.\n"
				+ "\n"
				+ "Add an instance of the handler to the "
				+ "root list.\n"),
	    new JSIPropertyInfo("handler",
				(1 << JSIMessage.DELETE_PROPERTY),
				"name:string",
				"Remove the handler from the root list.\n"),
	    new JSIPropertyInfo("handler",
				(1 << JSIMessage.GET_PROPERTY),
				"name:string",
				"Get the value of the handler.\n"),
	};

	/**
	 * The current domain pointer.
	 */
	private final Domain domain = Domain.currentDomain();

	public LoadClassHandler()
	{
	}

	public JSIHandler resolveSpecifier(JSIMessage jm, JSIPair jp)
	{
	    return this;
	}
	
	public void handleMessage(JSIHandlerContext jhc, JSIMessage jm)
	    throws ClassNotFoundException,
		   IllegalAccessException,
		   InstantiationException,
		   MissingPairException
	{
	    switch( jm.getWhat() )
	    {
	    case JSIMessage.CREATE_PROPERTY:
		if( jm.popSpecifier() == null )
		{
		    String name, from;
		    
		    name = jm.expect("name");
		    if( (from = jm.get("from")) != null )
		    {
			getJSIRoot().addHandler(name,
						getJSIRoot().getHandler(from));
			jm.reply("ok");
		    }
		    else
		    {
			Class cl;
			
			cl = domain.loadClass(jm.expect("class"));
			getJSIRoot().addHandler(name,
						(JSIHandler)cl.newInstance());
			jm.reply("ok");
		    }
		}
		else
		{
		    jm.error("Too many specifiers");
		}
		break;
	    case JSIMessage.DELETE_PROPERTY:
		if( jm.popSpecifier() == null )
		{
		    JSIPair jp;
		    
		    jm.restoreSpecifier();
		    jp = jm.popSpecifier();
		    if( getJSIRoot().remHandler(jp.getStringValue()) != null )
		    {
			jm.reply("ok");
		    }
		    else
		    {
			jm.error("Couldn't find handler named: "
				 + jp.getStringValue());
		    }
		}
		else
		{
		    jm.error("Too many specifiers");
		}
		break;
	    case JSIMessage.GET_PROPERTY:
		if( jm.popSpecifier() == null )
		{
		    JSIHandler jh;
		    JSIPair jp;
		    
		    jm.restoreSpecifier();
		    jp = jm.popSpecifier();
		    if( (jh = getJSIRoot().getHandler(jp.getStringValue()))
			!= null )
		    {
			jm.reply(jh.toString());
		    }
		    else
		    {
			jm.error("Couldn't find handler named: "
				 + jp.getStringValue());
		    }
		}
		else
		{
		    jm.error("Too many specifiers");
		}
		break;
	    default:
		throw new UnsupportedOperationException();
	    }
	}
	
	public void getSuites(JSIMessage jm)
	{
	    jm.add("result", properties);
	}

	public String toString()
	{
	    return "LoadClassHandler["
		+ "]";
	}
    }
    
    /**
     * A JSIHandler for Bees meta info.
     */
    private static class InfoHandler
	implements JSIHandler
    {
	/**
	 * The root level properties for InfoHandler.
	 */
	private static final JSIPropertyInfo JSI_PROPERTIES[] = {
	    new JSIPropertyInfo("info",
				(1 << JSIMessage.LIST_PROPERTIES),
				"List info's subproperties."),
	};

	/**
	 * The sub level properties for InfoHandler.
	 */
	private static final JSIPropertyInfo JSI_SUB_PROPERTIES[] = {
	    new JSIPropertyInfo("home",
				(1 << JSIMessage.GET_PROPERTY),
				"Get the Bees install directory."),
	};

	/**
	 * The current domain pointer.
	 */
	private final Domain domain = Domain.currentDomain();

	public InfoHandler()
	{
	    Properties props;

	    props = getInfoTree();
	    try
	    {
		ProtocolParameters.loadProperties(
			props,
			this.domain.
			readClassPathFile("bees/core/install.properties"));
	    }
	    catch(FileNotFoundException e)
	    {
		e.printStackTrace();
	    }
	}

	public JSIHandler resolveSpecifier(JSIMessage jm, JSIPair jp)
	{
	    return this;
	}
	
	public void handleMessage(JSIHandlerContext jhc, JSIMessage jm)
	    throws ClassNotFoundException,
		   IllegalAccessException,
		   InstantiationException,
		   MissingPairException
	{
	    JSIPair jp;
	    
	    switch( jm.getWhat() )
	    {
	    case JSIMessage.GET_PROPERTY:
		if( (jp = jm.popSpecifier()) != null )
		{
		    if( jp.getTag().equals("home") )
		    {
			jm.reply(getInfoTree().getProperty("bees.home"));
		    }
		    else
		    {
			jm.error("Unknown property: " + jp.getTag());
		    }
		}
		else
		{
		    jm.error("No property to get");
		}
		break;
	    case JSIMessage.GET_SUITES:
		jm.reply(JSI_SUB_PROPERTIES);
		break;
	    default:
		throw new UnsupportedOperationException();
	    }
	}
	
	public void getSuites(JSIMessage jm)
	{
	    jm.add("result", JSI_PROPERTIES);
	}

	public String toString()
	{
	    return "InfoHandler["
		+ "]";
	}
    }
    
    public static void main(String args[], ObjectTree capabilities)
    {
	final Domain domain = Domain.currentDomain();

	Assert.require(args);
	Assert.require(capabilities);

	if( (args.length % 2) == 0 )
	{
	    getJSIRoot().addHandler("handler", new LoadClassHandler());
	    getJSIRoot().addHandler("info", new InfoHandler());
	    parseArguments(domain, args);
	}
	else
	{
	    Node.println("Incorrect number of arguments.");
	    Node.println("Usage: XXX Fix me");
	}
    }

    private BeesKernel()
    {
    }
}
