/*
 * ConnectBDomainCoroutine.java
 *
 * Copyright (c) 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is part of the Janos Bees Software.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */

package bees.core.kernel;

import edu.utah.janosvm.sys.DeadTeamException;
import edu.utah.janosvm.sys.TeamEjectionException;

import edu.utah.janosvm.kit.comm.amiga.FaxPortHandle;

import bees.core.Assert;
import bees.core.Strict;
import bees.core.BeesEvent;
import bees.core.Coroutine;
import bees.core.StateResponder;
import bees.core.ResponderCoroutine;

/**
 * Coroutine used to transform a domain identifier into a usable port for
 * sending events.
 */
public final class ConnectBDomainCoroutine
    extends ResponderCoroutine
{
    /**
     * State used when waiting for a reply.
     */
    public static final int STATE_WAITING = STATE_USER;

    /**
     * A port connected to the domain.
     */
    private FaxPortHandle port;

    /**
     * Construct a ConnectBDomainCoroutine with the given parameters.
     *
     * @param sr A StateResponder object that should be executed on any state
     * transitions.
     */
    public ConnectBDomainCoroutine(StateResponder sr)
    {
	super(sr);
    }

    /**
     * Construct an empty ConnectBDomainCoroutine.
     */
    public ConnectBDomainCoroutine()
    {
	this(null);
    }

    /**
     * @return The port connected to the domain or NULL if the connection failed.
     */
    public synchronized FaxPortHandle getPort()
    {
	return this.port;
    }

    /**
     * Connect to the domain with the given identifier.
     *
     * @param sc The current domain's coroutine.
     * @param id The identifier of the domain whose port should be retrieved.
     */
    public synchronized void connect(BDomainCoroutine sc, byte id[])
    {
	try
	{
	    ConnectBDomainEvent cse;
	    FaxPortHandle fp;
	    
	    Assert.require(sc);
	    Assert.require(id);
	    
	    cse = new ConnectBDomainEvent(this, id);
	    /* Our BDomainCoroutine in the kernel will handle the message. */
	    cse.setServerCoroutineID(sc.getPeerID());
	    cse.setReplyPort(sc.getLocalPort());
	    try
	    {
		fp = BDomainRegisterHandler.currentPort();
	    }
	    catch(NullPointerException e)
	    {
		fp = sc.getRemotePort();
	    }
	    fp.putFacsimile(cse);
	    this.setState(STATE_WAITING);
	}
	catch(DeadTeamException e)
	{
	    this.setState(STATE_ERROR_CAGED);
	}
	catch(TeamEjectionException e)
	{
	    this.setState(STATE_ERROR_CAGED);
	}
	catch(Throwable th)
	{
	    super.setThrowable(th);
	}
    }

    public synchronized void handle(BeesEvent ae)
    {
	Assert.require(ae);
	
	switch( this.getState() )
	{
	case STATE_WAITING:
	    if( (ae instanceof ConnectBDomainEvent) && ae.isReply() )
	    {
		ConnectBDomainEvent cse = (ConnectBDomainEvent)ae;
		Throwable th;
		
		if( (th = cse.getThrowable()) != null )
		{
		    super.setThrowable(th);
		}
		else if( (this.port = cse.getPort()) != null )
		{
		    this.setState(STATE_COMPLETE);
		}
		else
		{
		    throw new KernelPanic("Bogus reply message, there's no "
					  + "port or exception.");
		}
	    }
	    else
	    {
		Strict.fail();
	    }
	    break;
	default:
	    throw new KernelPanic("Invalid state for coroutine: " + this);
	}
    }

    public String toString()
    {
	return "ConnectBDomainCoroutine["
	    + super.toString()
	    + "; port="
	    + this.port
	    + "]";
    }
}
