/*
 * KernelCapabilities.java
 *
 * Copyright (c) 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is part of the Janos Bees Software.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */

package bees.core.kernel;

import java.io.File;
import java.io.InputStream;
import java.io.IOException;
import java.io.EOFException;
import java.io.OutputStream;
import java.io.DataInputStream;
import java.io.FileInputStream;
import java.io.DataOutputStream;
import java.io.FileOutputStream;

import java.util.Vector;
import java.util.Enumeration;
import java.util.StringTokenizer;
import java.util.NoSuchElementException;

import java.util.zip.ZipFile;
import java.util.zip.ZipEntry;

import java.security.SecureRandom;
import java.security.GeneralSecurityException;

import javax.crypto.Cipher;
import javax.crypto.SecretKey;
import javax.crypto.SecretKeyFactory;
import javax.crypto.CipherInputStream;
import javax.crypto.CipherOutputStream;

import javax.crypto.spec.PBEKeySpec;
import javax.crypto.spec.PBEParameterSpec;

import edu.utah.janosvm.resources.ClassPathResource;

import edu.utah.janosvm.kit.comm.jsi.JSIPair;
import edu.utah.janosvm.kit.comm.jsi.JSIHandler;
import edu.utah.janosvm.kit.comm.jsi.JSIMessage;
import edu.utah.janosvm.kit.comm.jsi.JSIPropertyInfo;
import edu.utah.janosvm.kit.comm.jsi.JSIHandlerContext;
import edu.utah.janosvm.kit.comm.jsi.MissingPairException;

import edu.utah.janos.nodeos.Node;
import edu.utah.janos.nodeos.Domain;
import edu.utah.janos.nodeos.AddressFormatException;

import bees.core.Xdr;
import bees.core.Assert;
import bees.core.HashMap;
import bees.core.ObjectMap;
import bees.core.PairChain;
import bees.core.Capability;
import bees.core.ObjectTree;
import bees.core.XdrByteArray;
import bees.core.CapabilityFactory;
import bees.core.CapabilityEnvelope;
import bees.core.CapabilityStreamer;
import bees.core.MissingCapabilityException;

import bees.core.resource.Clock;
import bees.core.resource.Meter;
import bees.core.resource.Account;
import bees.core.resource.Console;
import bees.core.resource.PathSpec;
import bees.core.resource.AccountSpec;
import bees.core.resource.SystemClass;
import bees.core.resource.SystemPackage;
import bees.core.resource.FileSystemRoot;
import bees.core.resource.NetworkService;
import bees.core.resource.NetworkInterface;
import bees.core.resource.IPAddressPermission;

import bees.core.security.Permission;
import bees.core.security.PermissionSet;
import bees.core.security.FilePermission;
import bees.core.security.CountPermission;
import bees.core.security.FloatPermission;
import bees.core.security.RangePermission;
import bees.core.security.IntegerPermission;
import bees.core.security.EnumerationPermission;
import bees.core.security.MaximumFloatPermission;
import bees.core.security.MinimumFloatPermission;
import bees.core.security.MaximumIntegerPermission;
import bees.core.security.MinimumIntegerPermission;

import bees.core.bdomain.BDomain;
import bees.core.bdomain.BDomainLocal;

/**
 * Class used for tracking capabilities created in the kernel.
 */
public final class KernelCapabilities
    implements JSIHandler
{
    /**
     * The root level properties for this object.
     */
    private static final JSIPropertyInfo properties[] = {
	new JSIPropertyInfo("capability",
			    (1 << JSIMessage.CREATE_PROPERTY),
			    "name:string - The name of the sub-tree.\n"
			    + " - or -\n"
			    + "name:string - The name to store the capability "
			    + "under.\n"
			    + "id:int - The object store identifier.\n"
			    + "class:string - The name of the capability "
			    + "class.\n"
			    + "permission:string=string - Add a permission to "
			    + "the object.\n"
			    + "countPermission:string=int - Add a count "
			    + "permission to the object.\n"
			    + "\n"
			    + "Create a capability object with the given "
			    + "values and store it in the kernel.\n"),
	new JSIPropertyInfo("capability",
			    (1 << JSIMessage.GET_PROPERTY),
			    "name:string",
			    "Get the value of the capability.\n"),
	new JSIPropertyInfo("capability",
			    (1 << JSIMessage.DELETE_PROPERTY),
			    "name:string",
			    "Revoke the capability from the kernel.\n"),
	new JSIPropertyInfo("capability",
			    (1 << JSIMessage.LIST_PROPERTIES),
			    "List the capabilities held by the kernel.\n"),
    };

    /**
     * The object level properties for a capability object.
     */
    private static final JSIPropertyInfo capProperties[] = {
	new JSIPropertyInfo("revokePermissions",
			    (1 << JSIMessage.EXECUTE_PROPERTY),
			    "permission:string=string - Revoke a "
			    + "permission from the object.\n"
			    + "countPermission:string=int - Subtract an "
			    + "amount from a count permission.\n"
			    + "\n"
			    + "Revoke a set of permissions from a "
			    + "capability.\n"),
	new JSIPropertyInfo("grantPermissions",
			    (1 << JSIMessage.EXECUTE_PROPERTY),
			    "permission:string=string - Grant a "
			    + "permission to the object.\n"
			    + "countPermission:string=int - Add an "
			    + "amount to a count permission.\n"
			    + "\n"
			    + "Grant a set of permissions to a "
			    + "capability.\n"),
    };
    
    /**
     * The object level properties for a map object.
     */
    private static final JSIPropertyInfo mapProperties[] = {
	new JSIPropertyInfo("children",
			    (1 << JSIMessage.GET_PROPERTY),
			    "Get the set of children of a sub-tree.\n"),
	new JSIPropertyInfo("read",
			    (1 << JSIMessage.EXECUTE_PROPERTY),
			    "file:string - The file containing the "
			    + "objects.\n"
			    + "password:string - The password to use if "
			    + "the file is encrypted (Optional).\n"
			    + "\n"
			    + "Read the contents of the file and create "
			    + "the objects that it contains.\n"),
	new JSIPropertyInfo("write",
			    (1 << JSIMessage.EXECUTE_PROPERTY),
			    "file:string - The file containing the "
			    + "objects.\n"
			    + "password:string - The password to use when "
			    + "encrypting the file (Optional).\n"
			    + "\n"
			    + "Write the objects contained in the tree "
			    + "to the file.\n"),
    };
    
    /**
     * The object level properties for a map object.
     */
    private static final JSIPropertyInfo envelopeProperties[] = {
	new JSIPropertyInfo("revokePermissions",
			    (1 << JSIMessage.EXECUTE_PROPERTY),
			    "permission:string=string - Revoke a "
			    + "permission from the object.\n"
			    + "countPermission:string=int - Subtract an "
			    + "amount from a count permission.\n"
			    + "\n"
			    + "Revoke a set of permissions from a "
			    + "capability.\n"),
    };
    
    /**
     * The tree of capabilities held by the kernel.
     */
    private static final BDomainLocal tree = new BDomainLocal() {
	    public Object initialValue()
	    {
		return new ObjectTree();
	    }
	};

    public static ObjectTree getTree()
    {
	return (ObjectTree)tree.get();
    }

    public KernelCapabilities()
    {
	ObjectTree ot;

	ot = getTree();
	try
	{
	    BDomainCoroutine sc;
	    Capability cap;

	    sc = BDomain.currentBDomain().getBDomainCoroutine();
	    
	    cap = KernelObjectCoroutine.create(new Console(),
					       sc,
					       new byte[0],
					       null,
					       "Console",
					       Console.DEFAULT_PERMISSIONS);
	    ot.put("console", cap);
	    
	    cap = KernelObjectCoroutine.
		create(new FileSystemRoot(),
		       sc,
		       new PathSpec(System.getProperty("user.dir")).toBytes(),
		       null,
		       "Current working directory",
		       null);
	    ot.put("cwd", cap);
	    
	    cap = KernelObjectCoroutine.
		create(new Clock(),
		       sc,
		       new byte[0],
		       null,
		       "Clock",
		       Clock.DEFAULT_PERMISSIONS);
	    ot.put("clock", cap);
	    
	    cap = KernelObjectCoroutine.
		create(new Account(),
		       sc,
		       new AccountSpec(new PairChain()).toBytes(),
		       null,
		       "Machine account",
		       null);
	    ot.put("machine-account", cap);
	}
	catch(Throwable th)
	{
	    th.printStackTrace();
	}
    }

    private static RangePermission parseRangeString(String name, String args[])
    {
	String range, actions = "read,write,execute";
	RangePermission retval;
	StringTokenizer st;

	if( args.length == 2)
	{
	    actions = args[0];
	    range = args[1];
	}
	else
	{
	    range = args[0];
	}
	retval = new RangePermission(name, actions);
	st = new StringTokenizer(range, ",");
	while( st.hasMoreTokens() )
	{
	    StringTokenizer st2;
	    RangePermission rp;
	    long start, end;
	    
	    st2 = new StringTokenizer(st.nextToken(), "-");
	    switch( st2.countTokens() )
	    {
	    case 1:
		start = Long.decode(st2.nextToken()).longValue();
		end = start;
		break;
	    case 2:
		start = Long.decode(st2.nextToken()).longValue();
		end = Long.decode(st2.nextToken()).longValue();
		break;
	    default:
		throw new IllegalArgumentException("Malformed range: "
						   + range);
	    }
	    rp = new RangePermission(name, actions, start, end - start + 1);
	    retval.grant(rp);
	}
	return retval;
    }

    public static void toMessage(JSIMessage jm, PermissionSet ps)
    {
	Permission perms[];
	int lpc;

	if( ps == null )
	{
	    return;
	}
	perms = ps.toArray();
	for( lpc = 0; lpc < perms.length; lpc++ )
	{
	    Permission perm;
	    String value;
	    
	    perm = perms[lpc];
	    value = perm.getName() + "=" + perm.getAction();
	    if( perm instanceof MaximumIntegerPermission )
	    {
		IntegerPermission p = (IntegerPermission)perm;
		
		jm.add("maxIntPermission", value + " ;" + p.getValue());
	    }
	    else if( perm instanceof MinimumIntegerPermission )
	    {
		IntegerPermission p = (IntegerPermission)perm;
		
		jm.add("minIntPermission", value + " ;" + p.getValue());
	    }
	    else if( perm instanceof IntegerPermission )
	    {
		IntegerPermission p = (IntegerPermission)perm;
		
		jm.add("intPermission", value + " ;" + p.getValue());
	    }
	    else if( perm instanceof RangePermission )
	    {
		RangePermission p = (RangePermission)perm;
		
		jm.add("rangePermission", value + " ;" + p.getValue());
	    }
	    else if( perm instanceof MaximumFloatPermission )
	    {
		FloatPermission p = (FloatPermission)perm;
		
		jm.add("maxFloatPermission", value + " ;" + p.getValue());
	    }
	    else if( perm instanceof MinimumFloatPermission )
	    {
		FloatPermission p = (FloatPermission)perm;
		
		jm.add("minFloatPermission", value + " ;" + p.getValue());
	    }
	    else if( perm instanceof FloatPermission )
	    {
		FloatPermission p = (FloatPermission)perm;
		
		jm.add("floatPermission", value + " ;" + p.getValue());
	    }
	    else if( perm instanceof CountPermission )
	    {
		CountPermission p = (CountPermission)perm;
		
		jm.add("countPermission", value + " ;" + p.getMaxCount());
	    }
	    else if( perm instanceof EnumerationPermission )
	    {
		EnumerationPermission p = (EnumerationPermission)perm;
		
		jm.add("enumPermission", value + " ;" + p.getValue());
	    }
	    else if( perm instanceof IPAddressPermission )
	    {
		IPAddressPermission p = (IPAddressPermission)perm;
		
		jm.add("ipPermission", value + " ;" + p.getValue());
	    }
	    else if( perm instanceof FilePermission )
	    {
		FilePermission p = (FilePermission)perm;
		
		jm.add("filePermission", value);
	    }
	    else
	    {
		jm.add("permission", value);
	    }
	}
    }

    public static PermissionSet toPermissionSet(JSIMessage jm)
	throws NumberFormatException, AddressFormatException
    {
	Vector permissions = new Vector();
	PermissionSet retval;
	int lpc, size;
	Vector ids;

	ids = jm.getIdentifiers();
	size = ids.size();
	for( lpc = 0; lpc < size; lpc++ )
	{
	    JSIPair jp;

	    jp = (JSIPair)ids.elementAt(lpc);
	    if( jp.getTag().endsWith("ermission") )
	    {
		String type, name, value, args[];
		StringTokenizer st;
		int argc, index;

		type = jp.getTag();
		value = jp.getStringValue();
		index = value.indexOf("=");
		if( index < 0 )
		{
		    throw new IllegalArgumentException(
			"Format for permissions is `name=actions[; ...]'.");
		}
		name = value.substring(0, index);
		name = name.trim();
		st = new StringTokenizer(value.substring(index + 1), ";");
		args = new String[st.countTokens()];
		for( argc = 0; argc < args.length; argc++ )
		{
		    args[argc] = st.nextToken().trim();
		}
		if( type.equals("permission") )
		{
		    permissions.addElement(new Permission(name, args[0]));
		}
		else if( type.equals("countPermission") )
		{
		    permissions.addElement(
			new CountPermission(name,
					    Integer.decode(args[0]).
					    intValue()));
		}
		else if( type.equals("intPermission") )
		{
		    if( args.length == 2 )
		    {
			permissions.addElement(
			    new IntegerPermission(name,
						  args[0],
						  Integer.decode(args[1]).
						  intValue()));
		    }
		    else
		    {
			permissions.addElement(
			    new IntegerPermission(name,
						  Integer.decode(args[0]).
						  intValue()));
		    }
		}
		else if( type.equals("minIntPermission") )
		{
		    if( args.length == 2 )
		    {
			permissions.addElement(
			    new MinimumIntegerPermission(
				name,
				args[0],
				Integer.decode(args[1]).intValue()));
		    }
		    else
		    {
			permissions.addElement(
			    new MinimumIntegerPermission(
				name,
				Integer.decode(args[0]).intValue()));
		    }
		}
		else if( type.equals("maxIntPermission") )
		{
		    if( args.length == 2 )
		    {
			permissions.addElement(
			    new MaximumIntegerPermission(
				name,
				args[0],
				Integer.decode(args[1]).intValue()));
		    }
		    else
		    {
			permissions.addElement(
			    new MaximumIntegerPermission(
				name,
				Integer.decode(args[0]).intValue()));
		    }
		}
		else if( type.equals("rangePermission") )
		{
		    permissions.addElement(parseRangeString(name, args));
		}
		else if( type.equals("floatPermission") )
		{
		    if( args.length == 2 )
		    {
			permissions.addElement(
			    new FloatPermission(name,
						args[0],
						Double.parseDouble(args[1])));
		    }
		    else
		    {
			permissions.addElement(
			    new FloatPermission(name,
						Double.parseDouble(args[0])));
		    }
		}
		else if( type.equals("minFloatPermission") )
		{
		    if( args.length == 2 )
		    {
			permissions.addElement(
			    new MinimumFloatPermission(
				name,
				args[0],
				Double.parseDouble(args[1])));
		    }
		    else
		    {
			permissions.addElement(
			    new MinimumFloatPermission(
				name,
				Double.parseDouble(args[0])));
		    }
		}
		else if( type.equals("maxFloatPermission") )
		{
		    if( args.length == 2 )
		    {
			permissions.addElement(
			    new MaximumFloatPermission(
				name,
				args[0],
				Double.parseDouble(args[1])));
		    }
		    else
		    {
			permissions.addElement(
			    new MaximumFloatPermission(
				name,
				Double.parseDouble(args[0])));
		    }
		}
		else if( type.equals("enumPermission") )
		{
		    if( args.length == 2 )
		    {
			permissions.addElement(
				new EnumerationPermission(name,
							  args[0],
							  args[1]));
		    }
		    else
		    {
			permissions.addElement(
				new EnumerationPermission(name, args[0]));
		    }
		}
		else if( type.equals("ipPermission") )
		{
		    switch( args.length )
		    {
		    case 1:
			permissions.addElement(
				new IPAddressPermission(name, args[0]));
			break;
		    case 2:
			permissions.addElement(
				new IPAddressPermission(name,
							args[0],
							args[1]));
			break;
		    }
		}
		else if( type.equals("filePermission") )
		{
		    permissions.addElement(
			new FilePermission(name, args[0]));
		}
		else
		{
		    Node.println("Unknown permission type: " + type);
		}
	    }
	}
	return new PermissionSet(permissions);
    }

    /**
     * Convert a JSIMessage with `permission' and `countPermission' attributes
     * into a PermissionSet.
     *
     * @param jm The JSIMessage to extract the permissions from.
     * @return An initialized PermissionSet object.
     * @exception Throws a NumberFormatException with a CountPermission's
     * count value isn't a number value.
     */
    public static PermissionSet oldtoPermissionSet(JSIMessage jm)
	throws NumberFormatException, AddressFormatException
    {
	Vector permissions = new Vector();
	PermissionSet retval;
	String permPair;
	int lpc;

	Assert.require(jm);
	
	for( lpc = 0;
	     (permPair = jm.get("permission", lpc, null)) != null;
	     lpc++ )
	{
	    String permission, actions;
	    int index;
	    
	    index = permPair.indexOf("=");
	    if( index < 0 )
	    {
		throw new IllegalArgumentException(
			"Format for `permission' is `name=actions'.");
	    }
	    permission = permPair.substring(0, index);
	    actions = permPair.substring(index + 1);
	    permissions.addElement(new Permission(permission, actions));
	}
	for( lpc = 0;
	     (permPair = jm.get("countPermission", lpc, null)) != null;
	     lpc++ )
	{
	    String permission, count;
	    int index;
	    
	    index = permPair.indexOf("=");
	    if( index < 0 )
	    {
		throw new IllegalArgumentException(
			"Format for `countPermission' is `name=count'.");
	    }
	    permission = permPair.substring(0, index);
	    count = permPair.substring(index + 1);
	    permissions.addElement(new CountPermission(permission,
						       Integer.decode(count).
						       intValue()));
	}
	for( lpc = 0;
	     (permPair = jm.get("minIntPermission", lpc, null)) != null;
	     lpc++ )
	{
	    String permission, integer;
	    int index;
	    
	    index = permPair.indexOf("=");
	    if( index < 0 )
	    {
		throw new IllegalArgumentException(
			"Format for `countPermission' is `name=count'.");
	    }
	    permission = permPair.substring(0, index);
	    integer = permPair.substring(index + 1);
	    permissions.addElement(
		new MinimumIntegerPermission(permission,
					     Long.decode(integer).
					     intValue()));
	}
	for( lpc = 0;
	     (permPair = jm.get("maxIntPermission", lpc, null)) != null;
	     lpc++ )
	{
	    String permission, integer;
	    int index;
	    
	    index = permPair.indexOf("=");
	    if( index < 0 )
	    {
		throw new IllegalArgumentException(
			"Format for `countPermission' is `name=count'.");
	    }
	    permission = permPair.substring(0, index);
	    integer = permPair.substring(index + 1);
	    permissions.addElement(
		new MaximumIntegerPermission(permission,
					     Long.decode(integer).
					     intValue()));
	}
	for( lpc = 0;
	     (permPair = jm.get("enumPermission", lpc, null)) != null;
	     lpc++ )
	{
	    String permission, string;
	    int index;
	    
	    index = permPair.indexOf("=");
	    if( index < 0 )
	    {
		throw new IllegalArgumentException(
			"Format for `enumPermission' is `name=string'.");
	    }
	    permission = permPair.substring(0, index);
	    string = permPair.substring(index + 1);
	    permissions.addElement(new EnumerationPermission(permission,
							     string));
	}
	for( lpc = 0;
	     (permPair = jm.get("rangePermission", lpc, null)) != null;
	     lpc++ )
	{
	    String permission, range;
	    int index;
	    
	    index = permPair.indexOf("=");
	    if( index < 0 )
	    {
		throw new IllegalArgumentException(
			"Format for `rangePermission' is `name=range'.");
	    }
	    permission = permPair.substring(0, index);
	    range = permPair.substring(index + 1);
	    // permissions.addElement(parseRangeString(permission, range));
	}
	for( lpc = 0;
	     (permPair = jm.get("ipPermission", lpc, null)) != null;
	     lpc++ )
	{
	    String permission, range;
	    int index;
	    
	    index = permPair.indexOf("=");
	    if( index < 0 )
	    {
		throw new IllegalArgumentException(
			"Format for `ipPermission' is `name=ip'.");
	    }
	    permission = permPair.substring(0, index);
	    range = permPair.substring(index + 1);
	    permissions.addElement(new IPAddressPermission(permission, range));
	}
	for( lpc = 0;
	     (permPair = jm.get("filePermission", lpc, null)) != null;
	     lpc++ )
	{
	    String permission, string;
	    int index;
	    
	    index = permPair.indexOf("=");
	    if( index < 0 )
	    {
		throw new IllegalArgumentException(
			"Format for `filePermission' is `name=perms'.");
	    }
	    permission = permPair.substring(0, index);
	    string = permPair.substring(index + 1);
	    permissions.addElement(new FilePermission(permission, string));
	}
	retval = new PermissionSet(permissions);
	return retval;
    }

    /**
     * Use an ObjectMap as a result in a JSIMessage.
     *
     * @param jm The JSIMessage to add the results too.
     * @param om The ObjectMap whose key values are to be added to the
     * JSIMessage.
     */
    private void addObjectMap(JSIMessage jm, ObjectMap om)
    {
	synchronized( om )
	{
	    Enumeration enum;
	    
	    enum = om.keys();
	    while( enum.hasMoreElements() )
	    {
		Object obj;
		String key;
		
		key = (String)enum.nextElement();
		obj = om.get(key);
		if( obj instanceof ObjectMap )
		{
		    jm.add("result", key + "/");
		}
		else if( obj instanceof Capability )
		{
		    jm.add("result", key);
		}
		else if( obj instanceof CapabilityEnvelope )
		{
		    jm.add("result", key);
		}
		else
		{
		    throw new KernelPanic("Bogon in capability tree: " + obj);
		}
	    }
	}
    }

    /**
     * Terminator record type, this indicates the end of an object map listing
     * when processing records.
     */
    private static final int TYPE_TERMINATOR = 1;

    /**
     * Capability record, the capabilities contents are written.
     */
    private static final int TYPE_CAPABILITY = 2;

    /**
     * ObjectMap record, the map's children are written inline, ending with a
     * terminator record.
     */
    private static final int TYPE_OBJECT_MAP = 3;

    /**
     * ObjectMap record, the map's children are written inline, ending with a
     * terminator record.
     */
    private static final int TYPE_ENVELOPE = 4;

    /**
     * The password encrypted cipher algorithm.
     */
    private static final String KEY_CIPHER = "PBEWithSHAAnd3-KeyTripleDES-CBC";

    /**
     * The size of the salt we'll use to initialize the cipher.
     */
    private static final int KEY_SALT_SIZE = 20;

    /**
     * The minimum number of iterations used to initialize the cipher.
     */
    private static final int MIN_ITERATIONS = 1024;

    /**
     * Create and initialize a PBE based cipher.
     *
     * @param mode The encrypt/decrypt mode constant.
     * @param password The password used to cipher.
     * @param salt The salt value for the PBEParameterSpec.
     * @param iterationCount The iterationCount value for the PBEParameterSpec.
     * @return The initialized Cipher.
     * @exception Throws a number of crypto related exceptions.
     */
    private Cipher makePBECipher(int mode,
				 char password[],
				 byte salt[],
				 int iterationCount)
	throws GeneralSecurityException
    {
	PBEParameterSpec pbs;
	SecretKeyFactory skf;
	PBEKeySpec pks;
	Cipher retval;
	SecretKey sk;

	pks = new PBEKeySpec(password);
	pbs = new PBEParameterSpec(salt, iterationCount);
	skf = SecretKeyFactory.getInstance(KEY_CIPHER);
	retval = Cipher.getInstance(KEY_CIPHER);
	sk = skf.generateSecret(pks);
	retval.init(mode, sk, pbs);
	return retval;
    }

    /**
     * Create and initialize a CipherOutputStream that wraps the given stream
     * and encrypts the output using the given password.
     *
     * @param os The existing OutputStream.
     * @param password The password used to encrypt the stream.
     * @return An initialized CipherOutputStream
     * @exception Throws a number of crypto related exceptions and an
     * IOException if there's a problem writing the initialization vector to
     * the stream.
     */
    private OutputStream makeCipherOutputStream(DataOutputStream os,
						char password[])
	throws GeneralSecurityException,
	       IOException
    {
	SecureRandom sr = new SecureRandom();
	CipherOutputStream retval;
	Cipher cipher;
	byte salt[];

	salt = new byte[KEY_SALT_SIZE];
	sr.setSeed(System.currentTimeMillis());
	sr.nextBytes(salt);
	cipher = this.makePBECipher(Cipher.ENCRYPT_MODE,
				    password,
				    salt,
				    MIN_ITERATIONS);
	os.writeInt(salt.length);
	os.write(salt);
	os.writeInt(MIN_ITERATIONS);
	retval = new CipherOutputStream(os, cipher);
	return retval;
    }

    /**
     * Create and initialize a CipherInputStream that wraps the given stream
     * and decrypts the input using the given password.
     *
     * @param os The existing InputStream.
     * @param password The password used to decrypt the stream.
     * @return An initialized CipherInputStream
     * @exception Throws a number of crypto related exceptions and an
     * IOException if there's a problem reading the initialization vector from
     * the stream.
     */
    private InputStream makeCipherInputStream(DataInputStream is,
					      char password[])
	throws GeneralSecurityException,
	       IOException
    {
	CipherInputStream retval;
	int iterationCount;
	Cipher cipher;
	byte salt[];

	salt = new byte[is.readInt()];
	is.readFully(salt);
	iterationCount = is.readInt();
	cipher = this.makePBECipher(Cipher.DECRYPT_MODE,
				    password,
				    salt,
				    iterationCount);
	retval = new CipherInputStream(is, cipher);
	return retval;
    }

    /**
     * Write an ObjectMap's children to the stream.
     *
     * @param om The ObjectMap to write to the stream.
     * @param cs The streamer used to write the capabilities.
     * @param os The stream to write the capabilities too.
     * @exception Throws an IOException if there's a problem writing the data.
     */
    private void writeObjectMap(ObjectMap om,
				CapabilityStreamer cs,
				DataOutputStream os)
	throws IOException
    {
	synchronized( om )
	{
	    Enumeration keys;

	    keys = om.keys();
	    while( keys.hasMoreElements() )
	    {
		String name = (String)keys.nextElement();
		Object obj;
		
		obj = om.get(name);
		os.writeUTF(name);
		if( obj instanceof Capability )
		{
		    Capability cap = (Capability)obj;

		    os.writeInt(TYPE_CAPABILITY);
		    cs.write(cap, os);
		}
		else if( obj instanceof ObjectMap )
		{
		    os.writeInt(TYPE_OBJECT_MAP);
		    this.writeObjectMap((ObjectMap)obj, cs, os);
		}
		else if( obj instanceof CapabilityEnvelope )
		{
		    os.writeInt(TYPE_ENVELOPE);
		    CapabilityFactory.writeEnvelope(
			KernelObjectStore.getPrimary(),
			os,
			(CapabilityEnvelope)obj);
		}
		else
		{
		    throw new KernelPanic("Bogus object in capability tree: "
					  + obj);
		}
	    }
	    os.writeUTF("");
	    os.writeInt(TYPE_TERMINATOR);
	}
    }

    /**
     * Write an ObjectMap's children to a stream.
     *
     * @param om The ObjectMap to write to the stream.
     * @param jm The JSIMessage to use when setting everything up.
     * @exception Throws a MissingPairException if `jm' doesn't have the right
     * attributes, a number of crypto related exceptions, and an IOException if
     * there's a problem writing the data.
     */
    private synchronized void writeObjectMap(ObjectMap om, JSIMessage jm)
	throws GeneralSecurityException,
	       MissingPairException,
	       IOException
    {
	CapabilityStreamer cs;
	FileOutputStream fos;
	OutputStream os;
	String password;
	
	os = fos = new FileOutputStream(jm.expect("file"));
	cs = new CapabilityStreamer(KernelObjectStore.getPrimary());
	if( (password = jm.get("password")) != null )
	{
	    os = this.makeCipherOutputStream(new DataOutputStream(fos),
					     password.toCharArray());
	}
	os = new DataOutputStream(os);
	this.writeObjectMap(om, cs, (DataOutputStream)os);
	os.close();
	jm.reply("ok");
    }

    /**
     * Reads capabilities from a stream and adds them to the given ObjectMap.
     *
     * @param om The ObjectMap to add the decoded capabilities too.
     * @param cs The streamer used to read the capabilities.
     * @param is The stream to read the capabilities from.
     * @exception Throws an IOException if theres a problem reading from the
     * stream and throws a number of exceptions if theres a problem creating
     * the capability.
     */
    private synchronized void readObjectMap(ObjectMap om,
					    CapabilityStreamer cs,
					    DataInputStream is)
	throws IOException
    {
	boolean done = false;
	
	while( !done )
	{
	    Object value = null;
	    String name;
	    int type;
		
	    name = is.readUTF();
	    type = is.readInt();
	    switch( type )
	    {
	    case TYPE_TERMINATOR:
		done = true;
		break;
	    case TYPE_CAPABILITY:
		try
		{
		    value = cs.read(is);
		}
		catch(IllegalAccessException e)
		{
		    Node.println("Skipping: " + name);
		    Node.printStackTrace(e);
		    value = null;
		}
		catch(InstantiationException e)
		{
		    Node.println("Skipping: " + name);
		    Node.printStackTrace(e);
		    value = null;
		}
		catch(ClassNotFoundException e)
		{
		    Node.println("Skipping: " + name);
		    Node.printStackTrace(e);
		    value = null;
		}
		break;
	    case TYPE_OBJECT_MAP:
		{
		    HashMap hm;
		    
		    hm = new HashMap();
		    this.readObjectMap(hm, cs, is);
		    value = hm;
		}
		break;
	    case TYPE_ENVELOPE:
		try
		{
		    value = CapabilityFactory.readEnvelope(
			KernelObjectStore.getPrimary(),
			is);
		}
		catch(ClassNotFoundException e)
		{
		    Node.println("Skipping: " + name);
		    Node.printStackTrace(e);
		}
		break;
	    default:
		throw new IOException("Unknown type: " +
				      Integer.toHexString(type));
	    }
	    if( value != null )
	    {
		om.put(name, value);
	    }
	}
    }

    private void expandMap(ObjectMap om)
    {
	Enumeration enum;
	
	enum = om.keys();
	while( enum.hasMoreElements() )
	{
	    Object key, elem;

	    key = enum.nextElement();
	    elem = om.get(key);
	    if( elem instanceof Continuation )
	    {
		try
		{
		    om.put(key, ((Continuation)elem).run());
		}
		catch(Throwable th)
		{
		    throw new KernelPanic(th);
		}
	    }
	    else if( elem instanceof ObjectMap )
	    {
		this.expandMap((ObjectMap)elem);
	    }
	}
    }

    /**
     * Reads capabilities from a stream and adds them to the given ObjectMap.
     *
     * @param om The ObjectMap to add the decoded capabilities too.
     * @param jm The JSIMessage used to setup everything.
     * @exception Throws an IOException if theres a problem reading from the
     * stream, a number of exceptions if theres a problem creating the
     * capability, a MissingPairException if `jm' is missing some required
     * attributes, and a number of crypto related exceptions.
     */
    private synchronized void readObjectMap(ObjectMap om, JSIMessage jm)
	throws GeneralSecurityException,
	       IllegalAccessException,
	       InstantiationException,
	       ClassNotFoundException,
	       MissingPairException,
	       IOException
    {
	CapabilityStreamer cs;
	FileInputStream fis;
	String password;
	InputStream is;

	is = fis = new FileInputStream(jm.expect("file"));
	cs = new CapabilityStreamer(KernelObjectStore.getPrimary());
	if( (password = jm.get("password")) != null )
	{
	    is = this.makeCipherInputStream(new DataInputStream(fis),
					    password.toCharArray());
	}
	is = new DataInputStream(is);
	this.readObjectMap(om, cs, (DataInputStream)is);
	this.expandMap(om);
	is.close();
	jm.reply("ok");
    }

    public static void getChildren(JSIMessage jm, JSIPair jp, Class cl)
    {
	Enumeration enum;
	ObjectMap om;
	
	if( jp.getValue().length == 0 )
	{
	    om = getTree().getRootMap();
	}
	else
	{
	    om = (ObjectMap)getTree().get(jp.getStringValue());
	}
	jm.clear();
	jm.setWhat(JSIMessage.REPLY);
	synchronized( tree )
	{
	    enum = om.keys();
	    while( enum.hasMoreElements() )
	    {
		String name;
		Object obj;
		
		name = (String)enum.nextElement();
		obj = om.get(name);
		if( obj instanceof ObjectMap )
		{
		    jm.add("result", name + "/");
		}
		else if( cl.isAssignableFrom(obj.getClass()) )
		{
		    jm.add("result", name);
		}
		else
		{
		    /* Ignore it... */
		}
	    }
	}
    }
    
    public JSIHandler resolveSpecifier(JSIMessage jm, JSIPair jp)
    {
	return this;
    }
    
    public void handleMessage(JSIHandlerContext jhc, JSIMessage jm)
	throws GeneralSecurityException,
	       AddressFormatException,
	       ClassNotFoundException,
	       InstantiationException,
	       IllegalAccessException,
	       MissingPairException,
	       IOException
    {
	JSIPair jp;
	
	switch( jm.getWhat() )
	{
	case JSIMessage.CREATE_PROPERTY:
	    if( jm.popSpecifier() == null )
	    {
		Domain domain = Domain.currentDomain();
		String name;

		name = jm.expect("name");
		if( (jp = jm.getPair("id")) != null )
		{
		    PermissionSet ps;
		    String className;
		    KernelObject ko;
		    Capability cap;
		    Class cl;
		    long id;

		    id = jp.getIntValue();
		    className = jm.expect("class");
		    cl = domain.loadClass(className);
		    cap = (Capability)cl.newInstance();
		    ko = KernelObjectStore.getPrimary().get(id);
		    ps = this.toPermissionSet(jm);
		    if( CapabilityFactory.instantiate(cap,
						      id,
						      ko.getData(),
						      ps) == cap )
		    {
			getTree().put(name, cap);
		    }
		}
		else if( (jp = jm.getPair("capability")) != null )
		{
		    Capability cap;

		    cap = (Capability)getTree().get(jp.getStringValue());
		    getTree().put(name, new CapabilityEnvelope(cap));
		}
		else
		{
		    getTree().put(name, new HashMap());
		}
		jm.reply("ok");
	    }
	    else
	    {
		jm.error("Too many specifiers");
	    }
	    break;
	case JSIMessage.DELETE_PROPERTY:
	    if( jm.popSpecifier() == null )
	    {
		Capability cap;
		
		jm.restoreSpecifier();
		jp = jm.popSpecifier();
		getTree().remove(jp.getStringValue());
		// cap.revoke(); XXX Should we?
		jm.reply("ok");
	    }
	    else
	    {
		jm.error("Too many specifiers");
	    }
	    break;
	case JSIMessage.GET_PROPERTY:
	    try
	    {
		Object obj;
		
		jm.restoreSpecifier();
		jp = jm.popSpecifier();
		if( (jp.getValue().length == 0) ||
		    (jp.getStringValue().equals("/")) )
		{
		    obj = getTree().getRootMap();
		}
		else
		{
		    obj = getTree().get(jp.getStringValue());
		}
		if( (jp = jm.popSpecifier()) != null )
		{
		    if( jm.popSpecifier() != null )
		    {
			jm.error("Too many specifiers");
		    }
		    else if( obj instanceof ObjectMap )
		    {
			if( jp.getTag().equals("children") )
			{
			    jm.clear();
			    jm.setWhat(JSIMessage.REPLY);
			    this.addObjectMap(jm, (ObjectMap)obj);
			}
			else
			{
			    jm.error("Unknown property: " + jp.getTag());
			}
		    }
		    else if( obj instanceof Capability )
		    {
			if( jp.getTag().equals("type") )
			{
			    jm.reply(obj.getClass().getName());
			}
			else if( jp.getTag().equals("permissions") )
			{
			    jm.clear();
			    jm.setWhat(JSIMessage.REPLY);
			    toMessage(jm,
				      CapabilityFactory.
				      getPermissions((Capability)obj));
			}
			else
			{
			    jm.error("Unknown property: " + jp.getTag());
			}
		    }
		    else if( obj instanceof CapabilityEnvelope )
		    {
			if( jp.getTag().equals("permissions") )
			{
			    jm.clear();
			    jm.setWhat(JSIMessage.REPLY);
			    toMessage(jm,
				      CapabilityFactory.
				      getPermissions((CapabilityEnvelope)obj));
			}
			else
			{
			    jm.error("Unknown property: " + jp.getTag());
			}
		    }
		    else
		    {
			jm.error("Unknown property: " + jp.getTag());
		    }
		}
		else
		{
		    jm.reply(obj.toString());
		}
	    }
	    catch(NoSuchElementException e)
	    {
		jm.error(e.toString());
	    }
	    break;
	case JSIMessage.LIST_PROPERTIES:
	    jm.restoreSpecifier();
	    jp = jm.popSpecifier();
	    if( jm.popSpecifier() != null )
	    {
		jm.error("Too many specifiers");
	    }
	    else if( jp.getValue().length == 0 )
	    {
		jm.clear();
		jm.setWhat(JSIMessage.REPLY);
		synchronized( tree )
		{
		    ObjectMap om = getTree().getRootMap();

		    this.addObjectMap(jm, om);
		}
	    }
	    else
	    {
		Object obj;

		obj = getTree().get(jp.getStringValue());
		jm.clear();
		jm.setWhat(JSIMessage.REPLY);
		if( obj instanceof Capability )
		{
		    jm.add("result", "revokePermissions");
		    jm.add("result", "grantPermissions");
		}
		else if( obj instanceof ObjectMap )
		{
		    jm.add("result", "children");
		}
		else if( obj instanceof CapabilityEnvelope )
		{
		    jm.add("result", "revokePermissions");
		    jm.add("result", "boundPermissions");
		}
		else
		{
		    throw new KernelPanic("Bogon in capability tree: " + obj);
		}
	    }
	    break;
	case JSIMessage.EXECUTE_PROPERTY:
	    {
		Object obj;
		
		jm.restoreSpecifier();
		jp = jm.popSpecifier();
		if( jp.getValue().length == 0 )
		    obj = getTree().getRootMap();
		else
		    obj = getTree().get(jp.getStringValue());
		if( (jp = jm.popSpecifier()) != null )
		{
		    if( jm.popSpecifier() != null )
		    {
			jm.error("Too many specifiers");
		    }
		    else if( obj instanceof Capability )
		    {
			Capability cap = (Capability)obj;
			
			if( jp.getTag().equals("revokePermissions") )
			{
			    cap.revokePermissions(this.toPermissionSet(jm));
			    jm.reply("ok");
			}
			else if( jp.getTag().equals("grantPermissions") )
			{
			    CapabilityFactory.grant(cap,
						    this.toPermissionSet(jm));
			    jm.reply("ok");
			}
			else
			{
			    jm.error("Unknown property: " + jp.getTag());
			}
		    }
		    else if( obj instanceof CapabilityEnvelope )
		    {
			CapabilityEnvelope ce = (CapabilityEnvelope)obj;
			
			if( jp.getTag().equals("revokePermissions") )
			{
			    ce.revokePermissions(this.toPermissionSet(jm));
			    jm.reply("ok");
			}
			else if( jp.getTag().equals("boundPermissions") )
			{
			    ce.boundPermissions(this.toPermissionSet(jm));
			    jm.reply("ok");
			}
			else if( jp.getTag().equals("grantPermissions") )
			{
			    CapabilityFactory.getPermissions(ce).
				grantPermissions(this.toPermissionSet(jm));
			    jm.reply("ok");
			}
			else
			{
			    jm.error("Unknown property: " + jp.getTag());
			}
		    }
		    else if( obj instanceof ObjectMap )
		    {
			ObjectMap om = (ObjectMap)obj;

			if( jp.getTag().equals("write") )
			{
			    this.writeObjectMap(om, jm);
			}
			else if( jp.getTag().equals("read") )
			{
			    this.readObjectMap(om, jm);
			}
			else
			{
			    jm.error("Unknown property: " + jp.getTag());
			}
		    }
		    else
		    {
			throw new KernelPanic("Bogon in capability tree: "
					      + obj);
		    }
		}
		else
		{
		    jm.error("No property specified to execute.");
		}
	    }
	    break;
	case JSIMessage.GET_SUITES:
	    {
		Object obj;
		
		jm.restoreSpecifier();
		jp = jm.popSpecifier();
		if( jp.getValue().length == 0 )
		    obj = getTree().getRootMap();
		else
		    obj = getTree().get(jp.getStringValue());
		if( jm.popSpecifier() == null )
		{
		    if( obj instanceof Capability )
		    {
			jm.reply(capProperties);
		    }
		    else if( obj instanceof ObjectMap )
		    {
			jm.reply(mapProperties);
		    }
		    else if( obj instanceof CapabilityEnvelope )
		    {
			jm.reply(envelopeProperties);
		    }
		    else
		    {
			throw new KernelPanic("Bogon in capability tree: "
					      + obj);
		    }
		}
		else
		{
		    jm.error("Too many specifiers");
		}
	    }
	    break;
	default:
	    throw new UnsupportedOperationException();
	}
    }

    public void getSuites(JSIMessage jm)
    {
	jm.add("result", properties);
    }

    public String toString()
    {
	return "KernelCapabilities["
	    + "]";
    }
}
