/*
 * KernelObjectCoroutine.java
 *
 * Copyright (c) 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is part of the Janos Bees Software.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */

package bees.core.kernel;

import edu.utah.janosvm.sys.DeadTeamException;
import edu.utah.janosvm.sys.TeamEjectionException;

import bees.core.Assert;
import bees.core.Strict;
import bees.core.BeesEvent;
import bees.core.ByteArray;
import bees.core.Coroutine;
import bees.core.Capability;
import bees.core.StateResponder;
import bees.core.CapabilityFactory;
import bees.core.CapabilityEnvelope;
import bees.core.ResponderCoroutine;

import bees.core.security.PermissionSet;

/**
 * Coroutine used to manage the capability material stored in the kernel.
 */
public final class KernelObjectCoroutine
    extends ResponderCoroutine
{
    /**
     * State used when waiting for a reply.
     */
    public static final int STATE_WAITING = STATE_USER;
    public static final int STATE_LOADED = STATE_USER + 1;
    public static final int STATE_SUB_LOADED = STATE_USER + 2;

    /**
     * Helper function for creating a group of KernelObject's.
     *
     * @param sc The current domain's coroutine.
     * @param data An array of capability material to create in the kernel.
     * @param comment The comment to attach to all the KernelObject's.
     */
    public static boolean create(BDomainCoroutine sc,
				 Capability caps[],
				 byte data[][],
				 String comment)
    {
	KernelObjectCoroutine koc;
	boolean retval = true;
	int lpc;

	Assert.require(sc);
	Assert.require(data);

	for( lpc = 0; retval && (lpc < data.length); lpc++ )
	{
	    koc = new KernelObjectCoroutine();
	    koc.create(sc, caps[lpc], data[lpc], null, comment + ": " + lpc);
	    switch( koc.waitForState(STATE_LOADED) )
	    {
	    case STATE_LOADED:
		koc.instantiate(null);
		break;
	    default:
		retval = false;
		break;
	    }
	}
	if( !retval )
	{
	    for( ; lpc >= 0; lpc-- )
	    {
		/* XXX */
		// koc = new KernelObjectCoroutine();
		// koc.delete(sc, retval[lpc]);
	    }
	}
	return retval;
    }

    /**
     * Helper function for creating and initializing a Capability.
     *
     * @param cap The Capability to initialize.
     * @param sc The current domain's coroutine.
     * @param data The capability material.
     * @param comment The comment to attach to the KernelObject.
     * @param ps The PermissionSet to attach to the Capability.
     */
    private static Object create(Object cap,
				 BDomainCoroutine sc,
				 byte data[],
				 CapabilityEnvelope sub[],
				 String comment,
				 PermissionSet ps)
    {
	KernelObjectCoroutine koc;
	Object retval = null;

	Assert.require(cap);
	Assert.require(sc);
	Assert.require(data);

	koc = new KernelObjectCoroutine();
	koc.create(sc, cap, data, sub, ps, comment);
	switch( koc.waitForState(STATE_LOADED) )
	{
	case STATE_LOADED:
	    koc.instantiate(ps);
	    retval = koc.cap;
	    break;
	case STATE_ERROR:
	    try
	    {
		throw koc.getThrowable();
	    }
	    catch(RuntimeException re)
	    {
		throw re;
	    }
	    catch(Exception e)
	    {
		throw new ExceptionInInitializerError(e);
	    }
	    catch(Error e)
	    {
		throw e;
	    }
	    catch(Throwable th)
	    {
		th.printStackTrace();
	    }
	default:
	    break;
	}
	return retval;
    }

    public static Capability create(Capability cap,
				    BDomainCoroutine sc,
				    byte data[],
				    CapabilityEnvelope sub[],
				    String comment,
				    PermissionSet ps)
    {
	return (Capability)create((Object)cap, sc, data, sub, comment, ps);
    }
    
    public static Capability create(Capability cap,
				    BDomainCoroutine sc,
				    byte data[],
				    String comment,
				    PermissionSet ps)
    {
	return (Capability)create((Object)cap, sc, data, null, comment, ps);
    }
    
    public static Capability[] create(Capability cap[],
				      BDomainCoroutine sc,
				      byte data[],
				      String comment,
				      PermissionSet ps)
    {
	return (Capability[])create((Object)cap, sc, data, null, comment, ps);
    }
    
    /**
     * Helper function for creating and initializing a Capability.
     *
     * @param cap The Capability to initialize.
     * @param sc The current domain's coroutine.
     * @param data The capability material.
     * @param comment The comment to attach to the KernelObject.
     */
    public static Capability create(Capability cap,
				    BDomainCoroutine sc,
				    byte data[],
				    String comment)
    {
	return create(cap, sc, data, comment, null);
    }

    public static void delete(long id, BDomainCoroutine sc)
    {
	KernelObjectCoroutine koc;

	koc = new KernelObjectCoroutine();
	koc.delete(sc, id);
    }

    /**
     * The capability material.
     */
    private byte data[];

    private CapabilityEnvelope sub[];

    /**
     * The KernelObject's unique identifier.
     */
    private long storeID;

    /**
     * The created/found capability.
     */
    private Object cap;

    /**
     * Construct a KernelObjectCoroutine with the given value.
     *
     * @param sr A StateResponder that should be called when the Coroutine's
     * state changes.
     */
    public KernelObjectCoroutine(StateResponder sr)
    {
	super(sr);
    }

    /**
     * Construct an empty KernelObjectCoroutine.
     */
    public KernelObjectCoroutine()
    {
	this(null);
    }

    /**
     * @return The capability material.
     */
    public synchronized byte[] getData()
    {
	return this.data;
    }

    public synchronized CapabilityEnvelope[] getSubcapabilities()
    {
	return this.sub;
    }

    /**
     * @return The KernelObject's unique identifier.
     */
    public synchronized long getStoreID()
    {
	return this.storeID;
    }

    public synchronized Object getCapability()
    {
	return this.cap;
    }

    public synchronized void instantiate(PermissionSet ps)
    {
	switch( this.getState() )
	{
	case STATE_LOADED:
	case STATE_SUB_LOADED:
	    try
	    {
		if( this.cap instanceof Capability )
		{
		    this.cap = CapabilityFactory.
			instantiate((Capability)this.cap,
				    this.storeID,
				    this.data,
				    this.sub,
				    ps);
		}
		else if( this.cap instanceof Capability[] )
		{
		    Capability caps[] = (Capability[])this.cap;
		    int lpc;
		    
		    for( lpc = 0; lpc < caps.length; lpc++ )
		    {
			CapabilityFactory.instantiate(caps[lpc],
						      this.storeID,
						      this.data,
						      null,
						      ps);
		    }
		}
		this.setState(STATE_COMPLETE);
	    }
	    catch(Throwable th)
	    {
		super.setThrowable(th);
	    }
	    break;
	case STATE_COMPLETE:
	    break;
	default:
	    Strict.fail("Invalid state: " + this);
	    break;
	}
    }

    /**
     * Send a KernelObjectEvent to the kernel.
     *
     * @param sc The current domain's coroutine.
     * @param koe The KernelObjectEvent to send to the kernel.
     */
    private void sendEvent(BDomainCoroutine sc, KernelObjectEvent koe)
    {
	try
	{
	    Assert.require(sc);
	    Assert.require(koe);
	    
	    koe.setServerCoroutineID(sc.getPeerID());
	    koe.setReplyPort(sc.getLocalPort());
	    sc.getRemotePort().putFacsimile(koe);
	    this.setState(STATE_WAITING);
	}
	catch(DeadTeamException e)
	{
	    this.setState(STATE_ERROR);
	}
	catch(TeamEjectionException e)
	{
	    this.setState(STATE_ERROR);
	}
    }

    /**
     * Create a KernelObject in the kernel.
     *
     * @param sc The current domain's coroutine.
     * @param data The capability material.
     * @param comment The comment to attach to the KernelObject.
     */
    private synchronized void create(BDomainCoroutine sc,
				     Object cap,
				     byte data[],
				     CapabilityEnvelope sub[],
				     PermissionSet ps,
				     String comment)
    {
	try
	{
	    KernelObjectEvent koe;
	    
	    Assert.require(sc);
	    Assert.require(data);

	    this.cap = cap;
	    this.data = data;
	    this.sub = sub;
	    koe = new KernelObjectEvent(this,
					KernelObjectEvent.METHOD_CREATE,
					data,
					sub,
					ps,
					comment,
					KernelObjectStore.INVALID_ID);
	    this.sendEvent(sc, koe);
	}
	catch(Throwable th)
	{
	    super.setThrowable(th);
	}
    }
    
    public void create(BDomainCoroutine sc,
		       Capability cap,
		       byte data[],
		       PermissionSet ps,
		       String comment)
    {
	this.create(sc, (Object)cap, data, null, ps, comment);
    }
    
    public void create(BDomainCoroutine sc,
		       Capability cap,
		       byte data[],
		       CapabilityEnvelope sub[],
		       PermissionSet ps,
		       String comment)
    {
	this.create(sc, (Object)cap, data, sub, ps, comment);
    }
    
    public void create(BDomainCoroutine sc,
		       Capability cap[],
		       byte data[],
		       PermissionSet ps,
		       String comment)
    {
	this.create(sc, (Object)cap, data, null, ps, comment);
    }
    
    /**
     * Lookup the value of a KernelObject.
     *
     * @param sc The current domain's coroutine.
     * @param storeID The KernelObject's unique identifier.
     */
    private synchronized void lookup(BDomainCoroutine sc,
				     Object cap,
				     long storeID)
    {
	try
	{
	    KernelObjectEvent koe;
	    
	    Assert.require(sc);
	    Assert.require(storeID >= 0);

	    this.cap = cap;
	    this.storeID = storeID;
	    koe = new KernelObjectEvent(this,
					KernelObjectEvent.METHOD_LOOKUP,
					null,
					null,
					null,
					null,
					storeID);
	    this.sendEvent(sc, koe);
	}
	catch(Throwable th)
	{
	    super.setThrowable(th);
	}
    }

    public void lookup(BDomainCoroutine sc, Capability cap, long storeID)
    {
	this.lookup(sc, (Object)cap, storeID);
    }
    
    public void lookup(BDomainCoroutine sc, Capability cap[], long storeID)
    {
	this.lookup(sc, (Object)cap, storeID);
    }
    
    /**
     * Delete a KernelObject.
     *
     * @param sc The current domain's coroutine.
     * @param id The KernelObject's unique identifier.
     */
    public synchronized void delete(BDomainCoroutine sc, long id)
    {
	try
	{
	    KernelObjectEvent koe;
	    
	    Assert.require(sc);
	    Assert.require(id >= 0);

	    koe = new KernelObjectEvent(this,
					KernelObjectEvent.METHOD_DELETE,
					null,
					null,
					null,
					null,
					id);
	    this.sendEvent(sc, koe);
	    this.setState(STATE_COMPLETE);
	}
	catch(Throwable th)
	{
	    super.setThrowable(th);
	}
    }

    public synchronized void handle(BeesEvent ae)
    {
	switch( this.getState() )
	{
	case STATE_WAITING:
	    if( (ae instanceof KernelObjectEvent) && ae.isReply() )
	    {
		KernelObjectEvent koe = (KernelObjectEvent)ae;
		Throwable th;

		if( (th = koe.getThrowable()) != null )
		{
		    super.setThrowable(th);
		}
		else
		{
		    switch( koe.getMethod() )
		    {
		    case KernelObjectEvent.METHOD_CREATE:
			this.storeID = koe.getID();

			Assert.ensure(this.storeID >= 0);
			break;
		    case KernelObjectEvent.METHOD_LOOKUP:
			this.data = koe.getData();
			this.sub = koe.getSubcapabilities();
			break;
		    default:
			throw new KernelPanic("Invalid KernelObjectEvent: "
					      + koe);
		    }
		    this.setState(STATE_LOADED);
		}
	    }
	    else
	    {
		Strict.fail();
	    }
	    break;
	default:
	    throw new KernelPanic("Invalid state for coroutine: " + this);
	}
    }
    
    public String toString()
    {
	return "KernelObjectCoroutine["
	    + super.toString()
	    + "; data="
	    + ByteArray.toString(this.data, 32)
	    + "; storeID="
	    + this.storeID
	    + "]";
    }
}
