/*
 * KernelObjectStore.java
 *
 * Copyright (c) 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is part of the Janos Bees Software.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */

package bees.core.kernel;

import java.util.Hashtable;
import java.util.Enumeration;
import java.util.NoSuchElementException;

import edu.utah.janosvm.kit.util.SparseArray;
import edu.utah.janosvm.kit.util.SparseArrayElement;
import edu.utah.janosvm.kit.util.ObjectSparseArrayElement;

import bees.core.TypeID;
import bees.core.CapabilityFactory;
import bees.core.CapabilityEnvelope;

import bees.core.bdomain.BDomainLocal;

import bees.core.resource.Meter;

import bees.core.security.Permission;
import bees.core.security.PermissionSet;
import bees.core.security.CountPermission;

/**
 * A class used to store and track KernelObject's.
 */
public final class KernelObjectStore
{
    private static final BDomainLocal primary = new BDomainLocal() {
	    public Object initialValue()
	    {
		return new KernelObjectStore();
	    }
	};
    
    /**
     * The primary KernelObjectStore, everything gets put in here.
     */
    public static KernelObjectStore getPrimary()
    {
	return (KernelObjectStore)primary.get();
    }

    /**
     * The value of an invalid identifier.
     */
    public static final long INVALID_ID = SparseArray.INVALID_ID;

    public static boolean isValidID(long id)
    {
	return (id >= 0);
    }

    /**
     * The SparseArray used to store KernelObject's and generate their
     * unique identifiers.
     */
    private final SparseArray sa = new SparseArray();

    private final Hashtable externalToInternal = new Hashtable();

    /**
     * Construct an empty KernelObjectStore.
     */
    private KernelObjectStore()
    {
    }

    /**
     * @return An Enumeration containing the KernelObject's in this store.
     */
    Enumeration elements()
    {
	return this.sa.elements();
    }

    /**
     * Put a KernelObject in the store.
     *
     * @param ko A KernelObject to add to the store.
     * @return The KernelObject's unique identifier.
     */
    public synchronized long put(KernelObject ko)
    {
	long retval;
	Long l;
	
	if( (l = (Long)this.externalToInternal.get(ko.getHash())) == null )
	{
	    ObjectSparseArrayElement sae;
	    
	    if( (sae = (ObjectSparseArrayElement)this.sa.getElement())
		!= null )
	    {
		sae.setValue(ko);
		retval = sae.getID();
		this.externalToInternal.put(ko.getHash(), new Long(retval));
	    }
	    else
	    {
		throw new OutOfMemoryError();
	    }
	}
	else
	{
	    retval = l.longValue();
	}
	return retval;
    }

    /**
     * Put a KernelObject in the store.
     *
     * @param object The capability material.
     * @param comment The comment to attach to the KernelObject.
     * @return The KernelObject's unique identifier.
     */
    public synchronized long put(byte object[],
				 CapabilityEnvelope sub[],
				 PermissionSet ps,
				 Meter meter,
				 String comment)
    {
	return this.put(new KernelObject(this,
					 object,
					 sub,
					 ps,
					 meter,
					 comment));
    }

    public synchronized long put(byte object[],
				 PermissionSet ps,
				 String comment)
    {
	return this.put(new KernelObject(this, object, null, ps,
					 null, comment));
    }

    /**
     * Put a KernelObject in the store with a null comment.
     *
     * @param object The capability material.
     * @return The KernelObject's unique identifier.
     */
    public synchronized long put(byte object[], CapabilityEnvelope sub[])
    {
	return this.put(new KernelObject(this, object, sub, null, null,
					 (String)null));
    }

    /**
     * Put a KernelObject in the store with a null comment.
     *
     * @param object The capability material.
     * @return The KernelObject's unique identifier.
     */
    public synchronized long put(byte object[])
    {
	return this.put(new KernelObject(this, object, null, null, null,
					 (String)null));
    }

    /**
     * Get a KernelObject from the store using its unique identifier.
     *
     * @param id The KernelObject's unique identifier.
     * @return The matching KernelObject.
     * @exception Throws a NoSuchElementException if the id couldn't be found.
     */
    public synchronized KernelObject get(long id)
    {
	ObjectSparseArrayElement sae;
	KernelObject retval = null;

	if( (sae = (ObjectSparseArrayElement)this.sa.findElement(id)) != null )
	{
	    retval = (KernelObject)sae.getValue();
	}
	else
	{
	    throw new NoSuchElementException("Couldn't find object with ID: "
					     + Long.toString(id));
	}
	return retval;
    }

    public synchronized long putLate(TypeID hash)
    {
	long retval;
	Long l;
	
	if( (l = (Long)this.externalToInternal.get(hash)) == null )
	{
	    retval = this.put(new KernelObject(hash));
	}
	else
	{
	    retval = l.longValue();
	}
	return retval;
    }

    /**
     * Remove a KernelObject from the store.
     *
     * @param id The KernelObject's unique identifier.
     */
    public synchronized KernelObject remove(long id)
    {
	ObjectSparseArrayElement osae;
	KernelObject retval = null;
	
	if( (osae = (ObjectSparseArrayElement)
	     this.sa.findElement(id)) != null )
	{
	    Meter meter;
	    
	    retval = (KernelObject)osae.getValue();
	    if( (meter = retval.getMeter()) != null )
	    {
		CapabilityFactory.grant(meter, new PermissionSet(
		    new Permission[] {
			new CountPermission("capability.memory",
					    retval.computeSize()),
		    }));
	    }
	    this.sa.dropElement(osae);
	}
	return retval;
    }

    public String toString()
    {
	return "KernelObjectStore[sa="
	    + this.sa
	    + "]";
    }
}
