/*
 * ListBDomainCoroutine.java
 *
 * Copyright (c) 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is part of the Janos Bees Software.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */

package bees.core.kernel;

import java.util.Enumeration;

import edu.utah.janosvm.sys.DeadTeamException;
import edu.utah.janosvm.sys.TeamEjectionException;

import edu.utah.janosvm.kit.comm.amiga.FaxPort;
import edu.utah.janosvm.kit.comm.amiga.FaxPortHandle;

import bees.core.Assert;
import bees.core.Strict;
import bees.core.BeesEvent;
import bees.core.Coroutine;

import bees.core.bdomain.DeadBDomainException;

/**
 * Coroutine used for getting a list of the active domain's.
 */
public final class ListBDomainCoroutine
    extends Coroutine
{
    /**
     * State used when waiting for a reply.
     */
    public static final int STATE_WAITING = STATE_USER;

    /**
     * An array of the domain's KernelObject identifiers.
     */
    private long id[];

    /**
     * An array of the domain's identifiers.
     */
    private BDomainIdentifier si[];

    /**
     * Construct an empty ListBDomainCoroutine.
     */
    public ListBDomainCoroutine()
    {
    }

    /**
     * @return An array of the domain's KernelObject identifiers.
     */
    public synchronized long[] getStoreIDs()
    {
	return this.id;
    }

    /**
     * @return An array of the domain's identifiers.
     */
    public synchronized BDomainIdentifier[] getBDomainIdentifiers()
    {
	return this.si;
    }

    /**
     * Get a list of domain's from the kernel.
     *
     * @param local The current domain's EventHandler port.
     * @param remote The kernel's port.
     */
    public synchronized void list(FaxPortHandle local, FaxPortHandle remote)
    {
	try
	{
	    ListBDomainEvent lse;

	    lse = new ListBDomainEvent(this);
	    lse.setReplyPort(local);
	    remote.putFacsimile(lse);
	    this.setState(STATE_WAITING);
	}
	catch(DeadTeamException e)
	{
	    this.setState(STATE_ERROR_CAGED);
	}
	catch(TeamEjectionException e)
	{
	    this.setState(STATE_ERROR_CAGED);
	}
	catch(Throwable th)
	{
	    super.setThrowable(th);
	}
    }

    /**
     * Handle a ListBDomainEvent received by the kernel.
     *
     * @param lse The ListBDomainEvent to be handled.
     */
    private void handleNewList(ListBDomainEvent lse)
    {
	try
	{
	    BDomainRegister ar;

	    ar = (BDomainRegister)BDomainRegister.registry.get();
	    synchronized( ar )
	    {
		BDomainIdentifier bdomainIdentifiers[];
		Enumeration keys, elem;
		long storeIDs[];
		int lpc, size;
		
		size = ar.size();
		storeIDs = new long[size];
		bdomainIdentifiers = new BDomainIdentifier[size];
		keys = ar.keys();
		elem = ar.elements();
		for( lpc = 0; lpc < size; lpc++ )
		{
		    bdomainIdentifiers[lpc] = (BDomainIdentifier)
			keys.nextElement();
		    storeIDs[lpc] = ((BDomainCoroutine)elem.nextElement()).
			getStoreID();
		}
		lse.setStoreIDs(storeIDs);
		lse.setBDomainIdentifiers(bdomainIdentifiers);
	    }
	}
	catch(Throwable th)
	{
	    lse.setThrowable(th);
	}
	/* We're done. */
	this.setState(STATE_COMPLETE);
	try
	{
	    lse.sendReply();
	}
	catch(DeadBDomainException e)
	{
	}
    }

    public synchronized void handle(BeesEvent ae)
    {
	switch( this.getState() )
	{
	case STATE_NEWBORN:
	    if( (ae instanceof ListBDomainEvent) && !ae.isReply() )
	    {
		ListBDomainEvent lse = (ListBDomainEvent)ae;

		this.handleNewList(lse);
	    }
	    else
	    {
		Strict.fail();
	    }
	    break;
	case STATE_WAITING:
	    if( (ae instanceof ListBDomainEvent) && ae.isReply() )
	    {
		ListBDomainEvent lse = (ListBDomainEvent)ae;
		Throwable th;

		if( (th = lse.getThrowable()) != null )
		{
		    super.setThrowable(th);
		}
		else
		{
		    this.id = lse.getStoreIDs();
		    this.si = lse.getBDomainIdentifiers();
		    this.setState(STATE_COMPLETE);

		    Assert.ensure(this.id);
		    Assert.ensure(this.si);
		}
	    }
	    else
	    {
		Strict.fail();
	    }
	    break;
	default:
	    throw new KernelPanic("Invalid state for coroutine: " + this);
	}
    }

    public String toString()
    {
	return "ListBDomainCoroutine["
	    + super.toString()
	    + "; id="
	    + this.id
	    + "; si="
	    + this.si
	    + "]";
    }
}
