/*
 * AccountHandler.java
 *
 * Copyright (c) 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is part of the Janos Bees Software.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */

package bees.core.resource;

import edu.utah.janosvm.kit.comm.jsi.JSIPair;
import edu.utah.janosvm.kit.comm.jsi.JSIMessage;
import edu.utah.janosvm.kit.comm.jsi.JSIPropertyInfo;
import edu.utah.janosvm.kit.comm.jsi.MissingPairException;

import edu.utah.janos.nodeos.AddressFormatException;

import bees.core.PairChain;
import bees.core.Capability;
import bees.core.CapabilityEnvelope;

import bees.core.kernel.KernelCapabilities;
import bees.core.kernel.KernelObjectCoroutine;

import bees.core.security.EnumerationException;
import bees.core.security.EnumerationPermission;

import bees.core.bdomain.BDomain;

/**
 * The AccountHandler is a JSIHandler that can create Account capabilities.
 */
public final class AccountHandler
    extends CapabilityHandler
{
    /**
     * The set of supported properties to be added to a JSIMessage.GET_SUITES
     * request.
     */
    private static final JSIPropertyInfo JSI_PROPERTIES[] = {
	new JSIPropertyInfo("account",
			    (1 << JSIMessage.CREATE_PROPERTY),
			    "name:string - The capability name.\n"
			    + "pair:string - A tag/value pair to add to the "
			    + "account description.\n"
			    + "parent:string - The parent capability.  "
			    + "(Default: machine-account).\n"
			    + "comment:string - An optional comment for the "
			    + "capability.\n"
			    + "\n"
			    + "Create an account capability.\n"),
	new JSIPropertyInfo("account",
			    (1 << JSIMessage.GET_PROPERTY),
			    "name:string",
			    "Get the value of an Account capability.\n"),
    };
    
    /**
     * Construct an empty AccountHandler object.
     */
    public AccountHandler()
    {
    }
    
    /**
     * @see bees.core.resource.CapabilityHandler#getCapabilityClass()
     */
    public Class getCapabilityClass()
    {
	return Account.class;
    }

    /**
     * @see bees.core.resource.CapabilityHandler#createCapability(JSIMessage, String, String)
     */
    public Object createCapability(JSIMessage jm,
				   String name,
				   String comment)
	throws AddressFormatException
    {
	PairChain pc = new PairChain();
	String parent;
	JSIPair jp;
	int lpc;
	
	parent = jm.get("parent", "machine-account");
	for( lpc = 0; (jp = jm.getPair("pair", lpc)) != null; lpc++ )
	{
	    String type, tag, value, args[];
	    int argc, index;
	    
	    type = jp.getTag();
	    value = jp.getStringValue();
	    index = value.indexOf("=");
	    if( index < 0 )
	    {
		throw new IllegalArgumentException(
			"Format for pairs is `name=value'.");
	    }
	    tag = value.substring(0, index).trim();
	    value = value.substring(index + 1);
	    pc.add(tag, value);
	    
	}
	return (Capability)KernelObjectCoroutine.
	    create(new Account(),
		   BDomain.currentBDomain().getBDomainCoroutine(),
		   new AccountSpec(pc).toBytes(),
		   new CapabilityEnvelope[] {
		       new CapabilityEnvelope((Capability)
					      super.tree.get(parent)),
		   },
		   comment,
		   KernelCapabilities.toPermissionSet(jm));
    }

    /**
     * @see edu.utah.janosvm.kit.comm.jsi.JSIHandler#getSuites(JSIMessage)
     */
    public void getSuites(JSIMessage jm)
    {
	jm.add("result", JSI_PROPERTIES);
    }

    /**
     * @see java.lang.Object#toString()
     */
    public String toString()
    {
	return "AccountHandler["
	    + super.toString()
	    + "]";
    }
}
