/*
 * File.java
 *
 * Copyright (c) 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is part of the Janos Bees Software.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */

package bees.core.resource;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.FileNotFoundException;

import edu.utah.janosvm.kit.comm.jsi.JSIPair;
import edu.utah.janosvm.kit.comm.jsi.JSIHandler;
import edu.utah.janosvm.kit.comm.jsi.JSIMessage;
import edu.utah.janosvm.kit.comm.jsi.JSIPropertyInfo;
import edu.utah.janosvm.kit.comm.jsi.JSIHandlerContext;

import bees.core.CapabilityEnvelope;

import bees.core.security.Permission;
import bees.core.security.PermissionSet;
import bees.core.security.FilePermission;
import bees.core.security.PermissionException;

/**
 * Capability to govern access to a disk file.
 */
public final class File
    extends PathSpec
    implements JSIHandler
{
    /**
     * JSI Property info.
     */
    private static final JSIPropertyInfo JSI_PROPERTIES[] = {
	new JSIPropertyInfo("path",
			    (1 << JSIMessage.GET_PROPERTY),
			    "Get the file's path.\n"),
    };
    
    /**
     * Permission to read the file.
     */
    public static final Permission FILE_READ_PERMISSION =
	new Permission("file", "read");
    
    /**
     * Permission to write to the file.
     */
    public static final Permission FILE_WRITE_PERMISSION =
	new Permission("file", "write");

    /**
     * Permission to append to the file.
     */
    public static final FilePermission FILE_APPEND_PERMISSION =
	new FilePermission("file", "append");
    
    /**
     * Permission to get the file's length.
     */
    public static final Permission FILE_LENGTH_PERMISSION =
	new Permission("file.length", "read");
    
    /**
     * Permission to read the file's last modified time.
     */
    public static final Permission FILE_READ_LAST_MODIFIED_PERMISSION =
	new Permission("file.lastModified", "read");
    
    /**
     * Permission to write the file's last modified time.
     */
    public static final Permission FILE_WRITE_LAST_MODIFIED_PERMISSION =
	new Permission("file.lastModified", "write");
    
    /**
     * Permission to delete the file.
     */
    public static final Permission FILE_DELETE_PERMISSION =
	new Permission("file.delete", "execute");

    /**
     * The default permissions for a new file capability.
     */
    public static final PermissionSet DEFAULT_PERMISSIONS =
	new PermissionSet(new Permission[] {
	    new FilePermission("file", "read, write, append"),
	    new Permission("file.length", "read"),
	    new Permission("file.lastModified", "read, write"),
	    new Permission("file.delete", "execute"),
	});

    /**
     * The root this file is relative to.
     */
    private FileSystemRoot root;
    
    /**
     * The path to the file on disk.
     */
    private java.io.File path;

    /**
     * Construct an empty File capability.
     */
    public File()
    {
    }
    
    /**
     * @see bees.core.Capability#fromBytes(byte[], CapabilityEnvelope[])
     */
    protected synchronized void fromBytes(byte flattened[],
					  CapabilityEnvelope sub[])
	throws Exception
    {
	super.fromBytes(flattened);
	super.specFromBytes(flattened);

	this.root = (FileSystemRoot)sub[0].unseal();
	this.path = new java.io.File(this.root.getFile(), super.getPath());
	if( this.path.isDirectory() )
	{
	    throw new FileNotFoundException(super.getPath());
	}
    }

    /**
     * @see bees.core.Capability#revoke()
     */
    public synchronized void revoke()
    {
	super.clear();

	this.root = null;
	this.path = null;
    }

    /**
     * Get the last modified time of the file.  Since we cannot allow the
     * actual time to leak out, a clock capability must be provided in order
     * to form a stable base time.
     *
     * @param clock A clock from which to base the returned time value on.
     * @return If the clock is in uptime mode, then it is the clocks base time
     * minus the number of milliseconds since the file was last modified.
     * Otherwise, it is the absolute time it was last modified.
     */
    public synchronized long lastModified(Clock clock)
    {
	if( this.path != null )
	{
	    long offset, retval;
	    
	    super.checkPermission(FILE_READ_LAST_MODIFIED_PERMISSION);

	    offset = clock.getStartTime();
	    if( offset == -1 )
	    {
		throw new PermissionException("Invalid Clock capability");
	    }
	    if( offset == 0 )
	    {
		retval = this.path.lastModified();
	    }
	    else
	    {
		retval = offset - this.path.lastModified();
	    }
	    retval -= (retval % clock.getGranularity());
	    return retval;
	}
	else
	{
	    throw new NullPointerException();
	}
    }

    /**
     * @return The length of the file in bytes.
     */
    public synchronized long length()
    {
	if( this.path != null )
	{
	    super.checkPermission(FILE_LENGTH_PERMISSION);

	    return this.path.length();
	}
	else
	{
	    throw new NullPointerException();
	}
    }

    /**
     * @return An InputStream for reading from the file.
     * @throws FileNotFoundException If the file no longer exists in the local
     * file system.  XXX This really shouldn't be here.
     * @throws IOException If there is a problem opening the file.
     */
    public synchronized InputStream getInputStream()
	throws FileNotFoundException,
	       IOException
    {
	if( this.path != null )
	{
	    super.checkPermission(FILE_READ_PERMISSION);
	    
	    /*
	     * XXX We can't revoke/control these.
	     */
	    return new RevokableInputStream(new FileInputStream(this.path));
	}
	else
	{
	    throw new NullPointerException();
	}
    }

    /**
     * @param append True if the stream should start at the end of the file.
     * @return An OutputStream for writing to the file.
     * @throws FileNotFoundException If the file no longer exists in the local
     * file system.  XXX This really shouldn't be here.
     */
    public synchronized OutputStream getOutputStream(boolean append)
	throws FileNotFoundException
    {
	if( this.path != null )
	{
	    if( append )
	    {
		super.checkPermission(FILE_APPEND_PERMISSION);
	    }
	    else
	    {
		super.checkPermission(FILE_WRITE_PERMISSION);
	    }
	    
	    /*
	     * XXX We can't revoke/control these.
	     */
	    return new RevokableOutputStream(
		new FileOutputStream(this.path.getPath(), append));
	}
	else
	{
	    throw new NullPointerException();
	}
    }

    /**
     * @return An OutputStream for writing to the file.
     * @throws FileNotFoundException If the file no longer exists in the local
     * file system.  XXX This really shouldn't be here.
     */
    public OutputStream getOutputStream()
	throws FileNotFoundException
    {
	return this.getOutputStream(false);
    }
    
    /**
     * Delete the file on disk.
     */
    public synchronized void delete()
    {
	if( this.path != null )
	{
	    super.checkPermission(FILE_DELETE_PERMISSION);
	    
	    this.path.delete();
	    this.path = null;
	}
	else
	{
	    throw new NullPointerException();
	}
    }

    /**
     * @see edu.utah.janosvm.kit.comm.jsi.JSIHandler#resolveSpecifier(JSIMessage, JSIPair)
     */
    public JSIHandler resolveSpecifier(JSIMessage jm, JSIPair jp)
    {
	return this;
    }

    /**
     * @see edu.utah.janosvm.kit.comm.jsi.JSIHandler#handleMessage(JSIHandlerContext, JSIMessage)
     */
    public void handleMessage(JSIHandlerContext jhc, JSIMessage jm)
    {
	JSIPair jp;

	jm.restoreSpecifier();
	switch( jm.getWhat() )
	{
	case JSIMessage.GET_PROPERTY:
	    if( (jp = jm.popSpecifier()) != null )
	    {
		if( jp.getTag().equals("path") )
		{
		    jm.reply(super.getPath());
		}
		else
		{
		    jm.error("Unknown property: " + jp.getTag());
		}
	    }
	    else
	    {
		jm.reply(this.toString());
	    }
	    break;
	case JSIMessage.GET_SUITES:
	    jm.clear();
	    jm.setWhat(JSIMessage.REPLY);
	    this.getSuites(jm);
	    break;
	default:
	    throw new UnsupportedOperationException();
	}
    }

    /**
     * @see edu.utah.janosvm.kit.comm.jsi.JSIHandler#getSuites(JSIMessage)
     */
    public void getSuites(JSIMessage jm)
    {
	jm.add("result", JSI_PROPERTIES);
    }
    
    /**
     * @see java.lang.Object#toString()
     */
    public String toString()
    {
	return "File["
	    + super.toString()
	    + "; path="
	    + this.path
	    + "]";
    }
}
