/*
 * IPAddressPermission.java
 *
 * Copyright (c) 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is part of the Janos Bees Software.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */

package bees.core.resource;

import edu.utah.janos.nodeos.Address;
import edu.utah.janos.nodeos.AddressFormatException;

import bees.core.Xdr;
import bees.core.XdrByteArray;

import bees.core.security.Permission;

/**
 * An IPAddressPermission is used to encode an IPAddress as a permission
 * object.
 */
public class IPAddressPermission
    extends Permission
{
    /**
     * An IPv4 address encoded as an integer.
     */
    private int value;

    /**
     * The mask of bits that specify the valid part of the IPv4 address.
     */
    private int mask = 0xffffffff;

    /**
     * Construct an IPAddressPermission with the given values.
     *
     * @param name The permission name.
     * @param actions The actions granted by this permission.
     * @param value A partial IPv4 address encoded as an integer.
     * @param mask The mask of bits that specify the valid part of the IPv4
     * address.
     */
    public IPAddressPermission(String name, String actions,
			       int value, int mask)
    {
	super(name, actions);

	this.value = value;
	this.mask = mask;
    }

    /**
     * Construct an IPAddressPermission with the given values.
     *
     * @param name The permission name.
     * @param value A partial IPv4 address encoded as an integer.
     * @param mask The mask of bits that specify the valid part of the IPv4
     * address.
     */
    public IPAddressPermission(String name, int value, int mask)
    {
	this(name, "read,write", value, mask);
    }

    /**
     * Construct an IPAddressPermission with the given values.
     *
     * @param name The permission name.
     * @param value The IPv4 address encoded as an integer.
     */
    public IPAddressPermission(String name, int value)
    {
	this(name, value, 0xffffffff);
    }

    /**
     * Construct an IPAddressPermission with the given values.
     *
     * @param name The permission name.
     * @param actions The actions granted by this permission.
     * @param value An IPv4 address and optional mask encoded as a string.  If
     * the address covers a range of addresses then the string should have a
     * forward slash followed by the number of valid higher order bits
     * appended.
     * @throws AddressFormatException If there is a problem parsing the "value"
     * string.
     */
    public IPAddressPermission(String name, String actions, String value)
	throws AddressFormatException
    {
	super(name, actions);

	int index;

	if( (index = value.indexOf('/')) != -1 )
	{
	    try
	    {
		int sh;
		
		sh = Integer.parseInt(value.substring(index + 1));
		if( sh == 1 )
		{
		    this.mask = 0x80000000;
		}
		else if( sh > 0 )
		{
		    this.mask = (0x80000000 >> (sh - 1));
		}
		else
		{
		    this.mask = 0;
		}
		value = value.substring(0, index);
	    }
	    catch(NumberFormatException e)
	    {
		throw new AddressFormatException("Invalid mask in IP address: "
						 + value);
	    }
	}
	if( value.equals("*") )
	{
	    this.value = 0;
	    this.mask = 0;
	}
	else
	{
	    this.value = Address.fromString(value);
	}
    }

    /**
     * Construct an IPAddressPermission with the given values.
     *
     * @param name The permission name.
     * @param value An IPv4 address and optional mask encoded as a string.  If
     * the address covers a range of addresses then the string should have a
     * forward slash followed by the number of valid higher order bits
     * appended.
     * @throws AddressFormatException If there is a problem parsing the "value"
     * string.
     */
    public IPAddressPermission(String name, String value)
	throws AddressFormatException
    {
	this(name, "read,write", value);
    }
    
    /**
     * Construct an empty IPAddressPermission.
     */
    public IPAddressPermission()
    {
    }

    /**
     * @return The IPv4 address as an integer.
     */
    public int getValue()
    {
	return this.value;
    }

    /**
     * @return The mask of bits that specify the valid part of the IPv4
     * address.
     */
    public int getMask()
    {
	return this.mask;
    }

    /**
     * @return The IPv4 address as a string.
     */
    public String getStringValue()
    {
	return Address.toString(this.value);
    }

    /**
     * @see bees.core.security.Permission#getTypeID()
     */
    public int getTypeID()
    {
	return PERMISSION_IP_ADDRESS;
    }

    /**
     * @see bees.core.security.Permission#implies(Permission)
     */
    public boolean implies(Permission what)
    {
	boolean retval = false;

	if( super.implies(what) )
	{
	    if( what instanceof IPAddressPermission )
	    {
		IPAddressPermission ipap;

		ipap = (IPAddressPermission)what;
		if( (this.value & this.mask) ==
		    (ipap.value & ipap.mask & this.mask) )
		{
		    retval = true;
		}
	    }
	    else
	    {
		retval = true;
	    }
	}
	return retval;
    }

    /**
     * @see bees.core.security.Permission#marshalledLength()
     */
    public int marshalledLength()
    {
	return super.marshalledLength() + Xdr.INT;
    }

    /**
     * @see bees.core.security.Permission#marshal(XdrByteArray)
     */
    public void marshal(XdrByteArray xba)
    {
	super.marshal(xba);

	xba.PUT(this.value);
    }

    /**
     * @see bees.core.security.Permission#unmarshal(XdrByteArray)
     */
    public void unmarshal(XdrByteArray xba)
    {
	super.unmarshal(xba);

	this.value = xba.INT();
    }

    /**
     * @see java.lang.Object#toString()
     */
    public String toString()
    {
	return "IPAddressPermission["
	    + super.toString()
	    + "; value="
	    + Address.toString(this.value)
	    + "; mask="
	    + Address.toString(this.mask)
	    + "]";
    }
}
