/*
 * Logger.java
 *
 * Copyright (c) 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is part of the Janos Bees Software.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */

package bees.core.resource;

import java.io.PrintStream;
import java.io.OutputStream;
import java.io.FileNotFoundException;

import edu.utah.janosvm.kit.comm.jsi.JSIPair;
import edu.utah.janosvm.kit.comm.jsi.JSIHandler;
import edu.utah.janosvm.kit.comm.jsi.JSIMessage;
import edu.utah.janosvm.kit.comm.jsi.JSIHandlerContext;

import edu.utah.janos.nodeos.DomainLocal;

import edu.utah.janos.io.Log;
import edu.utah.janos.io.LogListener;

import bees.core.Capability;
import bees.core.CapabilityEnvelope;
import bees.core.MissingCapabilityException;

import bees.core.security.Permission;
import bees.core.security.PermissionSet;
import bees.core.security.EnumerationPair;
import bees.core.security.IntegerPermission;
import bees.core.security.PermissionException;
import bees.core.security.MaximumIntegerPermission;

import bees.core.bdomain.BDomain;

/**
 * The Logger capability provides logging functionality using another
 * capability that provides an OutputStream.  The only real advantage a Logger
 * has over the OutputStream is that it can put an unforgeable prefix on each
 * log entry, which is convenient for sharing log files.
 */
public final class Logger
    extends LoggerSpec
    implements JSIHandler
{
    /**
     * The permission containing the maximum LogLevel value for this Logger.
     */
    public static final Permission LEVEL_PERMISSION =
	new Permission("level", "write");

    /**
     * An array of IntegerPermissions corresponding to each LogLevel.
     */
    public static final IntegerPermission LEVEL_PERMISSIONS[] = {
	new IntegerPermission("level", "write", EMERG),
	new IntegerPermission("level", "write", ALERT),
	new IntegerPermission("level", "write", CRIT),
	new IntegerPermission("level", "write", ERROR),
	new IntegerPermission("level", "write", INFO),
	new IntegerPermission("level", "write", WARNING),
	new IntegerPermission("level", "write", NOTICE),
	new IntegerPermission("level", "write", DBG),
    };

    /**
     * An array of EnumerationPairs that map the LogLevel name to the LogLevel
     * integer value.
     */
    public static final EnumerationPair LEVEL_TYPES[] = {
	new EnumerationPair("emerg", EMERG),
	new EnumerationPair("emergency", EMERG),
	new EnumerationPair("alert", ALERT),
	new EnumerationPair("crit", CRIT),
	new EnumerationPair("critical", CRIT),
	new EnumerationPair("error", ERROR),
	new EnumerationPair("info", INFO),
	new EnumerationPair("warning", WARNING),
	new EnumerationPair("notice", NOTICE),
	new EnumerationPair("dbg", DBG),
	new EnumerationPair("debug", DBG),
    };

    /**
     * The default permissions for Logger capability.
     */
    public static final PermissionSet DEFAULT_PERMISSIONS =
	new PermissionSet(new Permission[] {
	    new MaximumIntegerPermission("level", "write", DBG),
	});
    
    /**
     * The current logger for this Domain.
     */
    private static final DomainLocal CURRENT = new DomainLocal() {
	    public Object initialValue()
	    {
		throw new PermissionException();
	    }
	};

    /**
     * Set the current logger for this Domain.
     *
     * @param logger The Logger to use as the default log for this Domain.
     */
    public static void setCurrent(Logger logger)
    {
	CURRENT.set(logger);
	logger.start();
    }

    /**
     * @return The default Logger for this Domain.
     */
    public static Logger currentLogger()
    {
	return (Logger)CURRENT.get();
    }

    /**
     * Log a message to the default domain logger at a given level of importance.
     *
     * @param level A LogLevel value.
     * @param msg The message to append to the log.
     */
    public static void log(int level, String msg)
    {
	try
	{
	    Logger logger = (Logger)CURRENT.get();
	    
	    logger.logMessage(level, msg);
	}
	catch(PermissionException e)
	{
	    /* Ignore... */
	}
    }

    /**
     * Log a message to the default domain logger at the debug level.
     *
     * @param msg The message to append to the log.
     */
    public static void log(String msg)
    {
	log(DBG, msg);
    }

    /**
     * Log a Throwable to the default domain logger at a given level of
     * importance.
     *
     * @param level A LogLevel value.
     * @param th The exception to log.
     */
    public static void log(int level, Throwable th)
    {
	try
	{
	    Logger logger = (Logger)CURRENT.get();
	    
	    logger.logThrowable(level, th);
	}
	catch(PermissionException e)
	{
	    /* Ignore... */
	}
    }

    /**
     * Log a Throwable to the default domain logger at the EMERGency level.
     *
     * @param th The exception to log.
     */
    public static void log(Throwable th)
    {
	log(EMERG, th);
    }

    /**
     * The capability to use as an output.
     */
    private Capability output;

    /**
     * The PrintStream derived from "output".
     */
    private PrintStream ps;

    /**
     * Construct an empty Logger capability.
     */
    public Logger()
    {
    }

    /**
     * @return The logger's output capability.
     */
    Capability getOutput()
    {
	return this.output;
    }

    /**
     * @see bees.core.Capability#revoke()
     */
    public synchronized void revoke()
    {
	super.revoke();

	this.output = null;
	this.ps = null;
    }

    /**
     * @see bees.core.Capability#fromBytes(byte[], CapabilityEnvelope[])
     */
    protected final void fromBytes(byte flattened[], CapabilityEnvelope sub[])
	throws Exception
    {
	OutputStream os = System.err /* XXX should be null */;
	
	super.fromBytes(flattened);

	super.specFromBytes(flattened);

	if( sub != null )
	{
	    try
	    {
		this.output = sub[0].unseal();
		/* Figure out what it is and get an OutputStream. */
		if( this.output instanceof Console )
		{
		    Console console;
		    
		    console = (Console)this.output;
		    os = console.getOutputStream();
		}
		else if( this.output instanceof File )
		{
		    File file;
		    
		    file = (File)this.output;
		    os = file.getOutputStream(true);
		}
		else
		{
		    throw new IllegalArgumentException(
			"Bad output capability");
		}
	    }
	    catch(FileNotFoundException e)
	    {
		/* XXX */
	    }
	    catch(MissingCapabilityException e)
	    {
		/* Ignore... */
	    }
	}
	if( os != null )
	{
	    this.ps = new PrintStream(os);
	}
    }

    /**
     * Start the logger by adding a LogListener to the internal jnodeos Log
     * object.
     * <p>
     * XXX This blows...
     */
    public synchronized void start()
    {
	this.log.addListener(new LogListener() {

		private final String domainPrefix =
		    (Logger.this.flags & LF_BDOMAIN_PREFIX) != 0 ?
		    "BDomain("
		    + BDomain.currentBDomain().getBDomainCoroutine().getStoreID()
		    + ") - " :
		    "";
		
		public void logMsg(int level, String msg)
		{
		    try
		    {
			Logger.this.checkPermission(LEVEL_PERMISSIONS[level]);
			try
			{
			    synchronized( Logger.this )
			    {
				Logger.this.ps.println(domainPrefix
						       + Log.LEVEL_NAMES[level]
						       + ": "
						       + msg);
			    }
			}
			catch(NullPointerException e)
			{
			    /* Revoked... */
			}
		    }
		    catch(PermissionException e)
		    {
			/* Ignore */
		    }
		}
	    });
    }

    /**
     * Log a message at the debug level.
     *
     * @param msg The message to append to the log.
     */
    public synchronized void logMessage(String msg)
    {
	this.log.logMsg(msg);
    }

    /**
     * Log a message at the given level of importance.
     *
     * @param level A LogLevel value.
     * @param msg The message to append to the log.
     * @throws IllegalArgumentException if the level value is not one
     * specified in the LogLevel interface.
     */
    public synchronized void logMessage(int level, String msg)
	throws IllegalArgumentException
    {
	if( (level < MIN) || (level > MAX) )
	{
	    throw new IllegalArgumentException("Invalid level value: "
					       + level);
	}
	this.log.logMsg(level, msg);
    }

    /**
     * Log a Throwable at the EMERGency level.
     *
     * @param th The exception to log.
     */
    public synchronized void logThrowable(Throwable th)
    {
	this.log.logThrowable(th);
    }

    /**
     * Log a Throwable at a given level of importance.
     *
     * @param level A LogLevel value.
     * @param th The exception to log.
     * @throws IllegalArgumentException If the level value is invalid.
     */
    public synchronized void logThrowable(int level, Throwable th)
	throws IllegalArgumentException
    {
	if( (level < MIN) || (level > MAX) )
	{
	    throw new IllegalArgumentException("Invalid level value: "
					       + level);
	}
	this.log.logThrowable(level, th);
    }

    /**
     * @see edu.utah.janosvm.kit.comm.jsi.JSIHandler#resolveSpecifier(JSIMessage, JSIPair)
     */
    public JSIHandler resolveSpecifier(JSIMessage jm, JSIPair jp)
    {
	return this;
    }

    /**
     * @see edu.utah.janosvm.kit.comm.jsi.JSIHandler#handleMessage(JSIHandlerContext, JSIMessage)
     */
    public void handleMessage(JSIHandlerContext jhc, JSIMessage jm)
    {
	switch( jm.getWhat() )
	{
	case JSIMessage.GET_PROPERTY:
	    if( jm.popSpecifier() == null )
	    {
		jm.reply(this.toString());
	    }
	    else
	    {
		jm.error("Too many specifiers");
	    }
	    break;
	default:
	    throw new UnsupportedOperationException();
	}
    }

    /**
     * @see edu.utah.janosvm.kit.comm.jsi.JSIHandler#getSuites(JSIMessage)
     */
    public void getSuites(JSIMessage jm)
    {
    }
    
    /**
     * @see java.lang.Object#toString()
     */
    public String toString()
    {
	return "Logger["
	    + super.toString()
	    + "; ps="
	    + this.ps
	    + "]";
    }
}
