/*
 * LoggerFactory.java
 *
 * Copyright (c) 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is part of the Janos Bees Software.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */

package bees.core.resource;

import bees.core.Capability;
import bees.core.CapabilitySpec;
import bees.core.CapabilityEnvelope;
import bees.core.SpecCapabilityFactory;

import bees.core.kernel.KernelObjectCoroutine;

import bees.core.security.PermissionSet;

import bees.core.bdomain.BDomain;

/**
 * A LoggerFactory is used to create new Logger capabilities from existing
 * capabilities.
 */
public class LoggerFactory
    extends SpecCapabilityFactory
{
    /**
     * Cached reference to the current BDomain.
     */
    private final BDomain bdomain;

    /**
     * Construct an empty Logger factory.
     */
    public LoggerFactory()
    {
	this.bdomain = BDomain.currentBDomain();
    }

    /**
     * Create a new Logger from the given capabilities.
     *
     * @param console The Console capability to be used as output.
     * @param ls The LoggerSpec that specifies the log prefix and level.
     * @param ps The permissions to grant the new Logger capability.
     * @param comment The comment to attach to capability in the kernel.
     * @return A new Logger capability.
     */
    public Logger create(Console console,
			 LoggerSpec ls,
			 PermissionSet ps,
			 String comment)
    {
	return (Logger)KernelObjectCoroutine.
	    create(new Logger(),
		   this.bdomain.getBDomainCoroutine(),
		   ls.toBytes(),
		   new CapabilityEnvelope[] {
		       new CapabilityEnvelope(console),
		   },
		   comment,
		   ps);
    }
    
    /**
     * Create a new Logger from the given capabilities.
     *
     * @param console The Console capability to be used as the output.
     * @param ls The LoggerSpec that specifies the log prefix and level.
     * @return A new Logger capability.
     */
    public Logger create(Console console, LoggerSpec ls)
    {
	return this.create(console, ls, Logger.DEFAULT_PERMISSIONS, "Logger");
    }
    
    /**
     * Create a new Logger from the given capabilities.
     *
     * @param file The File capability to be used as the output.
     * @param ls The LoggerSpec that specifies the log prefix and level.
     * @param ps The permissions to grant the new Logger capability.
     * @param comment The comment to attach to capability in the kernel.
     * @return A new Logger capability.
     */
    public Logger create(File file,
			 LoggerSpec ls,
			 PermissionSet ps,
			 String comment)
    {
	return (Logger)KernelObjectCoroutine.
	    create(new Logger(),
		   this.bdomain.getBDomainCoroutine(),
		   ls.toBytes(),
		   new CapabilityEnvelope[] {
		       new CapabilityEnvelope(file),
		   },
		   comment,
		   ps);
    }
    
    /**
     * Create a new Logger from the given capabilities.
     *
     * @param file The File capability to be used as the output.
     * @param ls The LoggerSpec that specifies the log prefix and level.
     * @return A new Logger capability.
     */
    public Logger create(File file, LoggerSpec ls)
    {
	return this.create(file, ls, Logger.DEFAULT_PERMISSIONS, "Logger");
    }
    
    /**
     * @see bees.core.SpecCapabilityFactory#create(Capability, CapabilitySpec)
     */
    public Capability create(Capability parent, CapabilitySpec sc)
    {
	return null;
    }
    
    /**
     * Create a new Logger from the given capabilities.
     *
     * @param parent The Parent capability to be used as the output and extra
     * prefix.
     * @param ls The LoggerSpec that specifies the log prefix and level.
     * @param ps The permissions to grant the new Logger capability.
     * @param comment The comment to attach to capability in the kernel.
     * @return A new Logger capability.
     */
    public Logger create(Logger parent,
			 LoggerSpec ls,
			 PermissionSet ps,
			 String comment)
    {
	ls.inherit(parent);
	return (Logger)KernelObjectCoroutine.
	    create(new Logger(),
		   this.bdomain.getBDomainCoroutine(),
		   ls.toBytes(),
		   new CapabilityEnvelope[] {
		       new CapabilityEnvelope(parent.getOutput()),
		   },
		   comment,
		   ps);
    }

    /**
     * Create a new Logger from the given capabilities.
     *
     * @param parent The Parent capability to be used as the output and extra
     * prefix.
     * @param prefix The prefix for the new Logger capability.
     * @param ps The permissions to grant the new Logger capability.
     * @return A new Logger capability.
     */
    public Logger create(Logger parent, String prefix, PermissionSet ps)
    {
	return this.create(parent,
			   new LoggerSpec(prefix),
			   ps,
			   "Logger");
    }
    
    /**
     * Create a new Logger from the given capabilities.
     *
     * @param parent The Parent capability to be used as the output and extra
     * prefix.
     * @param prefix The prefix for the new Logger capability.
     * @return A new Logger capability.
     */
    public Logger create(Logger parent, String prefix)
    {
	return this.create(parent,
			   new LoggerSpec(prefix, Logger.DBG),
			   Logger.DEFAULT_PERMISSIONS,
			   "Logger");
    }
    
    /**
     * @see java.lang.Object#toString()
     */
    public String toString()
    {
	return "LoggerFactory["
	    + super.toString()
	    + "]";
    }
}
