/*
 * NetworkInterface.java
 *
 * Copyright (c) 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is part of the Janos Bees Software.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */

package bees.core.resource;

import bees.core.Capability;
import bees.core.CapabilityEnvelope;

import bees.core.kernel.KernelObjectCoroutine;

import bees.core.security.PermissionSet;

import bees.core.bdomain.BDomain;

/**
 * A NetworkInterface governs access to a network interface.
 */
public final class NetworkInterface
    extends Capability
{
    /**
     * If not null, this refers to the local interface that is connected to
     * this remote interface.
     */
    private NetworkInterface local;

    /**
     * Construct an empty NetworkInterface capability.
     */
    public NetworkInterface()
    {
    }

    /**
     * @return If not null, this refers to the local interface that is
     * connected to this remote interface.
     */
    NetworkInterface getLocal()
    {
	return this.local;
    }

    /**
     * @see bees.core.Capability#fromBytes(byte[], CapabilityEnvelope[])
     */
    protected synchronized void fromBytes(byte flattened[],
					  CapabilityEnvelope sub[])
	throws Exception
    {
	super.fromBytes(flattened);
	
	if( sub != null )
	{
	    this.local = (NetworkInterface)sub[0].unseal();
	}
    }

    /**
     * @see bees.core.Capability#revoke()
     */
    public synchronized void revoke()
    {
	super.clear();

	this.local = null;
    }

    /**
     * Construct a NetworkInterface that refers to a remote interface connected
     * to this local interface.  Note:  If this interface isn't a local an
     * IllegalArgumentException will be thrown.
     *
     * @param ps The permissions to bind to the new interface.
     * @return A NetworkInterface capability that is "connected" to this
     * interface.
     */
    public synchronized NetworkInterface remoteInterface(PermissionSet ps)
    {
	NetworkInterface retval = null;

	if( this.local != null )
	{
	    throw new IllegalArgumentException(
		"Can only create remote interfaces from local interfaces");
	}
	
	if( super.valid() )
	{
	    retval = (NetworkInterface)KernelObjectCoroutine.
		create(new NetworkInterface(),
		       BDomain.currentBDomain().getBDomainCoroutine(),
		       new byte[0],
		       new CapabilityEnvelope[] {
			   new CapabilityEnvelope(this),
		       },
		       "NetworkInterface",
		       ps);
	}
	return retval;
    }

    /**
     * Construct a NetworkService that is available on this interface.
     *
     * @param ps The permissions to bind to the new service.
     * @return A NetworkService capability that functions over this interface.
     */
    public synchronized NetworkService service(PermissionSet ps)
    {
	NetworkService retval = null;

	if( super.valid() )
	{
	    retval = (NetworkService)KernelObjectCoroutine.
		create(new NetworkService(),
		       BDomain.currentBDomain().getBDomainCoroutine(),
		       new byte[0],
		       new CapabilityEnvelope[] {
			   new CapabilityEnvelope(this),
		       },
		       "NetworkService",
		       ps);
	}
	return retval;
    }
    
    /**
     * @see java.lang.Object#toString()
     */
    public String toString()
    {
	return "NetworkInterface["
	    + super.toString()
	    + "; local="
	    + this.local
	    + "]";
    }
}
