/*
 * NodeOSExternalChannel.java
 *
 * Copyright (c) 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is part of the Janos Bees Software.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */

package bees.core.resource;

import java.io.IOException;

import edu.utah.janos.nodeos.Domain;
import edu.utah.janos.nodeos.InChannel;
import edu.utah.janos.nodeos.OutChannel;
import edu.utah.janos.nodeos.DeliverFunc;
import edu.utah.janos.nodeos.BufferHandle;
import edu.utah.janos.nodeos.DemultiplexKey;
import edu.utah.janos.nodeos.ProtocolSpecFormatException;
import edu.utah.janos.nodeos.PacketAddressFormatException;

import bees.core.Xdr;
import bees.core.Leak;
import bees.core.Node;
import bees.core.Panic;
import bees.core.Capsule;
import bees.core.Capability;
import bees.core.XdrByteArray;
import bees.core.ExternalCapsule;
import bees.core.ExternalChannel;
import bees.core.ProtocolSession;

public final class NodeOSExternalChannel
    extends Capability
    implements ExternalChannel
{
    private final class ExternalReceiver
	implements DeliverFunc
    {
	public void receive(BufferHandle bh, int offset)
	{
	    ExternalCapsule ec = null;

	    ec = Leak.toCapsule(ec,
				NodeOSExternalChannel.this,
				NodeOSExternalChannel.this.ps,
				bh);
	    NodeOSExternalChannel.this.ps.receive(NodeOSExternalChannel.this.
						  node, ec);
	}
    }

    private Node node;

    private ProtocolSession ps;
    
    /**
     * The protocol spec for the InChannel.
     */
    private String incomingProtocolSpec;
    
    /**
     * The address spec for the InChannel.
     */
    private String incomingAddressSpec;
    
    /**
     * The protocol spec for the OutChannel.
     */
    private String outgoingProtocolSpec;
    
    /**
     * The address spec for the OutChannel.
     */
    private String outgoingAddressSpec;

    /**
     * The InChannel connected to the remote host.
     */
    private InChannel ic;

    /**
     * The OutChannel connected to the remote host.
     */
    private OutChannel oc;

    private Object local;

    public NodeOSExternalChannel()
    {
    }
    
    public void setLocal(Object obj)
    {
	this.local = obj;
    }

    public Object getLocal()
    {
	return this.local;
    }

    protected synchronized void fromBytes(byte flattened[])
	throws Exception
    {
	XdrByteArray xba;
	
	super.fromBytes(flattened);

	xba = new XdrByteArray(flattened);
	this.incomingProtocolSpec = xba.STRING();
	this.incomingAddressSpec = xba.STRING();
	this.outgoingProtocolSpec = xba.STRING();
	this.outgoingAddressSpec = xba.STRING();
    }

    public synchronized void revoke()
    {
	super.clear();

	this.incomingProtocolSpec = null;
	this.incomingAddressSpec = null;
	this.outgoingProtocolSpec = null;
	this.outgoingAddressSpec = null;
	/* XXX Kill ic and oc */
    }

    public synchronized void activate(Node node, ProtocolSession ps)
	throws IOException
    {
	try
	{
	    Domain domain = Domain.currentDomain();
	    int lpc;

	    this.node = node;
	    this.ps = ps;
	    this.oc = OutChannel.create(domain,
					this.outgoingProtocolSpec,
					this.outgoingAddressSpec);
	    this.ic = InChannel.create(domain,
				       1,
				       new DemultiplexKey(domain),
				       this.incomingProtocolSpec,
				       this.incomingAddressSpec,
				       new ExternalReceiver());
	    this.ic.addBuffers(5); /* XXX */
	}
	catch(PacketAddressFormatException e)
	{
	    throw new Panic("Address spec `"
			    + this.outgoingAddressSpec
			    + "' wasn't properly validated");
	}
	catch(ProtocolSpecFormatException e)
	{
	    throw new Panic("Protocol spec `"
			    + this.outgoingProtocolSpec
			    + "' wasn't properly validated");
	}
    }

    public synchronized void sendCapsule(ExternalCapsule ec)
	throws IOException
    {
	BufferHandle bh;

	bh = Leak.getBuffer(ec);
	this.oc.send(bh);
    }

    public synchronized void recycleCapsule(ExternalCapsule ec)
    {
	BufferHandle bh;

	bh = Leak.getBuffer(ec);
	this.ic.addBuffer(bh);
    }
    
    public String toString()
    {
	return "NodeOSNeighbor["
	    + super.toString()
	    + "; incomingProtocolSpec="
	    + this.incomingProtocolSpec
	    + "; incomingAddressSpec="
	    + this.incomingAddressSpec
	    + "; outgoingProtocolSpec="
	    + this.outgoingProtocolSpec
	    + "; outgoingAddressSpec="
	    + this.outgoingAddressSpec
	    + "; oc="
	    + this.oc
	    + "]";
    }
}
