/*
 * NodeOSNeighbor.java
 *
 * Copyright (c) 2002, 2003 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is part of the Janos Bees Software.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */

package bees.core.resource;

import java.util.Hashtable;
import java.util.Enumeration;

import java.io.IOException;

import java.security.InvalidKeyException;
import java.security.NoSuchAlgorithmException;

import edu.utah.janosvm.kit.comm.jsi.JSIPair;
import edu.utah.janosvm.kit.comm.jsi.JSIHandler;
import edu.utah.janosvm.kit.comm.jsi.JSIMessage;
import edu.utah.janosvm.kit.comm.jsi.JSIPropertyInfo;
import edu.utah.janosvm.kit.comm.jsi.JSIHandlerContext;

import edu.utah.janos.nodeos.Domain;
import edu.utah.janos.nodeos.InChannel;
import edu.utah.janos.nodeos.OutChannel;
import edu.utah.janos.nodeos.DeliverFunc;
import edu.utah.janos.nodeos.NetworkSpec;
import edu.utah.janos.nodeos.BufferHandle;
import edu.utah.janos.nodeos.DemultiplexKey;
import edu.utah.janos.nodeos.ProtocolSpecFormatException;
import edu.utah.janos.nodeos.PacketAddressFormatException;

import bees.core.Assert;
import bees.core.TypeID;
import bees.core.Capsule;
import bees.core.Neighbor;
import bees.core.NeighborImpl;
import bees.core.XdrByteArray;
import bees.core.CapsuleReceiver;

import bees.core.kernel.KernelPanic;

import bees.core.security.Permission;
import bees.core.security.PermissionSet;

import bees.util.MD5;

/**
 * A neighbor implementation that uses NodeOS in/out channels.
 */
public final class NodeOSNeighbor
    extends NeighborImpl
    implements JSIHandler
{
    /**
     * The per-object level properties of the NodeOSNeighbor handler.
     */
    private static final JSIPropertyInfo JSI_PROPERTIES[] = {
	new JSIPropertyInfo("incomingProtocolSpec",
			    (1 << JSIMessage.GET_PROPERTY),
			    "The NodeOS style InChannel protocol spec.\n"),
	new JSIPropertyInfo("incomingAddressSpec",
			    (1 << JSIMessage.GET_PROPERTY),
			    "The NodeOS style InChannel address spec.\n"),
	new JSIPropertyInfo("outgoingProtocolSpec",
			    (1 << JSIMessage.GET_PROPERTY),
			    "The NodeOS style OutChannel protocol spec.\n"),
	new JSIPropertyInfo("outgoingAddressSpec",
			    (1 << JSIMessage.GET_PROPERTY),
			    "The NodeOS style OutChannel address spec.\n"),
    };
    
    /**
     * Permission to read the incoming protocol spec.
     */
    public static final Permission INCOMING_PROTOCOL_SPEC__READ_PERMISSION =
	new Permission("incomingProtocolSpec", "read");
    
    /**
     * Permission to read the incoming address spec.
     */
    public static final Permission INCOMING_ADDRESS_SPEC__READ_PERMISSION =
	new Permission("incomingAddressSpec", "read");
    
    /**
     * Permission to read the outgoing protocol spec.
     */
    public static final Permission OUTGOING_PROTOCOL_SPEC__READ_PERMISSION =
	new Permission("outgoingProtocolSpec", "read");
    
    /**
     * Permission to read the outgoing address spec.
     */
    public static final Permission OUTGOING_ADDRESS_SPEC__READ_PERMISSION =
	new Permission("outgoingAddressSpec", "read");

    /**
     * Default permissions for a newly created NodeOSNeighbor object.
     */
    public static final PermissionSet DEFAULT_PERMISSIONS =
	new PermissionSet(new Permission[] {
	    new Permission("capsule", "read, write"),
	    new Permission("incomingProtocolSpec", "read"),
	    new Permission("incomingAddressSpec", "read"),
	    new Permission("outgoingProtocolSpec", "read"),
	    new Permission("outgoingAddressSpec", "read"),
	});

    /**
     * The default network spec passed to an OutChannel.
     */
    private static final NetworkSpec DEFAULT_NETWORK_SPEC = new NetworkSpec();

    /**
     * The protocol spec for the InChannel.
     */
    private String incomingProtocolSpec;
    
    /**
     * The address spec for the InChannel.
     */
    private String incomingAddressSpec;
    
    /**
     * The protocol spec for the OutChannel.
     */
    private String outgoingProtocolSpec;
    
    /**
     * The address spec for the OutChannel.
     */
    private String outgoingAddressSpec;

    /**
     * Finger print covering this neighbors data.
     */
    private TypeID fingerPrint;

    /**
     * The list of NodeOSReceivers currently receiving from the remote host.
     */
    private Hashtable receivers = new Hashtable();

    /**
     * The OutChannel connected to the remote host.
     */
    private OutChannel oc;

    /**
     * Construct an empty NodeOSNeighbor object.
     */
    public NodeOSNeighbor()
    {
    }

    /**
     * @return The protocol spec for the InChannel.
     * @throws PermissionException if the capability doesn't have
     * read permission.
     */
    public synchronized String getIncomingProtocolSpec()
    {
	super.checkPermission(INCOMING_PROTOCOL_SPEC__READ_PERMISSION);
	
	return this.incomingProtocolSpec;
    }

    /**
     * @return The address spec for the InChannel.
     * @throws PermissionException if the capability doesn't have
     * read permission.
     */
    public synchronized String getIncomingAddressSpec()
    {
	super.checkPermission(INCOMING_ADDRESS_SPEC__READ_PERMISSION);
	
	return this.incomingAddressSpec;
    }

    /**
     * @return The protocol spec for the OutChannel.
     * @throws PermissionException if the capability doesn't have
     * read permission.
     */
    public synchronized String getOutgoingProtocolSpec()
    {
	super.checkPermission(OUTGOING_PROTOCOL_SPEC__READ_PERMISSION);
	
	return this.outgoingProtocolSpec;
    }

    /**
     * @return The address spec for the OutChannel.
     * @throws PermissionException if the capability doesn't have
     * read permission.
     */
    public synchronized String getOutgoingAddressSpec()
    {
	super.checkPermission(OUTGOING_ADDRESS_SPEC__READ_PERMISSION);
	
	return this.outgoingAddressSpec;
    }

    /**
     * @see bees.core.Capability#fromBytes(byte[])
     */
    protected synchronized void fromBytes(byte flattened[])
	throws Exception
    {
	MD5 digest;
	
	Assert.require(flattened);
	
	super.fromBytes(flattened);
	
	XdrByteArray xba;
	
	xba = new XdrByteArray(flattened);
	this.incomingProtocolSpec = xba.STRING();
	this.incomingAddressSpec = xba.STRING();
	this.outgoingProtocolSpec = xba.STRING();
	this.outgoingAddressSpec = xba.STRING();
	digest = new MD5();
	digest.update(this.getClass().getName());
	digest.update(flattened);
	this.fingerPrint = new TypeID(digest.hash());
    }

    /**
     * Internal function to close the all the receivers and their InChannel's
     * when our read permission is revoked.
     */
    private void revokeReadPermission()
    {
	int lpc, size;
	
	if( this.receivers != null )
	{
	    Enumeration enum;

	    enum = this.receivers.elements();
	    while( enum.hasMoreElements() )
	    {
		Object obj;

		obj = enum.nextElement();
		if( obj instanceof NodeOSReceiver )
		{
		    NodeOSReceiver nosr = (NodeOSReceiver)obj;
		    
		    nosr.close();
		}
		else if( obj instanceof Hashtable )
		{
		    Hashtable ht = (Hashtable)obj;
		    Enumeration senum;

		    senum = ht.elements();
		    while( senum.hasMoreElements() )
		    {
			NodeOSReceiver nosr;

			nosr = (NodeOSReceiver)senum.nextElement();
			nosr.close();
		    }
		}
	    }
	}
	this.receivers = null;
    }

    /**
     * Internal function to close the OutChannel when our write permission is
     * revoked.
     */
    private void revokeWritePermission()
    {
	if( this.oc != null )
	{
	    this.oc.destroy();
	    this.oc = null;
	}
    }

    /**
     * @see bees.core.Capability#revokedPermission(Permission)
     */
    public synchronized void revokedPermission(Permission what)
    {
	Assert.require(what);
	
	if( what.implies(CAPSULE__WRITE_PERMISSION) )
	{
	    this.revokeWritePermission();
	}
	else if( what.implies(CAPSULE__READ_PERMISSION) )
	{
	    this.revokeReadPermission();
	}
    }

    /**
     * @see bees.core.Capability#revoke()
     */
    public synchronized void revoke()
    {
	super.clear();

	this.incomingProtocolSpec = null;
	this.incomingAddressSpec = null;
	this.outgoingProtocolSpec = null;
	this.outgoingAddressSpec = null;
	this.revokeReadPermission();
	this.revokeWritePermission();
    }

    /**
     * Create an OutChannel using the initialized specs.
     *
     * @param ns The NetworkSpec to pass to the OutChannel.
     * @return An initialized OutChannel object.
     * @throws IOException if the OutChannel couldn't be
     * constructed.
     */
    private OutChannel createOutChannel(NetworkSpec ns)
	throws IOException
    {
	Assert.require(ns);

	try
	{
	    OutChannel retval;
	    
	    retval = OutChannel.create(Domain.currentDomain(),
				       ns,
				       this.outgoingProtocolSpec,
				       this.outgoingAddressSpec);
	    return retval;
	}
	catch(PacketAddressFormatException e)
	{
	    throw new KernelPanic("Address spec `"
				  + this.outgoingAddressSpec
				  + "' wasn't properly validated");
	}
	catch(ProtocolSpecFormatException e)
	{
	    throw new KernelPanic("Protocol spec `"
				  + this.outgoingProtocolSpec
				  + "' wasn't properly validated");
	}
    }
    
    /**
     * Create an OutChannel using the initialized specs.
     *
     * @return An initialized OutChannel object.
     * @throws IOException if the OutChannel couldn't be
     * constructed.
     */
    private OutChannel createOutChannel()
	throws IOException
    {
	return this.createOutChannel(DEFAULT_NETWORK_SPEC);
    }

    /**
     * Create an InChannel using the initialized specs and the given values.
     *
     * @param dk The DemultiplexKey for this channel.
     * @param threadCount The number of threads used to process incoming
     * packets for this channel.
     * @param df The object used to process received packets.
     * @return An initialied InChannel object.
     * @throws IOException if the InChannel couldn't be
     * constructed.
     */
    private InChannel createInChannel(DemultiplexKey dk,
				      int threadCount,
				      DeliverFunc df)
	throws IOException
    {
	Assert.require(dk);
	Assert.require(threadCount > 0);
	Assert.require(df);
	
	try
	{
	    InChannel retval;
	    
	    retval = InChannel.create(Domain.currentDomain(),
				      threadCount,
				      dk,
				      this.incomingProtocolSpec,
				      this.incomingAddressSpec,
				      df);
	    return retval;
	}
	catch(PacketAddressFormatException e)
	{
	    throw new KernelPanic("Address spec `"
				  + this.incomingAddressSpec
				  + "' wasn't properly validated");
	}
	catch(ProtocolSpecFormatException e)
	{
	    throw new KernelPanic("Protocol spec `"
				  + this.incomingProtocolSpec
				  + "' wasn't properly validated");
	}
    }

    /**
     * A simple class used to handoff any packets that are received on the
     * InChannel connected to the neighbor.
     */
    private static class NodeOSReceiver
	implements DeliverFunc
    {
	/**
	 * The CapsuleReceiver object that will actually process the packets.
	 */
	private final CapsuleReceiver cr;

	/**
	 * The InChannel connected to the neighbor.
	 */
	private InChannel ic;

	/**
	 * Construct a NodeOSReceiver object with the given value.
	 *
	 * @param cr The CapsuleReceiver object that will actually process the
	 * packets.
	 */
	NodeOSReceiver(CapsuleReceiver cr)
	{
	    this.cr = cr;
	}

	/**
	 * @return The CapsuleReceiver object that will actually process the
	 * packets.
	 */
	CapsuleReceiver getReceiver()
	{
	    return this.cr;
	}

	/**
	 * @param ic The InChannel connected to the neighbor.
	 */
	void setInChannel(InChannel ic)
	{
	    Assert.require(ic);
	    
	    this.ic = ic;
	}

	/**
	 * @return The InChannel connected to the neighbor.
	 */
	InChannel getChannel()
	{
	    return this.ic;
	}
	
        /**
         * @see edu.utah.janos.nodeos.DeliverFunc#receive(BufferHandle, int)
         */
	public void receive(BufferHandle bh, int offset)
	{
	    Capsule cap;

	    try
	    {
		/* Extract the capsule from the buffer... */
		cap = this.cr.receive(bh, offset);
	    }
	    finally
	    {
		/* Always, return the buffer to the InChannel */
		this.ic.addBuffer(bh);
	    }
	    if( cap != null )
	    {
		/* Actually, evaluate the capsule */
		this.cr.evaluate(cap);
	    }
	}

	/**
	 * Destroy and drop the InChannel used by this receiver.
	 */
	synchronized void close()
	{
	    if( this.ic != null )
	    {
		this.ic.destroy();
		this.ic = null;
	    }
	}

        /**
         * @see java.lang.Object#toString()
         */
	public String toString()
	{
	    return "NodeOSNeighbor[cr="
		+ this.cr
		+ "]";
	}
    }

    /**
     * @see bees.core.NeighborImpl#receiveBuffer(CapsuleReceiver)
     */
    protected synchronized void receiveBuffer(CapsuleReceiver cr)
	throws IOException
    {
	Hashtable sub;
	TypeID key;

	if( (sub = (Hashtable)this.receivers.get(cr.getCapsuleID())) == null )
	{
	    sub = new Hashtable();
	    this.receivers.put(cr.getCapsuleID(), sub);
	}
	if( cr.getSpecializedCapsuleID() == null )
	{
	    key = cr.getCapsuleID();
	}
	else
	{
	    key = cr.getSpecializedCapsuleID();
	}
	if( !sub.containsKey(key) )
	{
	    int threadCount;
	    
	    super.checkPermission(CAPSULE__READ_PERMISSION);
	    
	    if( ((threadCount = cr.getThreadCount()) > 0) &&
		(cr.getBufferCount() > 0) )
	    {
		NodeOSReceiver nosr;
		InChannel ic;
		
		nosr = new NodeOSReceiver(cr);
		ic = this.createInChannel(cr.createDemultiplexKey(),
					  cr.getThreadCount(),
					  nosr);
		/*
		 * Need to set the channel before we add buffers since the
		 * receiver will reference the channel in the receive function.
		 */
		nosr.setInChannel(ic);
		ic.addBuffers(cr.getBufferCount());
		sub.put(key, nosr);
	    }
	}
    }

    /**
     * @see bees.core.NeighborImpl#cancelReceivers(Neighbor, TypeID)
     */
    protected synchronized void cancelReceivers(Neighbor neighbor,
						TypeID capsuleTypeID)
    {
	Enumeration enum;

	enum = this.receivers.elements();
	while( enum.hasMoreElements() )
	{
	    Enumeration senum;
	    Hashtable sub;

	    sub = (Hashtable)enum.nextElement();
	    senum = sub.elements();
	    while( senum.hasMoreElements() )
	    {
		CapsuleReceiver cr;
		NodeOSReceiver nr;

		nr = (NodeOSReceiver)senum.nextElement();
		cr = nr.getReceiver();
		if( (cr.getNeighbor() == neighbor) &&
		    cr.getCapsuleID().equals(capsuleTypeID) )
		{
		    nr.close();
		}
	    }
	}
    }

    /**
     * @see bees.core.NeighborImpl#sendBuffer(BufferHandle)
     */
    protected synchronized void sendBuffer(BufferHandle bh)
	throws IOException
    {
	try
	{
	    this.oc.send(bh);
	}
	catch(NullPointerException e)
	{
	    /* Either it hasn't been initialized or its been revoked */
	    super.checkPermission(CAPSULE__WRITE_PERMISSION);

	    this.oc = this.createOutChannel();
	    this.sendBuffer(bh);
	}
    }

    protected void setKey(java.security.Key sk, String algorithm)
	throws NoSuchAlgorithmException,
	       InvalidKeyException
    {
	try
	{
	    this.oc.setKey(sk, algorithm);
	}
	catch(NullPointerException e)
	{
	    try
	    {
		this.oc = this.createOutChannel();
		this.setKey(sk, algorithm);
	    }
	    catch(IOException e2)
	    {
		/* XXX ... */
	    }
	}
    }

    protected void enqueueKey(java.security.Key sk, String algorithm)
	throws NoSuchAlgorithmException,
	       InvalidKeyException
    {
	Enumeration enum;
	
	enum = this.receivers.elements();
	while( enum.hasMoreElements() )
	{
	    Enumeration senum;
	    Hashtable sub;

	    sub = (Hashtable)enum.nextElement();
	    senum = sub.elements();
	    while( senum.hasMoreElements() )
	    {
		NodeOSReceiver nr;

		nr = (NodeOSReceiver)senum.nextElement();
		nr.getChannel().enqueueKey(sk, algorithm);
	    }
	}
    }

    protected void dequeueKey()
    {
	Enumeration enum;
	
	enum = this.receivers.elements();
	while( enum.hasMoreElements() )
	{
	    Enumeration senum;
	    Hashtable sub;

	    sub = (Hashtable)enum.nextElement();
	    senum = sub.elements();
	    while( senum.hasMoreElements() )
	    {
		NodeOSReceiver nr;

		nr = (NodeOSReceiver)senum.nextElement();
		nr.getChannel().dequeueKey();
	    }
	}
    }

    /**
     * @see bees.core.NeighborImpl#getInterfaceName()
     */
    public synchronized String getInterfaceName()
    {
	String retval = "if";
	int index;

	if( (index = this.getIncomingProtocolSpec().indexOf('/')) != -1 )
	{
	    retval = this.incomingProtocolSpec.substring(0, index);
	}
	else if( this.incomingProtocolSpec.equals("") )
	{
	    retval = this.incomingProtocolSpec;
	}
	return retval;
    }

    /**
     * @see bees.core.NeighborImpl#getFingerPrint()
     */
    public TypeID getFingerPrint()
    {
	return this.fingerPrint;
    }

    /**
     * @see edu.utah.janosvm.kit.comm.jsi.JSIHandler#resolveSpecifier(JSIMessage, JSIPair)
     */
    public JSIHandler resolveSpecifier(JSIMessage jm, JSIPair jp)
    {
	return this;
    }

    /**
     * @see edu.utah.janosvm.kit.comm.jsi.JSIHandler#handleMessage(JSIHandlerContext, JSIMessage)
     */
    public void handleMessage(JSIHandlerContext jhc, JSIMessage jm)
    {
	JSIPair jp;
	
	switch( jm.getWhat() )
	{
	case JSIMessage.GET_PROPERTY:
	    if( (jp = jm.popSpecifier()) != null )
	    {
		if( jm.popSpecifier() != null )
		{
		    jm.error("Too many specifiers");
		}
		else if( jp.getTag().equals("incomingProtocolSpec") )
		{
		    jm.reply(this.getIncomingProtocolSpec());
		}
		else if( jp.getTag().equals("incomingAddressSpec") )
		{
		    jm.reply(this.getIncomingAddressSpec());
		}
		else if( jp.getTag().equals("outgoingProtocolSpec") )
		{
		    jm.reply(this.getOutgoingProtocolSpec());
		}
		else if( jp.getTag().equals("outgoingAddressSpec") )
		{
		    jm.reply(this.getOutgoingAddressSpec());
		}
		else 
		{
		    jm.error("Unknown specifier: " + jp.getTag());
		}
	    }
	    else
	    {
		jm.clear();
		jm.setWhat(JSIMessage.REPLY);
		jm.add("incomingProtocolSpec", this.getIncomingProtocolSpec());
		jm.add("incomingAddressSpec", this.getIncomingAddressSpec());
		jm.add("outgoingProtocolSpec", this.getOutgoingProtocolSpec());
		jm.add("outgoingAddressSpec", this.getOutgoingAddressSpec());
	    }
	    break;
	case JSIMessage.LIST_PROPERTIES:
	    if( jm.popSpecifier() == null )
	    {
		jm.clear();
		jm.setWhat(JSIMessage.REPLY);
		jm.add("result", "incomingProtocolSpec");
		jm.add("result", "incomingAddressSpec");
		jm.add("result", "outgoingProtocolSpec");
		jm.add("result", "outgoingAddressSpec");
	    }
	    else
	    {
		jm.error("Too many specifiers");
	    }
	    break;
	case JSIMessage.GET_SUITES:
	    if( jm.popSpecifier() == null )
	    {
		jm.reply(JSI_PROPERTIES);
	    }
	    else
	    {
		jm.error("Too many specifiers");
	    }
	    break;
	default:
	    throw new UnsupportedOperationException();
	}
    }

    /**
     * @see edu.utah.janosvm.kit.comm.jsi.JSIHandler#getSuites(JSIMessage)
     */
    public void getSuites(JSIMessage jm)
    {
    }

    /**
     * @return A more verbose description of this neighbor object.
     */
    public String toVerboseString()
    {
	return "NodeOSNeighbor["
	    + super.toString()
	    + "; incomingProtocolSpec="
	    + this.incomingProtocolSpec
	    + "; incomingAddressSpec="
	    + this.incomingAddressSpec
	    + "; outgoingProtocolSpec="
	    + this.outgoingProtocolSpec
	    + "; outgoingAddressSpec="
	    + this.outgoingAddressSpec
	    + "; receivers="
	    + this.receivers
	    + "; oc="
	    + this.oc
	    + "; fingerPrint="
	    + this.fingerPrint
	    + "]";
    }

    /**
     * @see java.lang.Object#toString()
     */
    public String toString()
    {
	return "NodeOSNeighbor["
	    + super.toString()
	    + "; incomingAddressSpec="
	    + this.incomingAddressSpec
	    + "; outgoingAddressSpec="
	    + this.outgoingAddressSpec
	    + "; fingerPrint="
	    + this.fingerPrint
	    + "]";
    }
}
