/*
 * EnumerationPermission.java
 *
 * Copyright (c) 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is part of the Janos Bees Software.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */

package bees.core.security;

import bees.core.Xdr;
import bees.core.XdrByteArray;

/**
 * An EnumerationPermission holds a string value, representing a symbol, that
 * can later be mapped to an integer.
 */
public class EnumerationPermission
    extends Permission
{
    /**
     * The symbol name.
     */
    private String value;

    /**
     * Construct an EnumerationPermission with the given arguments.
     *
     * @param name The name of the permission.
     * @param actions The actions granted by this permission.
     * @param value The symbol name.
     */
    public EnumerationPermission(String name, String actions, String value)
    {
	super(name, actions);

	this.value = value;
    }

    /**
     * Construct an EnumerationPermission with the given arguments.
     *
     * @param name The name of the permission.
     * @param value The symbol name.
     */
    public EnumerationPermission(String name, String value)
    {
	this(name, "read,write,execute", value);
    }

    /**
     * Construct an empty EnumerationPermission.
     */
    public EnumerationPermission()
    {
    }

    /**
     * @see bees.core.security.Permission#getTypeID()
     */
    public int getTypeID()
    {
	return PERMISSION_ENUMERATION;
    }

    /**
     * @return The symbol name.
     */
    public String getValue()
    {
	return this.value;
    }

    /**
     * Map the symbol name in this permission to an integer using the given
     * enumeration.  Note:  Names are considered case insensitive.
     *
     * @param ep The enumeration as an array of EnumerationPair objects.
     * @return The integer stored in the matching EnumerationPair object.
     * @throws EnumerationException if the symbol couldn't be matched.
     */
    public int mapValue(EnumerationPair ep[])
	throws EnumerationException
    {
	int lpc;

	for( lpc = 0; lpc < ep.length; lpc++ )
	{
	    if( ep[lpc].getString().equalsIgnoreCase(this.value) )
	    {
		return ep[lpc].getID();
	    }
	}
	throw new EnumerationException("Unknown symbol: " + this.value);
    }

    /**
     * @see bees.core.security.Permission#implies(Permission)
     */
    public boolean implies(Permission what)
    {
	boolean retval;

	if( (retval = super.implies(what)) )
	{
	    if( what instanceof EnumerationPermission )
	    {
		EnumerationPermission ep;

		ep = (EnumerationPermission)what;
		retval = this.value.equals(ep.value);
	    }
	}
	return retval;
    }

    /**
     * @see bees.core.security.Permission#marshalledLength()
     */
    public int marshalledLength()
    {
	return super.marshalledLength() + Xdr.STRING(this.value);
    }

    /**
     * @see bees.core.security.Permission#marshal(XdrByteArray)
     */
    public void marshal(XdrByteArray xba)
    {
	super.marshal(xba);

	xba.PUT(this.value);
    }

    /**
     * @see bees.core.security.Permission#unmarshal(XdrByteArray)
     */
    public void unmarshal(XdrByteArray xba)
    {
	super.unmarshal(xba);

	this.value = xba.STRING();
    }

    /**
     * @see java.lang.Object#equals(Object)
     */
    public boolean equals(Object obj)
    {
	boolean retval = false;

	if( super.equals(obj) && (obj instanceof EnumerationPermission) )
	{
	    EnumerationPermission ep;

	    ep = (EnumerationPermission)obj;
	    retval = this.value.equals(ep.value);
	}
	return retval;
    }
    
    /**
     * @see java.lang.Object#toString()
     */
    public String toString()
    {
	return "EnumerationPermission["
	    + super.toString()
	    + "; value="
	    + this.value
	    + "]";
    }
}
