/*
 * GuardedStringCapability.java
 *
 * Copyright (c) 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is part of the Janos Bees Software.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */

package bees.core.security;

import bees.core.Xdr;
import bees.core.Capability;
import bees.core.XdrByteArray;

/**
 * A GuardedStringCapability is a capability base class that wraps a string
 * that should not be leaked to untrusted code.
 */
class GuardedStringCapability
    extends Capability
{
    /**
     * Marshal a string into a byte array.
     *
     * @param value The string to flatten.
     * @return A byte array containing the flattened arguments.
     */
    static byte[] marshal(String value)
    {
	XdrByteArray xba;

	xba = new XdrByteArray(Xdr.STRING(value));
	xba.PUT(value);
	return xba.toBytes();
    }

    /**
     * The guarded string.
     */
    private String value;

    /**
     * @see bees.core.Capability#fromBytes(byte[])
     */
    protected final void fromBytes(byte flattened[])
	throws Exception
    {
	XdrByteArray xba;
	
	super.fromBytes(flattened);
	
	xba = new XdrByteArray(flattened);
	this.value = xba.STRING();
    }

    /**
     * @see bees.core.Capability#revoke()
     */
    public synchronized void revoke()
    {
	super.clear();
	
	this.value = null;
    }

    /**
     * @return The guarded string.
     */
    synchronized String getValue()
    {
	return this.value;
    }

    /**
     * @see java.lang.Object#toString()
     */
    public String toString()
    {
	return "GuardedStringCapability["
	    + super.toString()
	    + "]";
    }
}
