/*
 * KeyFactory.java
 *
 * Copyright (c) 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is part of the Janos Bees Software.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */

package bees.core.security;

import java.security.NoSuchAlgorithmException;

import java.security.spec.InvalidKeySpecException;

import bees.core.kernel.KernelObjectCoroutine;

import bees.core.bdomain.BDomain;

/**
 * A KeyFactory is used to create public and private keys from their respective
 * KeySpecs.
 */
public final class KeyFactory
{
    /**
     * Cached reference to the current domain.
     */
    private final BDomain bdomain;

    /**
     * Construct an empty KeyFactory.
     */
    public KeyFactory()
    {
	this.bdomain = BDomain.currentBDomain();
    }

    /**
     * Create a PrivateKey capability from the given values.
     *
     * @param ks The key specification.
     * @param comment A comment describing the key.
     * @return A new PrivateKey capability.
     * @throws NoSuchAlgorithmException if the key algorithm is not available.
     * @throws InvalidKeySpecException if the key spec is not valid.
     */
    public synchronized PrivateKey createPrivateKey(PrivateKeySpec ks,
						    String comment)
	throws NoSuchAlgorithmException,
	       InvalidKeySpecException
    {
	PrivateKey retval;

	retval = (PrivateKey)KernelObjectCoroutine.
	    create(new PrivateKey(),
		   this.bdomain.getBDomainCoroutine(),
		   ks.toBytes(),
		   comment);
	return retval;
    }

    /**
     * Create a PrivateKey capability from the given values.
     *
     * @param ks The key specification.
     * @return A new PrivateKey capability.
     * @throws NoSuchAlgorithmException if the key algorithm is not available.
     * @throws InvalidKeySpecException if the key spec is not valid.
     */
    public synchronized PrivateKey createPrivateKey(PrivateKeySpec ks)
	throws NoSuchAlgorithmException,
	       InvalidKeySpecException
    {
	return this.createPrivateKey(ks, "Private Key");
    }
    
    /**
     * Create a PublicKey capability from the given values.
     *
     * @param ks The key specification.
     * @param ps The permissions for the new capability.
     * @param comment A comment describing the key.
     * @return A new PublicKey capability.
     * @throws NoSuchAlgorithmException if the key algorithm is not available.
     * @throws InvalidKeySpecException if the key spec is not valid.
     */
    public synchronized PublicKey createPublicKey(PublicKeySpec ks,
						  PermissionSet ps,
						  String comment)
	throws NoSuchAlgorithmException,
	       InvalidKeySpecException
    {
	PublicKey retval;

	retval = (PublicKey)KernelObjectCoroutine.
	    create(new PublicKey(),
		   this.bdomain.getBDomainCoroutine(),
		   ks.toBytes(),
		   comment);
	return retval;
    }

    /**
     * Create a PublicKey capability from the given values.
     *
     * @param ks The key specification.
     * @param ps The permissions for the new capability.
     * @return A new PrivateKey capability.
     * @throws NoSuchAlgorithmException if the key algorithm is not available.
     * @throws InvalidKeySpecException if the key spec is not valid.
     */
    public synchronized PublicKey createPublicKey(PublicKeySpec ks,
						  PermissionSet ps)
	throws NoSuchAlgorithmException,
	       InvalidKeySpecException
    {
	return this.createPublicKey(ks, ps, "Public Key");
    }
    
    /**
     * Create a PublicKey capability from the given value with the default
     * permissions.
     *
     * @param ks The key specification.
     * @return A new PrivateKey capability.
     * @throws NoSuchAlgorithmException if the key algorithm is not available.
     * @throws InvalidKeySpecException if the key spec is not valid.
     */
    public synchronized PublicKey createPublicKey(PublicKeySpec ks)
	throws NoSuchAlgorithmException,
	       InvalidKeySpecException
    {
	return this.createPublicKey(ks,
				    PublicKey.DEFAULT_PERMISSIONS,
				    "Public Key");
    }
    
    /**
     * @see java.lang.Object#toString()
     */
    public String toString()
    {
	return "KeyFactory["
	    + "]";
    }
}
