/*
 * KeyPairGenerator.java
 *
 * Copyright (c) 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is part of the Janos Bees Software.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */

package bees.core.security;

import java.security.NoSuchAlgorithmException;

import bees.core.Capability;
import bees.core.CapabilityFactory;

import bees.core.kernel.KernelObjectCoroutine;

import bees.core.bdomain.BDomain;

/**
 * A KeyPairGenerator is used to manufacture new public and private key
 * capabilities.
 */
public final class KeyPairGenerator
{
    /**
     * Cached reference to the current domain.
     */
    private final BDomain bdomain;

    /**
     * Construct an empty KeyPairGenerator.
     */
    public KeyPairGenerator()
    {
	this.bdomain = BDomain.currentBDomain();
    }

    /**
     * Generate a new key pair from the given values.
     *
     * @param algorithm The type of key to generate.
     * @param keySize The suggested number of bits the key should have.
     * @param comment A comment to attach to the capability.
     * @return A KeyPair containing the public and private key capabilities.
     * @throws NoSuchAlgorithmException if "algorithm" is not available.
     */
    public synchronized KeyPair generateKeyPair(String algorithm,
						int keySize,
						String comment)
	throws NoSuchAlgorithmException
    {
	java.security.KeyPairGenerator keyGen;
	java.security.KeyPair jskp;
	KeyPair retval = null;
	byte data[][];
	
	keyGen = java.security.KeyPairGenerator.getInstance(algorithm);
	keyGen.initialize(keySize);
	jskp = keyGen.generateKeyPair();
	data = new byte[][] {
	    KeySpec.marshal(jskp.getPrivate()),
	    KeySpec.marshal(jskp.getPublic())
	};
	retval = new KeyPair(new PrivateKey(), new PublicKey());
	if( KernelObjectCoroutine.create(this.bdomain.getBDomainCoroutine(),
					 new Capability[] {
					     retval.getPrivate(),
					     retval.getPublic()
					 },
					 data,
					 comment) )
	{
	    CapabilityFactory.grant(retval.getPrivate(),
				    PrivateKey.DEFAULT_PERMISSIONS);
	    CapabilityFactory.grant(retval.getPublic(),
				    PublicKey.DEFAULT_PERMISSIONS);
	}
	else
	{
	    throw new OutOfMemoryError();
	}
	return retval;
    }

    /**
     * Generate a new key pair from the given values.
     *
     * @param algorithm The type of key to generate.
     * @param keySize The suggested number of bits the key should have.
     * @return A KeyPair containing the public and private key capabilities.
     * @throws NoSuchAlgorithmException if "algorithm" is not available.
     */
    public synchronized KeyPair generateKeyPair(String algorithm, int keySize)
	throws NoSuchAlgorithmException
    {
	return this.generateKeyPair(algorithm, keySize, "Key Pair");
    }

    /**
     * @see java.lang.Object#toString()
     */
    public String toString()
    {
	return "KeyPairGenerator["
	    + "]";
    }
}
