/*
 * KeyPairGeneratorHandler.java
 *
 * Copyright (c) 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is part of the Janos Bees Software.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */

package bees.core.security;

import java.security.NoSuchAlgorithmException;

import edu.utah.janos.nodeos.AddressFormatException;

import edu.utah.janosvm.kit.comm.jsi.JSIPair;
import edu.utah.janosvm.kit.comm.jsi.JSIMessage;
import edu.utah.janosvm.kit.comm.jsi.JSIPropertyInfo;
import edu.utah.janosvm.kit.comm.jsi.MissingPairException;

import bees.core.HashMap;
import bees.core.ObjectMap;
import bees.core.Capability;

import bees.core.kernel.KernelCapabilities;
import bees.core.kernel.KernelObjectCoroutine;

import bees.core.resource.CapabilityHandler;

import bees.core.bdomain.BDomain;

/**
 * The KeyPairGeneratorHandler is a JSIHandler that can create KeyPairs.
 */
public final class KeyPairGeneratorHandler
    extends CapabilityHandler
{
    /**
     * The set of supported properties to be added to a JSIMessage.GET_SUITES
     * request.
     */
    private static final JSIPropertyInfo JSI_PROPERTIES[] = {
	new JSIPropertyInfo("keyPairGenerator",
			    (1 << JSIMessage.CREATE_PROPERTY),
			    "name:string - The capability name.\n"
			    + "algorithm:string - The key algorithm.\n"
			    + "size:int - The key size (Default: 1024).\n"
			    + "comment:string - An optional comment for the "
			    + "capability.\n"
			    + "\n"
			    + "Create a public and private key pair.\n"),
    };
    
    /**
     * Construct an empty KeyPairGeneratorHandler object.
     */
    public KeyPairGeneratorHandler()
    {
    }

    /**
     * @see bees.core.resource.CapabilityHandler#getCapabilityClass()
     */
    public Class getCapabilityClass()
    {
	return ObjectMap.class;
    }

    /**
     * @see bees.core.resource.CapabilityHandler#createCapability(JSIMessage, String, String)
     */
    public Object createCapability(JSIMessage jm,
				   String name,
				   String comment)
	throws NoSuchAlgorithmException,
	       AddressFormatException,
	       MissingPairException
    {
	java.security.KeyPairGenerator kpg;
	ObjectMap retval = new HashMap();
	int size = 1024, signer = 1;
	java.security.KeyPair kp;
	String algorithm;
	JSIPair jp;
	long id;
	
	algorithm = jm.expect("algorithm");
	if( (jp = jm.getPair("size")) != null )
	{
	    size = (int)jp.getIntValue();
	}
	if( (jp = jm.getPair("signer")) != null )
	{
	    signer = (int)jp.getIntValue();
	}
	comment = jm.get("comment", "KeyPair");
	kpg = java.security.KeyPairGenerator.getInstance(algorithm);
	kpg.initialize(size);
	kp = kpg.generateKeyPair();
	retval.put("private",
		   KernelObjectCoroutine.
		   create(new PrivateKey(),
			  BDomain.currentBDomain().getBDomainCoroutine(),
			  KeySpec.marshal(kp.getPrivate()),
			  null,
			  comment,
			  new PermissionSet(new Permission[] {
			      signer != 0 ?
			      Signature.SIGN_PERMISSION :
			      Cipher.DECRYPT_PERMISSION,
			      Capability.DELETE_PERMISSION,
			  })));
	retval.put("public",
		   KernelObjectCoroutine.
		   create(new PublicKey(),
			  BDomain.currentBDomain().getBDomainCoroutine(),
			  KeySpec.marshal(kp.getPublic()),
			  null,
			  comment,
			  new PermissionSet(new Permission[] {
			      signer != 0 ?
			      Signature.VERIFY_PERMISSION :
			      Cipher.ENCRYPT_PERMISSION,
			      PublicKey.READ_ENCODED_PERMISSION,
			      Capability.DELETE_PERMISSION,
			  })));
	return retval;
    }

    /**
     * @see edu.utah.janosvm.kit.comm.jsi.JSIHandler#getSuites(JSIMessage)
     */
    public void getSuites(JSIMessage jm)
    {
	jm.add("result", JSI_PROPERTIES);
    }

    /**
     * @see java.lang.Object#toString()
     */
    public String toString()
    {
	return "KeyPairGeneratorHandler["
	    + "]";
    }
}
