/*
 * KeySpec.java
 *
 * Copyright (c) 2002, 2003 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is part of the Janos Bees Software.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */

package bees.core.security;

import java.security.NoSuchAlgorithmException;

import java.security.spec.InvalidKeySpecException;

import bees.core.Xdr;
import bees.core.Capability;
import bees.core.XdrByteArray;
import bees.core.CapabilitySpec;
import bees.core.SpecCapabilityFactory;

import edu.utah.janos.nodeos.Node;

/**
 * A KeySpec is a base class for the various cryptographic key classes.
 */
public abstract class KeySpec
    extends Capability
    implements CapabilitySpec
{
    /**
     * Marshal a given key into a byte array.
     *
     * @param key The key to flatten.
     * @return A byte array containing the flattened version of the key.
     */
    static byte[] marshal(java.security.Key key)
    {
	byte algorithm[], enc[];
	XdrByteArray xba;

	algorithm = key.getAlgorithm().getBytes();
	enc = key.getEncoded();
	xba = new XdrByteArray(Xdr.INT
			       + algorithm.length
			       + Xdr.INT
			       + enc.length);
	xba.PUT(algorithm.length);
	xba.PUT(algorithm);
	xba.PUT(enc.length);
	xba.PUT(enc);
	return xba.toBytes();
    }

    /**
     * Construct an empty KeySpec.
     */
    public KeySpec()
    {
    }

    /**
     * @see bees.core.CapabilitySpec#specFromBytes(byte[])
     */
    public final synchronized CapabilitySpec specFromBytes(byte flattened[])
	throws NoSuchAlgorithmException,
	       InvalidKeySpecException
    {
	XdrByteArray xba;
	String algorithm;
	byte enc[];

	xba = new XdrByteArray(flattened);
	algorithm = new String(xba.BYTES(xba.INT()));
	enc = xba.BYTES(xba.INT());
	this.createKey(algorithm, enc);
	return this;
    }

    /**
     * Call back used to create the actual key from the given values.
     *
     * @param algorithm The key algorithm.
     * @param encoding The encoded version of the key.
     * @throws NoSuchAlgorithmException if the algorithm is not supported.
     * @throws InvalidKeySpecException if the encoding value is invalid.
     */
    abstract void createKey(String algorithm, byte encoding[])
	throws NoSuchAlgorithmException,
	       InvalidKeySpecException;

    /**
     * Call back used to get the wrapped key.
     * 
     * @return The wrapped java.security.Key object.
     */
    abstract java.security.Key getKey();

    /**
     * @return The capability material.
     */
    abstract byte[] toBytes();

    /**
     * @see bees.core.CapabilitySpec#getFactory()
     */
    public SpecCapabilityFactory getFactory()
    {
	return null;
    }
    
    /**
     * @see java.lang.Object#toString()
     */
    public String toString()
    {
	return "KeySpec["
	    + super.toString()
	    + "]";
    }
}
