/*
 * PublicKey.java
 *
 * Copyright (c) 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is part of the Janos Bees Software.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */

package bees.core.security;

import java.security.NoSuchAlgorithmException;

import java.security.spec.InvalidKeySpecException;

import edu.utah.janosvm.kit.comm.jsi.JSIPair;
import edu.utah.janosvm.kit.comm.jsi.JSIHandler;
import edu.utah.janosvm.kit.comm.jsi.JSIMessage;
import edu.utah.janosvm.kit.comm.jsi.JSIPropertyInfo;
import edu.utah.janosvm.kit.comm.jsi.JSIHandlerContext;

import bees.core.TypeID;

import bees.util.MD5;

/**
 * A PublicKey capability wraps a java.security.PublicKey object.  This makes
 * it possible to pass the object to untrusted code without leaking the key
 * material.
 */
public final class PublicKey
    extends PublicKeySpec
    implements JSIHandler
{
    /**
     * The per-object level properties of the NodeOSNeighbor handler.
     */
    private static final JSIPropertyInfo JSI_PROPERTIES[] = {
	new JSIPropertyInfo("fingerPrint",
			    (1 << JSIMessage.GET_PROPERTY),
			    "The public key's finger print.\n"),
	new JSIPropertyInfo("algorithm",
			    (1 << JSIMessage.GET_PROPERTY),
			    "The public key's algorithm.\n"),
    };
    
    /**
     * Permission to read the encoded version of this key.
     */
    public static final Permission READ_ENCODED_PERMISSION =
	new Permission("encoded", "read");

    /**
     * The default permissions for a new PublicKey capability.
     */
    public static final PermissionSet DEFAULT_PERMISSIONS =
	new PermissionSet(new Permission[] {
	    new Permission("encoded", "read"),
	});

    /**
     * The finger print of this key.
     */
    private TypeID fingerPrint;

    /**
     * Construct an empty PublicKey capability.
     */
    public PublicKey()
    {
    }

    /**
     * @see bees.core.Capability#fromBytes(byte[])
     */
    protected final void fromBytes(byte flattened[])
	throws NoSuchAlgorithmException,
	       InvalidKeySpecException,
	       Exception
    {
	MD5 digest = new MD5();
	
	super.fromBytes(flattened);

	super.specFromBytes(flattened);

	digest.update(super.pk.getEncoded());
	this.fingerPrint = new TypeID(digest.hash());
    }

    /**
     * @return The capability material for this key.
     */
    public synchronized byte[] toBytes()
    {
	this.checkPermission(READ_ENCODED_PERMISSION);

	return super.toBytes();
    }

    /**
     * @return The encoded key material for this key.
     */
    public synchronized byte[] getEncoded()
    {
	this.checkPermission(READ_ENCODED_PERMISSION);

	return this.pk.getEncoded();
    }

    /**
     * @return The finger print for this key.
     */
    public synchronized TypeID getFingerPrint()
    {
	return this.fingerPrint;
    }

    /**
     * @see edu.utah.janosvm.kit.comm.jsi.JSIHandler#resolveSpecifier(JSIMessage, JSIPair)
     */
    public JSIHandler resolveSpecifier(JSIMessage jm, JSIPair jp)
    {
	return this;
    }

    /**
     * @see edu.utah.janosvm.kit.comm.jsi.JSIHandler#handleMessage(JSIHandlerContext, JSIMessage)
     */
    public void handleMessage(JSIHandlerContext jhc, JSIMessage jm)
    {
	JSIPair jp;
	
	switch( jm.getWhat() )
	{
	case JSIMessage.GET_PROPERTY:
	    if( (jp = jm.popSpecifier()) != null )
	    {
		if( jp.getTag().equals("fingerPrint") )
		{
		    jm.reply(this.fingerPrint.toPlainString());
		}
		else if( jp.getTag().equals("algorithm") )
		{
		    jm.reply(super.pk.getAlgorithm());
		}
		else
		{
		    jm.error("Unknown specifier: " + jp.getTag());
		}
	    }
	    else
	    {
		jm.clear();
		jm.setWhat(JSIMessage.REPLY);
		jm.add("fingerPrint", this.fingerPrint.toPlainString());
		jm.add("algorithm", super.pk.getAlgorithm());
	    }
	    break;
	case JSIMessage.LIST_PROPERTIES:
	    if( jm.popSpecifier() == null )
	    {
		jm.clear();
		jm.setWhat(JSIMessage.REPLY);
		jm.add("result", "algorithm");
		jm.add("result", "fingerPrint");
	    }
	    else
	    {
		jm.error("Too many specifiers");
	    }
	    break;
	case JSIMessage.GET_SUITES:
	    if( jm.popSpecifier() == null )
	    {
		jm.reply(JSI_PROPERTIES);
	    }
	    else
	    {
		jm.error("Too many specifiers");
	    }
	    break;
	default:
	    throw new UnsupportedOperationException();
	}
    }

    /**
     * @see edu.utah.janosvm.kit.comm.jsi.JSIHandler#getSuites(JSIMessage)
     */
    public void getSuites(JSIMessage jm)
    {
    }

    /**
     * @see java.lang.Object#toString()
     */
    public String toString()
    {
	return "PublicKey["
	    + super.toString()
	    + "; fingerPrint="
	    + this.fingerPrint
	    + "]";
    }
}
