/*
 * PublicKeySpec.java
 *
 * Copyright (c) 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is part of the Janos Bees Software.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */

package bees.core.security;

import java.security.KeyFactory;
import java.security.NoSuchAlgorithmException;

import java.security.spec.X509EncodedKeySpec;
import java.security.spec.InvalidKeySpecException;

/**
 * The PublicKeySpec class wraps a java.security.PublicKey object.  While this
 * class extends the Capability class it does not have to be a full capability
 * in order to be useful.  Its possible for untrusted code to create a
 * PublicKeySpec object from capability material and use it with a Signature
 * object.
 */
public class PublicKeySpec
    extends KeySpec
{
    /**
     * The wrapped key.
     */
    java.security.PublicKey pk;

    /**
     * Construct a PublicKeySpec with the given value.
     *
     * @param pk A PublicKey object to be wrapped.
     */
    public PublicKeySpec(java.security.PublicKey pk)
    {
	this.pk = pk;
    }
    
    /**
     * Construct an empty PublicKeySpec.
     */
    public PublicKeySpec()
    {
	this(null);
    }

    /**
     * @see bees.core.security.KeySpec#createKey(String, byte[])
     */
    public final void createKey(String algorithm, byte encoding[])
	throws NoSuchAlgorithmException,
	       InvalidKeySpecException
    {
	KeyFactory kf;
	
	kf = KeyFactory.getInstance(algorithm);
	this.pk = kf.generatePublic(new X509EncodedKeySpec(encoding));
    }
    
    /**
     * @see bees.core.security.KeySpec#getKey()
     */
    final java.security.Key getKey()
    {
	return this.pk;
    }
    
    /**
     * @see bees.core.security.KeySpec#toBytes()
     */
    public synchronized byte[] toBytes()
    {
	if( this.pk == null )
	{
	    throw new SecurityException("Invalid spec");
	}
	return marshal(this.pk);
    }

    /**
     * @see bees.core.Capability#revoke()
     */
    public synchronized void revoke()
    {
	super.clear();
	
	this.pk = null;
    }

    /**
     * @return The key algorithm.
     */
    public synchronized String getAlgorithm()
    {
	return this.pk.getAlgorithm();
    }

    /**
     * @return The format of the encoded key.
     */
    public synchronized String getFormat()
    {
	return this.pk.getFormat();
    }

    /**
     * @see java.lang.Object#toString()
     */
    public String toString()
    {
	return "PublicKeySpec["
	    + super.toString()
	    + "; pk="
	    + this.pk
	    + "]";
    }
}
