/*
 * Range.java
 *
 * Copyright (c) 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is part of the Janos Bees Software.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */

package bees.core.security;

import edu.utah.janosvm.sys.ListNode;

import bees.core.Xdr;
import bees.core.Panic;
import bees.core.Assert;
import bees.core.XdrByteArray;

/**
 * Range is a helper for the RangePermission class, it represents a single
 * unbroken range of integers.
 */
final class Range
    extends ListNode
    implements Cloneable
{
    /**
     * The range given to compare() is wholly less than this.
     */
    static final int CMP_LESS = -4;
    
    /**
     * The range given to compare() is less than this but ends where this
     * starts.
     */
    static final int CMP_LESS_EDGE = -3;
    
    /**
     * The range given to compare() is less than this but ends in the middle
     * of this.
     */
    static final int CMP_LESS_OVERLAP = -2;
    
    /**
     * The range given to compare() can completely contain this.
     */
    static final int CMP_EXTERNAL = -1;
    
    /**
     * The range given to compare() is the same as this.
     */
    static final int CMP_SAME = 0;
    
    /**
     * The range given to compare() can be completely contained by this.
     */
    static final int CMP_INTERNAL = 1;
    
    /**
     * The range given to compare() starts in the middle of this and continues
     * past the end.
     */
    static final int CMP_GREATER_OVERLAP = 2;
    
    /**
     * The range given to compare() starts where this ends.
     */
    static final int CMP_GREATER_EDGE = 3;
    
    /**
     * The range given to compare() wholly greater than this.
     */
    static final int CMP_GREATER = 4;

    /**
     * The start of the range.
     */
    private long start;

    /**
     * The number of integers in the range.
     */
    private long length;

    /**
     * @param start The start of the range.
     * @param length The number of integers in the range.
     */
    Range(long start, long length)
    {
	Assert.require(length >= 0);
	
	this.start = start;
	this.length = length;
    }

    /**
     * Construct an empty Range.
     */
    Range()
    {
	this(0, 0);
    }

    /**
     * @return The start of the range.
     */
    long getStart()
    {
	return this.start;
    }

    /**
     * @return The number of integers in the range.
     */
    long getLength()
    {
	return this.length;
    }

    /**
     * @param index The number to test for coverage by the range.
     * @return True if the number lies within the bounds of this range.
     */
    boolean inRange(long index)
    {
	boolean retval = false;

	if( (index >= this.start) && (index < (this.start + this.length)) )
	{
	    retval = true;
	}
	return retval;
    }

    /**
     * Union the given range into this one, they are expected to atleast
     * partially overlap.
     *
     * @param range The range to union.
     */
    void union(Range range)
    {
	if( range.start < this.start )
	{
	    long diff;

	    diff = this.start - range.start;
	    this.start = range.start;
	    this.length += diff;
	}
	if( (range.start + range.length) > (this.start + this.length) )
	{
	    this.length = (range.start + range.length) - this.start;
	}
    }

    /**
     * Intersect the given range with this one, they are expected to atleast
     * partially overlap.
     *
     * @param range The range to intersect.
     */
    void intersect(Range range)
    {
	if( range.start < this.start )
	{
	    this.length = (range.start + range.length) - this.start;
	}
	else
	{
	    long tend, rend;

	    tend = this.start + this.length - 1;
	    rend = range.start + range.length - 1;
	    this.start = range.start;
	    this.length = Math.min(tend, rend) - this.start + 1;
	}
    }

    /**
     * Remove the given range from this one, they are expected to atleast
     * partially overlap.
     *
     * @param range The range to remove.
     */
    void remove(Range range)
    {
	if( range.start < this.start )
	{
	    long diff;

	    diff = this.start - range.start;
	    this.start = range.start + range.length - 1;
	    this.length -= diff;
	}
	else
	{
	    this.length = range.start - this.start;
	}
    }

    /**
     * Split this range into two by removing the given middle chunk.
     *
     * @param range The range to remove from the middle.
     * @return A new Range covering the lower part of this range or null if
     * the lower part would have zero length.
     */
    Range split(Range range)
    {
	long end, sstart, slength;
	Range retval = null;

	sstart = this.start;
	slength = range.start - this.start;
	end = this.start + this.length - 1;
	this.start = range.start + range.length - 1;
	this.length = end - this.start + 1;
	if( slength > 0 )
	{
	    retval = new Range(sstart, slength);
	}
	return retval;
    }

    /**
     * Compare this range with the given one.
     *
     * @param range The range to compare against.
     * @return One of the CMP_* values defined above.
     */
    int compare(Range range)
    {
	int retval;
	
	if( range.start < this.start )
	{
	    long rend;

	    rend = range.start + range.length - 1;
	    if( (rend + 1) == this.start )
	    {
		retval = CMP_LESS_EDGE;
	    }
	    else if( rend < this.start )
	    {
		retval = CMP_LESS;
	    }
	    else if( rend < (this.start + this.length) )
	    {
		retval = CMP_LESS_OVERLAP;
	    }
	    else
	    {
		retval = CMP_EXTERNAL;
	    }
	}
	else if( range.start == this.start )
	{
	    if( range.length < this.length )
	    {
		retval = CMP_INTERNAL;
	    }
	    else if( range.length == this.length )
	    {
		retval = CMP_SAME;
	    }
	    else
	    {
		retval = CMP_EXTERNAL;
	    }
	}
	else if( range.start < (this.start + this.length) )
	{
	    long rend;

	    rend = range.start + range.length - 1;
	    if( rend < (this.start + this.length) )
	    {
		retval = CMP_INTERNAL;
	    }
	    else
	    {
		retval = CMP_GREATER_OVERLAP;
	    }
	}
	else if( range.start == (this.start + this.length) )
	{
	    retval = CMP_GREATER_EDGE;
	}
	else
	{
	    retval = CMP_GREATER;
	}
	return retval;
    }

    static int marshalledLength()
    {
	return Xdr.LONG + Xdr.LONG;
    }

    void marshal(XdrByteArray xba)
    {
	xba.PUT(this.start);
	xba.PUT(this.length);
    }

    void unmarshal(XdrByteArray xba)
    {
	this.start = xba.LONG();
	this.length = xba.LONG();
    }

    /**
     * @see java.lang.Object#clone()
     */
    public Object clone()
    {
	try
	{
	    return super.clone();
	}
	catch(CloneNotSupportedException e)
	{
	    throw new Panic(e);
	}
    }

    /**
     * @return A string of the form "<this.start>-<this.start + this.length -
     * 1>".
     */
    public String toNaturalString()
    {
	return Long.toString(this.start)
	    + "-"
	    + Long.toString(this.start + this.length - 1);
    }
    
    /**
     * @see java.lang.Object#toString()
     */
    public String toString()
    {
	return "Range["
	    + super.toString()
	    + "; start="
	    + this.start
	    + "; length="
	    + this.length
	    + "]";
    }
}
