/*
 * RangePermission.java
 *
 * Copyright (c) 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is part of the Janos Bees Software.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */

package bees.core.security;

import edu.utah.janosvm.sys.List;
import edu.utah.janosvm.sys.ListNode;

import bees.core.Xdr;
import bees.core.Panic;
import bees.core.Assert;
import bees.core.XdrByteArray;

/**
 * A RangePermission restricts access based on ranges of integers.
 */
public class RangePermission
    extends Permission
{
    /**
     * The list of ranges in this permission.
     */
    private final List rangeList = new List();

    /**
     * Construct a RangePermission with the given values and empty set of
     * ranges.
     *
     * @param name The name of the permission.
     * @param actions The actions allowed for this permission.
     */
    public RangePermission(String name, String actions)
    {
	super(name, actions);
    }

    /**
     * Construct a RangePermission with the given values.
     *
     * @param name The name of the permission.
     * @param actions The actions allowed for this permission.
     * @param start The start of the range.
     * @param length The length of the range.
     */
    public RangePermission(String name, String actions,
			   long start, long length)
    {
	this(name, actions);

	Assert.require(length >= 0);
	
	this.rangeList.addHead(new Range(start, length));
    }

    /**
     * Construct an empty RangePermission.
     */
    public RangePermission()
    {
    }

    /**
     * @see bees.core.security.Permission#getTypeID()
     */
    public int getTypeID()
    {
	return PERMISSION_RANGE;
    }

    public String getValue()
    {
	String retval = "";
	ListNode curr;

	curr = this.rangeList.head.succ;
	while( curr.succ != null )
	{
	    if( retval.length() != 0 )
	    {
		retval += ", ";
	    }
	    retval += ((Range)curr).toNaturalString();
	    curr = curr.succ;
	}
	return retval;
    }

    /**
     * @see bees.core.security.Permission#implies(Permission)
     */
    public boolean implies(Permission what)
    {
	boolean retval = false;

	Assert.require(what);

	if( (retval = super.implies(what)) )
	{
	    if( what instanceof RangePermission )
	    {
		RangePermission rp;
		ListNode curr;

		retval = false;
		rp = (RangePermission)what;
		curr = rp.rangeList.head.succ;
		while( (retval == false) && (curr.succ != null) )
		{
		    retval = this.intersects((Range)curr);
		    curr = curr.succ;
		}
	    }
	    else if( what instanceof IntegerPermission )
	    {
		IntegerPermission ip;

		ip = (IntegerPermission)what;
		retval = this.intersects(ip.value);
	    }
	}
	return retval;
    }

    /**
     * Test if a given range intersects with the set of ranges.
     *
     * @param nr The range to test.
     * @return True if the range is with any of the ranges.
     */
    private boolean intersects(Range nr)
    {
	boolean retval = false;
	ListNode curr;

	Assert.require(nr);

	curr = this.rangeList.head.succ;
	while( (retval == false) && (curr.succ != null) )
	{
	    Range range;

	    range = (Range)curr;
	    switch( range.compare(nr) )
	    {
	    case Range.CMP_LESS_OVERLAP:
	    case Range.CMP_EXTERNAL:
	    case Range.CMP_SAME:
	    case Range.CMP_INTERNAL:
	    case Range.CMP_GREATER_OVERLAP:
		retval = true;
		break;
	    }
	    curr = curr.succ;
	}
	return retval;
    }

    /**
     * Test if a given integer intersects with the set of ranges.
     *
     * @param index The integer to test.
     * @return True if the integer is with any of the ranges.
     */
    private boolean intersects(long index)
    {
	boolean retval = false;
	ListNode curr;

	curr = this.rangeList.head.succ;
	while( (retval == false) && (curr.succ != null) )
	{
	    Range range;

	    range = (Range)curr;
	    retval = range.inRange(index);
	    curr = curr.succ;
	}
	return retval;
    }

    /**
     * Intersect the given range with the internal set of ranges.
     *
     * @param nr The Range to intersect.
     */
    private void intersect(Range nr)
    {
	boolean modified = false, done = false;
	ListNode curr, succ;

	Assert.require(nr);
	
	curr = this.rangeList.head.succ;
	while( !done && (curr.succ != null) )
	{
	    Range range;

	    succ = curr.succ;
	    range = (Range)curr;
	    switch( range.compare(nr) )
	    {
	    case Range.CMP_LESS:
	    case Range.CMP_LESS_EDGE:
		done = true;
		break;
	    case Range.CMP_LESS_OVERLAP:
	    case Range.CMP_INTERNAL:
		range.intersect(nr);
		done = true;
		break;
	    case Range.CMP_EXTERNAL:
	    case Range.CMP_GREATER_EDGE:
	    case Range.CMP_GREATER:
		break;
	    case Range.CMP_SAME:
		done = true;
		break;
	    case Range.CMP_GREATER_OVERLAP:
		range.intersect(nr);
		modified = true;
		nr = range;
		break;
	    default:
		throw new Panic("Invalid compare() return value!");
	    }
	    curr = succ;
	}
    }

    /**
     * Union the given range with the internal set of ranges.
     *
     * @param nr The Range to union.
     */
    private void union(Range nr)
    {
	boolean modified = false, done = false;
	ListNode curr, succ;

	Assert.require(nr);
	
	curr = this.rangeList.head.succ;
	while( !done && (curr.succ != null) )
	{
	    Range range;

	    succ = curr.succ;
	    range = (Range)curr;
	    switch( range.compare(nr) )
	    {
	    case Range.CMP_LESS:
		if( !modified )
		{
		    curr.prepend((ListNode)nr.clone());
		}
		done = true;
		break;
	    case Range.CMP_LESS_EDGE:
	    case Range.CMP_LESS_OVERLAP:
		range.union(nr);
		done = true;
		if( modified )
		{
		    nr.remove();
		}
		break;
	    case Range.CMP_EXTERNAL:
	    case Range.CMP_GREATER_EDGE:
	    case Range.CMP_GREATER_OVERLAP:
		range.union(nr);
		if( modified )
		{
		    nr.remove();
		}
		modified = true;
		nr = range;
		break;
	    case Range.CMP_SAME:
	    case Range.CMP_INTERNAL:
		done = true;
		break;
	    case Range.CMP_GREATER:
		/* Nothing */
		break;
	    default:
		throw new Panic("Invalid compare() return value!");
	    }
	    curr = succ;
	}
	if( !done && !modified )
	{
	    this.rangeList.addTail((ListNode)nr.clone());
	}
    }

    /**
     * Remove the given range from the internal set of ranges.
     *
     * @param nr The Range to remove.
     */
    private void remove(Range nr)
    {
	boolean modified = false, done = false;
	ListNode curr, succ;

	Assert.require(nr);
	
	curr = this.rangeList.head.succ;
	while( !done && (curr.succ != null) )
	{
	    Range range;

	    succ = curr.succ;
	    range = (Range)curr;
	    switch( range.compare(nr) )
	    {
	    case Range.CMP_LESS:
	    case Range.CMP_LESS_EDGE:
		done = true;
		break;
	    case Range.CMP_LESS_OVERLAP:
		range.remove(nr);
		if( range.getLength() == 0 )
		{
		    range.remove();
		}
		done = true;
		break;
	    case Range.CMP_GREATER_OVERLAP:
		range.remove(nr);
		modified = true;
		nr = range;
		break;
	    case Range.CMP_EXTERNAL:
		curr.remove();
		break;
	    case Range.CMP_SAME:
		curr.remove();
		done = true;
		break;
	    case Range.CMP_INTERNAL:
		if( (nr = range.split(nr)) != null )
		{
		    curr.prepend(nr);
		}
		done = true;
		break;
	    case Range.CMP_GREATER_EDGE:
	    case Range.CMP_GREATER:
		break;
	    default:
		throw new Panic("Invalid compare() return value!");
	    }
	    curr = succ;
	}
	if( !done )
	{
	    /* Out of range */
	}
    }

    /**
     * @see bees.core.security.Permission#grant(Permission)
     */
    public void grant(Permission what)
    {
	Assert.require(what);
	
	super.grant(what);
	
	if( what instanceof RangePermission )
	{
	    RangePermission rp;
	    ListNode curr;

	    rp = (RangePermission)what;
	    curr = rp.rangeList.head.succ;
	    while( curr.succ != null )
	    {
		this.union((Range)curr);
		curr = curr.succ;
	    }
	}
	else if( what instanceof IntegerPermission )
	{
	    IntegerPermission ip;

	    ip = (IntegerPermission)what;
	    this.union(new Range(ip.getValue(), 1));
	}
    }
    
    /**
     * @see bees.core.security.Permission#revoke(Permission)
     */
    public void revoke(Permission what)
    {
	Assert.require(what);
	
	super.revoke(what);
	
	if( what instanceof RangePermission )
	{
	    RangePermission rp;
	    ListNode curr;

	    rp = (RangePermission)what;
	    curr = rp.rangeList.head.succ;
	    while( curr.succ != null )
	    {
		this.remove((Range)curr);
		curr = curr.succ;
	    }
	}
	else if( what instanceof IntegerPermission )
	{
	    IntegerPermission ip;

	    ip = (IntegerPermission)what;
	    this.remove(new Range(ip.getValue(), 1));
	}
    }

    /**
     * @see bees.core.security.Permission#bound(Permission)
     */
    public void bound(Permission what)
    {
	Assert.require(what);
	
	super.bound(what);
	
	if( what instanceof RangePermission )
	{
	    RangePermission rp;
	    ListNode curr;

	    rp = (RangePermission)what;
	    curr = rp.rangeList.head.succ;
	    while( curr.succ != null )
	    {
		this.intersect((Range)curr);
		curr = curr.succ;
	    }
	}
	else if( what instanceof IntegerPermission )
	{
	    IntegerPermission ip;

	    ip = (IntegerPermission)what;
	    this.intersect(new Range(ip.getValue(), 1));
	}
    }

    /**
     * @see bees.core.security.Permission#marshalledLength()
     */
    public int marshalledLength()
    {
	return super.marshalledLength() + Xdr.INT +
	    (this.rangeList.count() * Range.marshalledLength());
    }

    /**
     * @see bees.core.security.Permission#marshal(XdrByteArray)
     */
    public void marshal(XdrByteArray xba)
    {
	ListNode curr;

	Assert.require(xba);
	
	super.marshal(xba);

	xba.PUT(this.rangeList.count());
	curr = this.rangeList.head.succ;
	while( curr.succ != null )
	{
	    Range range;

	    range = (Range)curr;
	    range.marshal(xba);
	    curr = curr.succ;
	}
    }

    /**
     * @see bees.core.security.Permission#unmarshal(XdrByteArray)
     */
    public void unmarshal(XdrByteArray xba)
    {
	int lpc, count;

	Assert.require(xba);
	
	super.unmarshal(xba);

	count = xba.INT();
	for( lpc = 0; lpc < count; lpc++ )
	{
	    Range range;

	    range = new Range();
	    range.unmarshal(xba);
	    this.rangeList.addTail(range);
	}
    }
    
    /**
     * @see java.lang.Object#toString()
     */
    public String toString()
    {
	String retval = "RangePermission["
	    + super.toString()
	    + "; rangeList={ ";
	ListNode curr;

	curr = this.rangeList.head.succ;
	while( curr.succ != null )
	{
	    retval += curr + " ";
	    curr = curr.succ;
	}
	retval += "}]";
	return retval;
    }
}
