/*
 * SecretKeyFactory.java
 *
 * Copyright (c) 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is part of the Janos Bees Software.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */

package bees.core.security;

import java.security.NoSuchAlgorithmException;

import java.security.spec.InvalidKeySpecException;

import bees.core.kernel.KernelObjectCoroutine;

import bees.core.bdomain.BDomain;

/**
 * A SecretKeyFactory is used to create secret keys from SecretKeySpecs.
 */
public final class SecretKeyFactory
{
    /**
     * Cached reference to the current domain.
     */
    private final BDomain bdomain;
    
    /**
     * Construct an empty SecretKeyFactory.
     */
    public SecretKeyFactory()
    {
	this.bdomain = BDomain.currentBDomain();
    }

    /**
     * Create a SecretKey capability from the given values.
     *
     * @param ks The key specification.
     * @param comment A comment describing the key.
     * @return A new SecretKey capability.
     * @throws NoSuchAlgorithmException if the key algorithm is not available.
     * @throws InvalidKeySpecException if the key spec is not valid.
     */
    public synchronized SecretKey createKey(SecretKeySpec ks,
					    String comment)
	throws NoSuchAlgorithmException,
	       InvalidKeySpecException
    {
	SecretKey retval;
	
	retval = (SecretKey)KernelObjectCoroutine.
	    create(new SecretKey(),
		   this.bdomain.getBDomainCoroutine(),
		   ks.toBytes(),
		   comment);
	return retval;
    }

    /**
     * Create a SecretKey capability from the given values.
     *
     * @param ks The key specification.
     * @return A new SecretKey capability.
     * @throws NoSuchAlgorithmException if the key algorithm is not available.
     * @throws InvalidKeySpecException if the key spec is not valid.
     */
    public synchronized SecretKey createKey(SecretKeySpec ks)
	throws NoSuchAlgorithmException,
	       InvalidKeySpecException
    {
	return this.createKey(ks, "SecretKey");
    }
    
    /**
     * @see java.lang.Object#toString()
     */
    public String toString()
    {
	return "SecretKeyFactory["
	    + "]";
    }
}
