/*
 * Signature.java
 *
 * Copyright (c) 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is part of the Janos Bees Software.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */

package bees.core.security;

import java.security.SecureRandom;
import java.security.SignatureException;
import java.security.InvalidKeyException;
import java.security.NoSuchAlgorithmException;
import java.security.InvalidParameterException;
import java.security.InvalidAlgorithmParameterException;

import java.security.spec.AlgorithmParameterSpec;

import bees.core.Leak;
import bees.core.Panic;
import bees.core.ApplicationBDomain;
import bees.core.ProtocolSession;

/**
 * Signature is a thin veneer over the standard java.security.Signature class
 * that uses bees.core.security classes.
 * <p>
 * <i>Note:</i> In order to prevent downloaded code from producing signatures
 * for arbitrary data, the flow identifier is always added to the signature.
 */
public final class Signature
{
    /**
     * Permission to use this key for signing.
     */
    public static final Permission SIGN_PERMISSION =
	new Permission("sign", "execute");

    /**
     * Permission to use this key for verification.
     */
    public static final Permission VERIFY_PERMISSION =
	new Permission("verify", "execute");

    /**
     * The real Signature object.
     */
    private final java.security.Signature sig;

    /**
     * Construct a Signature object that uses the given algorithm type.
     *
     * @param algorithm The name of the algorithm to use when
     * signing/verifying.
     * @throws NoSuchAlgorithmException if the given algorithm is not
     * supported.
     */
    public Signature(String algorithm)
	throws NoSuchAlgorithmException
    {
	this.sig = java.security.Signature.getInstance(algorithm);
    }

    /**
     * Initialize the signature object for verification.
     *
     * @param publicKey The public key of the identity whose signature is going
     * to be verified.
     * @param ps A ProtocolSession from the flow this signature was bound to or
     * null if this domain is an application.
     * @throws SignatureException if there was a problem adding the flow
     * identifier to the Signature.
     * @throws InvalidKeyException if the given key is invalid.
     */
    public void initVerify(PublicKeySpec publicKey, ProtocolSession ps)
	throws InvalidKeyException,
	       SignatureException
    {
	if( publicKey instanceof PublicKey )
	{
	    publicKey.checkPermission(VERIFY_PERMISSION);
	}
	
	this.sig.initVerify(publicKey.pk);
	if( ps != null )
	{
	    this.sig.update(Leak.getFlowID(ps));
	}
    }

    /**
     * Initialize the signature object for verification.
     *
     * @param publicKey The public key of the identity whose signature is going
     * to be verified.
     * @throws InvalidKeyException if the given key is invalid.
     */
    public void initVerify(PublicKeySpec publicKey)
	throws InvalidKeyException
    {
	try
	{
	    this.initVerify(publicKey, null);
	}
	catch(SignatureException e)
	{
	    throw new Panic(e);
	}
    }

    /**
     * Initialize the signature object for signing.
     *
     * @param privateKey The private key of the identity whose signature is
     * going to be generated.
     * @throws SecurityException if this domain is not an application.
     * @throws InvalidKeyException if the given key is invalid.
     */
    public void initSign(PrivateKeySpec privateKey)
	throws SecurityException,
	       InvalidKeyException
    {
	ApplicationBDomain.currentApplication();
	if( privateKey instanceof PrivateKey )
	{
	    privateKey.checkPermission(SIGN_PERMISSION);
	}
	this.sig.initSign(privateKey.pk);
    }

    /**
     * Initialize the signature object for signing.
     *
     * @param privateKey The private key of the identity whose signature is
     * going to be generated.
     * @param random The source of randomness for this signature.
     * @throws SecurityException if this domain is not an application.
     * @throws InvalidKeyException if the given key is invalid.
     */
    public void initSign(PrivateKeySpec privateKey, SecureRandom random)
	throws SecurityException,
	       InvalidKeyException
    {
	ApplicationBDomain.currentApplication();
	if( privateKey instanceof PrivateKey )
	{
	    privateKey.checkPermission(SIGN_PERMISSION);
	}
	this.sig.initSign(privateKey.pk, random);
    }

    /**
     * Initialize the signature object for signing.
     *
     * @param privateKey The private key of the identity whose signature is
     * going to be generated.
     * @param ps A ProtocolSession from the flow this signature should be bound
     * to or null if this domain is an application.
     * @throws SecurityException if this domain is not an application.
     * @throws InvalidKeyException if the given key is invalid.
     * @throws SignatureException if there was a problem updating the signature
     * with the flow identifier.
     */
    public void initSign(PrivateKeySpec privateKey, ProtocolSession ps)
	throws SecurityException,
	       InvalidKeyException,
	       SignatureException
    {
	if( (ps == null) && (ApplicationBDomain.currentApplication() == null) )
	{
	    throw new IllegalArgumentException("Empty session argument");
	}
	if( privateKey instanceof PrivateKey )
	{
	    privateKey.checkPermission(SIGN_PERMISSION);
	}
	
	this.sig.initSign(privateKey.pk);
	if( ps != null )
	{
	    this.sig.update(Leak.getFlowID(ps));
	}
    }

    /**
     * Initialize the signature object for signing.
     *
     * @param privateKey The private key of the identity whose signature is
     * going to be generated.
     * @param ps A ProtocolSession from the flow this signature should be bound
     * to or null if this domain is an application.
     * @param random The source of randomness for this signature.
     * @throws SecurityException if this domain is not an application.
     * @throws InvalidKeyException if the given key is invalid.
     * @throws SignatureException if there was a problem updating the signature
     * with the flow identifier.
     */
    public void initSign(PrivateKeySpec privateKey,
			 SecureRandom random,
			 ProtocolSession ps)
	throws SecurityException,
	       InvalidKeyException,
	       SignatureException
    {
	if( (ps == null) && (ApplicationBDomain.currentApplication() == null) )
	{
	    throw new IllegalArgumentException("Empty session argument");
	}
	if( privateKey instanceof PrivateKey )
	{
	    privateKey.checkPermission(SIGN_PERMISSION);
	}
	
	this.sig.initSign(privateKey.pk, random);
	if( ps != null )
	{
	    this.sig.update(Leak.getFlowID(ps));
	}
    }

    public byte[] sign()
	throws SignatureException
    {
	return this.sig.sign();
    }
    
    public int sign(byte[] outbuf, int offset, int len)
	throws SignatureException
    {
	byte[] sigbits = this.sign();
	
	if( len < sigbits.length )
	{
	    throw new SignatureException();
	}
	System.arraycopy(sigbits, 0, outbuf, offset, sigbits.length);
	return sigbits.length;
    }

    public boolean verify(byte[] signature)
	throws SignatureException
    {
	return this.sig.verify(signature);
    }

    public void update(byte b)
	throws SignatureException
    {
	this.sig.update(b);
    }

    public void update(byte[] data)
	throws SignatureException
    {
	this.sig.update(data);
    }

    public void update(byte[] data, int off, int len)
	throws SignatureException
    {
	this.sig.update(data, off, len);
    }

    public String getAlgorithm()
    {
	return this.sig.getAlgorithm();
    }

    public void setParameter(String param, Object value)
	throws InvalidParameterException
    {
	this.sig.setParameter(param, value);
    }

    public void setParameter(AlgorithmParameterSpec params)
	throws InvalidAlgorithmParameterException
    {
	this.sig.setParameter(params);
    }

    public Object getParameter(String param)
	throws InvalidParameterException
    {
	return this.sig.getParameter(param);
    }
    
    public String toString()
    {
	return "Signature[sig="
	    + this.sig
	    + "]";
    }
}
