/*
 * Authenticator.java
 *
 * Copyright (c) 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is part of the Janos Bees Software.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */

package apps.authenticator;

import java.util.Enumeration;

import bees.core.Node;
import bees.core.TypeID;
import bees.core.ByteArray;
import bees.core.NodeCache;
import bees.core.ObjectMap;
import bees.core.Authorizer;
import bees.core.Capability;
import bees.core.Application;
import bees.core.EventChannel;
import bees.core.NeighborImpl;
import bees.core.ProtocolFlow;
import bees.core.CapabilityEnvelope;
import bees.core.CapabilityReference;
import bees.core.DLProtocolPresentation;
import bees.core.IDTranslatorPresentation;
import bees.core.MissingCapabilityException;
import bees.core.UnknownProtocolPresentation;
import bees.core.ExtendedProtocolPresentation;

import bees.core.resource.File;
import bees.core.resource.Meter;
import bees.core.resource.Logger;
import bees.core.resource.LoggerFactory;

import bees.core.security.PublicKey;
import bees.core.security.Permission;
import bees.core.security.PermissionSet;
import bees.core.security.CountPermission;
import bees.core.security.IntegerPermission;

import bees.core.bdomain.CapabilityTree;

import apps.discovery.DiscoveryProtocolPresentation;

import apps.liveness.LivenessPresentation;

/**
 * The Authenticator class provides a base for applications that wish to
 * authenticate protocol flows.  In addition, it will use the discovery and
 * liveness protocols to track neighbors on the local network.
 */
public abstract class Authenticator
    extends Application
{
    /**
     * Factory used to make Loggers for downloaded protocols.
     */
    private final LoggerFactory lf = new LoggerFactory();

    /**
     * Cached pointer to the current logger.
     */
    private final Logger currentLog = Logger.currentLogger();

    /**
     * Cached pointer to the current Node.
     */
    private Node node;

    /**
     * Our own private NodeCache.
     */
    private NodeCache nodeCache;

    /**
     * The parameters for this Authenticator.
     */
    private AuthenticatorParameters ap;

    /**
     * Object used to track Neighbors and notify protocols when they are found
     * or lost.
     */
    private NeighborManager nm;

    /**
     * A common set of capabilities for authenticated protocols.
     */
    private CapabilityTree common;

    /**
     * The set of resource meters.
     */
    private Meter meters[];

    /**
     * The set of initial capabilities passed to a bdomain.  Usually just a set
     * of resource meters.
     */
    private Capability initialCaps[];

    /**
     * A set of permissions that define what resources the bdomain should have.
     */
    private PermissionSet resourceSpecs;

    /**
     * The ProtocolPresentation used to communicate with the UnknownProtocol.
     */
    private UnknownProtocolPresentation upp;

    /**
     * The ProtocolPresentation used to communicate with the
     * IDTranslatorProtocol.
     */
    private IDTranslatorPresentation tpp;

    /**
     * The ProtocolPresentation used to communicate with the DiscoveryProtocol.
     */
    private AuthenticatorDiscoveryPresentation adp;

    /**
     * The ProtocolPresentation used to communicate with the LivenessProtocol.
     */
    private LivenessPresentation alp;

    /**
     * Construct an empty Authenticator object.
     */
    protected Authenticator()
    {
    }

    /**
     * @return Factory used to make Loggers for downloaded protocols.
     */
    public LoggerFactory getLoggerFactory()
    {
	return this.lf;
    }

    /**
     * @return Cached pointer to the current Node.
     */
    public Node getNode()
    {
	return this.node;
    }

    /**
     * @return The Authenticator's private NodeCache.
     */
    public NodeCache getNodeCache()
    {
	return this.nodeCache;
    }
    
    /**
     * @return The parameters for this Authenticator.
     */
    public AuthenticatorParameters getAuthenticatorParameters()
    {
	return this.ap;
    }

    /**
     * @return The object used to track Neighbors and notify protocols when
     * they are found or lost.
     */
    public NeighborManager getNeighborManager()
    {
	return this.nm;
    }

    /**
     * @return A common set of capabilities for authenticated protocols.
     */
    public CapabilityTree getCommonCapabilities()
    {
	return this.common;
    }

    /**
     * @return The set of initial capabilities for new BDomains.
     */
    public Capability[] getInitialCapabilities()
    {
	return this.initialCaps;
    }

    /**
     * @return The ProtocolPresentation used to communicate with the
     * UnknownProtocol.
     */
    public UnknownProtocolPresentation getUnknownProtocolPresentation()
    {
	return this.upp;
    }

    /**
     * @return The ProtocolPresentation used to communicate with the
     * IDTranslatorProtocol.
     */
    public IDTranslatorPresentation getIDTranslatorPresentation()
    {
	return this.tpp;
    }

    /**
     * @return The ProtocolPresentation used to communicate with the
     * DiscoveryProtocol.
     */
    public DiscoveryProtocolPresentation getDiscoveryPresentation()
    {
	return this.adp;
    }

    /**
     * @return The ProtocolPresentation used to communicate with the
     * LivenessProtocol.
     */
    public LivenessPresentation getLivenessPresentation()
    {
	return this.alp;
    }

    /**
     * @return An UnknownProtocolPresentation subclass that handles freshly
     * downloaded protocol flow descriptions.
     */
    protected abstract
	UnknownProtocolPresentation createUnknownProtocolPresentation();

    /**
     * A basic default IDTranslatorPresentation that immediately asks the
     * UnknownProtocolPresentation to download the flow description for the
     * mapped id.
     */
    private final class AuthenticatorIDTranslatorPresentation
	extends IDTranslatorPresentation
    {
        /**
         * @see bees.core.IDTranslatorPresentation#newFlow(CapabilityReference, TypeID)
         */
	public void newFlow(CapabilityReference neighbor, TypeID flowID)
	{
	    Logger.log("Downloading FlowID "
		       + flowID
		       + " from "
		       + neighbor);
	    Authenticator.this.upp.request(neighbor, flowID);

	    /*
	     * We should probably reject further requests for the same flow
	     * ID, but, multiple downloads of the flow tuple is no big deal.
	     */
	}

        /**
         * @see java.lang.Object#toString()
         */
	public String toString()
	{
	    return "AuthenticatorIDTranslatorPresentation["
		+ super.toString()
		+ "]";
	}
    }

    /**
     * @return An IDTranslatorPresentation subclass that handles mapped flow
     * identifiers.
     */
    protected IDTranslatorPresentation createIDTranslatorPresentation()
    {
	return new AuthenticatorIDTranslatorPresentation();
    }

    /**
     * @param fi The flow identifier being downloaded.
     * @return A DLProtocolPresentation subclass that handles bootstrapping
     * of new protocols.
     */
    protected DLProtocolPresentation createDLProtocolPresentation(TypeID fi)
    {
	return new AuthenticatorDLPresentation(this, fi);
    }

    /**
     * @param node The current Node object.
     * @return An AuthenticatorParameters that holds the interesting parameters
     * for this Authenticator.
     */
    protected AuthenticatorParameters createAuthenticatorParameters(Node node)
    {
	return new AuthenticatorParameters(node);
    }
    
    /**
     * @see bees.core.Application#start(Node, String[])
     */
    public void start(Node node, String args[])
	throws Throwable
    {
	PublicKey peerPublicKey = null;
	Logger currentLog;
	CapabilityTree ct;
	ProtocolFlow pf;

	this.node = node;
	this.nodeCache = new NodeCache(this.node, 15);
	this.nm = new NeighborManager();
	this.node.setAuthorizer(new Authorizer() {
		public void beginManagement(CapabilityEnvelope ce,
					    TypeID sessionID)
		{
		}
		
		public void beginExtendedManagement(
			final CapabilityEnvelope ce,
			final TypeID sessionID)
		{
		    final ExtendedProtocolPresentation epp;

		    epp = new ExtendedProtocolPresentation() {
			    public void userReceive()
			    {
			    }

			    public void disconnected()
			    {
				Authenticator.this.nm.
				    remInterestedProtocol(this);
			    }
			};
		    Authenticator.this.node.threadStart(new Runnable() {
			    public void run()
			    {
				try
				{
				    Logger.log("Managing: " + sessionID);
				    Authenticator.this.node.
					register(epp, ce, sessionID);
				    Authenticator.this.nm.
					addInterestedProtocol(epp);
				}
				catch(InstantiationException e)
				{
				    Logger.log(e);
				}
				catch(MissingCapabilityException e)
				{
				    Logger.log(e);
				}
			    }
			},
							"Managing thread.");
		}

		public void endManagement(CapabilityEnvelope ce,
					  final TypeID sessionID)
		{
		}
		
		public CapabilityTree authorize(CapabilityEnvelope req,
						String src,
						String dst,
						byte auth[])
		{
		    CapabilityTree retval = null;
		    
		    if( src.equals("neighbors") )
		    {
			retval = new CapabilityTree(dst,
						    Authenticator.
						    this.nm.namedNeighbors());
		    }
		    return( retval );
		}
	    });

	this.ap = this.createAuthenticatorParameters(node);

	/*
	 * Initialize the NeighborManager with our local material and
	 * access permissions.
	 */
	{
	    NeighborImpl bn[];
	    File file[];
	    int lpc;

	    bn = this.ap.getBroadcastNeighbor();
	    file = this.ap.getLocalMaterial();
	    for( lpc = 0; lpc < file.length; lpc++ )
	    {
		byte data[];
		
		data = ByteArray.fromFile(file[lpc]);
		this.nm.addLocalMaterial(bn[lpc],
					 AdvertiseMaterial.instantiate(data));
	    }
	}

	/*
	 * Initialize the set of bdomain initial capabilities.
	 */
	{
	    Enumeration enum;
	    ObjectMap om;
	    int lpc;

	    om = node.getCapabilityMap("meters");
	    enum = om.elements();
	    this.meters = new Meter[om.size()];
	    this.initialCaps = new Capability[this.meters.length];
	    for( lpc = 0; lpc < this.meters.length; lpc++ )
	    {
		this.meters[lpc] = (Meter)enum.nextElement();
		this.initialCaps[lpc] = this.meters[lpc];
	    }
	}

	/* Initialize the default resource specs. */
	this.resourceSpecs = new PermissionSet(new Permission[] {
	    new CountPermission("cpu.shares", 100),
	    new CountPermission("cpu.threads", 2),
	    new CountPermission("mem.pages", 1 * 1024 * 1024 / 4096),
	});

	/*
	 * Initialize the common set of capabilities that we want passed to
	 * every flow.
	 */
	this.common = new CapabilityTree(new CapabilityEnvelope[] {
	    new CapabilityEnvelope("node", node).
		revokePermissions(new PermissionSet(new Permission[] {
		    Node.THREAD_START,
		    Node.PROTOCOL_REGISTER,
		    Node.JSI_REGISTER,
		})),

	    new CapabilityEnvelope("user/clock",
				   node.getCapability("clock")),
	    new CapabilityEnvelope("dlversion",
				   node.getCapability("protocol/dlversion")),
	    new CapabilityEnvelope("unknownVersion", node.
				   getCapability("protocol/unknownVersion")),
	    new CapabilityEnvelope("idversion",
				   node.getCapability("protocol/idversion")),
	    
	    new CapabilityEnvelope("authenticator",
				   EventChannel.currentChannel()),
	}, new CapabilityTree[] {
	    new CapabilityTree(node.getCapabilityMap("protocol/classes")),
	});

	/* Create the capabilities for the UnknownProtocol flow. */
	ct = new CapabilityTree(new CapabilityEnvelope[] {
	    new CapabilityEnvelope("logger",
				   this.lf.create(this.currentLog,
						  "UnknownProtocol")),
	}, new CapabilityTree[] {
	    this.common
	});

	/* Register the UnknownProtocol flow. */
	this.upp = this.createUnknownProtocolPresentation();
	pf = new ProtocolFlow();
	pf.setPrimaryProtocol("bees.core.UnknownProtocol");
	node.register(this.upp, pf, this.initialCaps, this.resourceSpecs, ct);
	this.upp.register();
	this.nm.addInterestedProtocol(this.upp);
	
	/* Create the capabilities for the IDTranslator flow. */
	ct = new CapabilityTree(new CapabilityEnvelope[] {
	    new CapabilityEnvelope("logger",
				   this.lf.create(this.currentLog,
						  "IDTranslator")),
	}, new CapabilityTree[] {
	    this.common
	});
	
	/* Register the IDTranslation flow. */
	this.tpp = this.createIDTranslatorPresentation();
	pf = new ProtocolFlow();
	pf.setPrimaryProtocol("bees.core.IDTranslationProtocol");
	node.register(this.tpp, pf, this.initialCaps, this.resourceSpecs, ct);
	this.tpp.register();
	this.nm.addInterestedProtocol(this.tpp);

	/* Connect the Liveness and Discovery presentations. */
	this.adp = new AuthenticatorDiscoveryPresentation(this.ap, this.nm);
	this.alp = new AuthenticatorLivenessPresentation(this.nm, this.adp);

	/* Start the LivenessProtocol to monitor the neighbors. */
	pf = new ProtocolFlow();
	pf.setPrimaryProtocol("apps.liveness.LivenessProtocol");
	ct = new CapabilityTree(new CapabilityEnvelope[] {
	    new CapabilityEnvelope("logger",
				   node.getCapability("liveness/logger")),
	    new CapabilityEnvelope("user/liveVersion",
				   node.getCapability("liveVersion")),
	}, new CapabilityTree[] {
	    this.common,
	    new CapabilityTree(node.getCapabilityMap("liveness/classes")),
	});
	node.register(this.alp, pf, this.initialCaps, this.resourceSpecs, ct);
	this.alp.register();
	this.nm.addInterestedProtocol(this.alp);

	/* Start the DiscoveryProtocol to discover neighbors. */
	pf = new ProtocolFlow();
	pf.setPrimaryProtocol("apps.discovery.DiscoveryProtocol");
	ct = new CapabilityTree(new CapabilityEnvelope[] {
	    new CapabilityEnvelope("logger",
				   node.getCapability("discovery/logger")),
	    new CapabilityEnvelope("user/discoveryVersion",
				   node.getCapability("discoveryVersion")),
	    peerPublicKey == null ? null
	    : new CapabilityEnvelope("user/peerPublicKey", peerPublicKey),
	}, new CapabilityTree[] {
	    this.common,
	    new CapabilityTree(node.getCapabilityMap("discovery/classes")),
	});
	node.register(this.adp, pf, this.initialCaps, this.resourceSpecs, ct);
	this.adp.register();

	/* Start the advertisements */
	{
	    NeighborImpl bn[];
	    int lpc;

	    bn = this.ap.getBroadcastNeighbor();
	    for( lpc = 0; lpc < bn.length; lpc++ )
	    {
		/* Add the neighbor, the presentation will handle the rest. */
		this.adp.addNeighbor(bn[lpc]);
	    }
	}
    }

    /**
     * @see java.lang.Object#toString()
     */
    public String toString()
    {
	return "Everything["
	    + super.toString()
	    + "; lf="
	    + this.lf
	    + "; currentLog="
	    + this.currentLog
	    + "; ap="
	    + this.ap
	    + "; nm="
	    + this.nm
	    + "; common="
	    + this.common
	    + "; adp="
	    + this.adp
	    + "; alp="
	    + this.alp
	    + "]";
    }
}
