/*
 * AuthenticatorLivenessPresentation.java
 *
 * Copyright (c) 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is part of the Janos Bees Software.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */

package apps.authenticator;

import bees.core.Assert;
import bees.core.NeighborImpl;
import bees.core.CapabilityReference;
import bees.core.MissingCapabilityException;

import bees.core.resource.Logger;

import apps.discovery.DiscoveryProtocolPresentation;

import apps.liveness.LivenessPresentation;

/**
 * The AuthenticatorLivenessPresentation is a LivenessPresentation that\
 * propagates the lost neighbor callback to the Authenticator's
 * NeighborManager and DiscoveryProtocolPresentations died method.
 */
final class AuthenticatorLivenessPresentation
    extends LivenessPresentation
{
    /**
     * The NeighborManager to inform when a neighbor is lost.
     */
    private final NeighborManager nm;

    /**
     * The DiscoveryProtocolPresentation to inform when a neighbor is lost and
     * any cached data should be cancelled.
     */
    private final DiscoveryProtocolPresentation adp;

    /**
     * Construct an AuthenticatorLivenessPresentation with the given values.
     *
     * @param nm The NeighborManager to inform when a neighbor is lost.
     * @param adp The DiscoveryProtocolPresentation to inform when a neighbor
     * is lost and any cached data should be cancelled.
     */
    AuthenticatorLivenessPresentation(NeighborManager nm,
				      DiscoveryProtocolPresentation adp)
    {
	Assert.require(nm);
	Assert.require(adp);
	
	this.nm = nm;
	this.adp = adp;
    }

    /**
     * Executes the super's method and then immediately starts monitoring the
     * Neighbor.
     * 
     * @see bees.core.ExtendedProtocolPresentation#addNeighbor(NeighborImpl)
     */
    public synchronized void addNeighbor(NeighborImpl ni)
    {
	super.addNeighbor(ni);
	/* Immediately start monitoring. */
	super.monitor(ni);
    }

    /**
     * Removes the lost neighbor from the NeighborManager and tells the 
     * Discovery protocol to flush any advertising material from the Neighbor.
     * 
     * @see apps.liveness.LivenessPresentation#lost(CapabilityReference)
     */
    public synchronized void lost(CapabilityReference cr)
    {
	Assert.require(cr);

	try
	{
	    NeighborImpl ni;
	    
	    ni = (NeighborImpl)cr.dereference();
	    this.adp.died(this.nm.remNeighbor(ni));
	    Logger.log(Logger.INFO, "Lost neighbor: " + ni);
	}
	catch(MissingCapabilityException e)
	{
	    Logger.log(e);
	}
    }

    /**
     * @see java.lang.Object#toString()
     */
    public String toString()
    {
	return "AuthenticatorLivenessPresentation["
	    + super.toString()
	    + "; nm="
	    + this.nm
	    + "; adp="
	    + this.adp
	    + "]";
    }
}
