/*
 * AdvertiseCapsule.java
 *
 * Copyright (c) 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is part of the Janos Bees Software.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */

package apps.discovery;

import java.util.Hashtable;

import java.io.IOException;

import java.security.SignatureException;
import java.security.InvalidKeyException;
import java.security.NoSuchAlgorithmException;

import bees.core.Xdr;
import bees.core.Node;
import bees.core.Panic;
import bees.core.Assert;
import bees.core.TypeID;
import bees.core.Capsule;
import bees.core.ByteArray;
import bees.core.NodeCache;
import bees.core.CapsuleGroup;
import bees.core.ProtocolSession;
import bees.core.EncodingException;

import bees.core.resource.Logger;

import bees.core.security.Signature;
import bees.core.security.IntegerPermission;

import bees.util.MD5;

/**
 * An AdvertiseCapsule is used to carry broadcast data for a host.
 */
public final class AdvertiseCapsule
    extends Capsule
    implements NodeCache.EventHandler
{
    /**
     * Neo, this is the neighbor used to broadcast our pirate message, er,
     * advertisement.
     */
    private BroadcastNeighbor bn;
    
    /**
     * The material to broadcast to other nodes.
     */
    private byte material[];

    /**
     * Null or the signature covering the material.
     */
    private byte signature[];

    /**
     * The retransmit frequency expressed as a number of milliseconds.
     */
    private final IntegerPermission frequency;
    
    /**
     * Construct an empty AdvertiseCapsule.
     */
    public AdvertiseCapsule()
    {
	this.frequency = null;
    }

    /**
     * Constructor for telling other neighbors about this node.
     *
     * @param bn The neighor to use when broadcasting this advertisement.
     * @param material The material to advertise.
     * @param signature Null or the signature covering this material.
     * @param frequency The retransmit frequency expressed as a number of
     * milliseconds.
     */
    public AdvertiseCapsule(BroadcastNeighbor bn,
			    byte material[],
			    byte signature[],
			    IntegerPermission frequency)
    {
	Assert.require(material);
	Assert.require(frequency);

	this.bn = bn;
	/* Store our material */
	this.material = material;
	if( (signature != null) && (signature.length > 0) )
	{
	    this.signature = signature;
	}
	this.frequency = frequency;
    }

    /**
     * @return An MD5 hash covering the material.
     */
    public TypeID computeFingerPrint()
    {
	return new TypeID(MD5.hash(this.material));
    }

    /**
     * @param material The material to advertise.
     */
    public void setMaterial(byte material[])
    {
	this.material = material;
    }

    /**
     * @return The material to advertise.
     */
    public byte[] getMaterial()
    {
	return this.material;
    }

    /**
     * @param signature Null or the signature covering this material.
     */
    public void setSignature(byte signature[])
    {
	if( (signature != null) && (signature.length > 0) )
	{
	    this.signature = signature;
	}
	else
	{
	    this.signature = null;
	}
    }

    /**
     * @return Null or the signature covering this material.
     */
    public byte[] getSignature()
    {
	return this.signature;
    }

    /**
     * @return The retransmit frequency expressed as a number of milliseconds.
     */
    public IntegerPermission getFrequency()
    {
	return this.frequency;
    }

    /**
     * @see bees.core.NodeCache.EventHandler#removed(Node, ProtocolSession)
     */
    public void removed(Node node, ProtocolSession ps)
    {
    }

    /**
     * @see bees.core.NodeCache.EventHandler#collected(Node, ProtocolSession)
     */
    public IntegerPermission collected(Node node, ProtocolSession ps)
    {
	IntegerPermission retval = this.frequency;

	/* Timer event from NodeCache. */
	try
	{
	    DiscoveryProtocolSession dps;
	    
	    Logger.log(Logger.NOTICE, "Sending advertisement");
	    this.bn.send(this);
	}
	catch(EncodingException e)
	{
	    throw new Panic("Unable to encode AdvertiseCapsule?", e);
	}
	catch(IOException e)
	{
	    Logger.log(e);
	    retval = null;
	}
	catch(Throwable th)
	{
	    Logger.log(th);
	    retval = null;
	}
	return this.frequency;
    }
    
    /**
     * @see bees.core.Capsule#decode(Xdr)
     */
    public void decode(Xdr xdr)
    {
	Assert.require(xdr);
	
	this.material = xdr.BYTES(xdr.INT());
	this.signature = xdr.BYTES(xdr.INT());
    }

    /**
     * @see bees.core.Capsule#encode(Xdr)
     */
    public void encode(Xdr xdr)
    {
	int lpc;
	
	Assert.require(xdr);
	
	xdr.PUT(this.material.length);
	xdr.PUT(this.material);
	if( this.signature != null )
	{
	    xdr.PUT(this.signature.length);
	    xdr.PUT(this.signature);
	}
	else
	{
	    xdr.PUT(0);
	}
    }

    /**
     * @see bees.core.Capsule#evaluate(Node)
     */
    public void evaluate(Node node)
	throws NoSuchAlgorithmException,
	       InvalidKeyException,
	       SignatureException
    {
	DiscoveryProtocolSession dps;
	DiscoveryParameters dp;
	TypeID fp;

	dps = (DiscoveryProtocolSession)this.getSession();
	dp = dps.getDiscoveryParameters();
	this.bn = (BroadcastNeighbor)this.getSource();
	/* The fingerprint of the neighbor's material */
	fp = this.computeFingerPrint();
	if( dps.getLocalMaterial().get(fp) == null )
	{
	    boolean valid = true; // XXX User specified?
	    
	    if( dp.getPublicKey() != null )
	    {
		/* We're checking signatures. */
		
		if( this.signature != null )
		{
		    Signature verifier;
		    
		    /* Check the signature */
		    verifier = new Signature(dp.getAlgorithm());
		    verifier.initVerify(dp.getPublicKey());
		    verifier.update(this.material);
		    valid = verifier.verify(this.signature);
		}
		else
		{
		    valid = false;
		    Logger.log(Logger.ERROR,
			       "Received advertisement with no signature: "
			       + ByteArray.toString(this.material, 32));
		}
	    }
	    if( valid )
	    {
		NodeCache nc;

		/* Valid signature. */
		nc = dps.getCache();
		synchronized( nc )
		{
		    if( nc.refresh(fp, dp.getExpirationTime()) )
		    {
			/* Recently added/refreshed. */
			Logger.log(Logger.NOTICE,
				   "Received recently seen advertisement: fp="
				   + fp);
		    }
		    else
		    {
			Hashtable material;
			CapsuleGroup cg;

			Logger.log(Logger.INFO,
				   "Received new advertisement: fp="
				   + fp);
			material = dps.getDiscoveredMaterial();
			/* Store the mapping */
			material.put(fp, this.material);
			/* Tell any interested applications */
			dps.foundMaterial(fp, this);
			/*
			 * We haven't seen this advert before, so the neighbor
			 * probably hasn't seen any of ours.  Resend all the
			 * advertisements right away instead of waiting for the
			 * timer.
			 */
			cg = this.bn.getAdvertisements();
			cg.fromCapsule(this);
			try
			{
			    Logger.log(Logger.INFO,
				       "Resending local advertisements");
			    this.getSource().send(cg);
			}
			catch(EncodingException e)
			{
			    Logger.log(e);
			}
			catch(IOException e)
			{
			    Logger.log(e);
			}
			nc.put(fp,
			       new AdvertiseMaterial(fp),
			       dp.getExpirationTime());
		    }
		}
	    }
	    else
	    {
		Logger.log(Logger.WARNING,
			   "Invalid signature for material: "
			   + ByteArray.toString(this.material, 32)
			   + ByteArray.toString(this.signature, 32));
	    }
	}
	else
	{
	    /* Our advertisement loop back... */
	    Logger.log(Logger.NOTICE, "Received own advertisement.");
	}
	this.addToPool();
    }

    /**
     * @see java.lang.Object#toString()
     */
    public String toString()
    {
	return "AdvertiseCapsule["
	    + super.toString()
	    + "; material="
	    + ByteArray.toString(this.material, 32)
	    + "; signature="
	    + ByteArray.toString(this.signature)
	    + "; frequency="
	    + this.frequency
	    + "]";
    }
}
