/*
 * DiscoveryProtocol.java
 *
 * Copyright (c) 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is part of the Janos Bees Software.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */

package apps.discovery;

import java.io.FileNotFoundException;

import java.util.NoSuchElementException;

import bees.core.Node;
import bees.core.Panic;
import bees.core.Protocol;

import bees.core.resource.Logger;

import bees.core.security.Permission;
import bees.core.security.PermissionSet;
import bees.core.security.VersionString;
import bees.core.security.CountPermission;
import bees.core.security.MinimumIntegerPermission;

/**
 * The DiscoveryProtocol can be used to send and receive advertisements from
 * directly connected neighbors that haven't been told of each other's
 * existance.
 */
public final class DiscoveryProtocol
    extends Protocol
{
    /**
     * Construct a description of the DiscoveryProtocol.
     */
    public DiscoveryProtocol()
    {
	try
	{
	    Logger.log("Defining discovery protocol.");

	    this.startProtocolDefn();
	    {
		VersionString vs = null;
		Node node;

		node = super.getNode();
		try
		{
		    /* Get the version string to use with the capsules. */
		    vs = (VersionString)
			node.getCapability("discoveryVersion");
		}
		catch(NoSuchElementException e)
		{
		    Logger.log(Logger.WARNING,
			       "No \"discoveryVersion\" capability.  Using "
			       + "byte code to compute capsule identifiers.");
		}
		/* Add our properties file. */
		this.addFile("apps/discovery/protocol.properties");
		/* The main capsule for telling neighbors about this node. */
		this.addCapsule("apps.discovery.AdvertiseCapsule",
				vs,
				new PermissionSet(new Permission[] {
				    /*
				     * Create an advertisement from the
				     * original application request.
				     */
				    new CountPermission("initial", 1),
				    /*
				     * We create an external capsule for
				     * communicating with any interested
				     * applications.
				     */
				    new CountPermission("externals", 1),
				    /*
				     * We send the advertisements back through
				     * the source when we receive an unseen
				     * advert.
				     */
				    new CountPermission("returns", 1),
				    /*
				     * Can be used to prime CapsuleGroup.
				     */
				    new CountPermission("groups", 1),

				    /*
				     * We group the advertisements bound to the
				     * neighbors.
				     */
				    new CountPermission("grouped", 20),
				    
				    /*
				     * We put the advertisements in the cache
				     * to repeat broadcasts.
				     */
				    new CountPermission("cached", 20),
				    
				    new CountPermission("threadCount", 1),
				    new CountPermission("bufferCount", 2),

				    /*
				     * Minimum number of milliseconds between
				     * timer based replenishes.
				     */
				    new MinimumIntegerPermission("interval",
								 500),
				}));
		/* The boot capsule, sets everything in motion. */
		this.setBootStrap("apps.discovery.DiscoveryBootCapsule",
				  new PermissionSet(new Permission[] {
				      new CountPermission("threadCount", 0),
				      new CountPermission("bufferCount", 0),
				  }));
		/* An event handler for cleaning up stale material. */
		this.addCode("apps.discovery.AdvertiseMaterial");
		/* Our neighbor subclass. */
		this.addCode("apps.discovery.BroadcastNeighbor");
		/* The helper class that holds the parameters to the flow. */
		this.addCode("apps.discovery.DiscoveryParameters");
		/*
		 * The 'ProtocolSession', used to manage communications with
		 * applications.
		 */
		this.setSession("apps.discovery.DiscoveryProtocolSession");
	    }
	    this.endProtocolDefn();
	}
	catch(FileNotFoundException e)
	{
	    Logger.log(e);
	    throw new Panic("Unable to access file", e);
	}
	catch(ClassNotFoundException e)
	{
	    Logger.log(e);
	    throw new Panic("Unable to access classes", e);
	}
    }

    /**
     * @see java.lang.Object#toString()
     */
    public String toString()
    {
	return "DiscoveryProtocol["
	    + super.toString()
	    + "]";
    }
}
