/*
 * PathfinderSession.java
 *
 * Copyright (c) 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is part of the Janos Bees Software.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */

package apps.pathfinder;

import bees.core.Node;
import bees.core.Panic;
import bees.core.Capsule;
import bees.core.Neighbor;
import bees.core.NodeCache;
import bees.core.NeighborImpl;
import bees.core.ProtocolSession;
import bees.core.IDTranslatorSession;
import bees.core.MulticastNeighborhood;
import bees.core.ExtendedProtocolSession;

import bees.core.resource.Logger;

import bees.core.security.IntegerPermission;

/**
 * PathfinderSession tracks any state global to the pathfinder protocol.
 * Protocols using pathfinder as a companion will want to use the configure
 * and configureRoot methods.
 */
public class PathfinderSession
    extends ExtendedProtocolSession
    implements NodeCache.EventHandler
{
    /**
     * Major version of the protocol.
     */
    static final int VERSION_MAJOR = 0;
    
    /**
     * Minor version of the protocol.
     */
    static final int VERSION_MINOR = 5;

    /**
     * The set of neighbors that are assumed to be peers, this includes the
     * root neighbor.
     */
    private final MulticastNeighborhood peers = new MulticastNeighborhood();

    /**
     * The set of neighbors that are assumed to be children, this excludes the
     * root neighbor and any siblings.
     */
    private final MulticastNeighborhood children = new MulticastNeighborhood();
    
    /**
     * The ConfigurationCapsule that will be transmitted to the peers.
     */
    private final ConfigurationCapsule cc = new ConfigurationCapsule();

    /**
     * Cached reference to our ID translator.
     */
    private IDTranslatorSession idts;
    
    /**
     * The parameters for this flow.
     */
    private PathfinderParameters pp;

    /**
     * The discovered root neighbor.
     */
    private Neighbor rootNeighbor;

    /**
     * The smallest number of hops to the root we've found so far.
     */
    private int hops = Integer.MAX_VALUE;

    /**
     * Indicates whether or not we are the server node.
     */
    private boolean root = false;

    /**
     * Construct an empty PathfinderSession.
     */
    public PathfinderSession()
    {
    }

    /**
     * @param node The current Node object.
     * @param rootNeighbor The next neighbor on the path to the root node.
     */
    synchronized void setRootNeighbor(Node node, Neighbor rootNeighbor)
    {
	if( this.rootNeighbor != null )
	{
	    this.idts.attachNeighbor(node, this.rootNeighbor);
	    this.children.addNeighbor(this.rootNeighbor.getNeighborImpl());
	}
	this.rootNeighbor = rootNeighbor;
	this.cc.setRootNeighbor(rootNeighbor);
	if( rootNeighbor != null )
	{
	    NeighborImpl ni;

	    ni = rootNeighbor.getNeighborImpl();
	    this.children.removeNeighbor(ni);
	    this.idts.detachNeighbor(node, ni.getNeighbor(this.idts));
	}
    }

    /**
     * @return The next neighbor on the path to the root node.
     */
    public synchronized Neighbor getRootNeighbor()
    {
	return this.rootNeighbor;
    }

    /**
     * @return The set of neighbors that are assumed to be peers.
     */
    MulticastNeighborhood getPeers()
    {
	return this.peers;
    }

    /**
     * @return The set of neighbors that are assumed to be children.
     */
    MulticastNeighborhood getChildren()
    {
	return this.children;
    }

    /**
     * @return The ConfigurationCapsule that will be transmitted to the
     * neighborhood.
     */
    ConfigurationCapsule getConfigurationCapsule()
    {
	return this.cc;
    }

    /**
     * @return The parameters for this flow.
     */
    PathfinderParameters getPathfinderParameters()
    {
	return this.pp;
    }

    /**
     * @param pp The parameters for this flow.
     */
    void setPathfinderParameters(PathfinderParameters pp)
    {
	this.pp = pp;
    }

    /**
     * @return True if this node is the root, false otherwise.
     */
    public boolean isRoot()
    {
	return this.root;
    }

    /**
     * Configure the session object.  This entails creating the parameters
     * object and adding the ConfigurationCapsule to the NodeCache.
     *
     * @param node The current Node object.
     * @param cap A valid Capsule with a
     * "transformTo.apps.pathfinder.ConfigurationCapsule" permisson that can be
     * used as a specializer value
     */
    public void configure(Node node, Capsule cap)
    {
	if( this.pp == null )
	{
	    this.pp = new PathfinderParameters(this);
	    this.pp.initialize(node);
	    this.cc.specializeCapsule(cap);
	    try
	    {
		this.idts = (IDTranslatorSession)
		    this.cc.getSession("bees.core.IDTranslatorSession");
	    }
	    catch(ClassNotFoundException e)
	    {
		throw new Panic(e);
	    }
	    this.getCache().put(this.cc,
				this.cc,
				this.pp.getUpdateFrequency());
	}
    }

    /**
     * Configure the session object so that this is the root node.
     *
     * @param node The current Node object.
     * @param cap A valid Capsule with a
     * "transformTo.apps.pathfinder.ConfigurationCapsule" permisson that can be
     * used as a specializer value.
     */
    public void configureRoot(Node node, Capsule cap)
    {
	if( !this.root )
	{
	    this.root = true;
	    this.hops = 0;
	    this.cc.setHops(1);
	    this.configure(node, cap);
	}
    }

    /**
     * @return The smallest number of hops to the root we've found so far.
     */
    public int getHops()
    {
	return this.hops;
    }

    /**
     * @param hops The smallest number of hops to the root we've found so far.
     */
    public void setHops(int hops)
    {
	this.hops = hops;
	this.cc.setHops(hops + 1);
    }

    /**
     * @see bees.core.ProtocolSession#attachNeighbor(Node, Neighbor)
     */
    public synchronized void attachNeighbor(Node node, Neighbor neighbor)
    {
	NeighborImpl ni;

	ni = neighbor.getNeighborImpl();
	super.attachNeighbor(node, neighbor);
	this.peers.addNeighbor(ni);
	this.children.addNeighbor(ni);
    }

    /**
     * @see bees.core.ProtocolSession#detachNeighbor(Node, Neighbor)
     */
    public synchronized void detachNeighbor(Node node, Neighbor neighbor)
    {
	NeighborImpl ni;

	ni = neighbor.getNeighborImpl();
	if( (this.rootNeighbor != null) &&
	    (this.rootNeighbor.getNeighborImpl() == ni) )
	{
	    /* Lost the root neighbor... */
	    this.setRootNeighbor(node, null);
	    this.setHops(Integer.MAX_VALUE);
	}
	this.children.removeNeighbor(ni);
	this.peers.removeNeighbor(ni);
	super.detachNeighbor(node, neighbor);
    }

    /**
     * @see bees.core.NodeCache.EventHandler#removed(Node, ProtocolSession)
     */
    public void removed(Node node, ProtocolSession ps)
    {
    }

    /**
     * @see bees.core.NodeCache.EventHandler#collected(Node, ProtocolSession)
     */
    public synchronized IntegerPermission collected(Node node,
						    ProtocolSession ps)
    {
	this.setRootNeighbor(node, null);
	this.setHops(Integer.MAX_VALUE);
	return this.pp.getRootTimeout();
    }

    /**
     * @see java.lang.Object#toString()
     */
    public String toString()
    {
	return "PathfinderSession["
	    + super.toString()
	    + "; peers="
	    + this.peers
	    + "; children="
	    + this.children
	    + "; cc="
	    + this.cc
	    + "; idts="
	    + this.idts
	    + "; pp="
	    + this.pp
	    + "; rootNeighbor="
	    + this.rootNeighbor
	    + "; hops="
	    + this.hops
	    + "; root="
	    + this.root
	    + "]";
    }
}
