/*
 * ANEPPacket.java
 *
 * Copyright (c) 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is part of the Janos Bees Software.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */

package bees.core;

import edu.utah.janos.nodeos.Address;

import bees.core.anep.ANEPVersion;

/**
 * Class for handling/storing the relevant bits of an ANEP header in a packet.
 */
class ANEPPacket
    extends Packet
{
    /**
     * The offset of the marshalled version field.
     */
    static final int OFFSET_VERSION = 0;
    
    /**
     * The offset of the marshalled flags field.
     */
    static final int OFFSET_FLAGS = 1;
    
    /**
     * The offset of the marshalled type identifier field.
     */
    static final int OFFSET_TYPEID = 2;
    
    /**
     * The offset of the marshalled header length field.
     */
    static final int OFFSET_HEADER_LENGTH = 4;
    
    /**
     * The offset of the marshalled packet length field.
     */
    static final int OFFSET_PACKET_LENGTH = 6;
    
    /**
     * The multiplier applied to the header length field when
     * encoding/decoding.
     */
    private static final int HEADER_LENGTH_MULTIPLIER = Xdr.INT;
    
    /**
     * The multiplier applied to the option length field when
     * encoding/decoding.
     */
    private static final int OPTION_LENGTH_MULTIPLIER = Xdr.INT;

    /**
     * The byte length of an ANEP header with no options.
     */
    static final int NO_OPTION_HEADER_LENGTH =
	(Xdr.BYTE	/* Version */
	 + Xdr.BYTE	/* Flags */
	 + Xdr.SHORT	/* Type ID */
	 + Xdr.SHORT	/* ANEP Header Length */
	 + Xdr.SHORT);	/* ANEP Packet Length */

    /**
     * The byte length of an empty option.
     */
    static final int EMPTY_OPTION_LENGTH =
	(Xdr.SHORT	/* Flags + Type */
	 + Xdr.SHORT);	/* Length */

    /**
     * The byte length of an IPV4 address option.
     */
    static final int IPV4_OPTION_LENGTH =
	(EMPTY_OPTION_LENGTH
	 + Xdr.INT	/* Scheme */
	 + Xdr.INT);	/* Address */

    /**
     * Flag that indicates whether or not the node should discard the packet
     * if it doesn't recognize the type identifier.
     */
    static final byte FLAG_DISCARD = (byte)0x80;

    /**
     * The bitmask that covers the flag bits of an option type.
     */
    static final short OPTION_FLAG_MASK = (short)0xc000;
    
    /**
     * Option flag that indicates whether the option type is relative to the
     * type ID or not.
     */
    static final short OPTION_FLAG_INTERNAL = (short)0x8000;
    
    /**
     * Option flag that indicates whether the packet should be discarded if the
     * option type isn't known.
     */
    static final short OPTION_FLAG_DISCARD = 0x4000;
    
    /**
     * Option type identifier for the source address.
     */
    static final short OPTION_TYPE_SOURCE = 1;
    
    /**
     * Option type identifier for the destination address.
     */
    static final short OPTION_TYPE_DESTINATION = 2;

    /**
     * Option type identifier for the integrity checksum.
     */
    static final short OPTION_TYPE_CHECKSUM = 3;

    /**
     * Option type identifier for the N/N authentication. (XXX eh?)
     */
    static final short OPTION_TYPE_AUTHENTICATION = 4;

    /**
     * Use the IPv4 addressing scheme for source/destination options.
     */
    private static final int IPV4SCHEME = 1;

    /**
     * The version of the ANEP header.
     */
    private byte version = ANEPVersion.PACKET_VERSION;

    /**
     * Holds the various FLAG_* bits.
     */
    private byte flags;

    /**
     * The ANEP type identifier.
     */
    private short type;

    /**
     * The byte length of the header.
     */
    private int headerLength = NO_OPTION_HEADER_LENGTH;

    /**
     * The byte length of the body.
     */
    private short packetLength;

    /**
     * Cached value of the IPv4 destination option.
     */
    private int ipv4Destination;
    
    /**
     * Cached value of the IPv4 source option.
     */
    private int ipv4Source;

    /**
     * Construct an ANEPPacket with the default values.
     */
    protected ANEPPacket()
    {
    }

    /**
     * @param version The version of the ANEP header.
     * @return this
     */
    ANEPPacket setVersion(byte version)
    {
	this.version = version;
	return this;
    }

    /**
     * @return The version of the ANEP header.
     */
    byte getVersion()
    {
	return this.version;
    }

    /**
     * @param flags The various FLAG_* bits.
     * @return this
     */
    ANEPPacket setANEPFlags(byte flags)
    {
	this.flags = flags;
	return this;
    }

    /**
     * @return The various FLAG_* bits.
     */
    byte getANEPFlags()
    {
	return this.flags;
    }

    /**
     * @param type The ANEP type identifier.
     * @return this
     */
    ANEPPacket setType(short type)
    {
	this.type = type;
	return this;
    }

    /**
     * @return The ANEP type identifier.
     */
    short getType()
    {
	return this.type;
    }

    /**
     * @param len The byte length of the header.
     * @return this
     */
    ANEPPacket setHeaderLength(int len)
    {
	Assert.require(len >= NO_OPTION_HEADER_LENGTH);
	Assert.require((len % HEADER_LENGTH_MULTIPLIER) == 0);
	
	this.headerLength = len;
	return this;
    }

    /**
     * @return The byte length of the header.
     */
    int getHeaderLength()
    {
	Assert.ensure(this.headerLength >= NO_OPTION_HEADER_LENGTH);
	Assert.ensure((this.headerLength % HEADER_LENGTH_MULTIPLIER) == 0);
	
	return this.headerLength;
    }

    /**
     * @param len The byte length of the packet.
     * @return this
     */
    ANEPPacket setPacketLength(short len)
    {
	Assert.require(len >= 0);
	
	this.packetLength = len;
	return this;
    }

    /**
     * @return The byte length of the packet.
     */
    short getPacketLength()
    {
	Assert.require(this.packetLength >= 0);
	
	return this.packetLength;
    }

    /**
     * @param dst Cached value of the IPv4 destination option.
     * @return this
     */
    ANEPPacket setIPv4Destination(int dst)
    {
	this.ipv4Destination = dst;
	return this;
    }

    /**
     * @return Cached value of the IPv4 destination option.
     */
    int getIPv4Destination()
    {
	return this.ipv4Destination;
    }

    /**
     * @param src Cached value of the IPv4 source option.
     * @return this
     */
    ANEPPacket setIPv4Source(int src)
    {
	this.ipv4Source = src;
	return this;
    }

    /**
     * @return Cached value of the IPv4 source option.
     */
    int getIPv4Source()
    {
	return this.ipv4Source;
    }

    /**
     * @see bees.core.Packet#internalReset()
     */
    protected void internalReset()
    {
	super.internalReset();

	/* Clear optional data */
	this.ipv4Destination = 0;
	this.ipv4Source = 0;
    }

    /**
     * @see bees.core.Packet#internalDecode(Xdr)
     */
    protected void internalDecode(Xdr xdr)
	throws InvalidPacketException, DroppedPacketException
    {
	Assert.require(xdr);
	
	this.version = xdr.BYTE();
	this.flags = xdr.BYTE();
	this.type = xdr.SHORT();
	this.headerLength = xdr.SHORT() * HEADER_LENGTH_MULTIPLIER;
	if( this.headerLength < NO_OPTION_HEADER_LENGTH )
	{
	    throw new InvalidPacketException(
		"ANEP header length is too small");
	}
	this.packetLength = xdr.SHORT();
	if( this.packetLength < this.headerLength )
	{
	    throw new InvalidPacketException(
		"ANEP packet length is too small");
	}
    }

    /**
     * Skip the options in the ANEP header leaving the Xdr pointing to the
     * start of the payload.
     *
     * @param xdr The Xdr containing an encoded ANEP header.
     */
    protected void skipOptions(Xdr xdr)
    {
	Assert.require(xdr);
	
	xdr.SKIP(this.headerLength - NO_OPTION_HEADER_LENGTH);
    }

    /**
     * Decode the options in the ANEP header.
     * <p>
     * XXX This is woefully incomplete, we only pick up the IPv4 source and
     * destination options.  So its not going to forward any options either...
     *
     * @param xdr The Xdr containing an encoded ANEP header.
     * @throws InvalidPacketException if there is a problem in
     * decoding the options.
     * @throws DroppedPacketException if the packet should be
     * dropped because we don't understand an option.
     */
    void decodeOptions(Xdr xdr)
	throws InvalidPacketException,
	       DroppedPacketException
    {
	int start, curr, end;
	
	Assert.require(xdr);
	
	start = xdr.getReadCursor();
	curr = start;
	end = start + this.headerLength;
	/* Loop until we've reached the end of the header. */
	while( curr < end )
	{
	    short optType = xdr.SHORT();
	    /* Option length, minus the option header length */
	    short optLength = (short)(xdr.SHORT() - 1);

	    if( optLength < 0 )
	    {
		throw new InvalidPacketException("Option length is too small");
	    }
	    switch( optType )
	    {
	    case OPTION_TYPE_DESTINATION:
		switch( xdr.INT() )
		{
		case IPV4SCHEME:
		    this.ipv4Destination = xdr.INT();
		    break;
		default:
		    xdr.SKIP((optLength - 1) * OPTION_LENGTH_MULTIPLIER);
		    break;
		}
		break;
	    case OPTION_TYPE_SOURCE:
		switch( xdr.INT() )
		{
		case IPV4SCHEME:
		    this.ipv4Source = xdr.INT();
		    break;
		default:
		    xdr.SKIP((optLength - 1) * OPTION_LENGTH_MULTIPLIER);
		    break;
		}
		break;
	    default:
		if( ((optType & OPTION_FLAG_INTERNAL) != 0) &&
		    ((optType & OPTION_FLAG_DISCARD) != 0) )
		{
		    throw new DroppedPacketException(
			"Unknown option type, with discard bit set.");
		}
		xdr.SKIP(optLength * OPTION_LENGTH_MULTIPLIER);
		break;
	    }
	    curr = xdr.getReadCursor();
	}
    }

    /**
     * @see bees.core.Packet#internalEncode(Xdr)
     */
    void internalEncode(Xdr xdr)
    {
	Assert.require(xdr);
	
	xdr.PUT(this.version);
	xdr.PUT(this.flags);
	xdr.PUT(this.type);
	xdr.PUT((short)(this.headerLength / HEADER_LENGTH_MULTIPLIER));
	/*
	 * Note: We marshal this but don't verify its correct since it could be
	 * set later on.
	 */
	xdr.PUT(this.packetLength);
    }

    /**
     * Encode a source/destination option for IPv4 addresses.
     *
     * @param xdr The Xdr to use when encoding the options.
     * @param optionType The option type.
     * @param address The IPv4 address.
     */
    void encodeIPv4AddressOption(Xdr xdr,
				 short optionType,
				 int address)
    {
	Assert.require(xdr);
	
	xdr.PUT(optionType);
	xdr.PUT((short)(IPV4_OPTION_LENGTH / OPTION_LENGTH_MULTIPLIER));
	xdr.PUT(IPV4SCHEME);
	xdr.PUT(address);
    }

    /**
     * @see java.lang.Object#equals(Object)
     */
    public boolean equals(Object obj)
    {
	boolean retval = false;

	if( (obj instanceof ANEPPacket) &&
	    super.equals(obj) )
	{
	    ANEPPacket ap = (ANEPPacket)obj;

	    if( (this.version == ap.version) &&
		(this.flags == ap.flags) &&
		(this.type == ap.type) &&
		(this.headerLength == ap.headerLength) &&
		(this.packetLength == ap.packetLength) &&
		(this.ipv4Destination == ap.ipv4Destination) &&
		(this.ipv4Source == ap.ipv4Source) )
	    {
		retval = true;
	    }
	}
	return retval;
    }

    /**
     * @see java.lang.Object#toString()
     */
    public String toString()
    {
	return "ANEPPacket["
	    + super.toString()
	    + "; version="
	    + this.version
	    + "; flags=0x"
	    + Integer.toHexString(this.flags)
	    + "; type="
	    + Integer.toHexString(this.type)
	    + "; headerLength="
	    + this.headerLength
	    + "; packetLength="
	    + this.packetLength
	    + "; ipv4Destination="
	    + Address.toString(this.ipv4Destination)
	    + "; ipv4Source="
	    + Address.toString(this.ipv4Source)
	    + "]";
    }
}
