/* 
 * ByteArray.java
 *
 * Copyright (c) 1999, 2000, 2001, 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Bees distribution.
 * 
 * Permission to use, copy, modify, distribute, and sell this
 * software and its documentation for any purpose is hereby granted
 * without fee, provided that (i) the above copyright notice(s)
 * appear in all copies and (ii) that both the copyright notice(s)
 * and this permission notice appear in supporting documentation,
 * and (iii) that the name(s) of the copyright holders not be used
 * in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.  The
 * copyright holders make no representations about the suitability
 * of this software for any purpose.  It is provided "as is"
 * without express or implied warranty.
 * 
 * The maintainers of this code ask, but do not require, that
 * improvements and bug fixes be emailed to bees-dist@cs.utah.edu
 *         
 *  
 *  Copyright 1996 Massachusetts Institute of Technology
 * 
 *  Permission to use, copy, modify, distribute, and sell this software and
 *  its documentation for any purpose is hereby granted without fee, provided
 *  that the above copyright notice appear in all copies and that both that
 *  copyright notice and this permission notice appear in supporting
 *  documentation, and that the name of M.I.T. not be used in advertising or
 *  publicity pertaining to distribution of the software without specific,
 *  written prior permission.  M.I.T. makes no representations about the
 *  suitability of this software for any purpose.  It is provided "as is"
 *  without express or implied warranty.
 * 
 */
package bees.core;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.FileNotFoundException;

import bees.core.resource.File;

/**
 * Wrapper for an array of Bytes.  Incorporates offset and length.
 * <p>
 * Most of the methods in here are final for performance reasons.
 *
 * @author probably David Wetherall.
 */
public class ByteArray
{
    /**
     * The actual byte array this object wraps.
     */
    public byte[] buf;

    /**
     * The start of the byte array in `buf' that this object wraps.
     */
    public int offset;

    /**
     * The length of the byte array in `buf' that this object wraps.
     */
    public int len;

    /**
     * Construct a new ByteArray object that has the given length.
     *
     * @param len The length of the internal byte array to create.
     */
    public ByteArray(int len)
    {
	Assert.require(len >= 0);
	
	buf = new byte[len];
	offset = 0;
	this.len = len;
    }
    
    /**
     * Construct a new ByteArray object that wraps the given array.
     *
     * @param buf The byte array to wrap.
     */
    public ByteArray(byte[] buf)
    {
	Assert.require(buf);
	
	this.buf = buf;
	this.offset = 0;
	this.len = buf.length;
    }
    
    /**
     * Construct a new ByteArray object that wraps the given array.
     *
     * @param buf The byte array to wrap.
     * @param offset The start of the byte array in `buf' that this object
     * should wrap.
     */
    public ByteArray(byte[] buf, int offset)
    {
	this(buf, offset, buf.length-offset);
    }
    
    /**
     * Construct a new ByteArray object that wraps the given array.
     *
     * @param buf The byte array to wrap.
     * @param offset The start of the byte array in `buf' that this object
     * should wrap.
     * @param len The length of the byte array in `buf' that this object
     * should wrap.
     * @throws IndexOutOfBoundsException if the given range is outside the
     * given buffer.
     */
    public ByteArray(byte[] buf, int offset, int len) 
	throws IndexOutOfBoundsException
    {
	Assert.require(buf);
	Assert.require(offset >= 0);
	Assert.require(len >= 0);
	Assert.require((offset + len) <= buf.length);
	
	this.buf = buf;
	this.offset = offset;
	this.len = len;
    }

    /**
     * Construct a new ByteArray object that wraps the byte array encapsulated
     * by another ByteArray object.
     *
     * @param parent The ByteArray object that contains the byte array to
     * rewrap.
     * @param offset An extra offset to add to the parent's offset.
     */
    public ByteArray(ByteArray parent, int offset)
    {
	this(parent, offset, parent.len-offset);
    }
    
    /**
     * Construct a new ByteArray object that wraps the byte array encapsulated
     * by another ByteArray object.
     *
     * @param parent The ByteArray object that contains the byte array to
     * rewrap.
     * @param offset An extra offset to add to the parent's offset.
     * @param len The length of the wrapped byte array.
     * @throws IndexOutOfBoundsException if the given range is outside the
     * given buffer.
     */
    public ByteArray(ByteArray parent, int offset, int len)
	throws IndexOutOfBoundsException
    {
	Assert.require(parent);
	Assert.require(offset >= 0);
	Assert.require(len >= 0);
	Assert.require((offset + len) <= parent.len);
	
	buf = parent.buf;
	this.offset = parent.offset + offset;
	this.len = len;
    }

    /**
     * @return The length of the wrapped byte array.
     */
    public final int length()
    {
	return len;
    }

    /**
     * @param index The index of the byte to return.
     * @return The byte at the given index in the array.
     * @throws IndexOutOfBoundsException if `index' is out of
     * bounds of the array.
     */
    public final byte getByte(int index) throws IndexOutOfBoundsException
    {
	return buf[offset + index];
    }
    
    /**
     * @param index The index of the byte to return.
     * @param val The byte to put at the given index in the array.
     * @throws IndexOutOfBoundsException if `index' is out of
     * bounds of the array.
     */
    public final void setByte(int index, byte val) 
	throws IndexOutOfBoundsException
    {
	buf[offset+index] = val;
    }

    /**
     * Copy a range of bytes from the array into a basic byte array.
     *
     * @param srcBegin The offset into the byte array to start copying bytes.
     * @param dst The array to copy the bytes into.
     * @param dstBegin The offset in `dst' to start copying.
     * @param length The number of bytes to copy from the array.
     */
    public final void getBytes(int srcBegin, byte[] dst, int dstBegin, 
			       int length)
    {
	System.arraycopy(buf, offset + srcBegin, dst, dstBegin, length);
    }
    
    /**
     * Copy a range of bytes to the array from a basic byte array.
     *
     * @param dstBegin The offset into the byte array to start copying bytes.
     * @param src The array to copy the bytes from.
     * @param srcBegin The offset in `src' to start copying bytes.
     * @param length The number of bytes to copy to the array.
     */
    public final void setBytes(int dstBegin, byte[] src, int srcBegin,
			       int length)
    {
	System.arraycopy(src, srcBegin, buf, offset + dstBegin, length);
    }
    
    /**
     * Copy a range of bytes to the array from another byte array.
     *
     * @param dstBegin The offset into the byte array to start copying bytes.
     * @param src The array to copy the bytes from.
     * @param srcBegin The offset in `src' to start copying bytes.
     * @param length The number of bytes to copy to the array.
     */
    public final void setBytes(int dstBegin, ByteArray src, int srcBegin,
			       int length)
    {
	System.arraycopy(src.buf,
			 src.offset + srcBegin,
			 buf,
			 offset + dstBegin, 
			 length);
    }
    
    /* 
     * Note: setBytes/getBytes from String have been deleted because they 
     * used a deprecated API
     */
    
    /**
     * Returns a byte array exactly the right size for
     * the valid data in the buf.  Note that if this.buf
     * is the right side it is returned directly, otherwise
     * a new array is created of the right size.  (Weird.)
     *
     * @return array of bytes
     */
    public final byte[] toBytes()
    {
	if (offset == 0 && len == buf.length)
	{
	    return buf;
	}
	else 
	{
	    byte[] newbuf = new byte[len];
	    System.arraycopy(buf, offset, newbuf, 0, len);
	    this.buf = newbuf;
	    this.offset = 0;
	    return buf;
	}
    }

    /**
     * Create a byte array from the contents of the given file.
     *
     * @param file The file capability to read from.
     * @return The byte array containing the contents of the given file.
     * @throws FileNotFoundException if the file couldn't be found.
     * XXX This "shouldn't" happen since we have a capability to it.
     * @throws IOException if there is an error reading the file.
     */
    public static final byte[] fromFile(File file)
	throws FileNotFoundException,
	       IOException
    {
	byte retval[] = null;

	if( file != null )
	{
	    InputStream is;
	    
	    retval = new byte[(int)file.length()];
	    is = file.getInputStream();
	    try
	    {
		is.read(retval);
	    }
	    finally
	    {
		is.close();
	    }
	}
	return retval;
    }

    /**
     * Write the given byte array to the given file.
     *
     * @param file The file capability to write to.
     * @param bits The byte array to write to the file.
     * @throws FileNotFoundException if the file couldn't be found.
     * XXX This "shouldn't" happen since we have a capability to it.
     * @throws IOException if there is an error writing the file.
     */
    public static final void toFile(File file, byte bits[])
	throws FileNotFoundException,
	       IOException
    {
	OutputStream os;
	
	Assert.require(file);
	Assert.require(bits);

	os = file.getOutputStream();
	try
	{
	    os.write(bits);
	}
	finally
	{
	    os.close();
	}
    }

    /**
     * Concatenate an array of byte arrays together.
     *
     * @param ba The array of arrays to concatenate.  Any null elements will
     * be ignored.
     * @return A byte array consisting of the contents in ba.
     */
    public static final byte[] concat(byte ba[][])
    {
	int size = 0, lpc;
	byte retval[];

	Assert.require(ba);
	
	for( lpc = 0; lpc < ba.length; lpc++ )
	{
	    if( ba[lpc] != null )
	    {
		size += ba[lpc].length;
	    }
	}
	retval = new byte[size];
	size = 0;
	for( lpc = 0; lpc < ba.length; lpc++ )
	{
	    if( ba[lpc] != null )
	    {
		System.arraycopy(ba[lpc], 0,
				 retval, size, ba[lpc].length);
		size += ba[lpc].length;
	    }
	}
	return retval;
    }

    /**
     * Concatenate two byte arrays together.
     *
     * @param ba1 The first byte array.
     * @param ba2 The second byte array.
     * @return A byte array consisting of the contents of ba1 followed by the
     * contents of ba2.
     */
    public static final byte[] concat(byte ba1[], byte ba2[])
    {
	return concat(new byte[][] { ba1, ba2 });
    }

    /**
     * Convert a byte array to a string in a hexidecimal representation.
     *
     * @param bytes The byte array to format.
     * @param maxBytes The maximum number of bytes to format.
     * @return A plain hexadecimal string representing the values in bytes.
     */
    public static final String toPlainString(byte bytes[], int maxBytes)
    {
	String retval = "";

	if( bytes != null )
	{
	    StringBuffer sb;
	    int lpc;
	    
	    if( maxBytes > bytes.length )
	    {
		maxBytes = bytes.length;
	    }
	    sb = new StringBuffer(32 + maxBytes * 2);
	    for( lpc = 0; lpc < maxBytes; lpc++ )
	    {
		if( (bytes[lpc] & 0xf0) == 0 )
		{
		    sb.append("0");
		}
		sb.append(Integer.toHexString(bytes[lpc] & 0xff));
	    }
	    retval = sb.toString();
	}
	else
	{
	    retval = "(null)";
	}
	return retval;
    }

    /**
     * Convert a byte array to a string in a hexidecimal representation.
     *
     * @param bytes The byte array to format.
     * @return A plain hexadecimal string representing the values in bytes.
     */
    public static final String toPlainString(byte bytes[])
    {
	return toPlainString(bytes, Integer.MAX_VALUE);
    }
    
    /**
     * Convert a byte array to a string in a hexidecimal representation and
     * some extra formatting.
     *
     * @param bytes The bytes to turn into a string.
     * @param maxBytes The maximum number of bytes to format.
     * @return The formatted string.
     */
    public static final String toString(byte bytes[], int maxBytes)
    {
	String retval;

	if( bytes != null )
	{
	    retval = "byte[length=" + bytes.length + "; data={ ";
	    retval += toPlainString(bytes, maxBytes);
	    retval += " }]";
	}
	else
	{
	    retval = "(null)";
	}
	return retval;
    }
    
    /**
     * Convert a byte array to a string in a hexidecimal representation and
     * some extra formatting.
     *
     * @param bytes The bytes to turn into a string.
     * @return The formatted string.
     */
    public static final String toString(byte bytes[])
    {
	return toString(bytes, Integer.MAX_VALUE);
    }
    
    /**
     * @see java.lang.Object#toString()
     */
    public String toString()
    {
	return "ByteArray[buf="
	    + ByteArray.toString(this.buf)
	    + "; offset="
	    + this.offset
	    + "; len="
	    + this.len
	    + "]";
    }
}
