/*
 * ByteArrayResponseCapsule.java
 *
 * Copyright (c) 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is part of the Janos Bees Software.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */

package bees.core;

import bees.core.resource.Logger;

/**
 * A ByteArrayResponseCapsule is used to respond to a FlowByteArrayType request
 * from a peer node.
 */
final class ByteArrayResponseCapsule
    extends DataResponseCapsule
{
    /**
     * Construct an empty ByteArrayResponseCapsule.
     */
    ByteArrayResponseCapsule()
    {
    }

    /**
     * Construct a response to a ByteArrayRequestCapsule.
     *
     * @param req The request capsule.
     * @param ct The ByteArrayType containing the byte code.
     */
    ByteArrayResponseCapsule(ByteArrayRequestCapsule req, ByteArrayType ct)
    {
	super(req, ct.getByteArray());
    }

    /**
     * @see bees.core.Capsule#evaluate(Node)
     */
    public void evaluate(Node node)
    {
	TypeID specializer;

	if( (specializer = this.getSpecializer()) != null )
	{
	    FlowByteArrayType fct;
	    
	    if( (fct = FlowByteArrayType.
		 getBDomainFlowByteArray(specializer)) != null )
	    {
		DLProtocolSession dps;
		DataAggregator da;
		ByteArrayType ct;

		/* Remove the capsule from the cache to stop resends. */
		dps = (DLProtocolSession)this.getSession();
		dps.getCache().remove(specializer);
		ct = fct.getByteArrayType();
		da = DLProtocol.
		    getInProgressByteCode(ct, this.getTotalLength());
		if( super.merge(da) )
		{
		    TypeID flowID;

		    /* Got all the data for this class. */
		    DLProtocol.putByteCode(ct, da);
		    // XXX DLProtocol.removeInProgressByteCode(ct);
		    if( dps.getProgress().
			completedCode(this, specializer, da.getBytes()) )
		    {
			/* Got all the code for the protocol. */
			dps.finished(node);
		    }
		    else
		    {
			/* Still need some more code. */
		    }
		}
		else
		{
		    ByteArrayRequestCapsule crc;
		    
		    /* Still need more chunks. */
		    crc = new ByteArrayRequestCapsule(specializer);
		    crc.fromCapsule(this, specializer);
		    dps.getCache().put(specializer,
				       crc,
				       ReliableDataRequestCapsule.
				       DEFAULT_RETRY_INTERVAL);
		    crc.resend(node, dps);
		}
	    }
	    else
	    {
		Logger.log(Logger.WARNING,
			   "Unable to find flow code type for: "
			   + specializer);
	    }
	}
	else
	{
	    Logger.log(Logger.WARNING,
		       "Empty flow code response specializer");
	}
    }
    
    /**
     * @see java.lang.Object#toString()
     */
    public String toString()
    {
	return "ByteArrayResponseCapsule["
	    + super.toString()
	    + "]";
    }
}
