/*
 * CapsuleEncoder.java
 *
 * Copyright (c) 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is part of the Janos Bees Software.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */

package bees.core;

import edu.utah.janos.nodeos.BufferHandle;

/**
 * A class used to cache state used when encoding capsules for transmission.
 */
final class CapsuleEncoder
{
    /**
     * Number of buffers available for enconding capsules.
     */
    private static final int BUFFER_COUNT = 1;

    /**
     * Array of buffers to use when encoding capsules.
     */
    private final BufferHandle bh[] = new BufferHandle[BUFFER_COUNT];
    
    /**
     * The index into "bh" to use when encoding a capsule.
     */
    private int current;
    
    /**
     * BufferHandle used to encode the capsule in.
     */
    // private final BufferHandle bh = BufferHandle.create();

    /**
     * Xdr object used to encode capsules.
     */
    private final Xdr xdr = new Xdr();

    /**
     * The maximum buffer length available in "bh".
     */
    private final int maxBufferLength;

    /**
     * The offset into bh for the start of the user payload.  This value is used
     * when reencoding a capsule for a different neighbor.
     */
    private int userOffset;

    /**
     * Construct a CapsuleEncoder object.
     */
    CapsuleEncoder()
    {
	int lpc;

	for( lpc = 0; lpc < this.bh.length; lpc++ )
	{
	    this.bh[lpc] = BufferHandle.create();
	}
	this.maxBufferLength = this.bh[0].getLength();
    }

    /**
     * @return The BufferHandle that holds the encoded capsule.
     */
    BufferHandle getBuffer()
    {
	return this.bh[this.current];
    }

    /**
     * @return The Xdr used to encode capsules.
     */
    Xdr getXDR()
    {
	return this.xdr;
    }

    /**
     * @return The maximum buffer length available in the BufferHandle.
     */
    int getMaxBufferLength()
    {
	return this.maxBufferLength;
    }

    /**
     * Encode a capsule into the BufferHandle.
     *
     * @param cap The capsule to encode.
     * @return The BufferHandle with the encoded capsule.
     * @throws EncodingException if there was a problem
     * marshalling the capsule.
     */
    BufferHandle encode(Capsule cap)
	throws EncodingException
    {
	BufferHandle retval;
	int length;

	this.current += 1;
	this.current %= BUFFER_COUNT;
	retval = this.bh[this.current];
	retval.setLength(this.maxBufferLength);
	this.xdr.writeReset(retval);
	cap.internalEncode(this.xdr);
	this.userOffset = this.xdr.getWriteCursor();
	try
	{
	    cap.encode(this.xdr);
	}
	catch(Throwable th)
	{
	    throw new EncodingException(th);
	}
	length = this.xdr.getWriteCursor();
	retval.setLength(length);
	retval.setShort(ANEPPacket.OFFSET_PACKET_LENGTH, (short)length);
	return retval;
    }

    /**
     * Reencode a capsule for a different neighbor.
     * 
     * @param cap The capsule to reencode.
     * @param ni The neighbor to reencode the capsule for.
     * @throws EncodingException if there was a problem
     * marshalling the capsule.
     */
    void reencode(ReencodableCapsule cap, NeighborImpl ni)
	throws EncodingException
    {
	this.xdr.writeReset(this.bh[this.current], this.userOffset);
	try
	{
	    cap.reencode(this.xdr, ni);
	}
	catch(Throwable th)
	{
	    throw new EncodingException(th);
	}
    }

    /**
     * @see java.lang.Object#toString()
     */
    public String toString()
    {
	return "CapsuleEncoder[bh="
	    + this.bh
	    + "; xdr="
	    + this.xdr
	    + "; maxBufferLength="
	    + this.maxBufferLength
	    + "]";
    }
}
