/*
 * DLProtocol.java
 *
 * Copyright (c) 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is part of the Janos Bees Software.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */

package bees.core;

import java.util.Hashtable;

import bees.core.kernel.KernelPanic;

import bees.core.bdomain.BDomainLocal;

import bees.core.resource.Logger;

import bees.core.security.Permission;
import bees.core.security.PermissionSet;
import bees.core.security.VersionString;
import bees.core.security.CountPermission;
import bees.core.security.MinimumIntegerPermission;

/**
 * The DLProtocol is a companion protocol that will download its companion's
 * byte arrays.
 */
public final class DLProtocol
    extends Protocol
    implements BuiltinProtocol, ByteArrayCompanionProtocol
{
    /* XXX The stuff below should be moved into the protocol session. */
    
    private static final BDomainLocal domainInProgressByteCode = new BDomainLocal() {
	    protected Object initialValue()
	    {
		return new Hashtable();
	    }
	};

    static Hashtable getBDomainInProgressByteCode()
    {
	return (Hashtable)domainInProgressByteCode.get();
    }

    static void putInProgressByteCode(ByteArrayType ct, DataAggregator da)
    {
	Hashtable ht = (Hashtable)domainInProgressByteCode.get();

	synchronized( ht )
	{
	    if( !ht.containsKey(ct) )
		ht.put(ct, da);
	}
    }

    static DataAggregator getInProgressByteCode(ByteArrayType ct,
						int totalLength)
    {
	Hashtable ht = (Hashtable)domainInProgressByteCode.get();
	DataAggregator retval;

	synchronized( ht )
	{
	    if( ((retval = (DataAggregator)ht.get(ct)) == null) &&
		(totalLength >= 0) )
	    {
		retval = new DataAggregator(totalLength);
		ht.put(ct, retval);
	    }
	}
	return retval;
    }

    static DataAggregator getInProgressByteCode(ByteArrayType ct)
    {
	return getInProgressByteCode(ct, -1);
    }
    
    static void removeInProgressByteCode(ByteArrayType ct)
    {
	Hashtable ht = (Hashtable)domainInProgressByteCode.get();

	synchronized( ht )
	{
	    ht.remove(ct);
	}
    }
    
    private static final BDomainLocal domainByteCode = new BDomainLocal() {
	    protected Object initialValue()
	    {
		return new Hashtable();
	    }
	};

    static Hashtable getBDomainByteCode()
    {
	return (Hashtable)domainByteCode.get();
    }

    static void putByteCode(ByteArrayType ct, DataAggregator da)
    {
	Hashtable ht = (Hashtable)domainByteCode.get();

	synchronized( ht )
	{
	    if( !ht.containsKey(ct) )
		ht.put(ct, da);
	}
    }

    static DataAggregator getByteCode(ByteArrayType ct)
    {
	Hashtable ht = (Hashtable)domainByteCode.get();
	DataAggregator retval;

	retval = (DataAggregator)ht.get(ct);
	return retval;
    }

    /**
     * Construct a DLProtocol description.
     */
    public DLProtocol()
    {
	try
	{
	    Node node = super.getNode();
	    
	    this.startProtocolDefn();
	    {
		/* Capsule used to request a FlowByteArrayType. */
		this.addCapsule("bees.core.ByteArrayRequestCapsule",
				(VersionString)node.
				getSystemCapability("dlversion"),
				new PermissionSet(new Permission[] {
				    new CountPermission("initial", 1),
				    new CountPermission("returns", 1),
				    new CountPermission(
					"transformTo."
					+ "bees.core.ByteArrayResponseCapsule",
					1),
				    
				    new CountPermission("threadCount", 1),
				    new CountPermission("bufferCount", 1),

				    new CountPermission("cached", 20),
				    new MinimumIntegerPermission("interval",
								 500),
				}));
		/* Capsule used to respond to a FlowByteArrayType request. */
		this.addCapsule("bees.core.ByteArrayResponseCapsule",
				(VersionString)node.
				getSystemCapability("dlversion"),
				new PermissionSet(new Permission[] {
				    new CountPermission("returns", 1),
				    new CountPermission("externals", 1),
				    new CountPermission(
					"transformTo."
					+ "bees.core.ByteArrayRequestCapsule",
					1),
				    
				    new CountPermission("threadCount", 1),
				    new CountPermission("bufferCount", 1),
				}));
		this.setSession("bees.core.DLProtocolSession");
	    }
	    this.endProtocolDefn();
	}
	catch(ClassNotFoundException e)
	{
	    throw new Panic("Unable to access classes", e);
	}
    }

    /**
     * @see java.lang.Object#toString()
     */
    public String toString()
    {
	return "DLProtocol["
	    + super.toString()
	    + "]";
    }
}
