/*
 * DLProtocolSession.java
 *
 * Copyright (c) 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is part of the Janos Bees Software.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */

package bees.core;

import java.io.IOException;

import java.util.Vector;
import java.util.Hashtable;
import java.util.Enumeration;

import edu.utah.janos.nodeos.Domain;

import bees.core.kernel.KernelPanic;

import bees.core.resource.Logger;

import bees.core.bdomain.BDomainLocal;

/**
 * A DLProtocolSession manages application requests for downloading protocol
 * code from a peer node.
 */
public final class DLProtocolSession
    extends ExtendedProtocolSession
{
    /**
     * Start the download.
     */
    static final int METHOD_START = METHOD_USER + 1;
    static final int METHOD_BOOT_CAPSULE = METHOD_USER + 2;

    static final int STATE_FAILURE = -1;
    static final int STATE_UNKNOWN = 0;
    static final int STATE_BASIC = 1;
    static final int STATE_EXTENDED = 2;
    
    /**
     * Tracks the progress of the download.
     */
    private DLProtocolProgress dpp;

    private Capsule bootCapsule;

    /**
     * Construct an empty DLProtocolSession.
     */
    DLProtocolSession()
    {
    }

    public Neighbor getPreviousNeighbor()
    {
	Neighbor retval = null;

	if( this.dpp != null )
	    retval = this.dpp.getPreviousNeighbor();
	return retval;
    }

    /**
     * @return The object used to track the progress of the download.
     */
    DLProtocolProgress getProgress()
    {
	return this.dpp;
    }

    /**
     * The protocol download has finished.
     *
     * @param node The current node object.
     */
    void finished(Node node)
    {
	boolean success = true;
	ExternalCapsule ec;
	TypeID flowID;
	FlowType ft;
	Xdr xdr;

	ft = this.getSessionType().getFlowType();
	flowID = ft.getFlowID();
	try
	{
	    /*
	     * We have all the code now, so we can create the CapsuleType
	     * objects from the ProtocolType and ByteArrayType's.
	     */
	    ProtocolBDomain.createCapsuleTypes(ft, node.getCapabilities());
	    Logger.log("Starting up protocol: " + flowID);
	    try
	    {
		Domain domain = Domain.currentDomain();
		String bootStrapClassName;

		/* Evaluate the boot strap capsule. */
		if( (bootStrapClassName = ft.getPrimaryProtocolType().
		     getBootStrapClassName()) != null )
		{
		    Class cl;
		    
		    cl = domain.loadClass(bootStrapClassName);
		    this.bootCapsule = (Capsule)cl.newInstance();
		    this.bootCapsule.
			fromFlowType(ft, CapsuleType.NULL_SPECIALIZER);
		}
	    }
	    catch(Throwable th)
	    {
		Logger.log(th);
		success = false;
	    }
	}
	catch(IOException e)
	{
	    Logger.log(e);
	    success = false;
	}
	/* Notify the applications. */
	ec = new ExternalCapsule().init();
	xdr = ec.getXDR();
	xdr.PUT(METHOD_REPLY | METHOD_START);
	xdr.PUT(success);
	super.notifyInterests(ec);
    }

    public synchronized void userReceive(Node node,
					 int method,
					 ExternalCapsule ec)
    {
	Xdr xdr;

	xdr = ec.getXDR();
	switch( method )
	{
	case METHOD_START:
	    if( this.dpp == null )
	    {
		TypeID flowID;
		FlowType ft;

		flowID = this.getSessionType().getFlowType().getFlowID();
		/* Hasn't been started yet. */
		ft = FlowType.getBDomainFlow(flowID);
		this.dpp = new DLProtocolProgress(ft);
		Logger.log("Starting download for " + flowID);
		try
		{
		    dpp.setPreviousNeighbor(node, this);
		    /* Send the first request. */
		    dpp.requestCode(ec, null);
		}
		catch(Throwable th)
		{
		    super.reply(ec.getSource(), METHOD_START, th);
		}
	    }
	    else if( this.dpp.isComplete() )
	    {
		/*
		 * They might've registered after the notification was sent
		 * out.
		 */
		super.reply(ec.getSource(), METHOD_START);
	    }
	    else
	    {
		/* Still in progress. */
	    }
	    break;
	case METHOD_BOOT_CAPSULE:
	    try
	    {
		int state = STATE_BASIC;
		ExternalCapsule reply;
		
		if( this.bootCapsule.getSession() instanceof
		    ExtendedProtocolSession )
		{
		    state = STATE_EXTENDED;
		}
		this.bootCapsule.evaluate(bees.core.Node.currentNode());
		this.bootCapsule = null;
		reply = new ExternalCapsule().init();
		xdr = reply.getXDR();
		xdr.PUT(METHOD_REPLY | METHOD_BOOT_CAPSULE);
		xdr.PUT(true);
		xdr.PUT(state);
		xdr.PUT(this.getSessionType().getFlowType().
			getPrimaryProtocolType().getProtocolID());
		ec.getSource().sendCapsule(reply);
	    }
	    catch(Throwable th)
	    {
		Logger.log(th);
		super.reply(ec.getSource(), METHOD_BOOT_CAPSULE, th);
	    }
	    break;
	default:
	    Strict.fail();
	    break;
	}
	ec.getSource().recycleCapsule(ec);
    }

    public String toString()
    {
	return "DLProtocolSession["
	    + super.toString()
	    + "; dpp="
	    + this.dpp
	    + "]";
    }
}
