/*
 * DataResponseCapsule.java
 *
 * Copyright (c) 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is part of the Janos Bees Software.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */

package bees.core;

/**
 * A DataResponseCapsule is the complement of the DataRequestCapsule, it is
 * used to carry the response back to the requesting node.
 */
public abstract class DataResponseCapsule
    extends Capsule
{
    /**
     * The total length of the byte array being transferred.
     */
    private int totalLength;

    /**
     * The DataAggregator window that this response contains.
     */
    private int currentWindow;

    /**
     * The DataAggregator chunk set that this response contains.
     */
    private long windowChunks;

    /**
     * The byte array containing the requested chunks.
     */
    private byte chunkBytes[];

    /**
     * Construct an empty DataResponseCapsule.
     */
    protected DataResponseCapsule()
    {
    }

    /**
     * Construct a DataResponseCapsule with the given values.
     *
     * @param req The request capsule.
     * @param bytes The complete byte array that should be used to satisfy
     * the request.
     */
    protected DataResponseCapsule(DataRequestCapsule req, byte bytes[])
    {
	Assert.require(req);
	Assert.require(bytes);
	
	this.totalLength = bytes.length;
	this.currentWindow = req.getCurrentWindow();
	this.windowChunks = req.getWindowChunks();
	this.chunkBytes = DataAggregator.fillWindow(bytes,
						    this.currentWindow,
						    this.windowChunks);
    }

    /**
     * @return The total length of the byte array being transferred.
     */
    public int getTotalLength()
    {
	return this.totalLength;
    }

    /**
     * @return The bytes that make up the requested chunks.
     */
    public byte[] getChunkBytes()
    {
	return this.chunkBytes;
    }

    /**
     * @see bees.core.Capsule#decode(Xdr)
     */
    public void decode(Xdr xdr)
    {
	this.totalLength = xdr.INT();
	this.currentWindow = xdr.INT();
	this.windowChunks = xdr.LONG();
	this.chunkBytes = xdr.BYTES(xdr.INT());
    }

    /**
     * @see bees.core.Capsule#encode(Xdr)
     */
    public void encode(Xdr xdr)
    {
	xdr.PUT(this.totalLength);
	xdr.PUT(this.currentWindow);
	xdr.PUT(this.windowChunks);
	xdr.PUT(this.chunkBytes.length);
	xdr.PUT(this.chunkBytes);
    }

    /**
     * Merge the contents of this response into the given DataAggregator.
     *
     * @param da The DataAggregator that should be filled in.
     * @return True if the contents of DataAggregator has been completed.
     */
    public boolean merge(DataAggregator da)
    {
	return da.fromWindow(this.chunkBytes,
			     this.currentWindow,
			     this.windowChunks);
    }

    /**
     * @see java.lang.Object#toString()
     */
    public String toString()
    {
	return "DataResponseCapsule["
	    + super.toString()
	    + "; totalLength="
	    + this.totalLength
	    + "; currentWindow="
	    + this.currentWindow
	    + "; windowChunks=0x"
	    + Long.toHexString(this.windowChunks)
	    + "; chunkBytes="
	    + ByteArray.toString(this.chunkBytes)
	    + "]";
    }
}
