/*
 * ExtendedProtocolPresentation.java
 *
 * Copyright (c) 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is part of the Janos Bees Software.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */

package bees.core;

import java.util.Hashtable;

import edu.utah.janosvm.sys.DeadTeamException;
import edu.utah.janosvm.sys.TeamEjectionException;

import edu.utah.janosvm.kit.comm.amiga.FaxPortHandle;

import bees.core.bdomain.BDomain;
import bees.core.bdomain.BDomainLocal;
import bees.core.bdomain.DeadBDomainException;

import bees.core.resource.Logger;

import edu.utah.janosvm.kit.util.SparseArray;
import edu.utah.janosvm.kit.util.SparseArrayElement;
import edu.utah.janosvm.kit.util.ObjectSparseArrayElement;

/**
 * An ExtendedProtocolPresentation is a ProtocolPresentation that provides
 * functionality for interacting with a protocol that uses the
 * ExtendedProtocolSession class.
 */
public abstract class ExtendedProtocolPresentation
    extends ProtocolPresentation
{
    public static final int STATE_ERROR_TIMEOUT = -2;
    public static final int STATE_ERROR = -1;
    public static final int STATE_NEWBORN = 0;
    public static final int STATE_CONNECTED = 1;
    public static final int STATE_REGISTERED = 2;
    public static final int STATE_USER = 1000;
    public static final int STATE_COMPLETE = Integer.MAX_VALUE;
    
    protected int state = STATE_NEWBORN;
    private final SparseArray sa;

    /**
     * Construct an empty ExtendedProtocolPresentation.
     */
    public ExtendedProtocolPresentation()
    {
	this.sa = new SparseArray();
    }

    public void sendCapsule(ExternalCapsule ec)
    {
	if( this.state < STATE_CONNECTED )
	{
	    throw new IllegalStateException("Presentation not connected.");
	}
	super.sendCapsule(ec);
    }

    ProtocolPresentation setCoroutine(ExternalCoroutine ec)
    {
	super.setCoroutine(ec);
	this.state = STATE_CONNECTED;
	return this;
    }

    protected synchronized void setState(int newState)
    {
	if( (newState > this.state) || (newState <= STATE_ERROR) )
	{
	    this.state = newState;
	    this.notifyAll();
	    if( (newState <= STATE_ERROR) || (newState == STATE_COMPLETE) )
		this.drop();
	}
    }

    public final int getState()
    {
	return this.state;
    }

    public synchronized int waitForState(int nextState, long timeout)
    {
	long end;

	if( timeout == 0 )
	    end = Long.MAX_VALUE;
	else
	    end = System.currentTimeMillis() + timeout;
	while( (this.state > STATE_ERROR) && (this.state < nextState) )
	{
	    try
	    {
		this.wait(end - System.currentTimeMillis());
		if( System.currentTimeMillis() >= end )
		{
		    this.setState(STATE_ERROR_TIMEOUT);
		}
	    }
	    catch(InterruptedException e)
	    {
	    }
	}
	return this.state;
    }

    public final void drop()
    {
	/* XXX Revoke from protocol */
	this.setCoroutine(null);
    }
    
    protected long allocateMessageID(ExternalCapsule ec)
    {
	ObjectSparseArrayElement sae;

	sae = (ObjectSparseArrayElement)this.sa.getElement();
	sae.setValue(ec);
	return sae.getID();
    }

    protected ExternalCapsule lookupMessage(long id)
    {
	ExternalCapsule retval = null;
	ObjectSparseArrayElement sae;

	if( (sae = (ObjectSparseArrayElement)this.sa.findElement(id)) != null )
	{
	    retval = (ExternalCapsule)sae.getValue();
	}
	return retval;
    }

    protected void releaseMessageID(long id)
    {
	synchronized( this.sa )
	{
	    SparseArrayElement sae;
	    
	    if( (sae = this.sa.findElement(id)) != null )
	    {
		this.sa.dropElement(sae);
	    }
	}
    }

    public void register()
    {
	ExternalCapsule ec;
	Xdr xdr;

	ec = new ExternalCapsule().init();
	xdr = ec.getXDR();
	xdr.PUT(ExtendedProtocolSession.METHOD_REGISTER);
	xdr.PUT(this.allocateMessageID(ec));
	this.sendCapsule(ec);
    }

    public void registerReply(Node node, ExternalCapsule ec)
    {
    }

    public void unregister()
    {
	ExternalCapsule ec;
	Xdr xdr;

	ec = new ExternalCapsule().init();
	xdr = ec.getXDR();
	xdr.PUT(ExtendedProtocolSession.METHOD_UNREGISTER);
	xdr.PUT(this.allocateMessageID(ec));
	this.sendCapsule(ec);
    }

    public void unregisterReply(Node node, ExternalCapsule ec)
    {
    }

    public void addNeighbor(NeighborImpl ni)
    {
	ExternalCapsule ec;
	Xdr xdr;

	ec = new ExternalCapsule(new CapabilityEnvelope[] {
	    new CapabilityEnvelope(ni),
	}).init();
	xdr = ec.getXDR();
	xdr.PUT(ExtendedProtocolSession.METHOD_ADD_NEIGHBOR);
	xdr.PUT(this.allocateMessageID(ec));
	this.sendCapsule(ec);
    }

    public void addNeighborReply(Node node, CapabilityReference cr)
    {
    }

    public void remNeighbor(NeighborImpl ni)
    {
	ExternalCapsule ec;
	Xdr xdr;

	ec = new ExternalCapsule(new CapabilityReference[] {
	    new CapabilityReference(ni),
	}).init();
	xdr = ec.getXDR();
	xdr.PUT(ExtendedProtocolSession.METHOD_REM_NEIGHBOR);
	xdr.PUT(this.allocateMessageID(ec));
	this.sendCapsule(ec);
    }

    public void remNeighborReply(Node node, CapabilityReference cr)
    {
    }

    public void receive(Node node, ExternalCapsule ec)
    {
	int method;
	Xdr xdr;

	xdr = ec.getXDR();
	method = xdr.INT();
	if( (method & ExtendedProtocolSession.METHOD_REPLY) != 0 )
	{
	    boolean success;

	    switch( method & ~ExtendedProtocolSession.METHOD_REPLY )
	    {
	    case ExtendedProtocolSession.METHOD_REGISTER:
		this.registerReply(node, ec);
		break;
	    case ExtendedProtocolSession.METHOD_UNREGISTER:
		this.unregisterReply(node, ec);
		break;
	    case ExtendedProtocolSession.METHOD_ADD_NEIGHBOR:
		success = xdr.BOOLEAN();
		if( success )
		{
		    try
		    {
			this.addNeighborReply(node, ec.getReferences()[0]);
		    }
		    catch(Throwable th)
		    {
			Logger.log(th);
		    }
		}
		else
		{
		    this.addNeighborReply(node, null);
		}
		break;
	    case ExtendedProtocolSession.METHOD_REM_NEIGHBOR:
		success = xdr.BOOLEAN();
		if( success )
		{
		    this.remNeighborReply(node, ec.getReferences()[0]);
		}
		else
		{
		    this.remNeighborReply(node, null);
		}
		break;
	    default:
		this.userReceive(node, method, ec);
		break;
	    }
	}
	else
	{
	    this.userReceive(node, method, ec);
	}
    }

    protected void userReceive(Node node, int method, ExternalCapsule ec)
    {
    }

    public String toString()
    {
	return "ExtendedProtocolPresentation["
	    + super.toString()
	    + "]";
    }
}
