/*
 * ExtendedProtocolSession.java
 *
 * Copyright (c) 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is part of the Janos Bees Software.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */

package bees.core;

import java.io.IOException;

import java.util.Vector;
import java.util.Hashtable;
import java.util.Enumeration;

import bees.core.resource.Logger;

import bees.core.security.IntegerPermission;

import bees.core.bdomain.BDomainLocal;

/**
 * An ExtendedProtocolSession is a ProtocolSession that provides functionality
 * for tracking interested applications and managing the addition and removal
 * of neighbors.
 */
public class ExtendedProtocolSession
    extends ProtocolSession
{
    /**
     * Timeout for registered applications.
     */
    public static IntegerPermission REGISTER_TIMEOUT =
	new IntegerPermission("register-timeout", 60 * 60 * 1000);

    /**
     * Null method identifier.
     */
    public static final int METHOD_NONE		= 0x0000;

    /**
     * Report an error.
     */
    public static final int METHOD_ERROR	= 0x0001;

    /**
     * Register an application with the protocol.
     */
    public static final int METHOD_REGISTER	= 0x0002;

    /**
     * Unregister an application from a protocol.
     */
    public static final int METHOD_UNREGISTER	= 0x0003;

    /**
     * Add a neighbor object to the protocol.
     */
    public static final int METHOD_ADD_NEIGHBOR	= 0x0004;

    /**
     * Remove a neighbor object from the protocol.
     */
    public static final int METHOD_REM_NEIGHBOR	= 0x0005;

    /**
     * The start of user messages.
     */
    public static final int METHOD_USER		= 0x0100;

    /**
     * Flag indicating that the message is a reply to another message.
     */
    public static final int METHOD_REPLY	= 0x8000;

    /**
     * The set of applications interested in the protocol.
     */
    private final Hashtable interests = new Hashtable();

    /**
     * Construct an empty ExtendedProtocolSession.
     */
    public ExtendedProtocolSession()
    {
    }

    public Hashtable getNeighbors()
    {
	return this.neighbors;
    }

    /**
     * @return The table of external interests.
     */
    public Hashtable getInterests()
    {
	return this.interests;
    }

    /**
     * Broadcast an ExternalCapsule to any interested applications.
     *
     * @param ec The ExternalCapsule to broadcast.
     */
    public void notifyInterests(ExternalCapsule ec)
    {
	synchronized( this.interests )
	{
	    Enumeration enum;

	    enum = this.interests.elements();
	    while( enum.hasMoreElements() )
	    {
		ExternalChannel channel;

		channel = (ExternalChannel)enum.nextElement();
		try
		{
		    channel.sendCapsule(ec);
		}
		catch(IOException e)
		{
		}
	    }
	}
    }

    /**
     * Send a reply to a specific application.
     *
     * @param channel The destination for the reply.
     * @param method The method send a reply for.
     * @param th An exception or NULL if the operation succeeded.
     */
    public void reply(ExternalChannel channel, int method, Throwable th)
    {
	ExternalCapsule ec;
	Xdr xdr;
	
	ec = new ExternalCapsule().init();
	xdr = ec.getXDR();
	xdr.PUT(METHOD_REPLY | method);
	if( th == null )
	    xdr.PUT(true);
	else
	    xdr.PUT(false);
	try
	{
	    channel.sendCapsule(ec);
	}
	catch(IOException e)
	{
	    Logger.log(e);
	}
    }

    /**
     * Send a reply to a specific application.
     *
     * @param channel The destination for the reply.
     * @param method The method send a reply for.
     */
    public void reply(ExternalChannel channel, int method)
    {
	this.reply(channel, method, null);
    }

    /**
     * Handle a registration.
     *
     * @param node The current node.
     * @param cap The ExternalCapsule containing the registration message.
     * @param ec The ExternalChannel that should be registered.
     */
    public void register(Node node,
			 ExternalCapsule cap,
			 final ExternalChannel ec)
    {
	try
	{
	    this.interests.put(ec, ec);
	    
	    if( REGISTER_TIMEOUT != null )
	    {
		NodeCache nc;
		
		nc = this.getCache();
		nc.put(ec, new NodeCache.EventHandler() {
			private final ExternalChannel channel = ec;
			
			public void removed(Node n, ProtocolSession ps)
			{
			}
			
			public IntegerPermission collected(Node n,
							   ProtocolSession ps)
			{
			    ExtendedProtocolSession.this.interests.
				remove(this.channel);
			    return null;
			}
		    },
		       REGISTER_TIMEOUT);
	    }
	    this.reply(ec, METHOD_REGISTER);
	}
	catch(Throwable th)
	{
	    this.reply(ec, METHOD_REGISTER, th);
	}
    }

    /**
     * Handle an unregistration.
     *
     * @param node The current node.
     * @param cap The ExternalCapsule containing the unregistration message.
     * @param ec The ExternalChannel that should be unregistered.
     */
    public void unregister(Node node, ExternalCapsule cap, ExternalChannel ec)
    {
	try
	{
	    this.getCache().remove(ec);
	    this.interests.remove(ec);
	    this.reply(ec, METHOD_UNREGISTER);
	}
	catch(Throwable th)
	{
	    this.reply(ec, METHOD_UNREGISTER, th);
	}
    }
    
    /**
     * Handle a neighbor addition.
     *
     * @param node The current node.
     * @param cap The ExternalCapsule containing the message.
     * @param ni The NeighborImpl that should be added.
     */
    public void addNeighbor(Node node, ExternalCapsule ec, NeighborImpl ni)
    {
	try
	{
	    ExternalCapsule reply;
	    Xdr xdr;

	    super.addNeighbor(node, ni);
	    reply = new ExternalCapsule(new CapabilityReference[] {
		new CapabilityReference(ni),
	    }).init();
	    xdr = reply.getXDR();
	    xdr.PUT(METHOD_REPLY | METHOD_ADD_NEIGHBOR);
	    xdr.PUT(true);
	    this.notifyInterests(reply);
	}
	catch(Throwable th)
	{
	    Logger.log(th);
	    if( ec != null )
		this.reply(ec.getSource(), METHOD_ADD_NEIGHBOR, th);
	}
    }
    
    /**
     * Handle a neighbor removal.
     *
     * @param node The current node.
     * @param cap The ExternalCapsule containing the message.
     * @param ni The NeighborImpl that should be removed.
     */
    public void removeNeighbor(Node node, ExternalCapsule ec, NeighborImpl ni)
    {
	try
	{
	    ExternalCapsule reply;
	    Xdr xdr;

	    super.removeNeighbor(node, ni);
	    reply = new ExternalCapsule(new CapabilityReference[] {
		new CapabilityReference(ni),
	    }).init();
	    xdr = reply.getXDR();
	    xdr.PUT(METHOD_REPLY | METHOD_REM_NEIGHBOR);
	    xdr.PUT(true);
	    this.notifyInterests(reply);
	}
	catch(Throwable th)
	{
	    Logger.log(th);
	    if( ec != null )
		this.reply(ec.getSource(), METHOD_REM_NEIGHBOR, th);
	}
    }
    
    public void receive(Node node, ExternalCapsule ec)
    {
	NodeCache nc;
	int method;
	Xdr xdr;

	nc = this.getCache();
	// nc.refresh(ec.getSource(), REGISTER_TIMEOUT);
	xdr = ec.getXDR();
	method = xdr.INT();
	switch( method )
	{
	case METHOD_NONE:
	    /* XXX Error */
	    break;
	case METHOD_REGISTER:
	    this.register(node, ec, ec.getSource());
	    ec.getSource().recycleCapsule(ec);
	    break;
	case METHOD_UNREGISTER:
	    this.unregister(node, ec, ec.getSource());
	    ec.getSource().recycleCapsule(ec);
	    break;
	case METHOD_ADD_NEIGHBOR:
	    try
	    {
		NeighborImpl ni;

		ni = (NeighborImpl)ec.getEnvelopes()[0].unseal();
		this.addNeighbor(node, ec, ni);
	    }
	    catch(InstantiationException e)
	    {
		Logger.log(e);
	    }
	    catch(MissingCapabilityException e)
	    {
		Logger.log(e);
	    }
	    ec.getSource().recycleCapsule(ec);
	    break;
	case METHOD_REM_NEIGHBOR:
	    try
	    {
		NeighborImpl ni;

		ni = (NeighborImpl)ec.getReferences()[0].dereference();
		this.removeNeighbor(node, ec, ni);
	    }
	    catch(MissingCapabilityException e)
	    {
		Logger.log(e);
	    }
	    ec.getSource().recycleCapsule(ec);
	    break;
	default:
	    this.userReceive(node, method, ec);
	    break;
	}
    }

    protected void userReceive(Node node, int method, ExternalCapsule ec)
    {
	try
	{
	    ec.getSource().recycleCapsule(ec);
	}
	catch(Throwable th)
	{
	    Logger.log(th);
	}
    }
    
    public String toString()
    {
	return "ExtendedProtocolSession["
	    + super.toString()
	    + "; interests="
	    + this.interests
	    + "; neighbors()="
	    + this.neighbors
	    + "]";
    }
}
