/*
 * ExternalCoroutine.java
 *
 * Copyright (c) 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is part of the Janos Bees Software.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */

package bees.core;

import java.util.NoSuchElementException;

import bees.core.resource.Logger;

import bees.core.bdomain.BDomain;
import bees.core.bdomain.DeadBDomainException;

/**
 * An ExternalCoroutine is used to send ExternalCapsules between BDomains on
 * the same node.
 */
public final class ExternalCoroutine
    extends Coroutine
    implements ExternalChannel
{
    /**
     * The Coroutine is attempting to connect to its peer.
     */
    public static final int STATE_CONNECTING = STATE_USER;

    /**
     * The Coroutine is connected to its peer.
     */
    public static final int STATE_CONNECTED = STATE_USER + 1;

    /**
     * The event used to send ExternalCapsules to the peer coroutine.
     */
    private final ExternalCapsuleEvent ce = new ExternalCapsuleEvent();

    /**
     * The EventChannel that is connected to the peer BDomain.
     */
    private EventChannel ec;

    /**
     * The SessionType this coroutine is bound to.
     */
    private SessionType st;

    /**
     * The list of workers available for processing incoming ExternalCapsules.
     */
    private WorkerList wl;

    /**
     * The peer coroutine's identifier.
     */
    private long peerID;

    /**
     * An opaque pointer for user code.
     */
    private Object local;

    /**
     * Construct an empty ExternalCoroutine object.
     */
    public ExternalCoroutine()
    {
    }

    public void setLocal(Object obj)
    {
	this.local = obj;
    }

    public Object getLocal()
    {
	return this.local;
    }

    /**
     * @return The SessionType this coroutine is bound to.
     */
    SessionType getSessionType()
    {
	return this.st;
    }

    /**
     * @return The EventChannel that is connected to the peer BDomain.
     */
    EventChannel getChannel()
    {
	return this.ec;
    }

    /**
     * Wait for the coroutine to be connected to its peer.
     *
     * @return True if the coroutine was successfully connected to a peer.
     */
    public boolean waitForConnection()
    {
	boolean retval = false;

	if( this.waitForState(STATE_CONNECTED) == STATE_CONNECTED )
	{
	    retval = true;
	}
	return retval;
    }

    /**
     * Add more worker threads for processing incoming ExternalCapsules.
     *
     * @param ecr The object that should be notified when new capsules arrive.
     * @param count The number of worker threads to create.
     */
    private void addWorkers(final ExternalCapsuleReceiver ecr, int count)
    {
	Node node = Node.currentNode();
	int lpc;

	for( lpc = 0; lpc < count; lpc++ )
	{
	    node.systemThreadStart(new WorkerThread(node) {
		    public void run()
		    {
			try
			{
			    ExternalCapsule ec;
			    
			    while( (ec = (ExternalCapsule)
				    ExternalCoroutine.this.wl.getWork())
				   != null )
			    {
				ecr.receive(this.node, ec);
			    }
			}
			catch(InterruptedException ie)
			{
			    /* Dying... */
			}
			catch(Throwable th)
			{
			    Logger.log(th);
			}
			Logger.log("External worker dying: "
				   + Thread.currentThread().getName());
		    }

		    public String toString()
		    {
			return "ExternalCoroutineWorker["
			    + super.toString()
			    + "]";
		    }
		},
				   "External Worker: "
				   + BDomain.currentBDomain().getName());
	}
    }

    /**
     * Connect this domain/coroutine to a peer BDomain.
     *
     * @param pp The ProtocolPresentation that will process incoming capsules.
     * @param ec The channel used to send events to the peer BDomain.
     * @param st The session type this coroutine should be bound to.
     */
    public synchronized void connect(ProtocolPresentation pp,
				     EventChannel ec,
				     SessionType st)
    {
	try
	{
	    ConnectSessionEvent cse;

	    Assert.require(pp);
	    Assert.require(ec);
	    Assert.require(st);
	    
	    this.st = st;
	    cse = new ConnectSessionEvent(this,
					  EventChannel.currentChannel(),
					  st.getSessionID());
	    cse.useBDomainReplyPort();
	    this.wl = new WorkerList(5);
	    this.addWorkers(pp, 1);
	    this.ec = ec;
	    ec.sendEvent(cse);
	    this.setState(STATE_CONNECTING);
	}
	catch(DeadBDomainException e)
	{
	    this.setState(STATE_ERROR);
	}
	catch(Throwable th)
	{
	    super.setThrowable(th);
	}
    }

    /**
     * Disconnect this coroutine from its peer.
     */
    public synchronized void disconnect()
    {
	try
	{
	    DisconnectSessionEvent cse;

	    cse = new DisconnectSessionEvent(this);
	    cse.useBDomainReplyPort();
	    cse.setServerCoroutineID(this.peerID);
	    this.ec.sendEvent(cse);
	    this.setState(STATE_COMPLETE);
	}
	catch(DeadBDomainException e)
	{
	    this.setState(STATE_ERROR);
	}
	catch(Throwable th)
	{
	    super.setThrowable(th);
	}
    }

    public void disconnected()
    {
    }

    /**
     * Send a capsule to the peer coroutine.
     *
     * @param cap The capsule to send.
     */
    public synchronized void sendCapsule(ExternalCapsule cap)
    {
	switch( this.getState() )
	{
	case STATE_CONNECTED:
	    try
	    {
		this.ce.setBody(cap.getBuffer());
		this.ce.setEnvelopes(cap.getEnvelopes());
		this.ce.setReferences(cap.getReferences());
		// ce.setClientCoroutineID(this.getID());
		this.ce.setServerCoroutineID(this.peerID);
		this.ec.sendEvent(this.ce);
	    }
	    catch(DeadBDomainException e)
	    {
		this.setState(STATE_ERROR);
	    }
	    catch(Throwable th)
	    {
		th.printStackTrace();
		super.setThrowable(th);
	    }
	    break;
	default:
	    Strict.fail();
	    break;
	}
    }

    public synchronized void recycleCapsule(ExternalCapsule ec)
    {
    }

    public synchronized void cleanup()
    {
	this.disconnect();

	super.cleanup();
    }

    public synchronized void drop()
    {
	super.drop();

	if( this.wl != null )
	{
	    this.wl.kill();
	}
    }

    public synchronized void handle(BeesEvent ae)
    {
	switch( this.getState() )
	{
	case STATE_NEWBORN:
	    if( (ae instanceof ConnectSessionEvent) && !ae.isReply() )
	    {
		ConnectSessionEvent cse = (ConnectSessionEvent)ae;

		try
		{
		    if( (this.st = SessionType.
			 getBDomainSession(cse.getSessionID())) != null )
		    {
			ProtocolSession ps;
			
			ps = st.getSession();
			this.wl = new WorkerList(5);
			this.addWorkers(ps, 1);
			this.peerID = ae.getClientCoroutineID();
			/*
			 * XXX Should we catch the missing capability exception
			 * here?
			 */
			this.ec = cse.getEventChannel();
			this.setState(STATE_CONNECTED);
		    }
		    else
		    {
			Logger.log("Invalid session ID: "
				   + cse.getSessionID());
			throw new NoSuchElementException(); // XXX
		    }
		}
		catch(Throwable th)
		{
		    Logger.log(th);
		    cse.setThrowable(th);
		    super.setThrowable(th);
		}
		try
		{
		    cse.sendReply();
		}
		catch(DeadBDomainException e)
		{
		    super.setThrowable(e);
		}
	    }
	    else if( (ae instanceof DisconnectSessionEvent) && !ae.isReply() )
	    {
		this.setState(STATE_COMPLETE);
	    }
	    else
	    {
		Strict.fail();
	    }
	    break;
	case STATE_CONNECTING:
	    if( (ae instanceof ConnectSessionEvent) && ae.isReply() )
	    {
		ConnectSessionEvent cse = (ConnectSessionEvent)ae;
		Throwable th;

		if( (th = cse.getThrowable()) != null )
		{
		    super.setThrowable(th);
		}
		else
		{
		    this.peerID = ae.getServerCoroutineID();
		    this.setState(STATE_CONNECTED);
		}
	    }
	    else
	    {
		Strict.fail();
	    }
	    break;
	case STATE_CONNECTED:
	    if( (ae instanceof ExternalCapsuleEvent) && !ae.isReply() )
	    {
		ExternalCapsuleEvent ece = (ExternalCapsuleEvent)ae;
		ExternalCapsule ec;
		
		ec = new ExternalCapsule(ece.getEnvelopes(),
					 ece.getReferences());
		ec.fromEvent(this, ece);
		this.wl.addWork(ec);
	    }
	    else if( (ae instanceof DisconnectSessionEvent) && !ae.isReply() )
	    {
		this.setState(STATE_COMPLETE);
		this.disconnected();
	    }
	    else
	    {
		Strict.fail("Unexecpted event: " + ae);
	    }
	    break;
	default:
	    throw new Panic("Invalid state for coroutine: " + this);
	}
    }

    public String toString()
    {
	return "ExternalCoroutine["
	    + super.toString()
	    + "; ec="
	    + this.ec
	    + "; st="
	    + this.st
	    + "; wl="
	    + this.wl
	    + "; peerID="
	    + this.peerID
	    + "]";
    }
}
