/*
 * FlowType.java
 *
 * Copyright (c) 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is part of the Janos Bees Software.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */

package bees.core;

import java.util.Vector;
import java.util.Hashtable;

import bees.core.kernel.KernelPanic;

import bees.core.bdomain.BDomainLocal;

import bees.util.MD5;

/**
 * A FlowType describes a Bees network flow, which contains, a primary
 * protocol, its companion protocols, and an untyped set of user values.
 */
final class FlowType
{
    private static final BDomainLocal domainFlows = new BDomainLocal() {
	    protected Object initialValue()
	    {
		return new Hashtable();
	    }
	};

    static Hashtable getBDomainFlows()
    {
	return (Hashtable)domainFlows.get();
    }

    static void putBDomainFlow(FlowType ft)
    {
	Hashtable ht = (Hashtable)domainFlows.get();

	Assert.require(ft.flowID);

	synchronized( ht )
	{
	    if( !ht.containsKey(ft.flowID) )
		ht.put(ft.flowID, ft);
	}
    }
    
    static void putBDomainFlow(ProtocolType pt)
    {
	Hashtable ht = (Hashtable)domainFlows.get();
	
	synchronized( ht )
	{
	    if( !ht.containsKey(pt.getProtocolID()) )
	    {
		FlowType ft;
		
		ft = new FlowType();
		ft.setPrimaryProtocolType(pt);
		ft.setFlowID(pt.getProtocolID());
		ht.put(pt.getProtocolID(), ft);
	    }
	}
    }

    static FlowType getBDomainFlow(TypeID flowID)
    {
	Hashtable ht = (Hashtable)domainFlows.get();
	FlowType retval;

	retval = (FlowType)ht.get(flowID);
	return retval;
    }

    /**
     * The unique flow identifier.
     */
    private TypeID flowID;

    private final PairChain pc;
    private ProtocolType primaryProtocolType;
    private final Vector companions = new Vector();

    FlowType(PairChain pc)
    {
	this.pc = pc;
    }

    FlowType()
    {
	this(new PairChain());
    }

    PairChain getPairChain()
    {
	return this.pc;
    }
    
    FlowType setPrimaryProtocolType(ProtocolType pt)
    {
	this.primaryProtocolType = pt;
	this.pc.add(ProtocolFlow.PRIMARY_PROTOCOL, pt.toBytes());
	return this;
    }

    ProtocolType getPrimaryProtocolType()
    {
	return this.primaryProtocolType;
    }

    synchronized FlowType addCompanion(ProtocolType ct)
    {
	this.companions.addElement(ct);
	this.pc.add(ProtocolFlow.COMPANION_PROTOCOL, ct.toBytes());
	return this;
    }

    Vector getCompanions()
    {
	return this.companions;
    }

    FlowType updatePrimaryProtocolType(ProtocolType pt)
    {
	if( !this.primaryProtocolType.getProtocolID().
	    equals(pt.getProtocolID()) )
	{
	    throw new SecurityException();
	}
	this.primaryProtocolType = pt;
	return this;
    }

    FlowType setFlowID(TypeID flowID)
    {
	this.flowID = flowID;
	return this;
    }

    TypeID getFlowID()
    {
	return this.flowID;
    }

    void snapshotFlowType()
    {
	MD5 digest = new MD5();
	TypeID computedFlowID;
	int lpc, size;
	Pair pairs[];

	pairs = this.pc.toArray();
	digest.update(this.pc.toBytes());
	computedFlowID = new TypeID(digest.hash());
	if( this.flowID != null )
	{
	    if( !this.flowID.equals(computedFlowID) )
	    {
		throw new SecurityException("Computed FlowID "
					    + computedFlowID
					    + " != "
					    + this.flowID
					    + " for "
					    + this);
	    }
	}
	else
	{
	    this.flowID = computedFlowID;
	}
    }

    synchronized int marshalledLength()
    {
	return this.pc.marshalledLength();
    }

    void marshal(XdrByteArray xdr)
    {
	this.pc.marshal(xdr);
    }

    void unmarshal(XdrByteArray xdr)
    {
	Pair pair;
	int lpc;
	
	this.pc.unmarshal(xdr);
	this.primaryProtocolType =
	    ProtocolType.instantiate(this.pc.get(ProtocolFlow.
						 PRIMARY_PROTOCOL).
				     getValue());
	this.snapshotFlowType();
	for( lpc = 0;
	     (pair = this.pc.get(ProtocolFlow.COMPANION_PROTOCOL,
				 lpc)) != null;
	     lpc++ )
	{
	    ProtocolType pt = ProtocolType.instantiate(pair.getValue());

	    this.companions.addElement(pt);
	}
    }

    byte[] toBytes()
    {
	XdrByteArray xdr;

	xdr = new XdrByteArray(this.marshalledLength());
	this.marshal(xdr);
	if( xdr.index != xdr.len )
	{
	    new Panic("marshalledLength() doesn't match actual length");
	}
	return xdr.toBytes();
    }

    static FlowType instantiate(byte flattened[])
    {
	XdrByteArray xba;
	FlowType retval;

	retval = new FlowType();
	xba = new XdrByteArray(flattened);
	retval.unmarshal(xba);
	if( xba.index != xba.len )
	{
	    throw new IllegalArgumentException("Flattened FlowType is short");
	}
	return retval;
    }

    public String toString()
    {
	return "FlowType[pc="
	    + this.pc
	    + "; primaryProtocolType="
	    + this.primaryProtocolType
	    + "; companions="
	    + this.companions
	    + "; flowID="
	    + this.flowID
	    + "]";
    }
}
