/*
 * LocalFlowCoroutine.java
 *
 * Copyright (c) 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is part of the Janos Bees Software.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */

package bees.core;

import java.io.IOException;

import edu.utah.janos.nodeos.Domain;

import bees.core.kernel.KernelPanic;

import bees.core.resource.Logger;

import bees.core.bdomain.BDomain;
import bees.core.bdomain.DeadBDomainException;

public final class LocalFlowCoroutine
    extends Coroutine
{
    public static final int STATE_PREPARE_TO_LOAD = STATE_USER;
    
    private final EventChannel ec;
    
    public LocalFlowCoroutine(EventChannel ec)
    {
	this.ec = ec;
    }

    public LocalFlowCoroutine()
    {
	this(null);
    }

    public synchronized void loadFlow(byte pfBytes[])
    {
	try
	{
	    LocalFlowEvent lfe;

	    Assert.require(pfBytes);
	    Assert.require(this.ec);

	    lfe = new LocalFlowEvent(this, pfBytes);
	    lfe.useBDomainReplyPort();
	    this.ec.sendEvent(lfe);
	    this.setState(STATE_PREPARE_TO_LOAD);
	}
	catch(DeadBDomainException e)
	{
	    this.setState(STATE_ERROR_CAGED);
	}
	catch(Throwable th)
	{
	    super.setThrowable(th);
	}
    }

    public void loadFlow(ProtocolFlow pf)
    {
	Assert.require(pf);
	
	this.loadFlow(pf.toBytes());
    }

    public synchronized void handle(BeesEvent ae)
    {
	switch( this.getState() )
	{
	case STATE_NEWBORN:
	    if( (ae instanceof LocalFlowEvent) && !ae.isReply() )
	    {
		LocalFlowEvent lfe = (LocalFlowEvent)ae;
		
		try
		{
		    Domain domain = Domain.currentDomain();
		    ProtocolFlow pf;
		    FlowType ft;
		    
		    pf = new ProtocolFlow();
		    pf.unmarshal(new XdrByteArray(lfe.getProtocolFlow()));
		    // XXX should we?
		    domain.exportClass(pf.getPrimaryProtocol());
		    ft = pf.getFlowType();
		    if( FlowType.getBDomainFlow(ft.getFlowID()) == null )
		    {
			String bootStrapClassName;
			ObjectTree capabilities;
			Class cl;
			
			Logger.log("Starting up protocol: "
				   + pf.getPrimaryProtocol()
				   + "; flow="
				   + ft.getFlowID());
			
			capabilities = BDomain.currentBDomain().getCapabilities();
			ProtocolBDomain.createCapsuleTypes(ft, capabilities);
			if( (bootStrapClassName = ft.getPrimaryProtocolType().
			     getBootStrapClassName()) != null )
			{
			    Capsule cap;
			    
			    cl = domain.loadClass(bootStrapClassName);
			    cap = (Capsule)cl.newInstance();
			    cap.fromFlowType(ft, CapsuleType.NULL_SPECIALIZER);
			    cap.evaluate(bees.core.Node.currentNode());
			}
		    }
		    else
		    {
			Logger.log("Already started protocol: "
				   + pf.getPrimaryProtocol());
		    }
		    this.setState(STATE_COMPLETE);
		}
		catch(Throwable th)
		{
		    Logger.log(th);
		    lfe.setThrowable(th);
		}
		try
		{
		    lfe.sendReply();
		}
		catch(DeadBDomainException e)
		{
		}
	    }
	    else
	    {
		Strict.fail();
	    }
	    break;
	case STATE_PREPARE_TO_LOAD:
	    if( (ae instanceof LocalFlowEvent) && ae.isReply() )
	    {
		LocalFlowEvent lfe = (LocalFlowEvent)ae;
		Throwable th;
		
		if( (th = lfe.getThrowable()) != null )
		{
		    super.setThrowable(th);
		}
		else
		{
		    this.setState(STATE_COMPLETE);
		}
	    }
	    else
	    {
		Strict.fail();
	    }
	    break;
	default:
	    throw new KernelPanic("Invalid state for coroutine: " + this);
	}
    }
    
    public String toString()
    {
	return "LocalFlowCoroutine["
	    + super.toString()
	    + "; ec="
	    + this.ec
	    + "]";
    }
}
