/*
 * NeighborImpl.java
 *
 * Copyright (c) 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is part of the Janos Bees Software.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */

package bees.core;

import java.util.Hashtable;

import java.io.IOException;

import java.security.InvalidKeyException;
import java.security.NoSuchAlgorithmException;

import edu.utah.janos.nodeos.BufferHandle;

import bees.core.security.Permission;
import bees.core.security.PermissionSet;

/**
 * NeighborImpl is the base for classes that wish to implement neighbor
 * capabilities.  These capabilities can then be combined with Neighbor objects
 * to send and receive Capsules of a certain type.
 * <p>
 * Implementations are expected to capture the parameters for connecting to the
 * neighbor (e.g. IP address, UDP port) in this object.  Setting up capsule
 * senders and receivers is done by implementing the abstract methods in this
 * class which will be called by the Neighbor and Neighborhood classes.
 *
 * @see Neighbor
 * @see Neighborhood
 */
public abstract class NeighborImpl
    extends Capability
{
    /**
     * Permission to receive capsules from the neighbor.
     */
    public static final Permission CAPSULE__READ_PERMISSION =
	new Permission("capsule", "read");
    
    /**
     * Permission to send capsules to the neighbor.
     */
    public static final Permission CAPSULE__WRITE_PERMISSION =
	new Permission("capsule", "write");

    /**
     * The default permissions for a newly created NeighborImpl object.
     */
    public static final PermissionSet DEFAULT_PERMISSIONS =
	new PermissionSet(new Permission[] {
	    new Permission("capsule", "read, write")
	});

    /**
     * Map of protocol SessionTypes to Neighbor objects.
     */
    private final Hashtable neighbors = new Hashtable();

    /**
     * Remove the Neighbor for the given SessionType.
     *
     * @param st The SessionType whose Neighbor should be removed.
     * @return The Neighbor object that was created for the given SessionType
     * or null if it was previously removed.
     */
    synchronized Neighbor removeNeighbor(SessionType st)
    {
	return (Neighbor)this.neighbors.remove(st);
    }

    /**
     * Put a SessionType to Neighbor mapping.
     *
     * @param st The SessionType to be mapped.
     * @param neighbor The Neighbor the mapping should resolve to.
     */
    synchronized void putNeighbor(SessionType st, Neighbor neighbor)
    {
	this.neighbors.put(st, neighbor);
    }

    /**
     * Get a SessionType to Neighbor mapping.
     *
     * @param st The SessionType to be mapped.
     * @return The Neighbor the mapping should resolve to or null if there
     * is no mapping.
     */
    synchronized Neighbor getNeighbor(SessionType st)
    {
	return (Neighbor)this.neighbors.get(st);
    }

    /**
     * Get a ProtocolSession to Neighbor mapping.
     *
     * @param ps The ProtocolSession to be mapped.
     * @return The Neighbor the mapping should resolve to or null if there
     * is no mapping.
     */
    public Neighbor getNeighbor(ProtocolSession ps)
    {
	return this.getNeighbor(ps.getSessionType());
    }

    /**
     * Get an ExternalCapsule to Neighbor mapping.
     *
     * @param ps The ExternalCapsule to be mapped.
     * @return The Neighbor the mapping should resolve to or null if there
     * is no mapping.
     */
    public Neighbor getNeighbor(ExternalCapsule ec)
    {
	return this.getNeighbor(ec.getSessionType());
    }

    /**
     * Call back to be implemented by the class that will receive Capsules.
     * This method is used to setup the connection between the implementation
     * specific mechanism for receiving packets and the Bees mechanism for
     * decoding and dispatching them.
     *
     * @param cr The class that will receive and dispatch a type Capsule.
     * @throws IOException if there is a problem attaching the receiver.
     */
    protected abstract void receiveBuffer(CapsuleReceiver cr)
	throws IOException;
    
    /**
     * Call back to be implemented by the class that will receive Capsules.
     * This method is used to tear down the implementation specific receiver
     * for the given Capsule type.
     *
     * @param neighbor The Neighbor object that is being torn down.  This
     * should be matched to the Neighbor object held by the CapsuleReceiver
     * given in receiveBuffer.
     * @param capsuleTypeID The Capsule type whose receiver should be detached.
     */
    protected abstract void cancelReceivers(Neighbor neighbor,
					    TypeID capsuleTypeID);

    /**
     * Call back to be implemented by the class that will send Capsules.
     * This method is used to transmit an encoded buffer to the neighbor.
     *
     * @param bh The jnodeos BufferHandle to be transmitted.
     * @throws IOException if there is a problem sending the given buffer.
     */
    protected abstract void sendBuffer(BufferHandle bh)
	throws IOException;

    protected abstract void setKey(java.security.Key sk, String algorithm)
	throws NoSuchAlgorithmException,
	       InvalidKeyException;
    
    /**
     * Call back to be implemented by the class that will send and receive
     * Capsules.
     *
     * @param cr The class that will receive and dispatch a type Capsule.
     * @param sk The SecretKey to use when computing/verifying a MAC.
     * @param algorithm The MAC algorithm to use.
     * @throws NoSuchAlgorithmException if the MAC algorithm is not known.
     * @throws InvalidKeyException if the key is invalid.
     */
    protected abstract void enqueueKey(java.security.Key sk,
				       String algorithm)
	throws NoSuchAlgorithmException,
	       InvalidKeyException;

    /**
     *
     */
    protected abstract void dequeueKey();

    /**
     * @return The name of the interface this neighbor uses to communicate with
     * the network.
     */
    public abstract String getInterfaceName();

    /**
     * @return The FingerPrint for this neighbor.  This value should be based
     * on the data used to construct the connection between two neighbors.
     */
    public abstract TypeID getFingerPrint();

    public String toString()
    {
	return "NeighborImpl["
	    + super.toString()
	    + "]";
    }
}
