/*
 * NodeCacheCollector.java
 *
 * Copyright (c) 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is part of the Janos Bees Software.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */

package bees.core;

import java.util.Vector;

import edu.utah.janos.nodeos.Domain;

import bees.core.resource.Logger;

import bees.core.bdomain.BDomain;

final class NodeCacheCollector
    implements Runnable
{
    private Node owner;
    private boolean keepCollecting = true;
    private final Vector caches = new Vector();
    
    NodeCacheCollector(Node owner)
    {
	this.owner = owner;
	BDomain.currentBDomain().pushCleanupHandler(new Runnable() {
		public void run()
		{
		    synchronized( NodeCacheCollector.this )
		    {
			NodeCacheCollector.this.notifyAll();
		    }
		}
	    });
    }

    void add(NodeCache nc)
    {
	synchronized( this.caches )
	{
	    this.caches.addElement(nc);
	}
    }
    
    public void run()
    {
	long minTime = Long.MAX_VALUE, granularity;
	Domain domain = Domain.currentDomain();

	granularity = this.owner.getClock().getGranularity();
	while( !domain.destroyed() )
	{
	    long currentTime;
	    int lpc, size;
	    
	    /* Garbage collect, and then wait some more... */
	    currentTime = this.owner.time();
	    minTime = Long.MAX_VALUE;
	    synchronized( this.caches )
	    {
		size = this.caches.size();
		for( lpc = 0; lpc < size; lpc++ )
		{
		    NodeCache nc;
		    
		    nc = (NodeCache)this.caches.elementAt(lpc);
		    try
		    {
			minTime = Math.min(minTime, nc.clean(currentTime));
		    }
		    catch(ThreadDeath td)
		    {
			/* Dying... */
		    }
		    catch(Throwable th)
		    {
			Logger.log(th);
		    }
		}
	    }
	    if( minTime < Long.MAX_VALUE )
	    {
		minTime += granularity - 1;
		minTime -= (minTime % granularity);
	    }
	    /* Wait for a while... */
	    try
	    {
		synchronized(this)
		{
		    long start;
		    
		    /*
		     * If there aren't any elements we can
		     * just wait for forever, and let put()
		     * wake us up.  Otherwise, we wait
		     * until the next association expires.
		     */
		    start = this.owner.time();
		    this.wait(minTime);
		}
	    }
	    catch( InterruptedException e )
	    {
		;
	    }
	}
    }

    public String toString()
    {
	return "NodeCacheCollector[keepCollecting="
	    + this.keepCollecting
	    + "; caches="
	    + this.caches
	    + "]";
    }
}
