/*
 * Packet.java
 *
 * Copyright (c) 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is part of the Janos Bees Software.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */

package bees.core;

import bees.core.security.PermissionException;

/**
 * Base class for packets received by the node.
 */
abstract class Packet
{
    /**
     * The source of the packet.
     */
    private Neighbor source;

    /**
     * The set of permissions associated with this packet.
     */
    PacketAuthorization pa;

    /**
     * Construct an empty Packet.
     */
    protected Packet()
    {
    }

    /**
     * @param source The source of the packet.
     * @return this
     */
    Packet setSource(Neighbor source)
    {
	this.source = source;
	return this;
    }

    /**
     * @return The source of the packet.
     */
    public final Neighbor getSource()
    {
	return this.source;
    }

    /**
     * @param pa The set of permissions associated with this packet.
     * @return this
     */
    Packet setPacketAuthorization(PacketAuthorization pa)
    {
	this.pa = pa;
	return this;
    }

    /**
     * @return The set of permissions associated with this packet.
     */
    PacketAuthorization getPacketAuthorization()
    {
	return this.pa;
    }

    /**
     * Replenish the permissions of this packet with the authorization of
     * some external source.
     */
    void fromExternal()
    {
	this.source = null;
	this.pa.replenish();
    }
    
    /**
     * Replenish the authorization for this packet since it was received from
     * the given neighbor.
     *
     * @param neighbor The source of this packet.
     */
    void fromNeighbor(Neighbor source)
    {
	this.source = source;
	this.pa.replenish();
    }

    /**
     * Attach capabilities to this packet based on the authorization of the
     * given packet.
     *
     * @param packet The packet to use as authorization.
     */
    synchronized void fromPacket(Packet packet)
    {
	this.source = packet.source;
	this.pa.replenish();
    }

    /**
     * Reset any cached values so the capsule can be reused instead of left
     * to garbage collector.
     */
    protected void internalReset()
    {
	this.source = null;
    }

    /**
     * Decode the contents xdr into this packet object.
     *
     * @param xdr The Xdr containing the packet.
     */
    abstract void internalDecode(Xdr xdr)
	throws InvalidPacketException, DroppedPacketException;

    /**
     * Encode the contents of this packet object into the xdr.
     *
     * @param xdr The Xdr containing the packet.
     */
    abstract void internalEncode(Xdr xdr);

    public Object clone()
	throws CloneNotSupportedException
    {
	Object retval = super.clone();
	Packet packet;

	packet = (Packet)retval;
	/*
	 * XXX I'm not sure what to do here...  I think in the old Bees a clone
	 * would result in a Capsule object with the same TTL number,
	 * resulting in unbridled usage.  But here, they would just share
	 * the PacketAuthorization object so theres no distinction between
	 * the packet object's so its still controlled.
	 */
	this.source = null;
	return retval;
    }
    
    public String toString()
    {
	return "Packet[source="
	    + this.source
	    + "; pa="
	    + this.pa
	    + "]";
    }
}
