/*
 * Pair.java
 *
 * Copyright (c) 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is part of the Janos Bees Software.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */

package bees.core;

/**
 * A Pair is used to store a simple mapping between a name and byte array
 * value.  The class also provides a number of convenience functions for
 * converting the byte array value to a String, int, float, etc...  Grouping
 * Pairs together is done with a PairChain.
 *
 * @see PairChain
 */
public final class Pair
    implements Cloneable
{
    /**
     * The pair name.
     */
    private final String tag;

    /**
     * The pair value.
     */
    private final byte value[];

    /**
     * Construct a Pair with the given values.
     *
     * @param tag The pair name.
     * @param value The pair value.
     */
    public Pair(String tag, byte value[])
    {
	Assert.require(tag);
	Assert.require(value);
	
	this.tag = tag;
	this.value = value;
    }

    /**
     * Construct a Pair with the given values.
     *
     * @param tag The pair name.
     * @param value The pair value.
     */
    public Pair(String tag, String value)
    {
	this(tag, value.getBytes());
    }

    /**
     * @return The pair name.
     */
    public String getTag()
    {
	return this.tag;
    }

    /**
     * @return The pair value.
     */
    public byte[] getValue()
    {
	return this.value;
    }

    /**
     * @return The pair value as a String.
     */
    public String getStringValue()
    {
	return new String(this.value);
    }

    /**
     * @return The number of bytes needed to flatten this Pair.
     */
    public int marshalledLength()
    {
	return Xdr.STRING(this.tag) + Xdr.INT + this.value.length;
    }

    public void marshal(Xdr xdr)
    {
	Assert.require(xdr);
	
	xdr.PUT(this.tag);
	xdr.PUT(this.value.length);
	xdr.PUT(this.value);
    }

    public static Pair unmarshal(Xdr xdr)
    {
	Assert.require(xdr);
	
	return new Pair(xdr.STRING(), xdr.BYTES(xdr.INT()));
    }

    /**
     * @param xdr The destination for the flattened version of this Pair.
     */
    public void marshal(XdrByteArray xdr)
    {
	Assert.require(xdr);
	
	xdr.PUT(this.tag);
	xdr.PUT(this.value.length);
	xdr.PUT(this.value);
    }

    /**
     * @param xdr The source for the flattened version of a Pair.
     * @return A new Pair object initialized with the value in "xdr".
     */
    public static Pair unmarshal(XdrByteArray xdr)
    {
	Assert.require(xdr);
	
	return new Pair(xdr.STRING(), xdr.BYTES(xdr.INT()));
    }

    /**
     * @return A flattened version of this pair.
     */
    public byte[] toBytes()
    {
	XdrByteArray xdr;

	xdr = new XdrByteArray(this.marshalledLength());
	this.marshal(xdr);
	return xdr.toBytes();
    }

    public int hashCode()
    {
	return this.tag.hashCode(); // XXX
    }

    public boolean equals(Object obj)
    {
	boolean retval = false;

	if( obj instanceof Pair )
	{
	    Pair pair = (Pair)obj;

	    if( this.tag.equals(pair.tag) &&
		(this.value.length == pair.value.length) )
	    {
		int lpc;

		retval = true;
		for( lpc = 0; (lpc < this.value.length) && retval; lpc++ )
		{
		    retval = (this.value[lpc] == pair.value[lpc]);
		}
	    }
	}
	return retval;
    }

    public Object clone()
    {
	return new Pair(CloneFactory.clone(this.tag),
			(byte[])this.value.clone());
    }

    public String toString()
    {
	return "Pair[tag="
	    + this.tag
	    + "; value="
	    + ByteArray.toString(this.value)
	    + "]";
    }
}
