/*
 * Protocol.java
 *
 * Copyright (c) 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is part of the Janos Bees Software.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */

package bees.core;

import java.util.Vector;
import java.util.Enumeration;

import java.io.FileNotFoundException;

import edu.utah.janos.nodeos.Domain;

import bees.core.resource.Logger;

import bees.core.security.PermissionSet;
import bees.core.security.VersionString;

/**
 * Protocol is used to describe the classes, data, and companion protocols that
 * make up an Bees protocol.  Subclasses are expected to override the
 * constructor and use the functions in this class to build the protocol
 * description.  Once the constructor has finished executing, the resulting
 * Protocol should describe a completely self contained entity that can be
 * transferred between nodes.
 * <p>
 * Actual construction of the description is done on end nodes where the
 * protocol has access to files on the local file system and capabilities that
 * may not be available elsewhere.  Besides simple byte arrays, the protocol
 * can also specify "companion" protocols, which can be used to assist the
 * protocol perform its task.  For example, the system provides the
 * "DLProtocol" and "IDTranslationProtocol" companions which take care of
 * downloading the protocol to other nodes.  Once the protocol has been
 * successfully built, it can then be used as a companion for other protocols
 * (if it implements CompanionProtocol) or it can be used as the primary
 * protocol for a ProtocolFlow.
 * <p>
 * Starting a Protocol on a node is done by registering a ProtocolFlow through
 * a Node object or by externally starting up a ProtocolBDomain with the
 * Protocol class name as the first argument.  Once the flow has been
 * instantiated, the primary protocol will receive a "boot" capsule from
 * the system and the rest is up to the implementation.
 * <p>
 * Example:
 * <pre>
 * public class ExampleProtocol
 *   extends Protocol
 * {
 *   public ExampleProtocol()
 *   {
 *     super.startProtocolDefn();
 *     {
 *       super.addCapsule("MyCapsule", new PermissionSet(new Permission[] {
 *         new CountPermission("returns", 1),
 *       }));
 *       super.setBootStrap("MyBootCapsule", new PermissionSet());
 *       super.addCode("MyHelperClass");
 *       super.setSession("MySession");
 *       super.addCompanion(new DLProtocol());
 *       super.addCompanion(new IDTranslationProtocol());
 *     }
 *     super.endProtocolDefn();
 *   }
 * }
 * </pre>

 * @see Node
 * @see ProtocolFlow
 * @see DLProtocol
 * @see IDTranslationProtocol
 * @see CompanionProtocol
 */
public class Protocol
{
    /**
     * The current Node object.
     */
    private final Node node;

    /**
     * The ProtocolType that will contain the actual description of this
     * protocol.
     */
    private final ProtocolType pt = new ProtocolType();

    /**
     * Debugging helper that indicates whether the protocol object is in
     * declaration mode.
     */
    private boolean declaring = false;

    /**
     * Construct an empty Protocol object.
     */
    protected Protocol()
    {
	this.node = Node.currentNode();

	/* Flags first. */
	if( this instanceof FlowIndependentCompanionProtocol )
	{
	    this.pt.setFlags(ProtocolType.FLAG_COMPANION_FLOW_INDEPENDENT);
	}

	/* Companion type next. */
	if( this instanceof ByteArrayCompanionProtocol )
	{
	    this.pt.setCompanionType(ProtocolType.COMPANION_BYTEARRAY);
	}
	else if( this instanceof CapsuleCompanionProtocol )
	{
	    this.pt.setCompanionType(ProtocolType.COMPANION_CAPSULE);
	}
	else if( this instanceof FlowCompanionProtocol )
	{
	    this.pt.setCompanionType(ProtocolType.COMPANION_FLOW);
	}
    }

    /**
     * @return The ProtocolType that will contain the actual description of
     * this protocol.
     */
    ProtocolType getProtocolType()
    {
	return this.pt;
    }

    /**
     * @return The current Node object.
     */
    protected final Node getNode()
    {
	return this.node;
    }

    /**
     * Start the protocol definition, this is mostly for demarking the start
     * of the definition.
     */
    protected void startProtocolDefn()
    {
	this.declaring = true;
    }

    /**
     * End the protocol definition, this will demarcate the end of the
     * definition and finalize any internal hashed identifiers.
     *
     * @throws IllegalStateException if startProtocolDefn() wasn't called
     * before calling this method.
     */
    protected void endProtocolDefn()
	throws IllegalStateException
    {
	if( !this.declaring )
	{
	    throw new IllegalStateException("No startProtocolDefn before "
					    + "endProtocolDefn");
	}
	try
	{
	    if( this.pt.getSessionClassName() == null )
		this.setSession("bees.core.ProtocolSession");
	}
	catch(ClassNotFoundException e)
	{
	    throw new Panic(e);
	}
	this.declaring = false;
	this.pt.snapshotProtocolType();
    }

    /**
     * Add a plain byte array to the protocol.
     *
     * @param byteArrayName The name of the byte array, used when addressing it
     * through the ProtocolSession.
     * @param ba The actual byte array to store.
     * @param vs The optional VersionString to use when computing the hash id
     * of "ba".
     * @throws IllegalStateException if startProtocolDefn() wasn't called
     * before calling this method.
     */
    protected void addByteArray(String byteArrayName,
				byte ba[],
				VersionString vs)
	throws IllegalStateException
    {
    	if( this.declaring )
    	{
	    Domain domain = Domain.currentDomain();
	    ByteArrayType bat = new ByteArrayType();

	    bat.setType(ByteArrayType.TYPE_OPAQUE);
	    bat.setByteArrayName(byteArrayName);
	    bat.importByteArray(ba, vs);
	    this.pt.addByteArrayType(bat);
    	}
	else
	{
	    throw new IllegalStateException(
		"No startProtocolDefn before addCode");
	}
    }

    /**
     * Add a plain byte array to the protocol.
     *
     * @param byteArrayName The name of the byte array, used when addressing it
     * through the ProtocolSession.
     * @param ba The actual byte array to store.
     * @throws IllegalStateException if startProtocolDefn() wasn't called
     * before calling this method.
     */
    protected void addByteArray(String byteArrayName, byte ba[])
	throws IllegalStateException
    {
	this.addByteArray(byteArrayName, ba, null);
    }

    /**
     * Add a file from the class path to the protocol.
     *
     * @param fileName The name of the file, used to load the file and address
     * through the ProtocolSession.
     * @throws IllegalStateException if startProtocolDefn() wasn't called
     * before calling this method.
     */
    protected void addFile(String fileName)
	throws FileNotFoundException,
	       IllegalStateException
    {
	Domain domain = Domain.currentDomain();

	this.addByteArray(fileName, domain.readClassPathFile(fileName));
    }

    /**
     * Add a class to the protocol.
     *
     * @param type The type of code, regular or capsule.
     * @param className The name of the class to add.
     * @param versionID The optional VersionString to use when computing the
     * hash id of the class.
     * @param requiredPermissions The PermissionSet describing the propagation
     * properties of Capsule classes.
     * @throws IllegalStateException if startProtocolDefn() wasn't called
     * before calling this method.
     */
    private void addCode(short type,
			 String className,
			 VersionString versionID,
			 PermissionSet requiredPermissions)
	throws ClassNotFoundException,
	       IllegalStateException
    {
	if( this.declaring )
	{
	    CodeType ct = new CodeType();
	    Domain domain;
	    
	    domain = Domain.currentDomain();
	    ct.setType(type);
	    ct.setByteArrayName(className);
	    if( requiredPermissions != null )
		ct.setPermissions(requiredPermissions);
	    if( this instanceof BuiltinProtocol )
	    {
		/*
		 * This is a builtin protocol, the code should already be
		 * available on all hosts.
		 */
		try
		{
		    ct.importClass(domain.loadClass(className), versionID);
		}
		catch(ClassNotFoundException e)
		{
		    Logger.log(e);
		    throw new Panic("Cannot load builtin class?", e);
		}
	    }
	    else
	    {
		byte byteCode[];
		
		byteCode = domain.readByteCode(className);
		ct.importClass(byteCode, versionID);
	    }
	    this.pt.addByteArrayType(ct);
	}
	else
	{
	    throw new IllegalStateException(
		"No startProtocolDefn before addCode");
	}
    }

    /**
     * Add a class to the protocol.
     *
     * @param className The name of the class to add.
     * @param versionID The optional VersionString to use when computing the
     * hash id of the class.
     * @throws IllegalStateException if startProtocolDefn() wasn't called
     * before calling this method.
     */
    protected void addCode(String className, VersionString versionID)
	throws ClassNotFoundException,
	       IllegalStateException
    {
	this.addCode(ByteArrayType.TYPE_CODE, className, versionID, null);
    }

    /**
     * Add a class to the protocol.
     *
     * @param className The name of the class to add.
     * @throws IllegalStateException if startProtocolDefn() wasn't called
     * before calling this method.
     */
    protected void addCode(String className)
	throws ClassNotFoundException,
	       IllegalStateException
    {
	this.addCode(ByteArrayType.TYPE_CODE, className, null, null);
    }

    /**
     * Add a full capsule class to the protocol.  This method differs from
     * addCode() in that these types of Capsules can be received by Neighbor
     * objects.  For example, a subclass of Capsule that is added with
     * addCode() will be available as a base class, but capsules of this type
     * will not be received from neighbors.
     *
     * <p>
     * The PermissionSet that gets passed to this method describes the
     * properties of this Capsule type.  The propagation properties, encoded
     * as CountPermission objects, are:
     *
     * <ul>
     * <li>initial - Number of times a Capsule initialized from an
     * ExternalCapsule can be sent.
     * <li>returns - Number of times a Capsule can be sent back to its source.
     * <li>forwards - Number of times a Capsule can be forwarded to another
     * neighbor.
     * <li>externals - Number of times a Capsule can be used to create an
     * outgoing ExternalCapsule.
     * <li>neighborhood - Number of times a Capsule can be sent to a
     * neighborhood.
     * <li>groups - Number of times a Capsule can be used to replenish a
     * CapsuleGroup.
     * <li>transfers - Number of times a Capsule can be used to replenish a
     * Capsule of the same type.
     * <li>transformTo.<capsule class> - Number of times a Capsule can be used
     * to replenish a Capsule of the type, <capsule class>.
     * <li>interval - The minimum number of milliseconds allowed for an
     * expiration time on a Capsule added to a NodeCache.  <i>Note</i>: This
     * value should be specified using a MinimumIntegerPermission and not a
     * CountPermission.
     * </ul>
     *
     * The global properties, also encoded as CountPermissions, are:
     *
     * <ul>
     * <li>cached - The total number of Capsules of this type that can be
     * placed in NodeCaches.
     * <li>grouped - The total number of Capsules of this type that can be
     * placed in CapsuleGroups.
     * <li>threadCount - The maximum number of threads that can process
     * Capsules from a single neighbor.  (This number is passed to the
     * InChannel constructor.)
     * <li>bufferCount - The number of buffers to use when receiving Capsules
     * from a neighbor.
     * </ul>
     *
     * @see Capsule
     *
     * @param className The name of the class to add.
     * @param versionID The optional VersionString to use when computing the
     * hash id of the class.
     * @param ps The PermissionSet describing the propagation properties of
     * Capsule classes.
     * @throws IllegalStateException if startProtocolDefn() wasn't called
     * before calling this method.
     */
    protected void addCapsule(String className,
			      VersionString versionID,
			      PermissionSet ps)
	throws ClassNotFoundException,
	       IllegalStateException
    {
	this.addCode(ByteArrayType.TYPE_CAPSULE, className, versionID, ps);
    }

    /**
     * Add a full capsule class to the protocol.
     *
     * @see #addCapsule(String, VersionString, PermissionSet)
     *
     * @param className The name of the class to add.
     * @param requiredPermissions The PermissionSet describing the propagation
     * properties of Capsule classes.
     * @throws IllegalStateException if startProtocolDefn() wasn't called
     * before calling this method.
     */
    protected void addCapsule(String className, PermissionSet ps)
	throws ClassNotFoundException,
	       IllegalStateException
    {
	this.addCode(ByteArrayType.TYPE_CAPSULE, className, null, ps);
    }

    /**
     * Add a companion to the protocol.
     *
     * @param protocol A companion protocol.
     * @throws IllegalStateException if startProtocolDefn() wasn't called
     * before calling this method.
     */
    protected void addCompanion(CompanionProtocol protocol)
	throws IllegalStateException
    {
	if( this.declaring )
	{
	    this.pt.addCompanion(((Protocol)protocol).pt);
	}
	else
	{
	    throw new IllegalStateException(
		"No startProtocolDefn before addCode");
	}
    }

    /**
     * Set the ProtocolSession class name.
     *
     * @param className The name of the ProtocolSession class to use.
     * @throws IllegalStateException if startProtocolDefn() wasn't called
     * before calling this method.
     */
    protected final void setSession(String className)
	throws ClassNotFoundException,
	       IllegalStateException
    {
	this.addCode(ByteArrayType.TYPE_SESSION, className, null, null);
	this.pt.setSessionClassName(className);
    }

    /**
     * Set the boot strap capsule class name.  This will add the capsule class
     * and declare it the type to be executed when a protocol is booted.
     * <i>Note</i>: Only the primary protocol in a ProtocolFlow will receive a
     * boot capsule.
     *
     * @param className The name of the class to add.
     * @param requiredPermissions The PermissionSet describing the propagation
     * properties of Capsule classes.
     * @throws IllegalStateException if startProtocolDefn() wasn't called
     * before calling this method.
     */
    protected final void setBootStrap(String className, PermissionSet ps)
	throws ClassNotFoundException,
	       IllegalStateException
    {
	this.addCapsule(className, ps);
	this.pt.setBootStrapClassName(className);
    }

    /**
     * @return The number of bytes needed to encode this protocol.
     */
    public final int getByteArrayLength()
    {
	return this.pt.marshalledLength();
    }

    /**
     * Encode this Protocol object in the given byte array.
     *
     * @param xdr An XdrByteArray large enough to hold the flattened version
     * of this Protocol.
     */
    public final void toByteArray(XdrByteArray xdr)
    {
	Assert.require(xdr);
	
	this.pt.marshal(xdr);
    }

    /**
     * @return A flattened version of this Protocol object.
     */
    public final byte[] toByteArray()
    {
	XdrByteArray xdr;

	xdr = new XdrByteArray(this.pt.marshalledLength());
	this.toByteArray(xdr);
	return xdr.toBytes();
    }

    /**
     * @see java.lang.Object#toString()
     */
    public String toString()
    {
	return "Protocol[pt="
	    + this.pt
	    + "; declaring="
	    + this.declaring
	    + "]";
    }
}
