/*
 * ProtocolFlow.java
 *
 * Copyright (c) 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is part of the Janos Bees Software.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */

package bees.core;

import java.util.Vector;

import edu.utah.janos.nodeos.Domain;

/**
 * A ProtocolFlow is used to construct a description of a group of protocols,
 * the primary and its companions, and any user specified data.
 */
public class ProtocolFlow
{
    /**
     * The tag name given to the primary protocol in the PairChain.
     */
    public static final String PRIMARY_PROTOCOL = "primary-protocol";
    
    /**
     * The tag name given to any companion protocols in the PairChain.
     */
    public static final String COMPANION_PROTOCOL = "companion-protocol";

    /**
     * The name of the primary protocol.
     */
    private String primaryProtocol;

    /**
     * A vector containing the names of the companion protocols.
     */
    private final Vector companions = new Vector();

    /**
     * The PairChain containing the protocol descriptions and user data.
     */
    private final PairChain pc = new PairChain();

    /**
     * Cached FlowType object.
     */
    private FlowType ft;

    /**
     * Construct an empty ProtocolFlow object.
     */
    public ProtocolFlow()
    {
    }

    /**
     * @param primaryProtocol The name of the primary protocol.
     * @return this
     */
    public final ProtocolFlow setPrimaryProtocol(String primaryProtocol)
    {
	this.primaryProtocol = primaryProtocol;
	return this;
    }

    /**
     * @return The name of the primary protocol.
     */
    public final String getPrimaryProtocol()
    {
	return this.primaryProtocol;
    }

    /**
     * Add a companion protocol to the flow description.
     *
     * @param name The name of the companion protocol.
     * @return this
     */
    public final ProtocolFlow addCompanion(String name)
	throws ClassNotFoundException,
	       InstantiationException,
	       IllegalAccessException
    {
	this.companions.addElement(name);
	if( this.ft != null )
	{
	    Class cl;

	    cl = Domain.currentDomain().loadClass(name);
	    this.ft.addCompanion(((Protocol)cl.newInstance()).
				 getProtocolType());
	}
	return this;
    }

    /**
     * @param pair A pair to add to the description.
     * @return this
     */
    public final ProtocolFlow addPair(Pair pair)
    {
	this.pc.add(pair);
	return this;
    }

    /**
     * @param tag The name part of the pair.
     * @param value The value part of the pair.
     * @return this
     */
    public final ProtocolFlow addPair(String tag, String value)
    {
	this.addPair(new Pair(tag, value));
	return this;
    }

    /**
     * @param tag The name part of the pair.
     * @param value The value part of the pair.
     * @return this
     */
    public final ProtocolFlow addPair(String tag, byte value[])
    {
	this.addPair(new Pair(tag, value));
	return this;
    }

    /**
     * @param tag The pair name to get.
     * @param instance The instance of the name to get.
     * @return The Pair object if it was found or NULL.
     */
    public final Pair getPair(String tag, int instance)
    {
	return this.pc.get(tag, instance);
    }

    /**
     * @param tag The pair name to get.
     * @return The Pair object if it was found or NULL.
     */
    public final Pair getPair(String tag)
    {
	return this.pc.get(tag);
    }
    
    /* XXX 
    void exportClasses()
    {
	Domain domain = Domain.currentDomain();
	int lpc, size;
	
	domain.exportClass(this.primaryProtocol);
	size = this.companions.size();
	for( lpc = 0; lpc < size; lpc++ )
	{
	    domain.exportClass((String)this.companions.elementAt(lpc));
	}
    }
    */

    /**
     * @return The marshalled length of the description.
     */
    final int marshalledLength()
    {
	int retval, lpc, size;

	retval = Xdr.STRING(this.primaryProtocol);
	size = this.companions.size();
	retval += Xdr.INT;
	for( lpc = 0; lpc < size; lpc++ )
	{
	    retval += Xdr.STRING((String)this.companions.elementAt(lpc));
	}
	retval += this.pc.marshalledLength();
	return retval;
    }

    /**
     * @param xdr The object to marshal this description to.
     */
    final void marshal(XdrByteArray xdr)
    {
	int lpc, size;
	
	xdr.PUT(this.primaryProtocol);
	size = this.companions.size();
	xdr.PUT(size);
	for( lpc = 0; lpc < size; lpc++ )
	{
	    xdr.PUT((String)this.companions.elementAt(lpc));
	}
	this.pc.marshal(xdr);
    }

    /**
     * @param xdr The object to unmarshal a description from.
     */
    final void unmarshal(XdrByteArray xdr)
    {
	int lpc, size;
	
	this.primaryProtocol = xdr.STRING();
	size = xdr.INT();
	for( lpc = 0; lpc < size; lpc++ )
	{
	    this.companions.addElement(xdr.STRING());
	}
	this.pc.unmarshal(xdr);
    }

    /**
     * @param bytes A flattened ProtocolFlow object.
     * @return A new ProtocolFlow object initialized with the values in
     * "bytes".
     */
    public final static ProtocolFlow instantiate(byte bytes[])
    {
	ProtocolFlow retval;

	retval = new ProtocolFlow();
	retval.unmarshal(new XdrByteArray(bytes));
	return retval;
    }

    /**
     * @return A flattened version of this ProtocolFlow.
     */
    public final byte[] toBytes()
    {
	XdrByteArray xdr;
	byte retval[];

	xdr = new XdrByteArray(this.marshalledLength());
	this.marshal(xdr);
	retval = xdr.toBytes();
	return retval;
    }

    /**
     * @return A FlowType constructed from the values in this description.
     */
    final synchronized FlowType getFlowType()
	throws ClassNotFoundException,
	       InstantiationException,
	       IllegalAccessException
    {
	if( this.ft == null )
	{
	    Domain domain = Domain.currentDomain();
	    ProtocolType protocolType;
	    FlowType flowType;
	    int lpc, size;
	    Class cl;
	    
	    flowType = new FlowType();
	    cl = domain.loadClass(this.primaryProtocol);
	    protocolType = ((Protocol)cl.newInstance()).getProtocolType();
	    flowType.setPrimaryProtocolType(protocolType);
	    flowType.getPairChain().from(this.pc);
	    size = this.companions.size();
	    for( lpc = 0; lpc < size; lpc++ )
	    {
		String companion = (String)this.companions.elementAt(lpc);
		
		cl = domain.loadClass(companion);
		flowType.addCompanion(((Protocol)cl.newInstance())
				      .getProtocolType());
	    }
	    flowType.snapshotFlowType();
	    this.ft = flowType;
	}
	return this.ft;
    }

    /**
     * @return A PairChain constructed from the values in this description.
     */
    public final PairChain getFlowDesc()
	throws ClassNotFoundException,
	       InstantiationException,
	       IllegalAccessException
    {
	return this.getFlowType().getPairChain();
    }

    public String toString()
    {
	return "ProtocolFlow[primaryProtocol="
	    + this.primaryProtocol
	    + "; companions="
	    + this.companions
	    + "; pc="
	    + this.pc
	    + "]";
    }
}
