/*
 * RemoteFlowCoroutine.java
 *
 * Copyright (c) 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is part of the Janos Bees Software.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */

package bees.core;

import java.io.IOException;

import bees.core.kernel.KernelPanic;

import bees.core.resource.Logger;

import bees.core.bdomain.BDomain;
import bees.core.bdomain.DeadBDomainException;

public final class RemoteFlowCoroutine
    extends Coroutine
{
    public static final int STATE_PREPARE_TO_DOWNLOAD = STATE_USER;
    public static final int STATE_DOWNLOADING = STATE_USER + 1;
    
    private final EventChannel ec;
    
    public RemoteFlowCoroutine(EventChannel ec)
    {
	this.ec = ec;
    }

    public RemoteFlowCoroutine()
    {
	this(null);
    }

    public synchronized void downloadFlow(byte flowDesc[])
    {
	BeesEvent ae = new RemoteFlowEvent(this, flowDesc);
	
	ae.useBDomainReplyPort();
	ae.setClientCoroutineID(this.getID());
	try
	{
	    this.ec.sendEvent(ae);
	    this.setState(STATE_PREPARE_TO_DOWNLOAD);
	}
	catch(DeadBDomainException e)
	{
	    this.setState(STATE_ERROR);
	}
	catch(Throwable th)
	{
	    super.setThrowable(th);
	}
    }

    public void downloadFlow(PairChain pc)
    {
	this.downloadFlow(pc.toBytes());
    }

    public synchronized void handle(BeesEvent ae)
    {
	switch( this.getState() )
	{
	case STATE_NEWBORN:
	    if( ae instanceof RemoteFlowEvent && !ae.isReply() )
	    {
		RemoteFlowEvent rfe = (RemoteFlowEvent)ae;
		
		try
		{
		    FlowType ft = new FlowType();
		    ObjectTree capabilities;
		    
		    capabilities = BDomain.currentBDomain().getCapabilities();
		    ft.unmarshal(new XdrByteArray(rfe.getFlowDesc()));
		    Logger.log("Starting up remote: " + ft.getFlowID());
		    ProtocolBDomain.createCapsuleTypes(ft, capabilities);
		    this.setState(STATE_DOWNLOADING);
		}
		catch(IOException e)
		{
		    Logger.log(e);
		}
		catch(Throwable th)
		{
		    Logger.log(th);
		    rfe.setThrowable(th);
		}
		try
		{
		    rfe.sendReply();
		}
		catch(DeadBDomainException e)
		{
		}
	    }
	    else
	    {
		Strict.fail();
	    }
	    break;
	case STATE_PREPARE_TO_DOWNLOAD:
	    if( ae instanceof RemoteFlowEvent && ae.isReply() )
	    {
		RemoteFlowEvent rfe = (RemoteFlowEvent)ae;
		Throwable th;
		
		if( (th = rfe.getThrowable()) != null )
		{
		    super.setThrowable(th);
		}
		else
		{
		    this.setState(STATE_COMPLETE);
		}
	    }
	    else
	    {
		Strict.fail();
	    }
	    break;
	default:
	    throw new KernelPanic("Invalid state for coroutine: " + this);
	}
    }
    
    public String toString()
    {
	return "RemoteFlowCoroutine["
	    + super.toString()
	    + "; ec="
	    + this.ec
	    + "]";
    }
}
