/*
 * RevokeCapabilityCoroutine.java
 *
 * Copyright (c) 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is part of the Janos Bees Software.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */

package bees.core;

import java.util.Enumeration;

import edu.utah.janosvm.sys.DeadTeamException;
import edu.utah.janosvm.sys.TeamEjectionException;

import edu.utah.janosvm.kit.comm.amiga.FaxPortHandle;

import edu.utah.janosvm.kit.comm.jsi.JSIPair;
import edu.utah.janosvm.kit.comm.jsi.JSIHandler;
import edu.utah.janosvm.kit.comm.jsi.JSIMessage;
import edu.utah.janosvm.kit.comm.jsi.JSIPropertyInfo;
import edu.utah.janosvm.kit.comm.jsi.JSIHandlerContext;
import edu.utah.janosvm.kit.comm.jsi.MissingPairException;

import bees.core.kernel.KernelPanic;
import bees.core.kernel.BDomainCoroutine;
import bees.core.kernel.KernelCapabilities;

import bees.core.security.PermissionSet;

import bees.core.bdomain.BDomain;
import bees.core.bdomain.DeadBDomainException;

public final class RevokeCapabilityCoroutine
    extends Coroutine
{
    static  {
	JSIHandlerContext.addDefaultRootHandler("revokeCapability", new JSIHandler() {
		public JSIHandler resolveSpecifier(JSIMessage jm, JSIPair jp)
		{
		    return this;
		}
		
		public void handleMessage(JSIHandlerContext jhc, JSIMessage jm)
		    throws MissingPairException,
			   DeadBDomainException
		{
		    switch( jm.getWhat() )
		    {
		    case JSIMessage.EXECUTE_PROPERTY:
			{
			    RevokeCapabilityCoroutine rcc;
			    Capability capability;
			    EventChannel ec;
			    
			    ec = (EventChannel)KernelCapabilities.getTree().
				get(jm.expect("channel"));
			    capability = (Capability)KernelCapabilities.
				getTree().get(jm.expect("capability"));
			    rcc = new RevokeCapabilityCoroutine();
			    rcc.revoke(ec, capability);
			    jm.reply("ok");
			}
			break;
		    default:
			jm.error("Unsupported operation");
			break;
		    }
		}

		public void getSuites(JSIMessage jm)
		{
		}
	    });
	
	JSIHandlerContext.addDefaultRootHandler("deleteCapability", new JSIHandler() {
		public JSIHandler resolveSpecifier(JSIMessage jm, JSIPair jp)
		{
		    return this;
		}
		
		public void handleMessage(JSIHandlerContext jhc, JSIMessage jm)
		    throws MissingPairException,
			   DeadBDomainException
		{
		    switch( jm.getWhat() )
		    {
		    case JSIMessage.EXECUTE_PROPERTY:
			{
			    RevokeCapabilityCoroutine rcc;
			    Capability capability;
			    
			    capability = (Capability)KernelCapabilities.
				getTree().get(jm.expect("name"));
			    rcc = new RevokeCapabilityCoroutine();
			    rcc.deleteObject(BDomain.currentBDomain().
					     getBDomainCoroutine(),
					     capability);
			    jm.reply("ok");
			}
			break;
		    default:
			jm.error("Unsupported operation");
			break;
		    }
		}

		public void getSuites(JSIMessage jm)
		{
		}
	    });
    }

    public RevokeCapabilityCoroutine()
    {
    }

    private synchronized void deleteObject(FaxPortHandle fp, Capability cap)
	throws DeadBDomainException
    {
	try
	{
	    RevokeCapabilityEvent rce;
	    
	    rce = new RevokeCapabilityEvent(this, cap);
	    fp.putFacsimile(rce);
	}
	catch(DeadTeamException e)
	{
	    throw new DeadBDomainException();
	}
	catch(TeamEjectionException e)
	{
	    throw new DeadBDomainException();
	}
	finally
	{
	    this.setState(STATE_COMPLETE);
	}
    }

    public synchronized void deleteObject(BDomainCoroutine sc, Capability cap)
	throws DeadBDomainException
    {
	this.deleteObject(sc.getRemotePort(), cap);
    }
    
    public synchronized void revokePermissions(EventChannel ec,
					       Capability cap,
					       PermissionSet ps)
	throws DeadBDomainException
    {
	try
	{
	    RevokeCapabilityEvent rce;
	    
	    rce = new RevokeCapabilityEvent(this, cap, ps);
	    ec.sendEvent(rce);
	}
	finally
	{
	    this.setState(STATE_COMPLETE);
	}
    }
    
    public synchronized void revoke(EventChannel ec, Capability cap)
	throws DeadBDomainException
    {
	this.revokePermissions(ec, cap, null);
    }

    public synchronized void handle(BeesEvent ae)
    {
	switch( this.getState() )
	{
	case STATE_NEWBORN:
	    if( (ae instanceof RevokeCapabilityEvent) && !ae.isReply() )
	    {
		RevokeCapabilityEvent rce;
		CapabilityTracker ct;
		Capability cap;
		
		rce = (RevokeCapabilityEvent)ae;
		ct = CapabilityTracker.current();
		if( rce.getPermissions() != null )
		{
		    if( (cap = ct.get(rce.getStoreID(),
				      rce.getCapabilityClass())) != null )
		    {
			cap.getPermissions().
			    revokePermissions(rce.getPermissions());
		    }
		}
		else if( rce.getMethod() ==
			 RevokeCapabilityEvent.METHOD_REVOKE )
		{
		    if( rce.getCapabilityClass() != null )
		    {
			if( (cap = ct.remove(rce.getStoreID(),
					     rce.getCapabilityClass()))
			    != null )
			{
			    cap.revoke();
			}
			else
			{
			    System.out.println("couldn't find cap " + rce);
			}
		    }
		    else
		    {
			Enumeration enum;

			if( (enum = ct.remove(rce.getStoreID())) != null )
			{
			    while( enum.hasMoreElements() )
			    {
				cap = (Capability)enum.nextElement();
				cap.revoke();
			    }
			}
			else
			{
			    System.out.println("Couldn't find cap " + rce);
			}
		    }
		}
		else if( rce.getMethod() ==
			 RevokeCapabilityEvent.METHOD_DELETE )
		{
		}
		else
		{
		    throw new KernelPanic("");
		}
	    }
	    else
	    {
		Strict.fail();
	    }
	    break;
	default:
	    throw new KernelPanic("");
	}
	this.drop();
    }
    
    public String toString()
    {
	return "RevokeCapabilityCoroutine["
	    + super.toString()
	    + "]";
    }
}
