/*
 * TypeID.java
 *
 * Copyright (c) 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is part of the Janos Bees Software.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */

package bees.core;

import edu.utah.janosvm.sys.CrossTeamCloneable;

import bees.util.MD5;

import bees.core.kernel.KernelPanic;

/**
 * A TypeID is used to wrap the result of an MD5 hash.
 *
 * @see bees.util.MD5
 */
public final class TypeID
    implements Cloneable, CrossTeamCloneable
{
    /**
     * The MD5 hash bytes.
     */
    private byte hash[];

    /**
     * Package only constructor used when preallocating TypeID's.
     */
    TypeID()
    {
    }

    /**
     * Construct a TypeID with the given value.
     *
     * @param hash The MD5 hash bytes to be cloned and wrapped by the object.
     */
    public TypeID(byte hash[])
    {
	if( hash.length != MD5.HASH_LENGTH )
	{
	    throw new KernelPanic("TypeID hash bytes are "
				  + hash.length
				  + " long instead of "
				  + MD5.HASH_LENGTH);
	}
	this.hash = (byte[])hash.clone();
    }

    /**
     * Set the hash value.
     * NOTE: This will copy the hash value into a new array.
     *
     * @param hash A byte array containing the hash value.
     * @param offset The offset into the array where the hash is located.
     */
    void setHash(byte hash[], int offset)
    {
	if( (hash.length - offset) < MD5.HASH_LENGTH )
	{
	    throw new KernelPanic("TypeID hash bytes are "
				  + hash.length
				  + " long instead of "
				  + MD5.HASH_LENGTH);
	}
	this.hash = new byte[MD5.HASH_LENGTH];
	System.arraycopy(hash, offset,
			 this.hash, 0,
			 MD5.HASH_LENGTH);
    }

    /**
     * Set the hash value.
     *
     * @param hash A byte array containing the hash value.
     */
    void setHash(byte hash[])
    {
	this.hash = hash;
    }

    /**
     * @return The hash value.
     */
    byte[] getHash()
    {
	return this.hash;
    }

    public byte[] cloneHash()
    {
	return (byte[])this.hash.clone();
    }

    public boolean equals(Object obj)
    {
	boolean retval = false;

	if( obj instanceof TypeID )
	{
	    TypeID ti = (TypeID)obj;
	    int lpc;

	    retval = true;
	    for( lpc = 0; retval && (lpc < MD5.HASH_LENGTH); lpc++ )
	    {
		retval = (this.hash[lpc] == ti.hash[lpc]);
	    }
	}
	return retval;
    }

    public int hashCode()
    {
	return ((hash[0] + hash[4] + hash[8] + hash[12]) << 24)
	    + ((hash[1] + hash[5] + hash[9] + hash[13]) << 16)
	    + ((hash[2] + hash[6] + hash[10] + hash[14]) << 8)
	    + (hash[3] + hash[7] + hash[11] + hash[15]);
    }

    public Object clone()
    {
	try
	{
	    Object retval = super.clone();
	    TypeID ti;
	    
	    ti = (TypeID)retval;
	    if( this.hash != null )
		ti.hash = (byte[])this.hash.clone();
	    return retval;
	}
	catch(CloneNotSupportedException e)
	{
	    throw new KernelPanic("TypeID isn't Cloneable?");
	}
    }

    public String toPlainString()
    {
	return ByteArray.toPlainString(this.hash);
    }
    
    public String toString()
    {
	return "TypeID[hash="
	    + ByteArray.toString(this.hash)
	    + "]";
    }
}
