/*
 * UnknownProtocol.java
 *
 * Copyright (c) 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is part of the Janos Bees Software.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */

package bees.core;

import java.util.Hashtable;

import bees.core.kernel.KernelPanic;

import bees.core.bdomain.BDomainLocal;

import bees.core.security.Permission;
import bees.core.security.PermissionSet;
import bees.core.security.VersionString;
import bees.core.security.CountPermission;
import bees.core.security.MinimumIntegerPermission;

/**
 * The UnknownProtocol is used to download flow descriptions from unknown
 * flows on a neighbor.
 */
public final class UnknownProtocol
    extends Protocol
    implements BuiltinProtocol,
	       FlowCompanionProtocol,
	       FlowIndependentCompanionProtocol
{
    private static final BDomainLocal domainFlowDesc = new BDomainLocal() {
	    protected Object initialValue()
	    {
		return new Hashtable();
	    }
	};

    static Hashtable getBDomainFlowDesc()
    {
	return (Hashtable)domainFlowDesc.get();
    }

    static void putFlowDesc(TypeID fid, DataAggregator da)
    {
	Hashtable ht = (Hashtable)domainFlowDesc.get();

	synchronized( ht )
	{
	    if( !ht.containsKey(fid) )
	    {
		ht.put(fid, da);
	    }
	}
    }

    static DataAggregator getFlowDesc(TypeID fid)
    {
	Hashtable ht = (Hashtable)domainFlowDesc.get();
	DataAggregator retval;

	synchronized( ht )
	{
	    if( (retval = (DataAggregator)ht.get(fid)) == null )
	    {
		FlowType ft;

		if( (ft = FlowType.getBDomainFlow(fid)) != null )
		{
		    PairChain pc;
		    byte desc[];

		    pc = ft.getPairChain();
		    desc = pc.toBytes();
		    retval = new DataAggregator(desc);
		    ht.put(fid, retval);
		}
	    }
	}
	return retval;
    }
    
    static DataAggregator getFlowDesc(TypeID fid, int totalLength)
    {
	Hashtable ht = (Hashtable)domainFlowDesc.get();
	DataAggregator retval;

	synchronized( ht )
	{
	    if( (retval = (DataAggregator)ht.get(fid)) == null )
	    {
		retval = new DataAggregator(totalLength);
		ht.put(fid, retval);
	    }
	}
	return retval;
    }

    static boolean containsFlowDesc(TypeID fid)
    {
	Hashtable ht = (Hashtable)domainFlowDesc.get();
	
	synchronized( ht )
	{
	    return ht.containsKey(fid);
	}
    }

    static void removeFlowDesc(TypeID fid)
    {
	Hashtable ht = (Hashtable)domainFlowDesc.get();
	
	synchronized( ht )
	{
	    ht.remove(fid);
	}
    }

    /**
     * Construct an UnknownProtocol description.
     */
    public UnknownProtocol()
	throws ClassNotFoundException
    {
	try
	{
	    Node node = super.getNode();
	    
	    this.startProtocolDefn();
	    {
		this.addCapsule("bees.core.FlowRequestCapsule",
				(VersionString)node.
				getSystemCapability("unknownVersion"),
				new PermissionSet(new Permission[] {
				    new CountPermission("returns", 1),
				    new CountPermission("initial", 1),
				    new CountPermission(
					"transformTo."
					+ "bees.core.FlowResponseCapsule",
					1),

				    new CountPermission("threadCount", 1),
				    new CountPermission("bufferCount", 1),
				    
				    new CountPermission("cached", 20),
				    new MinimumIntegerPermission("interval",
								 500),
				}));
		this.addCapsule("bees.core.FlowResponseCapsule",
				(VersionString)node.
				getSystemCapability("unknownVersion"),
				new PermissionSet(new Permission[] {
				    new CountPermission("returns", 1),
				    new CountPermission("externals", 1),
				    new CountPermission(
					"transformTo."
					+ "bees.core.FlowRequestCapsule",
					1),
				    
				    new CountPermission("threadCount", 1),
				    new CountPermission("bufferCount", 1),
				}));
		this.setSession("bees.core.UnknownProtocolSession");
	    }
	    this.endProtocolDefn();
	}
	catch(ClassNotFoundException e)
	{
	    throw new Panic(e);
	}
    }

    public String toString()
    {
	return "UnknownProtocol["
	    + super.toString()
	    + "]";
    }
}
