/*
 * Xdr.java
 *
 * Copyright (c) 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is part of the Janos Bees Software.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */

package bees.core;

import edu.utah.janos.nodeos.BufferHandle;

import bees.util.MD5;

/**
 * An Xdr object wraps a system packet buffer for access by untrusted code.  It
 * also provides a number of utility functions for marshalling and
 * unmarshalling data.
 */
public final class Xdr
{
    /**
     * The length of a boolean in an xdr.
     */
    public final static int BOOLEAN = 1;
    
    /**
     * The length of a byte in an xdr.
     */
    public final static int BYTE = 1;
    
    /**
     * The length of a short in an xdr.
     */
    public final static int SHORT = 2;
    
    /**
     * The length of a int in an xdr.
     */
    public final static int INT = 4;
    
    /**
     * The length of a float in an xdr.
     */
    public final static int FLOAT = 4;
    
    /**
     * The length of a double in an xdr.
     */
    public final static int DOUBLE = 8;
    
    /**
     * The length of a long in an xdr.
     */
    public final static int LONG = 8;
    
    /**
     * The length of a typeid in an xdr.
     */
    public final static int TYPEID = MD5.HASH_LENGTH;

    private BufferHandle readBuffer;
    private BufferHandle writeBuffer;
    private int cursor;
    private int remaining;

    Xdr()
    {
    }

    void reset()
    {
	this.cursor = 0;
	this.readBuffer = null;
	this.writeBuffer = null;
    }

    void writeReset(BufferHandle bh, int cursor)
    {
	this.cursor = cursor;
	this.readBuffer = null;
	this.writeBuffer = bh;
	this.remaining = bh.getLength() - cursor;
    }

    void writeReset(BufferHandle bh)
    {
	this.cursor = 0;
	this.readBuffer = null;
	this.writeBuffer = bh;
	this.remaining = bh.getLength();
    }

    void readReset(BufferHandle bh, int cursor)
    {
	this.cursor = cursor;
	this.readBuffer = bh;
	this.writeBuffer = null;
	this.remaining = bh.getLength();
    }
    
    void readReset(BufferHandle bh)
    {
	this.cursor = 0;
	this.readBuffer = bh;
	this.writeBuffer = null;
	this.remaining = bh.getLength();
    }

    public int getUnreadLength()
    {
	return this.remaining - this.cursor;
    }

    public final static int STRING(String x)
    {
	if (x == null)
	{
	    return BYTE;
	}
	else
	{
	    Assert.check(x.length() < 128);
	    
	    return BYTE + x.length();
	}
    }

    public final static int BYTES(byte x[])
    {
	if( x == null )
	{
	    return 0;
	}
	else
	{
	    return x.length;
	}
    }
    
    public final static int BYTEARRAY(ByteArray x)
    {
	if (x == null)
	{
	    return SHORT;
	}
	else
	{
	    Assert.check(x.len < ((1 << 16)-1));
	
	    return SHORT + x.len;
	}
    }

    final BufferHandle getBuffer()
    {
	BufferHandle retval = null;
	
	if( this.writeBuffer != null )
	{
	    this.writeBuffer.setLength(this.cursor);
	    retval = this.writeBuffer;
	    this.writeBuffer = null;
	}
	else if( this.readBuffer != null )
	{
	    retval = this.readBuffer;
	    this.readBuffer = null;
	}
	return retval;
    }

    final int getWriteCursor()
    {
	return this.cursor;
    }

    final BufferHandle getWriteBuffer()
    {
	return this.writeBuffer;
    }

    public final int getReadCursor()
    {
	return this.cursor;
    }

    final BufferHandle getReadBuffer()
    {
	return this.readBuffer;
    }

    public final int getLength()
    {
	return this.readBuffer.getLength();
    }

    final void REWIND(int len)
    {
	this.cursor -= len;
    }

    public final void SKIP(int len)
    {
	if( len < 0 )
	    throw new NegativeArraySizeException();
	this.cursor += len;
    }

    public final void PUT(boolean x)
    {
	if( this.remaining < BYTE )
	    throw new ArrayIndexOutOfBoundsException();
	this.writeBuffer.setByte(this.cursor, x ? (byte)0xFF : (byte)0x00);
	this.cursor += BYTE;
	this.remaining -= BYTE;
    }

    public final void PUT(byte x)
    {
	if( this.remaining < BYTE )
	    throw new ArrayIndexOutOfBoundsException();
	this.writeBuffer.setByte(this.cursor, x);
	this.cursor += BYTE;
	this.remaining -= BYTE;
    }
    
    public final void PUT(short x)
    {
	if( this.remaining < SHORT )
	    throw new ArrayIndexOutOfBoundsException();
	this.writeBuffer.setShort(this.cursor, x);
	this.cursor += SHORT;
	this.remaining -= SHORT;
    }
    
    public final void PUT(int x)
    {
	if( this.remaining < INT )
	    throw new ArrayIndexOutOfBoundsException();
	this.writeBuffer.setInt(this.cursor, x);
	this.cursor += INT;
	this.remaining -= INT;
    }

    public final void PUTLE(short x)
    {
	if( this.remaining < SHORT )
	    throw new ArrayIndexOutOfBoundsException();
	this.writeBuffer.setLEShort(this.cursor, x);
	this.cursor += SHORT;
	this.remaining -= SHORT;
    }
    
    public final void PUTLE(int x)
    {
	if( this.remaining < INT )
	    throw new ArrayIndexOutOfBoundsException();
	this.writeBuffer.setLEInt(this.cursor, x);
	this.cursor += INT;
	this.remaining -= INT;
    }

    public final void PUT(float x)
    {
	this.PUT(Float.floatToRawIntBits(x));
    }
    
    public final void PUT(double x)
    {
	this.PUT(Double.doubleToRawLongBits(x));
    }
    
    public final void PUT(long x)
    {
	if( this.remaining < LONG )
	    throw new ArrayIndexOutOfBoundsException();
	this.writeBuffer.setInt(this.cursor, (int)((x >> 32) & 0xFFFFFFFFL));
	this.cursor += INT;
	this.writeBuffer.setInt(this.cursor, (int)((x      ) & 0xFFFFFFFFL));
	this.cursor += INT;
	this.remaining -= LONG;
    }

    public final void PUTLE(long x)
    {
	if( this.remaining < LONG )
	    throw new ArrayIndexOutOfBoundsException();
	this.writeBuffer.setLEInt(this.cursor, (int)((x      ) & 0xFFFFFFFFL));
	this.cursor += INT;
	this.writeBuffer.setLEInt(this.cursor, (int)((x >> 32) & 0xFFFFFFFFL));
	this.cursor += INT;
	this.remaining -= LONG;
    }

    public final void PUT(byte x[])
    {
	if( x != null )
	{
	    if( this.remaining < x.length )
		throw new ArrayIndexOutOfBoundsException();
	    this.writeBuffer.setBytes(this.cursor,
				      x, 0,
				      x.length);
	    this.cursor += x.length;
	    this.remaining -= x.length;
	}
    }

    public final void PUT(byte x[], int offset, int length)
    {
	if( this.remaining < length )
	    throw new ArrayIndexOutOfBoundsException();
	if( x != null )
	{
	    this.writeBuffer.setBytes(this.cursor,
				      x, offset,
				      length);
	    this.cursor += length;
	    this.remaining -= length;
	}
    }

    public final void PUT(ByteArray x)
    {
	if( x != null )
	{
	    byte xBytes[] = x.toBytes();
	    
	    this.PUT(xBytes.length);
	    this.PUT(xBytes);
	}
	else
	{
	    this.PUT((int)0);
	}
    }

    public final void PUT(String x)
    {
	if( x != null )
	{
	    byte xBytes[] = x.getBytes();

	    if( xBytes.length > 256 )
	    {
		throw new IllegalArgumentException(
			"More than 256 characters in string.");
	    }
	    else
	    {
		if( this.remaining < (BYTE + xBytes.length) )
		    throw new ArrayIndexOutOfBoundsException();
		this.writeBuffer.setByte(this.cursor, (byte)xBytes.length);
		this.cursor += BYTE;
		this.writeBuffer.setBytes(this.cursor,
					  xBytes,
					  0,
					  xBytes.length);
		this.cursor += xBytes.length;
		this.remaining -= BYTE + xBytes.length;
	    }
	}
	else
	{
	    if( this.remaining < BYTE )
		throw new ArrayIndexOutOfBoundsException();
	    this.writeBuffer.setByte(this.cursor, (byte)0);
	    this.cursor += BYTE;
	    this.remaining -= BYTE;
	}
    }

    public final void PUT(TypeID x)
    {
	if( this.remaining < TYPEID )
	    throw new ArrayIndexOutOfBoundsException();
	if( x != null )
	{
	    byte xBytes[] = x.getHash();
	    
	    this.writeBuffer.setBytes(this.cursor,
				      xBytes,
				      0,
				      TYPEID);
	}
	this.cursor += TYPEID;
	this.remaining -= TYPEID;
    }

    public final boolean BOOLEAN()
    {
	boolean retval = false;

	switch( this.readBuffer.getByte(this.cursor) )
	{
	case (byte)0xFF:
	    retval = true;
	    break;
	case 0x00:
	    break;
	default:
	    throw new Error("Oops"); // XXX better throw here
	}
	this.cursor += BYTE;
	return retval;
    }

    public final byte BYTE()
    {
	byte retval;

	retval = this.readBuffer.getByte(this.cursor);
	this.cursor += BYTE;
	return retval;
    }

    public final short SHORT()
    {
	short retval;

	retval = this.readBuffer.getShort(this.cursor);
	this.cursor += SHORT;
	return retval;
    }

    public final int INT()
    {
	int retval;

	retval = this.readBuffer.getInt(this.cursor);
	this.cursor += INT;
	return retval;
    }

    public final short LESHORT()
    {
	short retval;

	retval = this.readBuffer.getLEShort(this.cursor);
	this.cursor += SHORT;
	return retval;
    }

    public final int LEINT()
    {
	int retval;

	retval = this.readBuffer.getLEInt(this.cursor);
	this.cursor += INT;
	return retval;
    }

    public final float FLOAT()
    {
	return Float.intBitsToFloat(this.INT());
    }

    public final double DOUBLE()
    {
	return Double.longBitsToDouble(this.LONG());
    }

    public final long LONG()
    {
	long retval;

	retval = ((long)this.readBuffer.getInt(this.cursor)) << 32;
	this.cursor += INT;
	retval |= ((long)this.readBuffer.getInt(this.cursor)) &
	    0x00000000FFFFFFFFL;
	this.cursor += INT;
	return retval;
    }

    public final long LELONG()
    {
	long retval;

	retval = ((long)this.readBuffer.getLEInt(this.cursor)) &
	    0x00000000FFFFFFFFL;;
	this.cursor += INT;
	retval |= ((long)this.readBuffer.getLEInt(this.cursor)) << 32;
	this.cursor += INT;
	return retval;
    }

    public final byte[] BYTES(int length)
    {
	byte retval[] = new byte[length];

	this.readBuffer.getBytes(this.cursor,
				 retval, 0,
				 length);
	this.cursor += length;
	return retval;
    }

    public final void BYTES(byte buffer[], int offset, int length)
    {
	this.readBuffer.getBytes(this.cursor,
				 buffer,
				 offset,
				 length);
	this.cursor += length;
    }

    public final XdrByteArray BYTEARRAY()
    {
	XdrByteArray retval = null;
	int length;

	length = this.INT();
	if( length > 0 )
	{
	    byte bytes[];
	    
	    bytes = this.BYTES(length);
	    retval = new XdrByteArray(bytes);
	}
	return retval;
    }

    public final String STRING()
    {
	String retval = null;
	int len;

	len = this.readBuffer.getByte(this.cursor);
	this.cursor += BYTE;
	if( len > 0 )
	{
	    byte buf[];
	    
	    buf = this.BYTES(len);
	    retval = new String(buf);
	}
	return retval;
    }

    public final TypeID TYPEID()
    {
	TypeID retval;
	byte hash[];

	hash = this.BYTES(MD5.HASH_LENGTH);
	retval = new TypeID(hash);
	return retval;
    }

    public final Xdr subXDR()
    {
	Xdr retval = new Xdr();

	retval.writeReset(BufferHandle.create());
	return retval;
    }

    public String toString()
    {
	return "Xdr[readBuffer="
	    + this.readBuffer
	    + "; writeBuffer="
	    + this.writeBuffer
	    + "; cursor="
	    + this.cursor
	    + "]";
    }
}
