/*
 * XdrByteArray.java
 *
 * Copyright (c) 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is part of the Janos Bees Software.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */

package bees.core;

/**
 * A ByteArray subclass that supports xdr based encoding.
 */
public class XdrByteArray
    extends ByteArray
{
    /**
     * The current index into the byte array.  This is incremented whenever
     * data is written to or read from the array.
     */
    public int index;

    /**
     * Construct an XdrByteArray that can hold the given number of bytes.
     *
     * @param len The length of byte array to create.
     */
    public XdrByteArray(int len)
    {
	super(len);
    }

    /**
     * Construct an XdrByteArray that wraps the given buffer.
     *
     * @param buf The existing buffer to wrap.
     */
    public XdrByteArray(byte buf[])
    {
	super(buf);
    }
    
    /**
     * Construct an XdrByteArray that wraps the given buffer.
     *
     * @param buf The existing buffer to wrap.
     * @param offset The starting point in the given buffer.
     */
    public XdrByteArray(byte buf[], int offset)
    {
	super(buf, offset);
    }
    
    /**
     * Construct an XdrByteArray that wraps the given buffer.
     *
     * @param buf The existing buffer to wrap.
     * @param offset The starting point in the given buffer.
     * @param len The number of bytes in "buf" to wrap.
     */
    public XdrByteArray(byte buf[], int offset, int len)
    {
	super(buf, offset, len);
    }

    /**
     * @param offset The number of bytes to adjust the current index position
     * by.
     */
    public final void SKIP(int offset)
    {
	this.index += offset;
    }
    
    public final void PUT(boolean x)
    {
	int j = offset + index;
	
	if (index + Xdr.BOOLEAN > len)
	{
	    throw new IndexOutOfBoundsException(
		"Not enough space to encode boolean.");
	}
	if (x)
	    buf[j] = (byte)0xff;
	else
	    buf[j] = (byte)0;
	index += Xdr.BOOLEAN;
    }
    
    public final void PUT(byte x)
    {
	int j = offset + index;
	
	if (index + Xdr.BYTE > len)
	{
	    throw new IndexOutOfBoundsException(
		"Not enough space to encode byte: " +x);
	}
	buf[j] = x;
	index += Xdr.BYTE;
    }
    
    public final void PUT(short x)
    {
	int j = offset + index;
	
	if (index + Xdr.SHORT > len)
	{
	    throw new IndexOutOfBoundsException(
		"Not enough space to encode short: " + x);
	}
	buf[j++] = (byte)(x>>>8);
	buf[j] = (byte)x;
	index += Xdr.SHORT;
    }
    
    public final void PUT(int x)
    {
	int j = offset + index;
	
	if (index + Xdr.INT > len)
	{
	    throw new IndexOutOfBoundsException(
		"Not enough space to encode int: " + x);
	}
	buf[j++] = (byte)(x>>>24);
	buf[j++] = (byte)(x>>>16);
	buf[j++] = (byte)(x>>>8);
	buf[j] = (byte)x;
	index += Xdr.INT;
    }
    
    public final void PUTLE(short x)
    {
	int j = offset + index;
	
	if (index + Xdr.SHORT > len)
	{
	    throw new IndexOutOfBoundsException(
		"Not enough space to encode short: " + x);
	}
	buf[j++] = (byte)x;
	buf[j] = (byte)(x>>>8);
	index += Xdr.SHORT;
    }
    
    public final void PUTLE(int x)
    {
	int j = offset + index;
	
	if (index + Xdr.INT > len)
	{
	    throw new IndexOutOfBoundsException(
		"Not enough space to encode int: " + x);
	}
	buf[j++] = (byte)(x     );
	buf[j++] = (byte)(x>>> 8);
	buf[j++] = (byte)(x>>>16);
	buf[j++] = (byte)(x>>>24);
	index += Xdr.INT;
    }
    
    public final void PUT(float x)
    {
	this.PUT(Float.floatToRawIntBits(x));
    }
    
    public final void PUT(double x)
    {
	this.PUT(Double.doubleToRawLongBits(x));
    }
    
    final void PUT(TypeID x)
    {
	int j = offset + index;
	
	if (index + Xdr.TYPEID > len)
	{
	    throw new IndexOutOfBoundsException(
		"Not enough space to encode TypeID: " + x);
	}
	System.arraycopy(x.getHash(), 0, buf, j, Xdr.TYPEID);
	index += Xdr.TYPEID;
    }
    
    public final void PUT(long x)
    {
	int j = offset + index;
	
	if (index + Xdr.LONG > len)
	{
	    throw new IndexOutOfBoundsException(
		"Not enough space to encode long: " + x);
	}
	buf[j++] = (byte)(x>>>56);
	buf[j++] = (byte)(x>>>48);
	buf[j++] = (byte)(x>>>40);
	buf[j++] = (byte)(x>>>32);
	buf[j++] = (byte)(x>>>24);
	buf[j++] = (byte)(x>>>16);
	buf[j++] = (byte)(x>>>8);
	buf[j] = (byte)x;
	index += Xdr.LONG;
    }
    
    public final void PUTLE(long x)
    {
	int j = offset + index;
	
	if (index + Xdr.LONG > len)
	{
	    throw new IndexOutOfBoundsException(
		"Not enough space to encode long: " + x);
	}
	buf[j++] = (byte)x;
	buf[j++] = (byte)(x>>>8);
	buf[j++] = (byte)(x>>>16);
	buf[j++] = (byte)(x>>>24);
	buf[j++] = (byte)(x>>>32);
	buf[j++] = (byte)(x>>>40);
	buf[j++] = (byte)(x>>>48);
	buf[j++] = (byte)(x>>>56);
	index += Xdr.LONG;
    }
    
    public final void PUT(String x)
    {
	int j = offset + index;
	
	if (index + Xdr.STRING(x) > len)
	{
	    throw new IndexOutOfBoundsException(
		"Not enough space to encode String: " + x);
	}
	if (x == null)
	{
	    buf[j] = (byte)0;
	    index += Xdr.BYTE;
	}
	else
	{
	    byte [] xBytes = x.getBytes(); 
	    
	    buf[j++] = (byte)xBytes.length;
	    System.arraycopy(xBytes, 0, buf, j, xBytes.length);
	    index += Xdr.BYTE + xBytes.length;
	}
    }
    
    public final void PUT(ByteArray x)
    {
	int j = offset + index;
	
	if (index + Xdr.BYTEARRAY(x) > len)
	{
	    throw new IndexOutOfBoundsException(
		"Not enough space to encode ByteArray.");
	}
	if (x == null)
	{
	    buf[j++] = (byte)0; 
	    buf[j] = (byte)0;
	    index += Xdr.SHORT;
	}
	else
	{
	    int l = x.len;
	    
	    buf[j++] = (byte)(l>>>8);
	    buf[j++] = (byte)l;
	    if (x.buf != buf || x.offset != j)
		System.arraycopy(x.buf, x.offset, buf, j, l);
	    index += Xdr.SHORT + l;
	}
    }

    public final void PUT(byte[] x)
    {
	if( x != null )
	    this.PUT(x, 0, x.length);
    }
    
    public final void PUT(byte[] x, int start, int length)
    {
	int j = offset + index;
	
	if (index + length > len)
	{
	    throw new IndexOutOfBoundsException(
		"Not enough space to encode byte[].");
	}
	System.arraycopy(x, start, buf, j, length);
	index += length;
    }

    public final boolean BOOLEAN()
    {
	if (index + Xdr.BOOLEAN > len)
	    throw new IndexOutOfBoundsException("No more data left in xdr.");
	int j = offset + index;
	index += Xdr.BOOLEAN;
	if (buf[j] != 0)
	    return true;
	else
	    return false;
    }
    
    public final byte BYTE()
    { 
	if (index + Xdr.BYTE > len)
	    throw new IndexOutOfBoundsException("No more data left in xdr.");
	int j = offset + index;
	index += Xdr.BYTE;
	return (byte)(buf[j]&0xff);
    }
    
    public final short SHORT()
    {
	if (index + Xdr.SHORT > len)
	    throw new IndexOutOfBoundsException("No more data left in xdr.");
	int j = offset + index;
	index += Xdr.SHORT;
	return (short)((buf[j++]&0xff)<<8 | (buf[j]&0xff));
    }
    
    public final int INT()
    {
	if (index + Xdr.INT > len)
	    throw new IndexOutOfBoundsException("No more data left in xdr.");
	int j = offset + index;
	index += Xdr.INT;
	return (buf[j++]&0xff)<<24 | (buf[j++]&0xff)<<16 |
	    (buf[j++]&0xff)<<8 | buf[j]&0xff;
    }
    
    public final short LESHORT()
    {
	if (index + Xdr.SHORT > len)
	    throw new IndexOutOfBoundsException("No more data left in xdr.");
	int j = offset + index;
	index += Xdr.SHORT;
	return (short)(buf[j++]&0xff | (buf[j++]&0xff) << 8);
    }
    
    public final int LEINT()
    {
	if (index + Xdr.INT > len)
	    throw new IndexOutOfBoundsException("No more data left in xdr.");
	int j = offset + index;
	index += Xdr.INT;
	return (buf[j++]&0xff) | (buf[j++]&0xff)<<8 |
	    (buf[j++]&0xff)<<16 | (buf[j]&0xff)<<24;
    }
    
    public final float FLOAT()
    {
	return Float.intBitsToFloat(this.INT());
    }

    public final double DOUBLE()
    {
	return Double.longBitsToDouble(this.LONG());
    }

    public final long LONG()
    {
	if (index + Xdr.LONG > len)
	    throw new IndexOutOfBoundsException("No more data left in xdr.");
	int j = offset + index;
	index += Xdr.LONG;
	return ((long)buf[j++]&0xff)<<56 | ((long)buf[j++]&0xff)<<48 |
		((long)buf[j++]&0xff)<<40 | ((long)buf[j++]&0xff)<<32 |
		((long)buf[j++]&0xff)<<24 | ((long)buf[j++]&0xff)<<16 |
		((long)buf[j++]&0xff)<<8 | (long)buf[j]&0xff;
    }
    
    public final long LELONG()
    {
	if (index + Xdr.LONG > len)
	    throw new IndexOutOfBoundsException("No more data left in xdr.");
	int j = offset + index;
	index += Xdr.LONG;
	return ((long)buf[j++]&0xff) | ((long)buf[j++]&0xff)<<8 |
		((long)buf[j++]&0xff)<<16 | ((long)buf[j++]&0xff)<<24 |
		((long)buf[j++]&0xff)<<32 | ((long)buf[j++]&0xff)<<40 |
		((long)buf[j++]&0xff)<<48 | ((long)buf[j]&0xff)<<56;
    }
    
    final TypeID TYPEID()
    {
	TypeID retval = new TypeID();
	
	if (index + Xdr.TYPEID > len)
	    throw new IndexOutOfBoundsException("No more data left in xdr.");
	int j = offset + index;
	index += Xdr.TYPEID;
	retval.setHash(buf, j);
	return retval;
    }
    
    public final String STRING()
    {
	// XXX This is bogus
	if (index + Xdr.BYTE > len)
	    throw new IndexOutOfBoundsException("No more data left in xdr.");
	int j = offset + index;
	int l = buf[j++];
	if (index + Xdr.BYTE + l > len)
	    throw new IndexOutOfBoundsException("No more data left in xdr.");
	index += Xdr.BYTE + l;
	if (l == 0)
	    return null;
	else 
	    return new String(buf, j, l);
    }
    
    public final ByteArray BYTEARRAY()
    {
	if (index + Xdr.SHORT > len)
	    throw new IndexOutOfBoundsException("No more data left in xdr.");
	int j = offset + index;
	int l = (buf[j++]&0xff)<<8 | (buf[j++]&0xff);
	if (index + Xdr.SHORT + l > len)
	    throw new IndexOutOfBoundsException("No more data left in xdr.");
	index += Xdr.SHORT;
	ByteArray data = null;
	if (l != 0)
	    data = new ByteArray(buf, offset + index, l);
	index += l;
	return data;
    }

    public final void BYTES(byte[] x, int start, int length)
    {
	if (index + length > len)
	    throw new IndexOutOfBoundsException("No more data left in xdr.");
	int j = offset + index;
	System.arraycopy(buf, j, x, start, length);
	index += length;
    }

    public final byte[] BYTES(int length)
    {
	byte retval[] = new byte[length];

	this.BYTES(retval, 0, length);
	return retval;
    }

    public String toString()
    {
	return "XdrByteArray["
	    + super.toString()
	    + "; index="
	    + this.index
	    + "]";
    }
}
